#
#  Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
#
#
#  @(#)jcc.mk	1.75 00/07/20
#
# Makefile for building JavaCodeCompact(tm) and romjava.c

####################
# Exported variables
####################

# this is exported for the cleanjcc rule
CVM_JCC_CLASSPATH	= $(CVM_TOP)/build/classes.jcc

#CVM_ROMJAVA_O 		= romjava.o
CVM_ROMJAVA_O 		= libromjava.a
CVM_ROMJAVA_CPATTERN	= $(CVM_DERIVEDROOT)/javavm/runtime/romjava
CVM_ROMJAVA_LIST   	= $(CVM_ROMJAVA_CPATTERN)List
CVM_ROMJAVA_CLASSES_PER_FILE = 400

CVM_BUILDDIRS	+= $(CVM_JCC_CLASSPATH)

#################
# Local variables
#################

CVM_JCC_SRCPATH	= $(CVM_TOP)/src/share/javavm/jcc

##############################################
# We have a big set of files derived from 
# opcodes.list.
##############################################

CVM_GENOPCODE_DEPEND	= $(CVM_JCC_CLASSPATH)/GenOpcodes.class
CVM_OPCODE_LIST		= $(CVM_SHAREROOT)/javavm/include/opcodes.list

$(CVM_DERIVEDROOT)/javavm/include/opcodes.h \
$(CVM_DERIVEDROOT)/javavm/runtime/opcodes.c \
$(CVM_DERIVEDROOT)/javavm/runtime/bcattr.c \
$(CVM_DERIVEDROOT)/javavm/runtime/opcodelen.c \
$(CVM_DERIVEDROOT)/javavm/include/opcodeLabels.h \
$(CVM_DERIVEDROOT)/javavm/include/opcodeSimplification.h \
$(CVM_DERIVEDROOT)/javavm/runtime/OpcodeConst.java: \
    $(CVM_OPCODE_LIST) \
    $(CVM_GENOPCODE_DEPEND)
	export CLASSPATH; \
	CLASSPATH=$(CVM_JCC_CLASSPATH); \
	$(CVM_JAVA) GenOpcodes $(CVM_OPCODE_LIST) \
	    -h $(CVM_DERIVEDROOT)/javavm/include/opcodes.h \
	    -c $(CVM_DERIVEDROOT)/javavm/runtime/opcodes.c \
	    -bcAttr $(CVM_DERIVEDROOT)/javavm/runtime/bcattr.c \
	    -opcodeLengths $(CVM_DERIVEDROOT)/javavm/runtime/opcodelen.c \
	    -label $(CVM_DERIVEDROOT)/javavm/include/opcodeLabels.h \
	    -javaConst $(CVM_DERIVEDROOT)/javavm/runtime/OpcodeConst.java \
	    -simplification $(CVM_DERIVEDROOT)/javavm/include/opcodeSimplification.h

$(CVM_GENOPCODE_DEPEND)	= $(CVM_JCC_CLASSPATH)/GenOpcodes.class
$(CVM_GENOPCODE_DEPEND) :: $(CVM_JCC_CLASSPATH)
$(CVM_GENOPCODE_DEPEND) :: $(CVM_JCC_SRCPATH)/GenOpcodes.java \
		 $(CVM_JCC_SRCPATH)/text/*.java \
		 $(CVM_JCC_SRCPATH)/util/*.java \
		 $(CVM_JCC_SRCPATH)/JCCMessage.properties
	@echo "... $@"
	@CLASSPATH=$(CVM_JCC_SRCPATH); export CLASSPATH; \
	$(CVM_JAVAC) -d $(CVM_JCC_CLASSPATH) \
	    $(subst /,$(FILESEP),$(CVM_JCC_SRCPATH)/GenOpcodes.java)
	@rm -f $(CVM_JCC_CLASSPATH)/JCCMessage.properties; \
	cp $(CVM_JCC_SRCPATH)/JCCMessage.properties $(CVM_JCC_CLASSPATH)/JCCMessage.properties


CVM_JCC_DEPEND	= $(CVM_JCC_CLASSPATH)/JavaCodeCompact.class

# 
#  At this time, all natives would be JNI,
#  and we are ROM only -- no impure constant pools allowed,
#  so no constant-pool type-table required.
# 
CVM_GENERATE_OFFSETS	= $(patsubst %,-extraHeaders CVMOffsets %,$(CVM_OFFSETS_CLASSES))
CVM_GENERATE_NATIVES	= $(patsubst %,-nativesType CNI %,$(CVM_CNI_CLASSES))
CVM_JCC_OPTIONS = \
		  $(CVM_GENERATE_NATIVES) \
		  -nativesType JNI "-*" \
		  -headersDir CNI $(CVM_DERIVEDROOT)/cni \
		  -headersDir JNI $(CVM_DERIVEDROOT)/jni \
		  -headersDir CVMOffsets $(CVM_DERIVEDROOT)/offsets \
	          $(CVM_GENERATE_OFFSETS)

ifeq ($(CVM_NO_LOSSY_OPCODES), true)
CVM_JCC_OPTIONS += -qlossless
endif

ifneq ($(CVM_CLASSLOADING), true)
CVM_JCC_OPTIONS += -imageAttribute noClassLoading
endif

ifeq ($(CVM_NO_CODE_COMPACTION), true)
CVM_JCC_OPTIONS += -noCodeCompaction
endif

# NOTE: If class loading is enabled, then only the build time classes need be
#       preloaded.  If class loading is disabled, then it is assumed that
#       preload lib is enabled which means the entire library should be
#       preloaded.

CVM_JCC_INPUT += $(CVM_BUILDTIME_CLASSESZIP)
ifneq ($(CVM_CLASSLOADING), true)
    CVM_JCC_INPUT += $(CVM_TEST_CLASSESZIP)
endif

ifeq ($(CVM_DEBUG_CLASSINFO), true)
CVM_JCC_OPTIONS += -g
endif
# Allow breakpoints in ROMized code
ifeq ($(CVM_JVMDI), true)
CVM_JCC_OPTIONS += -imageAttribute noPureCode
endif

###########
# romjava.c files
###########

$(CVM_ROMJAVA_LIST): $(CVM_JCC_INPUT) $(CVM_JCC_DEPEND)
	@echo "... romjava.c files"
	$(AT)export CLASSPATH; \
	CLASSPATH=$(CVM_JCC_CLASSPATH); \
	$(CVM_JAVA) JavaCodeCompact $(CVM_JCC_OPTIONS) \
		-maxSegmentSize $(CVM_ROMJAVA_CLASSES_PER_FILE) \
		-o $(CVM_ROMJAVA_CPATTERN) $(CVM_JCC_INPUT)

###########
# romjava.o  is made by compiling all the .c files and linking the result
###########
#
# How to get from a list of C files to a single ROMJAVA_O
#
# We so dislike writing contorted make logic, that we'll force
# the compilation system to do a tiny bit more work.
# 1) make CVM_ROMJAVA_O depend on CVM_ROMJAVA_LIST, which contains
#    a list of all the C files created by a JCC run (when segmenting
#    output file.
# 2) provide a rule to get from LIST to O. This uses shell logic
#    to iterate over all the C files, compile them, and, given
#    successful compilation, link them all into a single .o file.
#

$(CVM_OBJDIR)/$(CVM_ROMJAVA_O): $(CVM_ROMJAVA_LIST)
	@for NAME in `cat $(CVM_ROMJAVA_LIST)` ; do  \
		target=`basename $${NAME} .c`.o ; \
		object=$(CVM_OBJDIR)/$${target} ; \
		echo ... $$target ;\
		$(CC) $(CFLAGS) -o $$object $${NAME} ; \
		namelist="$${namelist} $${object}" ; \
	done; \
	echo Linking $@ ; \
	echo $(AR) $(AR_CREATE_FLAGS) $@ $${namelist} ; \
	$(AR) $(AR_CREATE_FLAGS) $@ $${namelist} ; \
	echo $(AR) $(AR_UPDATE_FLAGS) $@ ; \
	$(AR) $(AR_UPDATE_FLAGS) $@

#################
# JavaCodeCompact
#################

CVM_JCC_CLASSES0 = \
	JavaCodeCompact.java \
	runtime/CVMOffsetsHeader.java \
	runtime/CNIHeader.java \
	runtime/JNIHeader.java \
	runtime/CVMWriter.java

CVM_JCC_CLASSES1 = $(patsubst %,$(CVM_JCC_SRCPATH)/%,$(CVM_JCC_CLASSES0))
CVM_JCC_CLASSES  = $(subst /,$(FILESEP),$(CVM_JCC_CLASSES1))

$(CVM_JCC_DEPEND) :: $(CVM_JCC_CLASSPATH)
$(CVM_JCC_DEPEND) :: $(CVM_DERIVEDROOT)/javavm/runtime/OpcodeConst.java
$(CVM_JCC_DEPEND) :: $(CVM_JCC_SRCPATH)/JavaCodeCompact.java \
		 $(CVM_JCC_SRCPATH)/components/*.java \
		 $(CVM_JCC_SRCPATH)/dependenceAnalyzer/*.java \
		 $(CVM_JCC_SRCPATH)/jcc/*.java \
		 $(CVM_JCC_SRCPATH)/runtime/*.java \
		 $(CVM_JCC_SRCPATH)/text/*.java \
		 $(CVM_JCC_SRCPATH)/util/*.java \
		 $(CVM_JCC_SRCPATH)/vm/*.java \
		 $(CVM_JCC_SRCPATH)/JCCMessage.properties
	@echo "... $@"
	@CLASSPATH=$(CVM_JCC_CLASSPATH):$(CVM_JCC_SRCPATH):$(CVM_DERIVEDROOT)/javavm/runtime; export CLASSPATH; \
	$(CVM_JAVAC) -d $(CVM_JCC_CLASSPATH) \
	    $(CVM_JCC_CLASSES)
	@rm -f $(CVM_JCC_CLASSPATH)/JCCMessage.properties; \
	cp $(CVM_JCC_SRCPATH)/JCCMessage.properties $(CVM_JCC_CLASSPATH)/JCCMessage.properties
