/*
 * @(#)exec_md.c	1.7 00/05/19
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include <assert.h>
#include <stdio.h>
#ifdef JDK
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include "sys.h"
#include "util.h"
#endif  

static char *skipWhitespace(char *p) {
#ifdef JDK
    while ((*p != '\0') && isspace(*p)) {
        p++;
    }
    return p;
#else
    fprintf(stderr, "exec-related jdb functionality "
	    "not supported on VxWorks\n");
    assert(0);
    return NULL;
#endif
}

static char *skipNonWhitespace(char *p) {
#ifdef JDK
    while ((*p != '\0') && !isspace(*p)) {
        p++;
    }
    return p;
#else
    fprintf(stderr, "exec-related jdb functionality "
	    "not supported on VxWorks\n");
    assert(0);
    return NULL;
#endif
}

int 
dbgsysExec(char *cmdLine)
{
#ifdef JDK
    int i;
    int argc;
    int pid = -1; /* this is the error return value */
    char **argv = NULL;
    char *p;
    char *args;

    /* Skip leading whitespace */
    cmdLine = skipWhitespace(cmdLine);

    args = jdwpAlloc(strlen(cmdLine)+1);
    if (args == NULL) {
        return SYS_NOMEM;
    }
    strcpy(args, cmdLine);

    p = args;

    argc = 0;
    while (*p != '\0') {
        p = skipNonWhitespace(p);
        argc++;
        if (*p == '\0') {
            break;
        }
        p = skipWhitespace(p);
    }

    /* calloc null terminates for us */
    argv = calloc((argc + 1) * sizeof(char *), 1);
    if (argv == 0) {
        jdwpFree(args);
        return SYS_NOMEM;
    }

    for (i = 0, p = args; i < argc; i++) {
        argv[i] = p;
        p = skipNonWhitespace(p);
        *p++ = '\0';
        p = skipWhitespace(p);
    }

    if ((pid = fork1()) == 0) {
        /* Child process */
        int i, max_fd;

        /* close everything */
        max_fd = sysconf(_SC_OPEN_MAX);
        for (i = 3; i < max_fd; i++) {
            close(i);
        }

        execvp(argv[0], argv);

        exit(-1);
    }
    jdwpFree(args);
    jdwpFree(argv);
    if (pid < 0) {
        return SYS_ERR;
    } else {
        return SYS_OK;
    }
#else
    fprintf(stderr, "exec-related jdb functionality "
	    "not supported on VxWorks\n");
    assert(0);
    return -1;
#endif
}

