package com.timesys;

import javax.realtime.*;
import java.util.*;


public class DMScheduler extends Scheduler
{
    private final int PERIODIC_PRIORITY_BAND_INITIAL = 11;
    private final int PERIODIC_PRIORITY_BAND_FINAL = 266;

    private int[]  priorityQuota;

    private static SchedulableWrapper comparator;
    private static DMScheduler theScheduler=null;
    TreeSet feasibleSet = null;
    TreeSet infeasibleSet = null;

    private class SchedulableWrapper implements Comparator
    {
	public Schedulable sched;
	public int possiblePriority;
	public PeriodicParameters periodic;

	public SchedulableWrapper()
	{
	    super();
	}

	public SchedulableWrapper(Schedulable s, PeriodicParameters periodic, int priority)
	{
	    sched = s;
	    this.periodic = periodic;
	    possiblePriority=priority;
	}

	public int compare(Object o1, Object o2) throws ClassCastException
	{
	    SchedulableWrapper w1,w2;
	    Schedulable s1,s2;

	    w1 = (SchedulableWrapper) o1;
	    w2 = (SchedulableWrapper) o2;

	    try
		{
		    return w1.periodic.getDeadline().compareTo(w2.periodic.getDeadline());
		}
	    catch(NullPointerException e)
		{
		    return 0;
		}
	}

	/*
	public boolean equals(Object o) throws ClassCastException
	{
	    SchedulableWrapper sw = (SchedulableWrapper) o;

	    return periodic.getDeadline().equals(sw.periodic.getDeadline());
	}
	*/
    }

    public DMScheduler()
    {
      comparator = new SchedulableWrapper();
      feasibleSet = new TreeSet((Comparator)comparator);
      infeasibleSet = new TreeSet((Comparator)comparator);
      priorityQuota = new int[PERIODIC_PRIORITY_BAND_FINAL - PERIODIC_PRIORITY_BAND_INITIAL];
    }

    public static DMScheduler instance()
    {
	if (theScheduler == null)
	  {
	    theScheduler = new DMScheduler();
	  }

	return theScheduler;
    }

    private float utilizationBound(SchedulableWrapper schedWrap, int taskSequenceNumber)
    {
	float delta = (schedWrap.periodic.getDeadline().getMilliseconds() * 1000000 + 
		       schedWrap.periodic.getDeadline().getNanoseconds()) /
	              (schedWrap.periodic.getPeriod().getMilliseconds() * 1000000 + 
		       schedWrap.periodic.getPeriod().getNanoseconds());

	if (delta >= 0.5)
	    return ( (float) (taskSequenceNumber 
		     * (java.lang.Math.pow((2 * delta),(1/taskSequenceNumber)) - 1)
		     + 1) -  delta );
	return delta;
    }

    protected boolean addToFeasibility(Schedulable schedulable)
    {
	SchedulableWrapper tmpWrap, schedWrap;
	Iterator iter;
	Schedulable tmpSched;
	float requestedCapacity = 0;
	int sequenceNumber = 1;


	/* if infeasibleSet is not already empty it means that adding this will
	 * not make it feasible. Therefore we just add it to the infeasibleSet
	 */
	if (!infeasibleSet.isEmpty())
	    {
		infeasibleSet.add(new SchedulableWrapper(schedulable,null,0));
		return false;
	    }

	/* Determine the release parameters I am using for the admission control*/

	AbsoluteTime start=null;
	RelativeTime period = null;
	RelativeTime cost = null;
	RelativeTime deadline = null;
	AsyncEventHandler dh = null;

	ReleaseParameters rp = schedulable.getReleaseParameters();
	ProcessingGroupParameters gp = schedulable.getProcessingGroupParameters();

	if (rp != null)
	    {		
		cost   = rp.getCost();
		deadline = rp.getDeadline();
		dh = rp.getDeadlineMissHandler();
		if (rp instanceof PeriodicParameters)
		    {
			period = ((PeriodicParameters)rp).getPeriod();
			start = (AbsoluteTime)((PeriodicParameters)rp).getStart();
		    }
		else if (rp instanceof SporadicParameters)
		    {
			period = ((SporadicParameters)rp).getMinimumInterarrival();
		    }
	    }

	/* for the Aperiodic let's see if we can combine them with the
	 * Processing Group and if that is not the case we will use
	 * the deadline as the period as agreed with the RTJEG
	 */
	if (gp != null)
	    {
		if (period == null)
		    period = gp.getPeriod();
		if (deadline == null)
		    deadline = gp.getDeadline();
		if (cost == null)
		    cost = gp.getCost();
		if (start == null)
		    start = (AbsoluteTime)gp.getStart();
		if (dh == null)
		    dh = gp.getDeadlineMissHandler();
	    }
	else
	    {
		/* Default values */

		if (period == null)
		    period = deadline;
		if (start == null)
		    start = new AbsoluteTime(0,0);
	    }
	
	/* Verify that we have all parameters */

	if (period == null ||
	    cost == null ||
	    deadline == null)
	    {
		/* not feasible because we don't have enough information */

		infeasibleSet.add(new SchedulableWrapper(schedulable,null,0));
		return false;	
	    }


	/* This scheduler will not allow deadline > period even though is possible to have it 
	 * and will be added later on
	 */
	if (deadline.compareTo(period)>0 )
	    {
		infeasibleSet.add(new SchedulableWrapper(schedulable,null,0));
		return false;		
	    }

	schedWrap = new SchedulableWrapper(schedulable, 
					   new PeriodicParameters(start,
								  period,
								  cost,
								  deadline,
								  null,
								  dh),
					   0);
	

	/* we are doing the capacity test */

	iter = feasibleSet.iterator();

	int currentPriority = PERIODIC_PRIORITY_BAND_INITIAL;

	schedWrap.possiblePriority = currentPriority;

	while (iter.hasNext())
	    {
		tmpWrap = (SchedulableWrapper) iter.next();

		try
		    {
			if (tmpWrap.compare(tmpWrap,schedWrap) <= 0)
			    {
				requestedCapacity += 
				    (tmpWrap.periodic.getCost().getMilliseconds() * 1000000 + 
				     tmpWrap.periodic.getCost().getNanoseconds()) /
				    (tmpWrap.periodic.getPeriod().getMilliseconds() * 1000000 + 
				     tmpWrap.periodic.getPeriod().getNanoseconds());

				sequenceNumber++;

				tmpWrap.possiblePriority = currentPriority;

				// We will verify whether the priority goes beyond the band
				currentPriority++;

				schedWrap.possiblePriority = currentPriority;
			    }
			else
			    {
				currentPriority++;
				tmpWrap.possiblePriority = currentPriority;
			    }
		    }
		catch(Exception e)
		    {
			/* a conversion exception happend = set not feasible*/
			infeasibleSet.add(schedWrap);
			return false;			
		    }
	    }
	if (requestedCapacity > utilizationBound(schedWrap, sequenceNumber))
	    {
		/* beyond capacity so not feasible*/
		infeasibleSet.add(schedWrap);
		return false ;
	    }

	feasibleSet.add(schedWrap);

	/* It was feasible so let's check the priority band */
	if (currentPriority > PERIODIC_PRIORITY_BAND_FINAL)
	    {
	      /* Adjust priorities */
	      int i, top = PERIODIC_PRIORITY_BAND_FINAL - PERIODIC_PRIORITY_BAND_INITIAL;
	      for (i=0;i<top; i++)
		{
		  priorityQuota[i] = currentPriority / top;
		  if (i > (top - (currentPriority % top)))
		    {
		      priorityQuota[i]++;
		    }
		}

	      iter = feasibleSet.iterator();
	      i=0;
	      currentPriority = PERIODIC_PRIORITY_BAND_INITIAL;

	      while (iter.hasNext())
		{
		  tmpWrap = (SchedulableWrapper) iter.next();
		  tmpWrap.possiblePriority = currentPriority;

		  priorityQuota[i]--;

		  if (priorityQuota[i] == 0)
		    currentPriority++;

		  i++;
		}

	    }

	/* Assign Priorities */ 
	iter = feasibleSet.iterator();

	while (iter.hasNext())
	    {
		tmpWrap = (SchedulableWrapper) iter.next();
		tmpWrap.sched.setSchedulingParameters((SchedulingParameters)new PriorityParameters(tmpWrap.possiblePriority));
	    }

	return true;
    }


    public boolean isFeasible()
    {
	/* I should try to see if the unfeasibleSet is empty
	 * if that is not the case then is possible that 
	 * the programmer remove some schedulables since the
	 * last time it ask. Therefore we will try to 
	 * re-add the unfeasibleSet to see if we have enough
	 * resources for that
	 */

	if (infeasibleSet.isEmpty())
	    return true;

	TreeSet iSet = (TreeSet)infeasibleSet.clone();
	infeasibleSet.clear();

	Iterator iSetIterator = iSet.iterator();
	while (iSetIterator.hasNext())
	    {
		addToFeasibility(((SchedulableWrapper)iSetIterator.next()).sched);
	    }

	return (infeasibleSet.isEmpty());
    }

    public boolean changeIfFeasible(Schedulable schedulable,
				    ReleaseParameters release,
				    MemoryParameters memory)
    {
      ReleaseParameters tmpRP = schedulable.getReleaseParameters();
      MemoryParameters tmpMP = schedulable.getMemoryParameters();

      schedulable.setReleaseParameters(release);
      schedulable.setMemoryParameters(memory);
      addToFeasibility(schedulable);
      if (!isFeasible())
	{
	  removeFromFeasibility(schedulable);
	  schedulable.setReleaseParameters(tmpRP);
	  schedulable.setMemoryParameters(tmpMP);
	  return false;
	}
      return true;
    }

    private boolean removeFromSet(Schedulable sched, TreeSet set)
    {
	Iterator iter = set.iterator();
	SchedulableWrapper wrap;
	while(iter.hasNext())
	    {
		wrap = (SchedulableWrapper) iter.next();
		if (wrap.sched.equals(sched))
		    {
			set.remove(wrap);
			return true;
		    }
	    }
	return false;
    }

    protected boolean removeFromFeasibility(Schedulable schedulable)
    {
	synchronized(this)
	    {
		if (!removeFromSet(schedulable, infeasibleSet))
		    removeFromSet(schedulable,feasibleSet);
		return isFeasible();
	    }
    }

    public static Scheduler getDefaultScheduler()
    {
	return (Scheduler) instance();
    }

    public java.lang.String getPolicyName()
    {
	return new String("Deadline Monotonic");
    }

    public void fireSchedulable(Schedulable schedulable)
    {
	System.out.println("DMScheduler.java: firing schedulable");
	if (schedulable instanceof NoHeapRealtimeThread)
	    ((NoHeapRealtimeThread)schedulable).start();
	else if (schedulable instanceof RealtimeThread)
	    ((RealtimeThread)schedulable).start();
	else
	    {
		RealtimeThread t = new RealtimeThread(schedulable.getSchedulingParameters(),
						      schedulable.getReleaseParameters(),
						      schedulable.getMemoryParameters(),
						      null, // memory area
						      schedulable.getProcessingGroupParameters(),
						      schedulable);
		addToFeasibility(t);
		//if (isFeasible())
		t.start();
		// else
		// What should I do? 
	    }
    }    
}
