//package javax.realtime;
package com.timesys; 

import javax.realtime.*;

/**
 * A generic filter implementation.
 */
public class DefaultPhysicalMemoryFilter implements PhysicalMemoryTypeFilter 
{
    //private Object type;
    private long base;
    private long size;

    // Highly Linux dependent values! 
    // To be used as attributes for mmap
    static final int PROT_READ  = 1;     // Page can be read.  
    static final int PROT_WRITE = 2;     // Page can be written.  
    static final int PROT_EXEC  = 4;     // Page can be executed.  
    static final int PROT_NONE  = 0;     // Page can not be accessed.  

    static final int MAP_SHARED = 1;        // Share changes.  
    static final int MAP_PRIVATE= 2;        // Changes are private.  
    static final int MAP_FIXED  = 16;       // Interpret addr exactly.  
    
 
    public DefaultPhysicalMemoryFilter(long base, long size) //throws DuplicateFilterException
    {
      //  this.type = type;
        this.base = base;
        this.size = size;
        //PhysicalMemoryManager.registerFilter(type, this);
    }
    
    /** 
     * Search for memory of the right type.
     *
     * @param base start searching here
     * @param size look for at least this much memory
     * @return The address where memory was found or -1 if it was not found.
     */
    public long find(long base, long size)
    {
        long reqEnd = base +size;
        long myEnd = this.base+this.size;    
        
        if(this.size < size)
        {
            // there's no way we can fulfill the request
            //System.out.println("DPMF: this.size<size");
            return -1;
        }
        
        /*if(this.base > base)
        {
            System.out.println("DPMF: this.base>base");
            return -1;
        }*/
 
        if(base >= this.base && reqEnd <= myEnd)
        {
            // the requested area is within my limits. Return the base passed.
            // We accept.
            return base;
        }

        return this.base;
    }
    
    /** Search for virtual memory of the right type.
     *  This is important for systems where attributes are associated with 
     *  particular ranges of virtual memory.
     *  @param base start searching here
     *  @param size look for at least this much memory
     *  @return the address where memory was found or -1 if it was not found.
     */
    public long vFind(long base, long size)
    {
        return base;
    }
    
    /** If configuration is required for memory to fit the attribute of this object,
     *  do the configuration here.
     *  @param base Address of the beginning of the physical memory region.
     *  @param vBase Address of the beginning of the virtual memory region.
     *  @param size Size of the memory region
     *  @throws IllegalArgumentException Thrown if the base and size do not fall 
     *  into this type of memory
     */
    public void initialize(long base, long vBase, long size)
    {
    }
    
    /** If this type of memory is removable, return true.
     * @return true if this type of memory is removable.
     */
    public boolean isRemovable()
    {
        // No. I will not allow you to remove this memory. :)
        return false;
    }
    
    /** Does the specified range of memory contain any of this type?
     *  @param base physical address of the beginning of the memory region
     *  @param size size of the memory region
     *  @return true if the specified range contains ANY of this type of memory.
     */
    public boolean contains(long base, long size)
    {
        return true;
    }
    
    /** Arrange for the specified AEH to be called if any memory in the specified
     *  range is removed.
     *  @param physical base address of the beginning of the memory region
     *  @param size size of the memory region
     *  @param aeh Run this if any memory in the specified range is removed.
     *  @exception IllegalArgumentException if the base and size do not fall into
     *  this type of memory
     */
    public void onRemoval(long base, long size, AsyncEventHandler aeh)
    {
    }
    
    /** Arrange for the specified AEH to be called if any memory in the specified
     *  range is inserted.
     *  @param physical base address of the beginning of the memory region
     *  @param size size of the memory region
     *  @param aeh Run this if any memory in the specified range is inserted.
     *  @exception IllegalArgumentException if the base and size do not fall into
     *  this type of memory
     */
    public void onInsertion(long base, long size, AsyncEventHandler aeh)
    {
    }
    
    /** Is all of the specified range of physical memory present in the system?
     *  (If any of it has been removed, return false.)
     *  @param physical base address of the beginning of the memory region
     *  @param size size of the memory region
     *  @param aeh Run this if any memory in the specified range is inserted.
     *  @exception IllegalArgumentException if the base and size do not fall into
     *  this type of memory
     */
    public boolean isPresent(long base, long size)
    {
        return true;
    }
    
    public int getVMAttributes()
    {
        return PROT_READ | 
                PROT_WRITE;
    }

    public int getVMFlags()
    {
        return MAP_SHARED;
    }

}
    
