package com.timesys;

import javax.realtime.*;

import java.util.*;

public class RMScheduler extends Scheduler
{
    /*
    static {
	System.loadLibrary("RMScheduler");
    }
    */

    static RMScheduler theScheduler = null;

    static HashSet feasibleSet = null;
    static HashSet unfeasibleSet = null;
  
    public native int deleteResourceSet0(int rs);
    public native int createResourceSet0();
    public native int createCPUReserve0(int rs, long start_milliseconds, 
					int start_nanosecons, 
					long period_milliseconds,
					int period_nanoseconds,
					long cost_milliseconds,
					int cost_nanoseconds,
					long deadline_milliseconds,
					int deadline_nanoseconds);
    public RMScheduler()
    {
    }

    public void fireSchedulable(Schedulable schedulable)
    {
    }

    protected void addToFeasibility(Schedulable schedulable)
    {
	ProcessingGroupParameters gp;
	ReleaseParameters rp;

	/* Create the resource set and reserve and
	 * store them in the execution environment of the thread
	 */
	int rs =0 ;
	int cpuRsv =0;

	synchronized (this)
	    {
		if (feasibleSet == null)
		    {
			feasibleSet = new HashSet();
		    }
		if (unfeasibleSet == null)
		    {
			unfeasibleSet = new HashSet();
		    }
	    }

	synchronized(feasibleSet)
	    {
		if (feasibleSet.contains(schedulable))
		    return; 
	    }

	synchronized(unfeasibleSet)
	    {
		if (unfeasibleSet.contains(schedulable))
		    return;
	    }

	rp = schedulable.getReleaseParameters();
	gp = schedulable.getProcessingGroupParameters();

	if (gp != null)
	    {
		if(!((rp instanceof AperiodicParameters) || 
		     (rp instanceof SporadicParameters)))
		    {
			/* error */
			return ; // throw(new com.timesys.WrongArgumentsException());
		    }
		/* create the cpu reserve */
		if (gp.getRsHandle() == 0)
		    {
			rs = createResourceSet0();
			cpuRsv = createCPUReserve0(rs, gp.getStart().getMilliseconds(), 
						   gp.getStart().getNanoseconds(), 
						   gp.getPeriod().getMilliseconds(), 
						   gp.getPeriod().getNanoseconds(), 
						   gp.getCost().getMilliseconds(), 
						   gp.getCost().getNanoseconds(), 
						   gp.getDeadline().getMilliseconds(), 
						   gp.getDeadline().getNanoseconds());		
			
			if (cpuRsv == -1)
			    {
				deleteResourceSet0(rs);
				synchronized(unfeasibleSet)
				    {
					unfeasibleSet.add(schedulable);
				    }

				return ; //throw(new ResourceLimitError());
			    }

			/* attach it to the ProcessingGroup */
			gp.setRsHandle(rs);
		    }
		else
		    rs = gp.getRsHandle();
		schedulable.setResourceSet(rs);
	    }
	else if (rp instanceof PeriodicParameters)
	    {
		PeriodicParameters pp = (PeriodicParameters) rp;

		rs = createResourceSet0();
		if( rs == -1)
		    {
			synchronized(unfeasibleSet)
			    {
				unfeasibleSet.add(schedulable);
			    }
			return;
		    }
		
		/* Uncomment for debugging 
		   System.out.println("RMScheduler.java: calling createCPUReserve0 with:");
		   System.out.println("rs("+rs+") start.millis("+
		   pp.getStart().getMilliseconds()+
		   ") start.nanos("+pp.getStart().getNanoseconds()+
		   ") period.millis("+pp.getPeriod().milliseconds+
		   ") period.nanos ("+pp.getPeriod().nanoseconds+
		   ") cost.millis ("+pp.getCost().milliseconds+
		   ") cost.nanos ("+pp.getCost().nanoseconds +
		   ") deadline.millis ("+pp.getDeadline().milliseconds+
		   ") deadline.nanos ("+pp.getDeadline().nanoseconds+
		   ")"
		   );
		*/
		
		cpuRsv = createCPUReserve0(rs, pp.getStart().getMilliseconds(), 
					   pp.getStart().getNanoseconds(), 
					   pp.getPeriod().getMilliseconds(), 
					   pp.getPeriod().getNanoseconds(), 
					   pp.getCost().getMilliseconds(), 
					   pp.getCost().getNanoseconds(), 
					   pp.getDeadline().getMilliseconds(), 
					   pp.getDeadline().getNanoseconds());
		
		if (cpuRsv == -1)
		    {
			/* delete the resource set */
			deleteResourceSet0(rs);
			synchronized(unfeasibleSet)
			    {
				unfeasibleSet.add(schedulable);
			    }
			return;
			// throw(new ResourceLimitError());
		    }

		schedulable.setResourceSet(rs);
	    }
        synchronized(feasibleSet)
	{
	    feasibleSet.add(schedulable);
	}

    }

    public boolean changeIfFeasible(Schedulable schedulable,
				    ReleaseParameters release,
				    MemoryParameters memory)
    {
	ReleaseParameters tmpRP = schedulable.getReleaseParameters();
	MemoryParameters tmpMP = schedulable.getMemoryParameters();
	schedulable.setReleaseParameters(release);
	schedulable.setMemoryParameters(memory);
	addToFeasibility(schedulable);
	if (isFeasible())
	    return true;

	schedulable.setReleaseParameters(tmpRP);
	schedulable.setMemoryParameters(tmpMP);
	unfeasibleSet.remove(schedulable);

	return false;
    }

    public static Scheduler getDefaultScheduler()
    {
	if (theScheduler == null)
	    theScheduler = new RMScheduler();
	return theScheduler;
    }

    public java.lang.String getPolicyName()
    {
	return new String("RK CPU Reservation");
    }

    public synchronized boolean isFeasible()
    {
	/* I should try to see if the unfeasibleSet is empty
	 * if that is not the case then is possible that 
	 * the programmer remove some schedulables since the
	 * last time it ask. Therefore we will try to 
	 * re-add the unfeasibleSet to see if we have enough 
	 * resources for that
	 */
	if (unfeasibleSet.isEmpty())
	    return true;

	HashSet uSet = unfeasibleSet;
	unfeasibleSet = null;
	Iterator uSetIterator = uSet.iterator();
	for (Schedulable s = (Schedulable) uSetIterator.next(); 
	     uSetIterator.hasNext() ; 
	     s = (Schedulable) uSetIterator.next())
	    {
		uSetIterator.remove();
		addToFeasibility(s);
	    }
	
	return unfeasibleSet.isEmpty();
    }

    protected void removeFromFeasibility(Schedulable schedulable)
    {
	ProcessingGroupParameters gp;

	/* verify if it is in the set */
 	synchronized (this)
	    {
		if (feasibleSet == null)
		    {
		    feasibleSet = new HashSet();
		    }
		if (unfeasibleSet == null)
		    {
			unfeasibleSet = new HashSet();
		    }
	    }

	synchronized(feasibleSet)
	    {
		if (feasibleSet.contains(schedulable))
		    {
		      /* delete the rs only if is not the one in the ProcessingGroupParameter */ 
		      gp = schedulable.getProcessingGroupParameters();
		      if (gp != null)
			  {
			      if ((schedulable.getResourceSet() != gp.getRsHandle()) &&
				  schedulable.getResourceSet() != 0)
				  {
				      /* delete the rs */
				      deleteResourceSet0(schedulable.getResourceSet());
				  }
				  
			  }
		      /* delete from set */
		      feasibleSet.remove(schedulable);
		      return;
		    }
	    }

	synchronized(unfeasibleSet)
	    {
		unfeasibleSet.remove(schedulable);
	    }
   }

}


