/*
 * -----------------------------------------------------------------
 * GarbageCollector.java : Abstract java class encapsulating the
 * characteristics of the VM garbage collector.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
 * The system shall provide dynamic and static information 
 * characterizing the temporal behavior and imposed overhead of any 
 * garbage collection algorithm provided by the system.  This 
 * information shall be made available to applications via methods on 
 * subclasses of <code>GarbageCollector</code>. Implementations are 
 * allowed to provide any set of methods in subclasses as long as the
 * temporal behavior and overhead are sufficiently categorized.  The 
 * implementations are also required to fully document the subclasses.
 * In addition, the method(s) in <code>GarbageCollector</code> shall be
 * made available by all implementations.  
 *
 * @see MarkAndSweepCollector
 * @author Pratik Solanki
 */

public abstract class GarbageCollector
{

  public GarbageCollector()
  {
  }

  /**
   * Preemption latency is a measure of the maximum time a {@link RealtimeThread}
   * may have to wait for the collector to reach a preemption-safe point.
   * Instances of {@link RealtimeThread} are allowed to preeempt
   * the garbage colllector (instances of {@link NoHeapRealtimeThread}
   * preempt immediately but instances of {@link RealtimeThread}
   * must wait until the collector reaches a preemption-safe point). 
   *
   * @return The preempting latency of this if applicable. May return 0
   * if there is no collector available
   */
  public abstract RelativeTime getPreemptionLatency();
}
