/*
 * -----------------------------------------------------------------
 * OneShotTimer.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
 * A timed AsyncEvent that is driven by a clock. It will fire off once,
 * when the clock time reaches the timeout time. If the clock time has
 * already passed the timeout time, it will fire immediately.
 */


public class OneShotTimer extends Timer{

	/** Create an instance of AsyncEvent that will excute its fire method at the
	* expiration of the given time.
	* @param  time - After timeout time units from 'now' fire will be excuted
	* @param  handler - The AsyncEventHandler that will be scheduled when fire is excuted
	*/
	public OneShotTimer(HighResolutionTime time ,AsyncEventHandler handler){

		super(time,null,handler);
		if(this.time==null)
			this.time = Clock.getRealtimeClock().getTime();


	}

	/**
	* Create an instance of AsyncEvent, based on the given clock, that will excute
	* its fire method at the expiration of the given time.
	* @param  start   start time for timer
	* @param  clock   The timer will increment based on this clock
	* @param  handler The AsyncEventHandler that will be scheduled when fire is excuted
	*/
	public OneShotTimer(HighResolutionTime start, Clock clock, AsyncEventHandler handler){

		super(start,clock,handler);
		if(this.time==null)
			this.time  = Clock.getRealtimeClock().getTime();

	}

	/**
	*  Called from super class start method
	*/
	protected void Init(){//OneShot

	    while(true){

		//sleep until action time..!
		at = time.absolute(clock,at);
		AbsoluteTime now = clock.getTime();
	
		//System.out.println(at.toString());
		//System.out.println(now.toString());

		RelativeTime rt = at.subtract(now);
		long milli = rt.getMilliseconds();
		int nano = rt.getNanoseconds();
		//System.out.println(rt.toString());

		if((milli<0 || nano<0 || (milli==0 && nano==0)) && isEnabled){
			fire(); //AsyncEvent's method
			return;
		}

		if(isEnabled){
			try{
				waitLock.wait(milli,nano);
				if(rescheduled){
					rescheduled = false;
					continue;
				}
			}
			catch(Exception e){
				e.printStackTrace();
			}
			fire(); //AsyncEvent's method
			return;
		}
             }//endof while loop.
	
	}


}


