/*
 * -----------------------------------------------------------------
 * PriorityCeilingEmulation.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
  Monitor control class specifying use of the priority ceiling emulation
  protocol for monitor objects.  Objects under the influence of this
  protocol have the effect that a thread entering the monitor has its
  effective priority -- for priority-based dispatching -- raised to the
  ceiling on entry, and is restored to its previous effective priority when
  it exits the monitor.

  See also {@link MonitorControl} and {@link PriorityInheritance}.
  
 */

public class PriorityCeilingEmulation extends MonitorControl {
   /**
      The priority ceiling value.
     */
   int ceiling = 0;
    
    /**
      Create a <code>PriorityCeilingEmulation</code> object with a given ceiling.
      @param ceiling Priority ceiling value.
     */
    public PriorityCeilingEmulation(int ceiling)
    {
	this.ceiling = ceiling;
    }
    
    /**
      Get the priority ceiling for this <code>PriorityCeilingEmulation</code> object.
      */
    public int getDefaultCeiling()
    {
	return ceiling;
    }
}
