/*
 * -----------------------------------------------------------------
 * RealtimeSecurity.java : Security policy object for real-time 
 * specific issues.  Primarily used to control access 
 * to physical memory.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

import java.io.*;
import java.util.StringTokenizer;

/** 
  Security policy object for real-time specific issues.  Primarily used to control
  access to physical memory.
  */
public class RealtimeSecurity
{
	boolean ret = true;

    /* Constructor: currently does nothing */
	public RealtimeSecurity()
	{
	}
    
    /** 
      Check whether the application is allowed to access physical memory.
      @exception	SecurityException	the application doesn't have permission.
      */
	public void checkAccessPhysical() throws SecurityException
	{
		try {
			FileInputStream fs = new FileInputStream("/dev/mem");
		}catch(Exception e) {
			throw new SecurityException();
		}
	}
	
    /** 
      Check whether the application is allowed to access physical memory
      within the specified range.
      @exception	SecurityException	the application doesn't have permission.
      */
	public void checkAccessPhysicalRange(long base, long size) throws SecurityException
	{
		String s;
		try{
			BufferedReader in = new BufferedReader(new FileReader("/proc/meminfo"));
		in.readLine();
		s = in.readLine();
		}catch(FileNotFoundException fnf){
			return;
		}catch(IOException ie){
			return;
		}

		StringTokenizer st = new StringTokenizer(s);
		st.nextToken();
		String mem = st.nextToken();
		int memSize = new Integer(mem).intValue();
		
		if(memSize < base)
			throw new SecurityException();
		if(memSize < base + size)
			throw new SecurityException();
	}
	
    /** 
      Check whether the application is allowed to set filter objects.
      @exception	SecurityException	the application doesn't have permission.
      */
	public void checkSetFilter() throws SecurityException
	{
		if(!ret)
			throw new SecurityException();
		checkAccessPhysical();
		return;
	}

    /** 
      Check whether the application is allowed to set factory objects.
      @exception	SecurityException	the application doesn't have permission.
      */
	void checkSetFactory() throws SecurityException
	{
		if(!ret)
			throw new SecurityException();
		checkAccessPhysical();
		return;
	}
	
    /** 
      Check whether the application is allowed to set the scheduler.
      @exception	SecurityException	the application doesn't have permission.
      */
	public void checkSetScheduler() throws SecurityException
	{
		if(!ret)
			throw new SecurityException();
		checkAccessPhysical();
		return;
	}
}
