#ifndef _INCLUDED_MEM_REGISTRY_H
#define _INCLUDED_MEM_REGISTRY_H


struct FinalizableObject
{
	CVMObject *reference;
	struct FinalizableObject *next;	
};

struct GlobalRefList
{
    CVMObjectICell *ref;
    struct GlobalRefList *next;
};

struct MemoryAreaInfo
{
  CVMObjectICell *memoryArea;
  char *memBase;
  char *currentPointer;
  CVMJavaLong memSize;
  CVMJavaLong memAllocated;
  CVMJavaLong memRemaining;

  // This is the ScopedMemory portal object
  CVMObject *portal;

//  MemoryType memType;
  enum 
  {
    MEMTYPE_UNKNOWN,
    MEMTYPE_HEAP,
    MEMTYPE_IMMORTAL,
    MEMTYPE_SCOPED
  } memType;  
  struct MemoryAreaInfo *parent;
  //struct MemoryAreaInfo *child;
  struct FinalizableObject *finalizeListHead;
  struct GlobalRefList *reflistHead;
  struct MemoryAreaInfo *next;
  struct MemoryAreaInfo *prev;
};

struct PerThreadMemoryInfo
{
  struct MemoryAreaInfo *info;
  struct PerThreadMemoryInfo *upInStack;
  struct PerThreadMemoryInfo *downInStack;
};


typedef struct MemoryAreaInfo MemoryAreaInfo;
typedef struct PerThreadMemoryInfo PerThreadMemoryInfo;
typedef struct FinalizableObject FinalizableObject;
typedef struct GlobalRefList GlobalRefList;


/* Function Declarations */
CVMBool CVMJTregistryRegisterMemoryArea(CVMExecEnv *ee, CVMObjectICell *memoryArea,
										CVMJavaLong base, CVMJavaLong size, int memType);

CVMBool CVMJTregistryUnregisterMemoryArea(CVMExecEnv *ee, CVMObjectICell *memoryArea);

CVMObject* CVMJTregistryInvokeNewInstance(CVMExecEnv *ee, CVMObjectICell *memoryAreaicell, CVMUint32 size);

int CVMJTregistryScanAndWalkAllRoots(void *arg);

int CVMJTregistryHandleReference(void *arg);

#define JT_gcunsafeIsSame(ee_, icell1_, icell2_, res_)   \
{ \
      /*CVMObject *obj1_;*/ \
      /*CVMObject *obj2_;*/ \
      /*if(icell1_ == icell2_);*/ \
      CVMObject *obj1_ = CVMID_icellDirect(ee_, icell1_); \
      CVMObject *obj2_ = CVMID_icellDirect(ee_, icell2_); \
      res_ = (obj1_==obj2_); \
}
    
/* Reference checking functions */
CVMBool CVMJTrcAllowReferenceAssignment(CVMExecEnv *ee, CVMObject *dstObj, CVMObject *srcObj);
CVMBool CVMJTrcAllowLocalVarAssignment(CVMExecEnv *ee, CVMObject *directObj);

CVMBool CVMJTregistryAddFinalizableObject(CVMExecEnv *ee, CVMObject *newInstance);
void CVMJTregistryInvokeAllFinalizers(CVMExecEnv *ee, CVMObjectICell *mem);
CVMBool CVMJTisHeapReference(CVMExecEnv *ee, CVMObject *obj);

PerThreadMemoryInfo* CVMJTregistryCreatePTMI(CVMExecEnv *ee);
void CVMJTregistryDeletePTMI(PerThreadMemoryInfo *info);

CVMObjectICell* CVMJTregistryGetMemoryAreaForObject(CVMExecEnv *ee, CVMObjectICell *obj);
MemoryAreaInfo* CVMJTregistryGetMemoryAreaInfoForObject(CVMExecEnv *ee, CVMObjectICell *obj);

CVMBool CVMJTtreeAddEdge(CVMExecEnv *ee, MemoryAreaInfo *child, 
                            MemoryAreaInfo *parent, int refCount);
CVMBool CVMJTtreeRemoveEdge(CVMExecEnv* ee, MemoryAreaInfo *child, MemoryAreaInfo *parent);

MemoryAreaInfo* CVMJTregistryLookupMemoryArea(CVMExecEnv *ee, CVMObjectICell *memoryAreaicell);

void initialiseMemoryRegistry(CVMExecEnv *ee);
void CVMJTregistryCreateHeapMemoryAreaInfo(CVMExecEnv* ee, CVMObjectICell* theHeap);
void CVMJTregistryReclaimMemory(CVMExecEnv *ee, CVMObjectICell *memoryArea);

void CVMJTregistryRegisterImmortal(CVMExecEnv *ee, CVMObjectICell *immortalMemory,
                                    CVMJavaLong base, CVMJavaLong size, int immortalObjSize);


CVMJavaLong CVMJTregistryGetMemoryAllocated(CVMExecEnv *ee, CVMObjectICell *memoryAreaicell);
CVMJavaLong CVMJTregistryGetMemoryRemaining(CVMExecEnv *ee, CVMObjectICell *memoryAreaicell);

// ScopedMemory get/set Portal functions
CVMObjectICell* CVMJTsmGetPortal(CVMExecEnv* ee, CVMObjectICell* scopedMemory);
void CVMJTsmSetPortal(CVMExecEnv* ee, CVMObjectICell* scopedMemory, CVMObjectICell* portalObject);

    
/* Declaration of some global variables.
 */
extern CVMObjectICell *jtImmortalMemory;


#define DECLARE_VARS_FOR_MA_SLIDING(ee_) \
    PerThreadMemoryInfo *saved = NULL; \
    PerThreadMemoryInfo *savedDownInStack = NULL; \
    PerThreadMemoryInfo *traverser = NULL; \
    CVMBool deletePTMI = CVM_FALSE; \
    MemoryAreaInfo *info = NULL


// Macro that will run 'action_' in ImmortalMemory.
// Sort of like the C equivalent of calling enter on ImmortalMemory.
#define ExecuteInImmortal(ee_, action_) \
    { \
        extern int immortalInitialized; \
        int isInitialized = immortalInitialized; \
     /*   CVMObjectICell oldMemoryArea; */\
        CVMBool oldCheckConstraints; \
        if(isInitialized) \
        { \
            DECLARE_VARS_FOR_MA_SLIDING(ee_); \
            \
            oldCheckConstraints = CVMcheckForConstraints(ee_);\
            CVMdisableCheckForConstraints(ee_);\
            \
            SLIDE_UP_TO_MA(ee_, jtImmortalMemory); \
            CVMassert(!CVMlocalExceptionOccurred(ee_)); \
            action_; \
            SLIDE_BACK_DOWN(ee_); \
            \
            CVMsetCheckForConstraints(ee_, oldCheckConstraints);\
            \
        } \
        else \
        { \
            action_; \
        } \
    }

            
/*        { \
            CVMID_icellAssign(ee_, &oldMemoryArea, CVMmemoryAreaICell(ee_)); \
            CVMID_icellAssign(ee_, CVMmemoryAreaICell(ee_), jtImmortalMemory); \
            oldCheckConstraints = CVMcheckForConstraints(ee_);\
            CVMdisableCheckForConstraints(ee_);\
        } \
        action_; \
        if(isInitialized) \
        { \
            CVMID_icellAssign(ee_, CVMmemoryAreaICell(ee_), &oldMemoryArea); \
            CVMsetCheckForConstraints(ee_, oldCheckConstraints);\
        } \
    }*/

#define ExecuteInImmortalGcUnsafe(ee_, action_) \
    { \
        extern int immortalInitialized; \
        int isInitialized = immortalInitialized; \
       /* CVMObjectICell oldMemoryArea; */\
        CVMBool oldCheckConstraints; \
        if(isInitialized) \
        { \
            DECLARE_VARS_FOR_MA_SLIDING(ee_); \
            \
            oldCheckConstraints = CVMcheckForConstraints(ee_);\
            CVMdisableCheckForConstraints(ee_);\
            \
            SLIDE_UP_TO_MA(ee_, jtImmortalMemory); \
            CVMassert(!CVMlocalExceptionOccurred(ee_)); \
            action_; \
            SLIDE_BACK_DOWN(ee_); \
            \
            CVMsetCheckForConstraints(ee_, oldCheckConstraints);\
            \
        } \
        else \
        { \
            action_; \
        } \
    }


      /*      CVMID_icellAssignDirect(ee_, &oldMemoryArea, CVMmemoryAreaICell(ee_)); \
            CVMID_icellAssignDirect(ee_, CVMmemoryAreaICell(ee_), jtImmortalMemory); \
            oldCheckConstraints = CVMcheckForConstraints(ee_);\
            CVMdisableCheckForConstraints(ee_);\
        } \
        action_; \
        if(isInitialized) \
        { \
            CVMID_icellAssignDirect(ee_, CVMmemoryAreaICell(ee_), &oldMemoryArea); \
            CVMsetCheckForConstraints(ee_, oldCheckConstraints);\
        } \
    }*/

#define CheckAndExecuteInImmortal(ee_,action_) \
{ \
    if(CVMD_isgcUnsafe(ee_)) \
    { \
        ExecuteInImmortalGcUnsafe(ee_,action_) \
    } \
    else \
    { \
        ExecuteInImmortal(ee_, action_) \
    } \
}

   
#define SLIDE_UP_TO_MA(ee_, memoryAreaICell_) \
    CVMassert(!CVMlocalExceptionOccurred(ee_));\
    saved = ee_->perThreadStack; \
    traverser = saved; \
    info = CVMJTregistryLookupMemoryArea(ee_, memoryAreaICell_);\
    \
    if(info->memType == MEMTYPE_HEAP ||     \
       info->memType == MEMTYPE_IMMORTAL) \
    { \
        /* No need to travel up the stack. Just save the stack and start afresh */ \
        /*CVMassert(info->memType == MEMTYPE_HEAP ||     \
                   info->memType == MEMTYPE_IMMORTAL); */\
        traverser = NULL; \
    } \
    else \
    { \
        /* We need to slide up */ \
        while(traverser != NULL) \
        { \
          CVMBool result = CVM_FALSE;\
          if(CVMD_isgcSafe(ee_))    \
          { \
              CVMID_icellSameObject(ee_, (traverser->info->memoryArea), (memoryAreaICell_), result);\
          } \
          else \
          { \
              JT_gcunsafeIsSame(ee_, traverser->info->memoryArea, (memoryAreaICell_), result) \
          } \
          if(result) \
              break; \
          traverser = traverser->upInStack; \
        } \
    } \
    if(traverser == NULL) \
    { \
        if(info->memType == MEMTYPE_SCOPED) \
        { \
            /* Error */ \
            /*CVMthrowMemoryAccessError(ee_, "Unable to slide to Scope that isn't on stack.");   */ \
            CVMthrowInaccessibleAreaException(ee_, "Unable to slide to Scope that isn't on stack.");    \
        } \
        else \
        { \
            PerThreadMemoryInfo *newRoot = CVMJTregistryCreatePTMI(ee_); \
            newRoot->info = info; \
            deletePTMI = CVM_TRUE; \
            traverser = newRoot; \
        } \
    } \
    if(!CVMlocalExceptionOccurred(ee_)) \
    { \
        savedDownInStack = traverser->downInStack; \
        traverser->downInStack = NULL; \
        ee_->perThreadStack = traverser; \
        if(CVMD_isgcUnsafe(ee_)) \
            { CVMID_icellAssignDirect(ee_, CVMmemoryAreaICell(ee_), ee_->perThreadStack->info->memoryArea); } \
        else \
            { CVMID_icellAssign(ee_, CVMmemoryAreaICell(ee_), ee_->perThreadStack->info->memoryArea); } \
    }


#define SLIDE_BACK_DOWN(ee_) \
    CVMassert(ee->perThreadStack->downInStack == NULL); \
    ee_->perThreadStack->downInStack = savedDownInStack; \
    if(deletePTMI) \
        CVMJTregistryDeletePTMI(ee_->perThreadStack); \
    ee_->perThreadStack = saved; \
    if(saved != NULL) \
    { \
        if(CVMD_isgcUnsafe(ee_)) \
            { CVMID_icellAssignDirect(ee_, CVMmemoryAreaICell(ee_), ee_->perThreadStack->info->memoryArea); } \
        else \
            { CVMID_icellAssign(ee_, CVMmemoryAreaICell(ee_), ee_->perThreadStack->info->memoryArea); } \
    } \
    else \
    { \
        CVMID_icellSetNull(CVMmemoryAreaICell(ee_))\
    }


#define SLIDE_UP_AND_EXECUTE(ee_, memoryAreaICell_, action_) \
    { \
        DECLARE_VARS_FOR_MA_SLIDING(ee_); \
        /*MemoryAreaInfo* memArea_ = CVMJTregistryLookupMemoryArea(ee_, memoryAreaICell_); \
        if(memArea_ == NULL) \
        { \
            printf("I COULDNT FIND THE MEMORY AREA\n"); \
        }    \
        if(memArea_->memType == MEMTYPE_IMMORTAL || \
                memArea_->memType == MEMTYPE_HEAP) \
        { \
            dont_slide_ = CVM_TRUE; \
        } */\
        /*printf("***Before sliding up\n"); */\
        SLIDE_UP_TO_MA(ee_, memoryAreaICell_); \
    /*    printf("***Successfully slid up\n");*/ \
        if(!CVMlocalExceptionOccurred(ee_))\
        { \
            action_; \
  /*          printf("***Successfully called action\n"); */\
            SLIDE_BACK_DOWN(ee_); \
/*            printf("***Successfully slid down\n"); */\
        } \
    } \



#endif //#ifdef MEM_REGISTRY_H__INCLUDED
