/*
 * @(#)threads_md.c	1.8 00/11/07
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/porting/float.h"
#include "javavm/include/porting/threads.h"
#include "javavm/include/porting/sync.h"
#include <pthread.h>
#include <signal.h>
#include <assert.h>
#include <stdio.h>
#include <rk/rk.h>

void
CVMthreadYield(void)
{
    thr_yield();
}

void
CVMthreadSuspend(CVMThreadID *t)
{
    linuxSyncSuspend(t);
}

void
CVMthreadResume(CVMThreadID *t)
{
    linuxSyncResume(t);
}

#include <sys/resource.h>
#include <unistd.h>

static void
LINUXcomputeStackTop(CVMThreadID *self)
{
    pid_t pid = getpid();
    char buf[32];
    FILE *fp;
    void *sp = &self;

    sprintf(buf, "/proc/%d/maps", pid);
    fp = fopen(buf, "r");
    if (fp != NULL) {
	unsigned int last_high = 0;
	unsigned int low, high, offset;
	struct rlimit rlim;

	getrlimit(RLIMIT_STACK, &rlim);
	while (1) {
	    int c;
	    int tokens = fscanf(fp, "%x-%x %*4c %x %*d:%*d %*d",
		&low, &high, &offset);
	    if (tokens == 3) {
		void *sp_low = (void *)low;
		void *sp_high = (void *)high;

		while ((c = fgetc(fp)) != '\n') {
		    continue;
		}

		if (sp_low < sp && sp_high > sp) {
		    /* found stack region */

		    /*
		     * The current version of LinuxThreads fails to create
		     * guard pages, so the stack limit is really the
		     * kernel limit.
		     */
		    self->stackTop = (void *)(high - rlim.rlim_cur);
		    /*
		     * Assuming entries are sorted, check if the
		     * previous map encroaches on the default
		     * range.  This is what would happen if LinuxThreads
		     * successfully created a guard page.
		     */
		    if ((void *)last_high > self->stackTop &&
			last_high < high)
		    {
			self->stackTop = (void *)last_high;
		    }
		    break;
		}
		last_high = high;
	    } else {
		break;
	    }
	}
	fclose(fp);
    }
}

void
CVMthreadAttach(CVMThreadID *self)
{
    pthread_mutexattr_t attr;

    pthread_mutex_init(&self->lock, NULL);
    pthread_mutex_init(&self->wait_mutex, NULL);
    pthread_cond_init(&self->wait_cv, NULL);
    pthread_cond_init(&self->suspend_cv, NULL);

    pthread_mutexattr_init(&attr);
    pthread_mutexattr_setprotocol(&attr, PTHREAD_PRIO_INHERIT);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_NORMAL);

    pthread_mutex_init(&self->wait_until_mtx, &attr);
    pthread_cond_init(&self->wait_until_cv, NULL);

    self->sharedTimerReferenceCount=0;
    self->waitUntilTime = 0L;
    self->waitUntilDataValid = 0;
    self->waitingOn = NULL;
    self->waitingQnext = NULL;
    self->waitingQprev_p = (CVMThreadID**) &(self->waitingQprev_p);
 
    setFPMode();
    POSIXthreadAttach(self);
    LINUXcomputeStackTop(self);
}

void
CVMthreadDetach(CVMThreadID *self)
{
    pthread_mutex_lock(&self->wait_until_mtx);
  
    while (self->sharedTimerReferenceCount >0)
      {
	pthread_mutex_unlock(&self->wait_until_mtx);
	usleep(10000);
	pthread_mutex_lock(&self->wait_until_mtx);
      }
    
    pthread_mutex_unlock(&self->wait_until_mtx);

    POSIXthreadDetach(self);
    //#ifdef LINUX_THREADS_BUG
    pthread_cond_destroy(&self->wait_cv);
    pthread_mutex_destroy(&self->wait_mutex);
    //#endif
    pthread_mutex_destroy(&self->lock);
}

void
CVMthreadSetPriority(CVMThreadID *t, CVMInt32 priority)
{
  // printf ("ISHIN - CVMthreadSetPriority %d\n", priority);

    t->lastPriority = priority;
    POSIXthreadSetPriority(t, priority);
}

CVMBool
CVMthreadStackCheck(CVMThreadID *self, CVMUint32 redZone)
{
    return (char *)self->stackTop + redZone < (char *)&self;
}

void
CVMthreadInterruptWait(CVMThreadID *thread)
{
    thread->interrupted = CVM_TRUE;
    linuxSyncInterruptWait(thread);
}

CVMBool
CVMthreadIsInterrupted(CVMThreadID *thread, CVMBool clearInterrupted)
{
    if (clearInterrupted) {
	CVMBool wasInterrupted;
	assert(thread == CVMthreadSelf());
	wasInterrupted = thread->interrupted;
	thread->interrupted = CVM_FALSE;
	return wasInterrupted;
    } else {
	return thread->interrupted;
    }
}
