/*
 * -----------------------------------------------------------------
 *  HeapMemoryAdapter.c :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


#include "com_timesys_CoreExtension_HeapMemoryAdapter.h"
#include "jvmcei.h"
#include "HeapMemoryAdapter.h"

#include "javavm/include/interpreter.h"  //For CVMExecEnv
#include "javavm/include/indirectmem.h"  //For all the CVMID macros
#include "javavm/include/directmem.h"    //For the CVMD_isgcSafe macro
#include "javavm/include/typeid.h"
#include "javavm/include/jni_impl.h"

#include "javavm/include/mem_registry.h"
#include "javavm/include/utils.h"
#include <stdio.h>

/*
 * ALL MEMORY FILTER FUNCTIONS SHOULD CHECK FOR GCSAFE/UNSAFEness.
 *     This is because the code can be called from a gc-safe code or 
 * gc-unsafe code. Because of this the code behaves differently based on
 * gc-safe/unsafe mode.
 */

/*
 * Returns 1 if the memory should be allocated from the heap for the 
 * CVMExecEnv passed.
 * else returns 0
 */

int CVMJTisHeapMemoryArea(void *p)
{
  
  CVMExecEnv* ee = (CVMExecEnv*) p;
  CVMBool res1, res2, res3;

  if(CVMD_isgcSafe(ee))
    {
      CVMD_gcUnsafeExec(ee, 
      {
	//return true (or 1) if ee->memoryArea is NULL because 
	//if the memoryArea is not set, we need to allocate from heap.
	
	res2 = (ee->memoryArea == NULL) || CVMID_icellIsNull(ee->memoryArea);
	res3 = (jtHeapMemoryArea == NULL) || CVMID_icellIsNull(jtHeapMemoryArea);
	
	if(res2 || res3) // Things haven't been set up yet, use normal heap.
	  {
	    res1 = CVM_TRUE;
	  }
	else
	  CVMID_icellSameObject(ee, CVMmemoryAreaICell(ee), jtHeapMemoryArea, res1);
      });
    }
  else  // I am already gcunsafe
    {
	  res2 = (ee->memoryArea == NULL) || CVMID_icellIsNull(ee->memoryArea);
      res3 = (jtHeapMemoryArea == NULL) || CVMID_icellIsNull(jtHeapMemoryArea);
      
      if(res2 || res3) // Things haven't been set up yet, use normal heap.
	{
	  res1 = CVM_TRUE;
	}
      else
	{
	  res1 = (CVMID_icellDirect(ee, CVMmemoryAreaICell(ee)) == CVMID_icellDirect(ee, jtHeapMemoryArea));
	}
    }
  
  if(res1)
    return 1;
  return 0;
}


int CVMJTinitExecEnv(void* arg)
{
  CVMExecEnv* ee = (CVMExecEnv*) arg;
  if(ee == NULL)
    return -1;
  //  if(CVMD_isgcSafe(ee))
  CVMID_icellAssign(ee, CVMmemoryAreaICell(ee), jtHeapMemoryArea);
  //ee->ptrNewInstance = CVMgcimplAllocObject;
  return 0;
}

int CVMJTdestroyExecEnv(void *arg)
{
  return 1;
}

/*
 * This function is going to be invoked primarily from a gc-unsafe code -
 * CVMgcAllocNewinstance()
 */ 

/*
 * This function is used for both, allocating an instance and 
 * allocating an array.
 */
int CVMJTallocNewInstance(void* argp)
{
  struct arg_t
  {
    CVMExecEnv* ee;
    CVMJavaInt size;
    CVMObject* newInstance;
  };

  struct arg_t* arg = (struct arg_t*) argp;
  CVMExecEnv* ee = arg->ee;
  //JNIEnv *env = CVMexecEnv2JniEnv(ee);
  CVMObjectICell *currentMemoryArea = CVMmemoryAreaICell(ee);
  
//  CVMClassBlock *cb;
  
  if(currentMemoryArea == NULL)
    {
      CVMdebugPrintf(("arg->ee->memoryArea is NULL..Huh!?\n"));
	  arg->newInstance = 0;
      return -1;
    }

  if(CVMD_isgcUnsafe(ee))
  {
//	cb = CVMobjectGetClass(CVMID_icellDirect(ee, currentMemoryArea));
	arg->newInstance = CVMJTregistryInvokeNewInstance(ee, currentMemoryArea, (CVMUint32) arg->size);
  }
  else
  {
	CVMD_gcUnsafeExec(ee, {
//		cb = CVMobjectGetClass(CVMID_icellDirect(ee, currentMemoryArea));
		arg->newInstance = CVMJTregistryInvokeNewInstance(ee, currentMemoryArea, (CVMUint32) arg->size);
	});
  }
  
 return 0;
}

/* This fn is never used. I should get rid of it 
 * - Pratik
 */

/*int CVMJTallocNewArray(void* p)
{
  struct arg_t {
    CVMExecEnv* ee; 
    CVMJavaInt size; 
    CVMArrayOfAnyType* newArray; 
  };

  struct arg_t* args = (struct arg_t*)p;

  if(args->ee->memoryArea == NULL)
    {
      printf("CVMJTallocNewArray(): args->ee->memoryArea is NULL..Huh!?\n");
      args->newArray = 0;
      return -1;
    }

  //Okay... Go ahead..
  {
    JNIEnv* env = CVMexecEnv2JniEnv(args->ee);
    
    // Sanity check. ee->memoryArea should be a MemoryArea object.
    jobject memObj = args->ee->memoryArea;
    jclass clsMemoryArea, clsMyMemoryArea;//, objToAllocate;
    jmethodID midNew;
    printf("CVMJTallocNewArray():Trying to allocate an object in non-heap\n");
    clsMemoryArea = CVMjniFindClass(env, "javax/realtime/MemoryArea");
    clsMyMemoryArea = CVMjniGetObjectClass(env, memObj);
    
    if(CVMjniIsInstanceOf(env, memObj, clsMemoryArea) == JNI_FALSE)
      {
	printf("CVMgcAllocNewInstance - PANIC! memory area is not javax.realtime.MemoryArea!!!\n");
	//exit(2);
	args->newArray = 0;
	return -1;
      }
    
    // Now try to call MemoryArea.newArray()
    midNew = CVMjniGetMethodID(env, clsMyMemoryArea, "newArray0", "(I)Ljava/lang/Object;");

    args->newArray = (CVMArrayOfAnyType*) CVMjniCallObjectMethod(env, memObj, midNew, args->size);
    return 0;
  }
}*/

/* The p->memoryArea parameter passed should be initialised.  */
/* i.e. it should not be NULL. */

int CVMJTgetMemoryArea(void *p)
{
  struct arg_t
  {
    CVMExecEnv *ee;
    CVMObjectICell *memoryArea;
  };
  struct arg_t *arguments = (struct arg_t*)p;
  CVMExecEnv *ee = arguments->ee;

  //  arguments->memoryArea = CVMID_NonNullICellPtrFor(arguments->memoryArea);
  if(CVMD_isgcSafe(ee))
    {
      //gcsafe code
      CVMID_icellAssign(ee, arguments->memoryArea, CVMmemoryAreaICell(ee));
    }
  else
    {
      //gcunsafe code
      CVMID_icellAssignDirect(ee, arguments->memoryArea, CVMmemoryAreaICell(ee));
    }
  
  return 1;
}

int CVMJTsetMemoryArea(void *p)
{
  struct arg_t
  {
    CVMExecEnv *ee;
    CVMObjectICell *memoryArea;
  };
  struct arg_t *arguments = (struct arg_t*)p;
  CVMExecEnv *ee = arguments->ee;

  arguments->memoryArea = CVMID_NonNullICellPtrFor(arguments->memoryArea);
  if(CVMD_isgcSafe(ee))
    {
      //gcsafe code
      CVMID_icellAssign(ee, CVMmemoryAreaICell(ee), arguments->memoryArea);
    }
  else
    {
      //gcunsafe code
      CVMID_icellAssignDirect(ee, CVMmemoryAreaICell(ee), arguments->memoryArea);
    }
  
  return 1;
}

/*
 * Class:     com_timesys_CoreExtension_HeapMemoryAdapter
 * Method:    bootNative
 * Signature: (Ljava/lang/Object;)I
 */
JNIEXPORT jint JNICALL Java_com_timesys_CoreExtension_HeapMemoryAdapter_bootNative
  (JNIEnv *env, jclass thisObj, jobject heap)
{
  //JNIEnv* theEnv = env;
  struct jvmcei* theJVMCei;

//  printf("Trying to initialise Heap Memory Area...\n");
//  CVMJTinitMemoryArea(theEnv);

  
//  printf("HeapMemoryAdapter: getting the CE interface\n");
  theJVMCei = (*env)->GetJVMCeiInterface(env);

  
  /* install the filter */
  CVMdebugPrintf(("HeapMemoryAdapter: installing the filter\n"));
  //  theJVMCei->installFilter(JVMCEF_INIT_HEAP_MEMORY, CVMJTinitMemoryArea);
  //jtHeapMemoryArea = (*env)->NewGlobalRef(env, heap);

// Now create the heap memory area info
  //CVMJTregistryCreateHeapMemoryAreaInfo(CVMjniEnv2ExecEnv(env), jtHeapMemoryArea);

  theJVMCei->installFilter(JVMCEF_IS_HEAP_MEMORY, CVMJTisHeapMemoryArea);
  theJVMCei->installFilter(JVMCEF_ALLOC_NEW_INSTANCE, CVMJTallocNewInstance);
  theJVMCei->installFilter(JVMCEF_INIT_EXECENV, CVMJTinitExecEnv);
  theJVMCei->installFilter(JVMCEF_ALLOC_NEW_ARRAY, CVMJTallocNewInstance);

  theJVMCei->installFilter(JVMCEF_GET_MEMORY_AREA, CVMJTgetMemoryArea);
  theJVMCei->installFilter(JVMCEF_SET_MEMORY_AREA, CVMJTsetMemoryArea);
  theJVMCei->installFilter(JVMCEF_GC_SCAN_ALL_MEMORY_AREAS, CVMJTregistryScanAndWalkAllRoots);
  theJVMCei->installFilter(JVMCEF_GC_HANDLE_REFERENCE, CVMJTregistryHandleReference);
  theJVMCei->installFilter(JVMCEF_DESTROY_EXEC_ENV, CVMJTdestroyExecEnv);
  
  /* set up done. Everything else is taken care of by the filter
   * handlers
   */
#ifdef CVM_DEBUG
  printf("HeapMemoryAdapter - Native Initialized\n");
#endif

  return 0;
}

