/*
 * -----------------------------------------------------------------
 * RTThreadAdapter.c : Filter functions for RealtimeThreads.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


#include "com_timesys_CoreExtension_RTThreadAdaptor.h"
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <sched.h>

#include "jni.h"
#include "jvmcei.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/utils.h"
#include <rk/rk.h>
#include <rk/rk_error.h>
//#include <rk/timespec.h>
#include <semaphore.h>

static JNIEnv *theEnv;
static struct jvmcei * theJVMCei;

/* Semaphore used to avoid race condition. VM crahses if 
 * afterSysThreadCreationFilterHandler gets invoked before 
 * JNI_CreateJavaVM completes.
 * - Pratik, Kunal
 */
sem_t vmWait;

void javax_realtime_realtimethread_sleepUntil(CVMThreadID *self, long long time);

int beforeSysThreadCreationFilterHandler(void *p)
{
 jclass threadClazz;
 jmethodID mID;
 //int rs;
 jobject threadObject;
 JNIEnv *env;
 CVMExecEnv *ee;
 

 struct  arg_t {
   JNIEnv *env; 
   jobject tObject;
   CVMInt16 *ppolicy;
   CVMInt16 *ppriority;
 } *arg;

 arg = (struct arg_t *) p;

 env = arg->env;
 threadObject = arg->tObject;
 ee = CVMjniEnv2ExecEnv(env);

 threadClazz = (*env)->FindClass(env, "javax/realtime/RealtimeThread");

 if (threadClazz == NULL)
   {
     CVMdebugPrintf(("Thread Create Error getting class javax.realtime.RealtimeThread\n"));
     return -1 ;
   }

 if ((*env)->IsInstanceOf(env, threadObject, threadClazz) == JNI_TRUE)
   {
     /* set the priority to the one the current thread has */
     struct sched_param p;

     sched_getparam(getpid(), &p);


     mID = (*env)->GetMethodID(env, threadClazz, "getRTPriority","()I");
     if (mID == NULL)
       {
	 CVMdebugPrintf(("RTThreadAdapter: Could not find MethodID: getRTPriority\n"));
	 return -1;
       }

     (*(arg->ppriority)) = (*env)->CallIntMethod(env, threadObject, mID);
     if ((*(arg->ppriority)) > 0)
       {
	 (*(arg->ppolicy)) = SCHED_FIFO;
/* 	 (*(arg->ppolicy)) = SCHED_FIFO; ISHIN */
       }
     else
       {
	 (*(arg->ppolicy)) = SCHED_OTHER;
       }

#ifdef CVM_DEBUG
     CVMdebugPrintf(("************************************\n"));
     CVMdebugPrintf(("* RTThread prio(%d), policy(%d)*\n",(*(arg->ppolicy)),(*(arg->ppolicy)) ));
     CVMdebugPrintf(("************************************\n"));
#endif
     
/*       (*(arg->ppolicy)) = sched_getscheduler(getpid()); */
/*       (*(arg->ppriority)) = p.sched_priority; */

/*       (*(arg->ppolicy)) = SCHED_FIFO; */
/*       (*(arg->ppriority)) = JVMCEI_SYSTEM_PRIORITY; */
   }
 else
   {
     *(arg->ppolicy) = SCHED_OTHER;
   }
 return JVMCEI_CONTINUE_CURRENT_CALL;
}

int afterSysThreadCreationFilterHandler(void *p)
{
 jclass threadClazz;
 /*Added by Kunal for NoHeapRealtimeThread*/
 jclass noHeapthreadClazz;
 //int rs;
 jobject threadObject;
 JNIEnv *env;
 CVMExecEnv *ee;


 struct  arg_t {JNIEnv *env; jobject tObject; PerThreadMemoryInfo *inheritedStack;} *arg;

 arg = (struct arg_t *) p;

 env = arg->env;
 threadObject = arg->tObject;
 ee = CVMjniEnv2ExecEnv(env);
 
 if (threadObject == NULL)
   {
     CVMdebugPrintf(("UGG the theadobject is NULL\n"));
     return 0;
   }

 /* initialize variable in ExecEnv*/
 ee = CVMjniEnv2ExecEnv(env);

 ee->canChangeScheduler = 0;

 /* uncomment for debug
 printf("Inside rtthreadadapter threadObject(%x)\n",
	(unsigned int) threadObject);
 */

 /* verify if this thread is an RealtimeThread */

 // Confused by the following two lines???
 // Contact Pratik or Kunal. :-)
 sem_wait(&vmWait);
 sem_post(&vmWait);

 // enable constraints checking.
 CVMenableCheckForConstraints(ee);

 threadClazz = (*env)->FindClass(env, "javax/realtime/RealtimeThread");
 noHeapthreadClazz = (*env)->FindClass(env, "javax/realtime/NoHeapRealtimeThread");

 if (threadClazz == NULL)
   {
     CVMdebugPrintf(("Thread Create Error getting class javax.realtime.RealtimeThread\n"));
     return -1 ;
   }

 if ((*env)->IsInstanceOf(env, threadObject, noHeapthreadClazz) == JNI_TRUE)
   CVMsetNoHeapRealtimeThread(ee);

 if ((*env)->IsInstanceOf(env, threadObject, threadClazz) == JNI_TRUE)
   {
     jlong msecs, nsecs;
     jmethodID mID;
     struct timespec period, start;
     struct timespec r_deadline; /*ISHIN*/
     struct sched_param p;


     mID = (*env)->GetMethodID(env, threadClazz, "getPeriodMillis","()J");
     if (mID == NULL)
       {
	 CVMdebugPrintf(("RTThreadAdapter: Could not find MethodID: getPeriodMillis\n"));
	 return-1;
       }
     msecs = (*env)->CallLongMethod(env, threadObject, mID);
			 
     mID = (*env)->GetMethodID(env, threadClazz, "getPeriodNanos","()J");
     if (mID == NULL)
       {
	 CVMdebugPrintf(("RTThreadAdapter: Could not find MethodID: getPeriodNanos\n"));
	 return-1;
       }
     nsecs = (*env)->CallLongMethod(env, threadObject, mID);

     period.tv_sec = msecs / 1000;
     period.tv_nsec = (msecs % 1000) * 1000 * 1000 + nsecs;


     mID = (*env)->GetMethodID(env, threadClazz, "getStartMillis","()J");
     if (mID == NULL)
       {
	 CVMdebugPrintf(("RTThreadAdapter: Could not find MethodID: getStartMillis\n"));
	 return-1;
       }
     msecs = (*env)->CallLongMethod(env, threadObject, mID);
			 
     mID = (*env)->GetMethodID(env, threadClazz, "getStartNanos","()I");
     if (mID == NULL)
       {
	 CVMdebugPrintf(("RTThreadAdapter: Could not find MethodID: getStartNanos\n"));
	 return-1;
       }
     nsecs = (*env)->CallIntMethod(env, threadObject, mID);



     /*CVMdebugPrintf(("start millis(%lld) nanos(%lld)\n",msecs, nsecs);*/

     {
       unsigned long long Lstart, Lcurrent;
       unsigned long long diff;
       struct timeval t;
       start.tv_sec = msecs / 1000;
       start.tv_nsec = (msecs % 1000) * 1000 * 1000  + nsecs;
       
       gettimeofday(&t,NULL);
       
       Lstart = (long long) start.tv_sec * (long long) 1000000;
       Lstart += (long long) start.tv_nsec / (long long) 1000;

       Lcurrent = (long long) t.tv_sec * (long long) 1000000;
       Lcurrent += (long long) t.tv_usec;

       
       if  (Lstart > Lcurrent) 
	 {
	   diff =  Lstart - Lcurrent ;
	   
	   /*CVMdebugPrintf(("rtthreadadpater waiting for starttime %Lu\n",Lstart);*/
	   javax_realtime_realtimethread_sleepUntil(&ee->threadInfo, Lstart);	   
	   //usleep(diff);
	   ee->startNextPeriodMicros = Lstart;
           ee->startNextPeriodMicros +=  ((long long )period.tv_sec * (long long)1000000) ;
	   ee->startNextPeriodMicros += ((long long)period.tv_nsec / (long long )1000);
	   /*CVMdebugPrintf(("rtthreadadapter after waiting for starttime nextperiodmicros(%Lu)))\n",
		  ee->startNextPeriodMicros);*/
	 }
       else
	 {
	   ee->startNextPeriodMicros = Lcurrent ;
           ee->startNextPeriodMicros +=  ((long long )period.tv_sec * (long long)1000000) ;
	   ee->startNextPeriodMicros += ((long long)period.tv_nsec / (long long )1000);	

	   /* try */
	   // Lstart = Lcurrent;
	 }

       /* set the startOfPeriodMicros*/
       {
	 jfieldID fID = (*env)->GetFieldID(env, threadClazz, "startOfPeriodMicros","J");
#ifdef CVM_DEBUG
	 printf("RTThreadAdapter.c: setting startOfPeriodMicros(%Lu)\n",Lstart);
#endif
	 if (fID == NULL)
	   {
	     CVMdebugPrintf(("RTThreadAdapter: could not get the FID for startOfPeriodMicros\n"));
	   }
	 else
	   {
	     (*env)->SetLongField(env, threadObject, fID, (jlong)Lstart);
	   }
       }
     }	







     /*********** ISHIN **********/
     mID = (*env)->GetMethodID(env, threadClazz, "getDeadlineMillis","()J");
     if (mID == NULL)
       {
	 printf(("RTThreadAdapter: Could not find MethodID: getDeadlineMillis\n"));
	 return-1;
       }
     msecs = (*env)->CallLongMethod(env, threadObject, mID);
			 
     mID = (*env)->GetMethodID(env, threadClazz, "getDeadlineNanos","()J");
     if (mID == NULL)
       {
	 printf(("RTThreadAdapter: Could not find MethodID: getPeriodNanos\n"));
	 return-1;
       }
     nsecs = (*env)->CallLongMethod(env, threadObject, mID);

     r_deadline.tv_sec = msecs / 1000;
     r_deadline.tv_nsec = (msecs % 1000) * 1000 * 1000 + nsecs;
     ee->deadline = 0;
     ee->deadline += ((long long)r_deadline.tv_sec * (long long)1000000) ;
     ee->deadline += ((long long)r_deadline.tv_nsec / (long long )1000);
     /*********** ISHIN **********/

     
     /* we are setting the priority at creation time now */

     mID = (*env)->GetMethodID(env, threadClazz, "getRTPriority","()I");
     if (mID == NULL)
       {
	 CVMdebugPrintf(("RTThreadAdapter: Could not find MethodID: getRTPriority\n"));
	 return -1;
       }

     p.sched_priority = (*env)->CallIntMethod(env, threadObject, mID);


     /* Assign the real-time priority */ 
     if (p.sched_priority > 0)
	 {
	   CVMdebugPrintf(("assigning fixed priority %d\n",p.sched_priority));
	   if (sched_setscheduler(getpid(),SCHED_FIFO, &p) < 0)
	     CVMdebugPrintf(("!!!! ERROR the fixed priority could not be assigned to %d\n", getpid()));
	 }
     else
       {
	 /* even if we don't have an rt priority we need to return the thread to normal
	  * priority given that we raise the priority to avoid interruptions in the
	  * startup
	  */

	 p.sched_priority = 0;
	 if (sched_setscheduler(getpid(),SCHED_OTHER, &p) < 0)
	   CVMdebugPrintf(("!!!! ERROR the fixed priority could not be assigned to %d\n", getpid()));	 
       }
     
    

     /* start the deadline Handler timer if any */

#ifdef CVM_DEBUG
     CVMdebugPrintf(("RTThreadAdapter.c: before startDeadlineTimer\n"));
#endif

     mID = (*env)->GetMethodID(env, threadClazz, "startDeadlineTimer","()V");
     if (mID == NULL)
       {
	 CVMdebugPrintf(("RTThreadAdapter: Could not find MethodID: startDeadlineTimer\n"));
     	 return-1;
       }
     (*env)->CallVoidMethod(env, threadObject, mID);

     if ((*env)->ExceptionOccurred(env))
       {
	 (*env)->ExceptionDescribe(env);
	 (*env)->ExceptionClear(env);
       }

#ifdef CVM_DEBUG
     CVMdebugPrintf(("RTThreadAdapter.c: after startDeadlineTimer\n"));
#endif

     ee->perThreadStack = arg->inheritedStack;
     if(ee->perThreadStack == NULL)
     {
        MemoryAreaInfo *root;
        root = CVMJTregistryGetMemoryAreaInfoForObject(ee, threadObject);
        CVMassert(root != NULL && 
                    (root->memType == MEMTYPE_IMMORTAL || root->memType == MEMTYPE_HEAP));
        
        ee->perThreadStack = CVMJTregistryCreatePTMI(ee);
        ee->perThreadStack->info = root;
     }
    
     mID = (*env)->GetMethodID(env, threadClazz, "enterMemoryArea","()V");
	 if(mID == NULL)
	 {
	   CVMdebugPrintf(("RTThreadAdapter: Could not find MethodID: enterMemoryArea\n"));
	   return -1;
	 }
	 (*env)->CallVoidMethod(env, threadObject, mID);
     if((*env)->ExceptionCheck(env))
     {
         //An exception occured. The thread cannot be allowed to start.
         CVMconsolePrintf("Unable to start thread. Printing exception\n");
         (*env)->ExceptionDescribe(env);
         (*env)->ExceptionClear(env);
         return -1;
     }
   }
 
 
 return JVMCEI_CONTINUE_CURRENT_CALL;
}

int beforeSysThreadDestroyFilterHandler(void *p)
{
//#if 0 
 //jfieldID fID;
 jclass threadClazz;
 //int rs;
 jobject threadObject;
 JNIEnv *env;

 struct  arg_t {JNIEnv *env; jobject tObject;} *arg;

 arg = (struct arg_t *) p;

 env = arg->env;
 threadObject = arg->tObject;

 if (threadObject == NULL)
   {
     CVMdebugPrintf(("UGG the theadobject is NULL\n"));
     return 0;
   }

 /* uncomment for debug
 CVMdebugPrintf(("Inside rtthreadadapter threadObject(%x)))\n", 
	(unsigned int) threadObject);
 */

 /* verify if this thread is an RealtimeThread */
	 
 threadClazz = (*env)->FindClass(env, "javax/realtime/RealtimeThread");
	 
 if (threadClazz == NULL)
   {
     CVMdebugPrintf(("Thread Create Error getting class javax.realtime.RealtimeThread\n"));
     return -1 ;
   }
	 
 if ((*env)->IsInstanceOf(env, threadObject, threadClazz) == JNI_TRUE)
   {    
     jmethodID mID;

     mID = (*env)->GetMethodID(env, threadClazz, "stopDeadlineTimer","()V");
	 if(mID == NULL)
	 {
	   CVMdebugPrintf(("RTThreadAdapter: Could not find MethodID: stopDeadlineTimer\n"));
	   return -1;
	 }
	 (*env)->CallVoidMethod(env, threadObject, mID);


     mID = (*env)->GetMethodID(env, threadClazz, "removeFromFeasibility","()Z");
	 if(mID == NULL)
	 {
	   CVMdebugPrintf(("RTThreadAdapter: Could not find MethodID: removeFromFeasibility\n"));
	   return -1;
	 }
	 (*env)->CallBooleanMethod(env, threadObject, mID);

     mID = (*env)->GetMethodID(env, threadClazz, "exitMemoryArea","()V");
	 if(mID == NULL)
	 {
	   CVMdebugPrintf(("RTThreadAdapter: Could not find MethodID: exitMemoryArea\n"));
	   return -1;
	 }
	 (*env)->CallVoidMethod(env, threadObject, mID);

   }	 
//#endif
 return JVMCEI_CONTINUE_CURRENT_CALL; 
}

void javax_realtime_realtimethread_initSleepQ();

JNIEXPORT jint JNICALL Java_com_timesys_CoreExtension_RTThreadAdaptor_bootNative
  (JNIEnv * env, jclass clazz)
{
  theEnv = env;
#ifdef CVM_DEBUG
  CVMdebugPrintf(("RTThreadAdaptor: getting the CE interface\n"));
#endif
  theJVMCei = (*env)->GetJVMCeiInterface();

  /* install the filter */
#ifdef CVM_DEBUG
  CVMdebugPrintf(("RTThreadAdaptor: installing the filter\n"));
#endif
  theJVMCei->installFilter(JVMCEF_BEFORE_CREATE_SYS_THREAD, beforeSysThreadCreationFilterHandler);
  theJVMCei->installFilter(JVMCEF_AFTER_CREATE_SYS_THREAD, afterSysThreadCreationFilterHandler);
  theJVMCei->installFilter(JVMCEF_BEFORE_DESTROY_SYS_THREAD, beforeSysThreadDestroyFilterHandler);
  
  /* set up done. Everything else is taken care of by the filter
   * handlers
   */

  /* initialize the thread sleep queue */
  javax_realtime_realtimethread_initSleepQ();
#ifdef CVM_DEBUG
  CVMdebugPrintf(("RTThreadAdaptor - Native Initialized\n"));
#endif
  return 0;
}

