/*
 * -----------------------------------------------------------------
 * MemoryArea.c : The native implementation of MemoryAreas.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


#include "javax_realtime_MemoryArea.h"

#include "javavm/include/interpreter.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/directmem.h"
#include "javavm/include/directmem.h"
#include "javavm/include/globalroots.h"
#include "javavm/include/common_exceptions.h"
#include "javavm/include/utils.h"
#include "javavm/export/jvmcei.h"
#include "javavm/include/mem_registry.h"

/*
 * Class:     javax_realtime_MemoryArea
 * Method:    getCurrentMemoryArea
 * Signature: ()Ljavax/realtime/MemoryArea;
 */
JNIEXPORT jobject JNICALL Java_javax_realtime_MemoryArea_getCurrentMemoryArea
  (JNIEnv *env, jobject thisObj)
{
  // Cei declarations
  extern int JVMCeiInitialized;
  extern intFuncPtr JVMCeiFilters[];

  if(JVMCeiInitialized && JVMCeiFilters[JVMCEF_GET_MEMORY_AREA] != NULL)
    {
      CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
      
      struct arg_t
      {
	CVMExecEnv *ee;
	CVMObjectICell *memoryArea;
      };
      struct arg_t arguments;
      arguments.ee = ee;
      arguments.memoryArea = CVMjniCreateLocalRef(ee);
      if(JVMCeiFilters[JVMCEF_GET_MEMORY_AREA](&arguments))
	return arguments.memoryArea;
    }
  
  //return a null object
  
  return NULL;
}

/*
 * Class:     javax_realtime_MemoryArea
 * Method:    setCurrentMemoryArea
 * Signature: (Ljavax/realtime/MemoryArea;)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_MemoryArea_setCurrentMemoryArea
  (JNIEnv *env, jobject thisObj, jobject memory)
{
  extern int JVMCeiInitialized;
  extern intFuncPtr JVMCeiFilters[];

  if(JVMCeiInitialized && JVMCeiFilters[JVMCEF_SET_MEMORY_AREA] != NULL)
    {
      CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
      
      struct arg_t
      {
	CVMExecEnv *ee;
	CVMObjectICell *memoryArea;
      };
      struct arg_t arguments;
      
      arguments.ee = ee;
      arguments.memoryArea = CVMjniNewLocalRef(env, memory);
      JVMCeiFilters[JVMCEF_SET_MEMORY_AREA](&arguments);
	
    }
  return;
}


/*
 * Class:     javax_realtime_MemoryArea
 * Method:    getMemoryAreaForObject 
 * Signature: (Ljava/lang/Object;)Ljavax/realtime/MemoryArea;
 */
JNIEXPORT jobject JNICALL Java_javax_realtime_MemoryArea_getMemoryAreaForObject
  (JNIEnv *env, jobject thisObj, jobject obj)
{
  return (*env)->NewLocalRef(env, CVMJTregistryGetMemoryAreaForObject(CVMjniEnv2ExecEnv(env) ,obj));
}


/*
 * Class:     javax_realtime_MemoryArea
 * Method:    getMemoryConsumed0
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_javax_realtime_MemoryArea_getMemoryConsumed0
  (JNIEnv *env, jobject thisObj)
{
  return CVMJTregistryGetMemoryAllocated(CVMjniEnv2ExecEnv(env), thisObj);
}
  
/*
 * Class:     javax_realtime_MemoryArea
 * Method:    getMemoryRemaining0
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_javax_realtime_MemoryArea_getMemoryRemaining0
  (JNIEnv *env, jobject thisObj)
{
  return CVMJTregistryGetMemoryRemaining(CVMjniEnv2ExecEnv(env), thisObj);
}


/*
 * Class:     javax_realtime_MemoryArea
 * Method:    freeMemory
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_javax_realtime_MemoryArea_freeMemory
  (JNIEnv *env, jobject thisObj)
{
  CVMJTregistryUnregisterMemoryArea(CVMjniEnv2ExecEnv(env), thisObj);
  return;
}


/*
 * Class:     javax_realtime_MemoryArea
 * Method:    registerMemoryArea0
 * Signature: (JJI)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_MemoryArea_registerMemoryArea0
  (JNIEnv *env, jobject thisObj, jlong base, jlong size, jint memType)
{
  if(!CVMJTregistryRegisterMemoryArea(CVMjniEnv2ExecEnv(env), thisObj, base, size, memType))
  {
	printf("PANIC: Unable to register memory area in MemoryArea.registerMemoryArea");
  }
  return;
}

/*
 * Class:     javax_realtime_MemoryArea
 * Method:    runAllFinalizers
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_javax_realtime_MemoryArea_runAllFinalizers
  (JNIEnv *env, jobject thisObj)
{
	CVMJTregistryInvokeAllFinalizers(CVMjniEnv2ExecEnv(env), thisObj);
}

/*
 * Class:     javax_realtime_MemoryArea
 * Method:    newInstance0
 * Signature: (Ljava/lang/Class;)Ljava/lang/Object;
 */
JNIEXPORT jobject JNICALL Java_javax_realtime_MemoryArea_newInstance0
  (JNIEnv *env, jobject thisObj, jclass type)
{
    jobject  objAllocated = NULL;
    // CVMClassBlock *cbType;
    // jobject oldMemoryArea;
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    jmethodID constructorID;
    
    // Get the constructor's method id.
    constructorID = (*env)->GetMethodID(env, type, "<init>", "()V");
    if(constructorID == NULL)
    {
        jclass clazz2;
        // A NoSuchmethodError was already thrown. Clear that exception before 
        // throwing InstantiationException.
        if((*env)->ExceptionCheck(env))
            (*env)->ExceptionClear(env);
        
        //Throw InstatiationException
        clazz2 = (*env)->FindClass(env, "java/lang/InstantiationException");
        (*env)->ThrowNew(env, clazz2, "Could not find a no-argument constructor");
	    return NULL;
    }
  
    
    // Set current memory area to self.
    //oldMemoryArea = Java_javax_realtime_MemoryArea_getCurrentMemoryArea(env, thisObj);
    //Java_javax_realtime_MemoryArea_setCurrentMemoryArea(env, thisObj, thisObj);
    
    // Get CVMClassBlock* for 'type'
    // ee = CVMjniEnv2ExecEnv(env);
    // cbType = CVMgcSafeClassRef2ClassBlock(ee, type);
    SLIDE_UP_AND_EXECUTE(ee, thisObj, {
        // Allocate the object thru JNI.
        objAllocated = (*env)->AllocObject(env, type);
        
        if(objAllocated != NULL)
        {
            if((*env)->ExceptionCheck(env))
            {
                CVMdebugPrintf(("AN EXCEPTION WAS THROWN\n"));
                CVMdebugPrintf(("exception is %I\n", CVMlocalExceptionICell(ee)));
            }    
            
            // Now invoke the constructor
            (*env)->CallVoidMethod(env, objAllocated, constructorID);
        
            // Set the old memory area back.
            //Java_javax_realtime_MemoryArea_setCurrentMemoryArea(env, thisObj, oldMemoryArea);
        }
    });

    if((*env)->ExceptionCheck(env))
    {
        CVMdebugPrintf(("AN EXCEPTION WAS THROWN\n"));
        CVMdebugPrintf(("exception is %I\n", CVMlocalExceptionICell(ee)));
    }
    return objAllocated;
   
}

/* FIXME: Need to change this implementation to use SLIDE_UP_AND_EXECUTE 
 * - Pratik.
 */
/*
 * Class:     javax_realtime_MemoryArea
 * Method:    newInstance1
 * Signature: (Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;
 */
JNIEXPORT jobject JNICALL Java_javax_realtime_MemoryArea_newInstance1
  (JNIEnv *env, jobject thisObj, jobject constructor, jobject argarray)
{
    jobject oldMemoryArea, objAllocated;
    jmethodID newInstanceID;
    jclass constrClazz;
    //check if constructor is null
    if(constructor == NULL)
    {
        printf("Constructor passed was NULL!!!!\n");
        return NULL;
    }
    constrClazz = (*env)->GetObjectClass(env, constructor);
    
    newInstanceID = (*env)->GetMethodID(env, constrClazz, "newInstance", "([Ljava/lang/Object;)Ljava/lang/Object;");

    if(newInstanceID == NULL)
    {
        printf("Could not get method ID for newInstance()!!!\n"); 
        //maybe I should throw an exception over here.
        return NULL;
    }
    
    oldMemoryArea = Java_javax_realtime_MemoryArea_getCurrentMemoryArea(env, thisObj);
    Java_javax_realtime_MemoryArea_setCurrentMemoryArea(env, thisObj, thisObj);

    //Now invoke newInstance()
    objAllocated = (*env)->CallObjectMethod(env, constructor, newInstanceID, argarray);

    Java_javax_realtime_MemoryArea_setCurrentMemoryArea(env, thisObj, oldMemoryArea);

    if((*env)->ExceptionCheck(env))
    {
        //an exception occurred that was allocated in thisObj. 
        //Reallocate the exceptin if necessary in the mempry area 
        //active when the newInstance() was invoked.
        jobject exMemoryArea;
        jclass exClazz;
        // Get teh exception.
        jthrowable exception = (*env)->ExceptionOccurred(env);
        exClazz = (*env)->GetObjectClass(env,exception);

        //clear it for now.
        (*env)->ExceptionClear(env);
        exMemoryArea = CVMJTregistryGetMemoryAreaForObject(CVMjniEnv2ExecEnv(env), exception);
        //se if we really need to rethrow a new exception
        if((*env)->IsInstanceOf(env, exMemoryArea, (*env)->FindClass(env, "javax/realtime/ScopedMemory")))
        {
            // reallocate.
            // First get the message of the exception
            const char *cmesg;
            jboolean isCopy;
            jstring message = (*env)->CallObjectMethod(env, exception, 
                                        (*env)->GetMethodID(env, exClazz, 
                                                            "getMessage", "()Ljava/lang/Object;"));
            
            cmesg = (*env)->GetStringUTFChars(env, message, &isCopy);
            (*env)->ThrowNew(env, exClazz, cmesg); 
            if(isCopy == JNI_TRUE)
            {
                (*env)->ReleaseStringUTFChars(env, message, cmesg);
            }
            
        }
        else
        {
            (*env)->Throw(env, exception);
        }
    }
    return objAllocated; 
}


/*
 * Class:     javax_realtime_MemoryArea
 * Method:    newArray0
 * Signature: (Ljava/lang/Class;I)Ljava/lang/Object;
 */
JNIEXPORT jobject JNICALL Java_javax_realtime_MemoryArea_newArray0
  (JNIEnv *env, jobject thisObj, jclass type, jint num)
{
//    jobject oldMemoryArea;
    jarray arrayAllocated = NULL;
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    
    // Set current memory area to self.
    //oldMemoryArea = Java_javax_realtime_MemoryArea_getCurrentMemoryArea(env, thisObj);
    //Java_javax_realtime_MemoryArea_setCurrentMemoryArea(env, thisObj, thisObj);
    
    SLIDE_UP_AND_EXECUTE(ee, thisObj, {
        // Allocate the array thru JNI.
        arrayAllocated = (*env)->NewObjectArray(env, num, type, NULL);
        
        // Set the old memory area back.
        //Java_javax_realtime_MemoryArea_setCurrentMemoryArea(env, thisObj, oldMemoryArea);
    });
    
    if((*env)->ExceptionCheck(env))
    {
        CVMdebugPrintf(("AN EXCEPTION WAS THROWN\n"));
        CVMdebugPrintf(("exception is %I\n", CVMlocalExceptionICell(ee)));
    }
    // Don't need to check if arrayAllocated is NULL because there isn't 
    // anything I can do anyway. An exception would have been thrown and
    // it would be checked on return to Java code.
    return arrayAllocated;

}

/*
 * Class:     javax_realtime_MemoryArea
 * Method:    executeInArea0()
 * Signature: (Ljava/lang/Object;)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_MemoryArea_executeInArea0
  (JNIEnv *env, jobject thisObj, jobject logic)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    jclass logicClazz;
    jmethodID runMethodID = NULL;
    
    logicClazz = (*env)->GetObjectClass(env, logic);
    // Get the method id.
    runMethodID = (*env)->GetMethodID(env, logicClazz, "run", "()V");
    if(runMethodID == NULL)
    {
        // NoSuchMethodException is already thrown
	    return ;
    }
  
//    printf("MemoryArea.c: About to slide up to Memory Area\n");    
    SLIDE_UP_AND_EXECUTE(ee, thisObj, {
            // Now invoke the method
            (*env)->CallVoidMethod(env, logic, runMethodID);
    });

    if((*env)->ExceptionCheck(env))
    {
        CVMdebugPrintf(("AN EXCEPTION WAS THROWN\n"));
        CVMdebugPrintf(("exception is %I\n", CVMlocalExceptionICell(ee)));
    }
    return;
 
}



// *********
// HeapMemory native functions

/*
 * Class:     javax_realtime_HeapMemory
 * Method:    initHeap
 * Signature: ()V
 */
JNIEXPORT jobject JNICALL Java_javax_realtime_HeapMemory_initHeap
  (JNIEnv *env, jclass clazz)
{
    extern jobject jtHeapMemoryArea, jtImmortalMemory;
    CVMExecEnv *ee;
    CVMClassBlock *cbHeap;
    jmethodID constructorID;
    int heapObjSize;
    CVMObject *heap;
    
    ee = CVMjniEnv2ExecEnv(env);
    cbHeap = CVMgcSafeClassRef2ClassBlock(ee, clazz);
    heapObjSize = CVMcbInstanceSize(cbHeap);
    
    jtHeapMemoryArea = CVMID_getGlobalRoot(ee);
    
    // Black magic!!! I have to manually create HeapMemory object....
    CVMD_gcUnsafeExec(ee, {
            heap = CVMJTregistryInvokeNewInstance(ee, jtImmortalMemory, heapObjSize);
            
            memset((void*)heap, 0, heapObjSize);
            heap->hdr.clas = cbHeap;
            CVMobjMonitorSet(heap, 0, CVM_LOCKSTATE_UNLOCKED);

            CVMassert(CVMobjectVariousWord(heap) ==
                        CVM_OBJECT_DEFAULT_VARIOUS_WORD);
 
            CVMID_icellSetDirect(ee, jtHeapMemoryArea, heap);
        });

    // Call constructor.
    constructorID = (*env)->GetMethodID(env, clazz, "<init>", "()V");
    if(constructorID == NULL)
    {
      //Throw InternalError and halt JVM
      jclass clazz2 = (*env)->FindClass(env, "java/lang/InternalError");
      (*env)->ThrowNew(env, clazz2, "Could not find constructor of javax.realtime.HeapMemory!!!");
	  return NULL;
    }
  
    (*env)->CallVoidMethod(env, jtHeapMemoryArea, constructorID);
    
    // Now create the heap memory area info
    CVMJTregistryCreateHeapMemoryAreaInfo(ee, jtHeapMemoryArea);
    return jtHeapMemoryArea;
}
