/*
 * -----------------------------------------------------------------
 * RawMemoryAccess.c :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


#include "javax_realtime_RawMemoryAccess.h"
#include "javax_realtime_RawMemoryFloatAccess.h"

//#include "javax_realtime_DefaultPhysicalMemoryFactory.h"

#include <malloc.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

// This is the file descriptor for /dev/mem
//
// This is a global variable. I might need to put a 
// lock around it. - Pratik.
static int memFileFD = -1;

/* Helper functin for throwing an exception 
 */
static void throwException(JNIEnv *env, const char *className, const char *msg)
{
  jclass cls;

  cls = (*env)->FindClass(env, className);
  (*env)->ThrowNew(env, cls, msg);
  (*env)->DeleteLocalRef(env, cls);
  return;
}

/*
 * Internal function that gets called by RawMemoryAccess.map0() and 
 * DefaultPhysicalMemoryFactory.map0(). This ensures that both the 
 * classes use the same file descriptor for /dev/mem. Plus it saves 
 * duplicating the code in both the native functions.
 * -Pratik
 */
static void * internalMMap(JNIEnv *env, jobject thisObj, jlong base, jlong size)
{
  //jclass cls;
//  jfieldID fid;
//  jint fd;
  void *ptr;
  jlong result;

  //  char msg[70];
  //char *buf = (char*) malloc(sizeof(char) * size);
  int pageSize;

  // base should be a multiple of page size.
  // If not throw exception.

  pageSize = sysconf(_SC_PAGESIZE);
  
  
  /*printf("RawMemoryAccess::map0() - page size is %d\n", pageSize);*/
  
  if(base%pageSize != 0)
    {
      // ERROR! Throw exception
      throwException(env, "javax/realtime/SizeOutOfBoundsException", "Base is not a multiple of pagesze");
      return (void *) -1;
    }
  
  // Now see if the file is open
  //cls = (*env)->GetObjectClass(env, thisObj);
  //fid = (*env)->GetFieldID(env, cls, "fd", "I");
  //fd = (*env)->GetIntField(env, thisObj, fid);
  
  if(memFileFD < 0)
    {
      // File has never been opened.
      //
	memFileFD = open("/dev/mem", O_RDWR);
	if(memFileFD == -1)
	  {
#ifdef CVM_DEBUG
	    printf("RawMemoryAccess::map0() - Error while opening /dev/mem.\nAre you superuser?\n");
#endif
	    throwException(env, "java/lang/SecurityException", "Cannot access physical memory. Are you superuser?");
	    return (void *) -1;
	  }
	
	//(*env)->SetIntField(env, thisObj, fid, fd);
	
    }
  
  ptr = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, memFileFD, base);
  if(ptr == MAP_FAILED)
  {
      //check error and throw corresponding exception
      switch(errno)
      {
        case EINVAL:
        {
          throwException(env, "javax/realtime/OffsetOutOfBoundsException", "System did not like the offset or base address passed");
          return (void*) -1;
        }
        default:
        {
          // FIXME: Maybe I should throw another exception here?
          throwException(env, "javax/realtime/UnsupportedPhysicalMemoryException", "An error occured which I didn't understand.");
          return (void*)-1;
        }
      }
  }

  // This is a workaround to get rid of the warning that the compiler 
  // gives me because of incompatible data type sizes.
  // - Pratik.
  memset(&result,0,sizeof(jlong));
  memcpy(&result + ((sizeof(jlong) -4)), &ptr,4);
  //return result;
  // By passing the above memcpy/memset... returning a ptr instead
  return ptr;
}


/* Class:     javax_realtime_RawMemoryAccess
 * Method:    map0
 * Signature: (JJ)V
 */
JNIEXPORT jlong JNICALL Java_javax_realtime_RawMemoryAccess_map0
(JNIEnv *env, jobject thisObj, jlong base, jlong size)
{
  return (jlong) ((int)internalMMap(env, thisObj, base, size));
}


/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    unmap0
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_unmap0
  (JNIEnv *env, jobject thisObj)
{
  jclass cls;
  jfieldID /*fdId,*/ vaddrId, sizeId;
  jint /*fd,*/ retVal;
  jlong virtAddress, size;
  //void *convAddress=NULL;

  cls = (*env)->GetObjectClass(env, thisObj);
  //fdId = (*env)->GetFieldID(env, cls, "fd", "I");
  //fd = (*env)->GetIntField(env, thisObj, fdId);

  vaddrId = (*env)->GetFieldID(env, cls, "virtAddress", "J");
  virtAddress = (*env)->GetLongField(env, thisObj, vaddrId);

  sizeId = (*env)->GetFieldID(env, cls, "size", "J");
  size = (*env)->GetLongField(env, thisObj, sizeId);

  //memcpy to get rid of the typecast warnings

  //memcpy(&convAddress, &virtAddress +(sizeof(jlong) - sizeof(void*)), sizeof(int)); 
  //unmap virtual memory
  retVal = munmap((void *)((int)virtAddress), (size_t) size);
  if(retVal == -1)
    {
      perror("RawMemoryAccess::unmap0() - munmap failed");
      return;
    }

  /*retVal = close(fd);
  if(retVal == -1)
  {
    perror("RawMemoryAccess::unmap0() - close failed");
  }*/

  //Set the fd and virtAddress to -1
  //(*env)->SetIntField(env, thisObj, fdId, -1);
  (*env)->SetLongField(env, thisObj, vaddrId, -1);
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    getByte0
 * Signature: (J)B
 */
JNIEXPORT jbyte JNICALL Java_javax_realtime_RawMemoryAccess_getByte0
(JNIEnv *env, jobject thisObj, jlong virtAddress)
{
  jbyte *ptr;
  // base is long (8 bytes), and pointer (void *) is 4 bytes. Problems???
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));
  ptr = (jbyte *) ((int)virtAddress);//convAddress;
  return *ptr;
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    setByte0
 * Signature: (JB)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_setByte0
(JNIEnv *env, jobject thisObj, jlong virtAddress, jbyte value)
{
  jbyte *ptr;
  // base is long (8 bytes), and pointer (void *) is 4 bytes. Problems???
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));

  ptr = (jbyte *) ((int)virtAddress);// convAddress;
  *ptr = value;
  return;
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    getInt0
 * Signature: (J)I
 */
JNIEXPORT jint JNICALL Java_javax_realtime_RawMemoryAccess_getInt0
  (JNIEnv *env, jobject thisObj, jlong virtAddress)
{
  jint *ptr;
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));

  ptr = (jint *) ((jint) virtAddress); //convAddress;
  return *ptr;

}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    setInt0
 * Signature: (JI)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_setInt0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jint value)
{
  jint *ptr;
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));
  
  ptr = (jint *) ((int)virtAddress); //convAddress;
  *ptr = value;
  return;
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    getLong0
 * Signature: (J)J
 */
JNIEXPORT jlong JNICALL Java_javax_realtime_RawMemoryAccess_getLong0
  (JNIEnv *env, jobject thisObj, jlong virtAddress)
{
  jlong *ptr;
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));
  ptr = (jlong *) ((int)virtAddress); //convAddress;
  return *ptr;
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    setLong0
 * Signature: (JJ)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_setLong0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jlong value)
{
  jlong *ptr;
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));
  ptr = (jlong *) ((int)virtAddress); //convAddress;
  *ptr = value;
  return;
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    getShort0
 * Signature: (J)S
 */
JNIEXPORT jshort JNICALL Java_javax_realtime_RawMemoryAccess_getShort0
  (JNIEnv *env, jobject thisObj, jlong virtAddress)
{
  jshort *ptr;
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));
  ptr = (jshort *) ((int)virtAddress); //convAddress;
  return *ptr;
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    setShort0
 * Signature: (JS)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_setShort0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jshort value)
{
  jshort *ptr;
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));
  ptr = (jshort *) ((int)virtAddress); //convAddress;
  *ptr = value;
  return;
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    getBytes0
 * Signature: (J[BII)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_getBytes0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jbyteArray bytes, jint low, jint number)
{
  jboolean isCopy;
  jbyte* byteArrayPtr;

  int counter;

  //Get the array as a pointer
  byteArrayPtr = (*env)->GetByteArrayElements(env, bytes, &isCopy);

  for(counter=0; counter<number; counter++)
    {
      byteArrayPtr[low+counter] = Java_javax_realtime_RawMemoryAccess_getByte0(env, 
									       thisObj, 
									       virtAddress+(counter*sizeof(jbyte)));
    }
  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseByteArrayElements(env, bytes, byteArrayPtr, 0);
    }
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    setBytes0
 * Signature: (J[BII)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_setBytes0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jbyteArray bytes, jint low, jint number)
{
  jboolean isCopy;
  jbyte* byteArrayPtr;
  
  int counter;

  //Get the array as a pointer
  byteArrayPtr = (*env)->GetByteArrayElements(env, bytes, &isCopy);

  for(counter=0; counter<number; counter++)
    {
      Java_javax_realtime_RawMemoryAccess_setByte0(env, 
						   thisObj, 
						   virtAddress+(counter*sizeof(jbyte)),
						   byteArrayPtr[low+counter]);
    }

  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseByteArrayElements(env, bytes, byteArrayPtr, JNI_ABORT); //JNI_ABORT because the array passed should never be modified
    }

  
}


/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    getInts0
 * Signature: (J[III)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_getInts0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jintArray ints, jint low, jint number)
{
  jboolean isCopy;
  jint* intArrayPtr;

  int counter;

  //Get the array as a pointer
  intArrayPtr = (*env)->GetIntArrayElements(env, ints, &isCopy);

  for(counter=0; counter<number; counter++)
    {
      intArrayPtr[low+counter] = Java_javax_realtime_RawMemoryAccess_getInt0(env, 
									     thisObj, 
									     virtAddress+(counter*sizeof(jint)));
    }
  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseIntArrayElements(env, ints, intArrayPtr, 0);
    }
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    setInts0
 * Signature: (J[III)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_setInts0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jintArray ints, jint low, jint number)
{
  jboolean isCopy;
  jint* intArrayPtr;
  
  int counter;
  
  //Get the array as a pointer
  intArrayPtr = (*env)->GetIntArrayElements(env, ints, &isCopy);
  
  for(counter=0; counter<number; counter++)
    {
      Java_javax_realtime_RawMemoryAccess_setInt0(env, 
						  thisObj, 
						  virtAddress+(counter*sizeof(jint)),
						  intArrayPtr[low+counter]);
    }
  
  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseIntArrayElements(env, ints, intArrayPtr, JNI_ABORT); //JNI_ABORT because the array passed should never be modified
    }

  
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    getLongs0
 * Signature: (J[JII)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_getLongs0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jlongArray longs, jint low, jint number)
{
  jboolean isCopy;
  jlong* longArrayPtr;

  int counter;

  //Get the array as a pointer
  longArrayPtr = (*env)->GetLongArrayElements(env, longs, &isCopy);

  for(counter=0; counter<number; counter++)
    {
      longArrayPtr[low+counter] = Java_javax_realtime_RawMemoryAccess_getLong0(env, 
									       thisObj, 
									       virtAddress+(counter*sizeof(jlong)));
    }
  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseLongArrayElements(env, longs, longArrayPtr, 0);
    }
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    setLongs0
 * Signature: (J[JII)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_setLongs0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jlongArray longs, jint low, jint number)
{
  jboolean isCopy;
  jlong* longArrayPtr;
 
  int counter;

  //Get the array as a pointer
  longArrayPtr = (*env)->GetLongArrayElements(env, longs, &isCopy);

  for(counter=0; counter<number; counter++)
    {
      Java_javax_realtime_RawMemoryAccess_setLong0(env, 
						   thisObj, 
						   virtAddress+(counter*sizeof(jlong)),
						   longArrayPtr[low+counter]);
    }

  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseLongArrayElements(env, longs, longArrayPtr, JNI_ABORT); //JNI_ABORT because the array passed should never be modified
    }

  
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    getShorts0
 * Signature: (J[SII)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_getShorts0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jshortArray shorts, jint low, jint number)
{
  jboolean isCopy;
  jshort* shortArrayPtr;

  int counter;

  //Get the array as a pointer
  shortArrayPtr = (*env)->GetShortArrayElements(env, shorts, &isCopy);

  for(counter=0; counter<number; counter++)
    {
      shortArrayPtr[low+counter] = Java_javax_realtime_RawMemoryAccess_getShort0(env, 
										 thisObj, 
										 virtAddress+(counter*sizeof(jshort)));
    }
  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseShortArrayElements(env, shorts, shortArrayPtr, 0);
    }
}

/*
 * Class:     javax_realtime_RawMemoryAccess
 * Method:    setShorts0
 * Signature: (J[SII)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryAccess_setShorts0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jshortArray shorts, jint low, jint number)
{
  jboolean isCopy;
  jshort* shortArrayPtr;
  
  int counter;

  //Get the array as a pointer
  shortArrayPtr = (*env)->GetShortArrayElements(env, shorts, &isCopy);

  for(counter=0; counter<number; counter++)
    {
      Java_javax_realtime_RawMemoryAccess_setShort0(env, 
						   thisObj, 
						   virtAddress+(counter*sizeof(jshort)),
						   shortArrayPtr[low+counter]);
    }

  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseShortArrayElements(env, shorts, shortArrayPtr, JNI_ABORT); //JNI_ABORT because the array passed should never be modified
    }

  
}


/* *** RawMemoryFloatAccess functions *** */

/*
 * Class:     javax_realtime_RawMemoryFloatAccess
 * Method:    getDouble0
 * Signature: (J)D
 */
JNIEXPORT jdouble JNICALL Java_javax_realtime_RawMemoryFloatAccess_getDouble0
  (JNIEnv *env, jobject thisObj, jlong virtAddress)
{
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));
  return *((jdouble *) ((int)virtAddress)); //convAddress);
}

/*
 * Class:     javax_realtime_RawMemoryFloatAccess
 * Method:    getFloat0
 * Signature: (J)F
 */
JNIEXPORT jfloat JNICALL Java_javax_realtime_RawMemoryFloatAccess_getFloat0
  (JNIEnv *env, jobject thisObj, jlong virtAddress)
{
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));
  return *((jfloat *) ((int)virtAddress)); //convAddress);
}

/*
 * Class:     javax_realtime_RawMemoryFloatAccess
 * Method:    setDouble0
 * Signature: (JD)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryFloatAccess_setDouble0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jdouble value)
{
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));
  *((jdouble*) ((int)virtAddress)) = value;
  return;
}

/*
 * Class:     javax_realtime_RawMemoryFloatAccess
 * Method:    setFloat0
 * Signature: (JF)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryFloatAccess_setFloat0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jfloat value)
{
  //void *convAddress = NULL;
  // Do a memcpy to get rid of the warnings
  //memcpy(&convAddress, &virtAddress + (sizeof(jlong)-sizeof(void*)), sizeof(void*));
  *((jfloat*) ((int)virtAddress)) = value;
}

/*
 * Class:     javax_realtime_RawMemoryFloatAccess
 * Method:    getDoubles0
 * Signature: (J[DII)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryFloatAccess_getDoubles0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jdoubleArray doubles, jint low, jint number)
{
  jboolean isCopy;
  jdouble* doubleArrayPtr;

  int counter;

  //Get the array as a pointer
  doubleArrayPtr = (*env)->GetDoubleArrayElements(env, doubles, &isCopy);

  for(counter=0; counter<number; counter++)
    {
      doubleArrayPtr[low+counter] = Java_javax_realtime_RawMemoryFloatAccess_getDouble0(env, 
										   thisObj, 
										   virtAddress+(counter*sizeof(jdouble)));
    }
  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseDoubleArrayElements(env, doubles, doubleArrayPtr, 0);
    }
}

/*
 * Class:     javax_realtime_RawMemoryFloatAccess
 * Method:    setDoubles0
 * Signature: (J[DII)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryFloatAccess_setDoubles0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jdoubleArray doubles, jint low, jint number)
{
  jboolean isCopy;
  jdouble* doubleArrayPtr;
  
  int counter;

  //Get the array as a pointer
  doubleArrayPtr = (*env)->GetDoubleArrayElements(env, doubles, &isCopy);

  for(counter=0; counter<number; counter++)
    {
      Java_javax_realtime_RawMemoryFloatAccess_setDouble0(env, 
						     thisObj, 
						     virtAddress+(counter*sizeof(jdouble)),
						     doubleArrayPtr[low+counter]);
    }

  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseDoubleArrayElements(env, doubles, doubleArrayPtr, JNI_ABORT); //JNI_ABORT because the array passed should never be modified
    }

  
}

/*
 * Class:     javax_realtime_RawMemoryFloatAccess
 * Method:    getFloats0
 * Signature: (J[FII)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryFloatAccess_getFloats0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jfloatArray floats, jint low, jint number)
{
  jboolean isCopy;
  jfloat* floatArrayPtr;

  int counter;

  //Get the array as a pointer
  floatArrayPtr = (*env)->GetFloatArrayElements(env, floats, &isCopy);

  for(counter=0; counter<number; counter++)
    {
      floatArrayPtr[low+counter] = Java_javax_realtime_RawMemoryFloatAccess_getFloat0(env, 
										 thisObj, 
										 virtAddress+(counter*sizeof(jfloat)));
    }
  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseFloatArrayElements(env, floats, floatArrayPtr, 0);
    }
}

/*
 * Class:     javax_realtime_RawMemoryFloatAccess
 * Method:    setFloats0
 * Signature: (J[FII)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_RawMemoryFloatAccess_setFloats0
  (JNIEnv *env, jobject thisObj, jlong virtAddress, jfloatArray floats, jint low, jint number)
{
  jboolean isCopy;
  jfloat* floatArrayPtr;
  
  int counter;

  //Get the array as a pointer
  floatArrayPtr = (*env)->GetFloatArrayElements(env, floats, &isCopy);

  for(counter=0; counter<number; counter++)
    {
      Java_javax_realtime_RawMemoryFloatAccess_setFloat0(env, 
						    thisObj, 
						    virtAddress+(counter*sizeof(jfloat)),
						    floatArrayPtr[low+counter]);
    }
  
  
  if(isCopy == JNI_TRUE)
    {
      (*env)->ReleaseFloatArrayElements(env, floats, floatArrayPtr, JNI_ABORT); //JNI_ABORT because the array passed should never be modified
    }

  
}

/******* DefaultPhysicalMemoryFactory *********/

/*
 * Class:     javax_realtime_DefaultPhysicalMemoryFactory
 * Method:    map0
 * Signature: (JJ)J
 */
/*JNIEXPORT jlong JNICALL Java_javax_realtime_DefaultPhysicalMemoryFactory_map0
  (JNIEnv *env, jobject thisObj, jlong base, jlong size)
{
  return (jlong) ((int)internalMMap(env, thisObj, base, size)); 
}*/
