/*
 * -----------------------------------------------------------------
 * ScopedMemory.c  :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


#include "javax_realtime_ScopedMemory.h"

#include "javavm/include/indirectmem.h"
#include "javavm/include/globalroots.h"

#include "javavm/include/mem_registry.h"

/*
 * Class:     javax_realtime_ScopedMemory
 * Method:    initScoped
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_ScopedMemory_initScoped
  (JNIEnv *env, jobject thisObj, jlong memSize)
{
    char *scopedMemBase = (char *) malloc(memSize);
    if(scopedMemBase == NULL)
    {
      //Throw OutOfMemoryError and halt JVM
      jclass clazz = (*env)->FindClass(env, "java/lang/OutOfMemoryError");
      (*env)->ThrowNew(env, clazz, "Could not allocate initial memory for javax.realtime.ScopedMemory");
      return;
    }
//  currentPtr = scopedMemBase;
//  immortalMemorySize = memSize;
  // register with central database
    CVMJTregistryRegisterMemoryArea(CVMjniEnv2ExecEnv(env), thisObj,
      (jlong) ((int)scopedMemBase), memSize, 3);

}

/*
 * Class:     javax_realtime_ScopedMemory
 * Method:    reclaimMemoryAllocated
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_javax_realtime_ScopedMemory_reclaimMemoryAllocated
  (JNIEnv *env, jobject thisObj)
{
    CVMJTregistryReclaimMemory(CVMjniEnv2ExecEnv(env), thisObj);
}

/*
 * Class:     javax_realtime_ScopedMemory
 * Method:    getPortal0
 * Signature: ()Ljava/lang/Object;
 */
JNIEXPORT jobject JNICALL Java_javax_realtime_ScopedMemory_getPortal0
  (JNIEnv *env, jobject thisObj)
{
    // CVMsmGetPortal will create a new local ref to the portal object 
    // for us
    return CVMJTsmGetPortal(CVMjniEnv2ExecEnv(env), thisObj);
}

/*
 * Class:     javax_realtime_ScopedMemory
 * Method:    setPortal0
 * Signature: (Ljava/lang/Object;)V
 */
JNIEXPORT void JNICALL Java_javax_realtime_ScopedMemory_setPortal0
  (JNIEnv *env, jobject thisObj, jobject portalObject)
{
    CVMJTsmSetPortal(CVMjniEnv2ExecEnv(env), thisObj, portalObject);
}

/*
 * Class:     javax_realtime_ScopedMemory
 * Method:    getOuterScope0
 * Signature: ()Ljava/lang/Object;
 */
/*JNIEXPORT jobject JNICALL Java_javax_realtime_ScopedMemory_getOuterScope0
  (JNIEnv *env, jobject thisObj)
{
    
}*/
