import javax.realtime.*;

/* 
   This is the first in a series of basic AIE tests designed as a starting
   point for more comprehensive testing. Each program tests an individual
   feature of the the AsynchronouslyInterruptedException (AIE) class and
   it's features.

   AIE1.java tests basic remote interuption of non-deferred methods. The
   parent thread sleeps for five seconds then interrupts it's child with
   the system generic AIE.

   The correct output is as follows:
   
   * AIE1 Test commencing
   * starting target thread
   * Target thread started
   * Target thread spinning in interruptible method
   * Parent sleeping for five seconds
   * Parent interrupting Target
   * Target caught system generic AIE
   javax.realtime.AsynchronouslyInterruptedException
           at javax.realtime.MemoryArea.newInstance0(Ljava/lang/Class;)Ljava/lang/Object;(Native Method)
	   at javax.realtime.MemoryArea.newInstance(Ljava/lang/Class;)Ljava/lang/Object;(Unknown Source)
	   at javax.realtime.AsynchronouslyInterruptedException.getGeneric()Ljavax/realtime/AsynchronouslyInterruptedException;(Unknown Source)
	   at javax.realtime.RealtimeThread.interrupt()V(Unknown Source)
	   at AIE1.main([Ljava/lang/String;)V(Unknown Source)
   * Test completed successfully


*/


public class AIE1
{
  public static void main(String[] argv) throws InterruptedException
  {
    RealtimeThread t = new RealtimeThread(null, null, null, null, null,
					  new Target1());
    
    System.out.print("* AIE1 Test commencing\n");
    System.out.print("* starting target thread\n");
    
    t.start();

    System.out.print("* Parent sleeping for five seconds\n");
    Thread.sleep(5000);

    System.out.print("* Parent interrupting Target\n");
    t.interrupt();
  }
}

class Target1 implements Runnable
{
  
  public void run()
  {
    System.out.print("* Target thread started\n");
    try{
      loop();
    }
    catch(Throwable e){
      if(e == AsynchronouslyInterruptedException.getGeneric()){
	System.out.print("* Target caught system generic AIE\n");
	e.printStackTrace();
	((AsynchronouslyInterruptedException)e).happened(false);
	System.out.print("* Test completed successfully\n");
      }
      else{
	System.out.print("* Target did not catch generic AIE\n");
	e.printStackTrace();
	System.out.print("* Test failed\n");
      }
    }
  }

  public void loop() throws AsynchronouslyInterruptedException
  {
    System.out.print("* Target thread spinning in interruptible method\n");
    while(true);
  }
  
}
