import javax.realtime.*;

/* 
   This is the fifth in a series of basic AIE tests designed as a starting
   point for more comprehensive testing. Each program tests an individual
   feature of the the AsynchronouslyInterruptedException (AIE) class and
   it's features.

   AIE5.java tests interruption of java.io.* methods. The parent thread sleeps
   for five seconds then interrupts it's child with the system generic
   AIE. Meanwhile, the child is blocked in System.in.read() (a blocking 
   system.io.* method).

   The correct output is as follows:
  
   * AIE5 Test commencing
   * starting target thread
   * Parent sleeping for five seconds
   * Target thread started
   * Target calling blocking IO method, don't type anything
   * Parent interrupting Target
   * Target caught system generic AIE
   javax.realtime.AsynchronouslyInterruptedException
           at javax.realtime.MemoryArea.newInstance0(Ljava/lang/Class;)Ljava/lang/Object;(Native Method)
	   at javax.realtime.MemoryArea.newInstance(Ljava/lang/Class;)Ljava/lang/Object;(Unknown Source)
	   at javax.realtime.AsynchronouslyInterruptedException.getGeneric()Ljavax/realtime/AsynchronouslyInterruptedException;(Unknown Source)
	   at javax.realtime.RealtimeThread.interrupt()V(Unknown Source)
	   at AIE5.main([Ljava/lang/String;)V(Unknown Source)
   * Test completed successfully
*/


public class AIE5
{
  public static void main(String[] argv) throws InterruptedException
  {
    RealtimeThread t = new RealtimeThread(null, null, null, null, null,
					  new Target5());
    
    System.out.print("* AIE5 Test commencing\n");
    System.out.print("* starting target thread\n");
    
    t.start();
    
    System.out.print("* Parent sleeping for five seconds\n");
    Thread.sleep(5000);

    System.out.print("* Parent interrupting Target\n");
    t.interrupt();
  }
}

class Target5 implements Runnable
{

  public void run()
  {
    System.out.print("* Target thread started\n");
    try{
      loop();
    }
    catch(Throwable e){
      if(e == AsynchronouslyInterruptedException.getGeneric()){
	System.out.print("* Target caught system generic AIE\n");
	e.printStackTrace();
	((AsynchronouslyInterruptedException)e).happened(false);
	System.out.print("* Test completed successfully\n");
      }
      else{
	System.out.print("* Target did not catch generic AIE\n");
	e.printStackTrace();
	System.out.print("* Test failed\n");
      }
    }
  }

  public void loop() throws AsynchronouslyInterruptedException
  {
    System.out.print("* Target calling blocking IO method, don't type anything\n");
    try{
      System.in.read();
    }
    catch(java.io.IOException e){
      System.out.print("* Target thread has caught instance of java.io.IOException\n");
    }
    System.out.print("* Target thread has left IO method without throwing aie.\n");
    System.out.print("* Target thread spinning in interruptible method\n");
    while(true);
  }
  
}
