import javax.realtime.*;

/* 
   This is the sixth in a series of basic AIE tests designed as a starting
   point for more comprehensive testing. Each program tests an individual
   feature of the the AsynchronouslyInterruptedException (AIE) class and
   it's features.

   AIE6.java tests AIE replacment of propigaing exceptions. The parent thread
   sleeps for two seconds then interrupts it's child with the system generic
   AIE. Meanwhile, the child is spinning (not sleeping, as sleep has special
   behaviour with AIE's) in a non-ai method before it throws an unrelated
   exception.

   Depending on the testing machine, the nature of the build and other timing
   factors, Target2.count may have to be adjusted to ensure the AIE gets
   posted during the 'defer' method.

   The correct output is as follows:

   * AIE6 Test commencing
   * starting target thread
   * Parent sleeping for two seconds
   * Target thread started
   * Target thread busey waiting in deferred section
   * Parent interrupting Target
   * Target throwing UnrelatedException
   * Target caught system generic AIE
   javax.realtime.AsynchronouslyInterruptedException
           at javax.realtime.MemoryArea.newInstance0(Ljava/lang/Class;)Ljava/lang/Object;(Native Method)
	   at javax.realtime.MemoryArea.newInstance(Ljava/lang/Class;)Ljava/lang/Object;(Unknown Source)
	   at javax.realtime.AsynchronouslyInterruptedException.getGeneric()Ljavax/realtime/AsynchronouslyInterruptedException;(Unknown Source)
	   at javax.realtime.RealtimeThread.interrupt()V(Unknown Source)
	   at AIE6.main([Ljava/lang/String;)V(Unknown Source)
   * Test completed successfully

*/


public class AIE6
{
  public static void main(String[] argv) throws InterruptedException
  {
    RealtimeThread t = new RealtimeThread(null, null, null, null, null,
					  new Target6());
    
    System.out.print("* AIE6 Test commencing\n");
    System.out.print("* starting target thread\n");
    
    t.start();
    
    System.out.print("* Parent sleeping for two seconds\n");
    Thread.sleep(2000);

    System.out.print("* Parent interrupting Target\n");
    t.interrupt();
  }
}

class Target6 implements Runnable
{

  public int count = 50000000;

  public void run()
  {
    System.out.print("* Target thread started\n");
    try{
      loop();
    }
    catch(Throwable e){
      if(e == AsynchronouslyInterruptedException.getGeneric()){
	System.out.print("* Target caught system generic AIE\n");
	e.printStackTrace();
	((AsynchronouslyInterruptedException)e).happened(false);
	System.out.print("* Test completed successfully\n");
      }
      else{
	System.out.print("* Target did not catch generic AIE\n");
	e.printStackTrace();
	System.out.print("* Test failed\n");
      }
    }
  }

  public void defer() throws UnrelatedException
  {
    System.out.print("* Target thread busey waiting in deferred section\n");
    for(int i=0; i<count; i++)
      ;
    System.out.print("* Target throwing UnrelatedException\n");
    throw new UnrelatedException();
  }

  public void loop() throws AsynchronouslyInterruptedException, UnrelatedException
  {
    defer();
    System.out.print("* Target thread has left deferred method without throwing aie.\n");
    System.out.print("* Target thread spinning in interruptible method\n");
    while(true);
  }
}

class UnrelatedException extends Exception
{
}
