import javax.realtime.*;

import java.util.Vector;

public class MemoryParametersTest extends RealtimeThread
{
    public static void usage()
    {
        System.out.println("usage: ./cvm -Djava.class.path=<path> MemoryParametersTest <testno>");
        System.exit(0);
    }

    
    public static void main(String[] args)
    {
        if(args.length == 0)
            usage();
        int testnum = 0;
        
        try
        {
            testnum = Integer.parseInt(args[0]);
        }
        catch(NumberFormatException nme)
        {
            nme.printStackTrace();
            usage();
        }
        
        switch(testnum)
        {
            case 1:
            {
                System.out.println("Starting Test 1");
                test1();
                break;
            }
            
            case 2:
            {
                System.out.println("Starting Test 2");
                test2();
                break;
            }

            case 3:
            {
                System.out.println("Starting Test 3");
                test3();
                break;
            }
            
            default:
            {
                System.out.println("Unknown test number"); 
                usage();
            }
        }
    }

    public MemoryParametersTest(MemoryParameters mp, MemoryArea ma)
    {
        super(null,null, mp, ma, null, null);
    }
   
    public void run()
    {
        for (int i=0; i<10; i++)
        {
            Vector v = new Vector();
        }
        System.out.println("memory allocated");
        try
        {
            Thread.sleep(10000);
        }
        catch(InterruptedException e)
        { e.printStackTrace(); }
            
    }
    
    public static void test1()
    {
        System.out.println("This test tries to set maxImmortal memory to a number less than what the thread has already allocated");
        Vector v = new Vector(); // To make sure classloader has loaded Vector
        MemoryParameters param = new MemoryParameters(10000, 10000);
        MemoryParametersTest m = new MemoryParametersTest(param, ImmortalMemory.instance());
        m.start();
        try
        {
            Thread.sleep(500);
        }
        catch(InterruptedException e)
        { e.printStackTrace(); }

        boolean ret = param.setMaxImmortal(20);
        if(ret)
            System.out.println("FAILED: Tried to set maxImmortal to 20 and it returned true");
        else
            System.out.println("PASS: Tried to set maxImmortal to 20 and it returned false");
    }

    public static void test2()
    {
        System.out.println("This test tries to set maxMemoryArea to a number less than what the thread has already allocated");
        Vector v = new Vector(); // To make sure classloader has loaded Vector
        MemoryParameters param = new MemoryParameters(10000, 10000);
        MemoryParametersTest m = new MemoryParametersTest(param, new LTMemory(20000, 20000));
        m.start();
        try
        {
            Thread.sleep(500);
        }
        catch(InterruptedException e)
        { e.printStackTrace(); }

        boolean ret = param.setMaxMemoryArea(20);
        if(ret)
            System.out.println("FAILED: Tried to set maxMemoryArea to 20 and it returned true");
        else
            System.out.println("PASS: Tried to set maxMemoryArea to 20 and it returned false");
   }
    
    public static void test3()
    {
        Vector v = new Vector(); // To make sure classloader has loaded Vector
        MemoryParameters param = null; //new MemoryParameters(10000, 10000);
        MemoryParametersTest m = new MemoryParametersTest(param, new LTMemory(20000, 20000));
        m.start();
    }

}
