/* InterruptibleThread - This class is a subclass of RealtimeThread used in
**		         the AsynchronouslyInterruptedExceptionTest to provide
**			 a thread containing interruptible logic. A shared
**			 semaphore is passed into the constructor and then
**			 into the interruptible logic constructor so that
**			 the master test can wait for and be notified by the
**			 interruptible logic.  The testcase passed in is used
**			 by the interruptible logic to perform the appropriate
**			 logic for the subtest being run.
*/
import javax.realtime.*;

public class InterruptibleThread extends RealtimeThread 
{
    AsynchronouslyInterruptedException aie;
    Semaphore semaphore;
    int testCase;

    public InterruptibleThread(AsynchronouslyInterruptedException shared_aie,
			       Semaphore shared_semaphore,int tcase)
    {
	super();
	System.out.println("InterruptibleThread: Creating an interruptible "+
			   "thread");
	aie = shared_aie;
	semaphore = shared_semaphore;
	testCase = tcase;
    }

    public void run() {

	System.out.println("InterruptibleThread: run()");

	try {
	    MyInterruptible mi = new MyInterruptible(semaphore,testCase);
	    boolean x = aie.doInterruptible(mi);
	} catch(Exception e) {
	    System.out.println("InterruptibleThread: aie.doInterruptible(mi) "+
			       "in Subtest " + testCase + " of the "+
			       "AsynchronouslyInterruptedExceptionTest threw "+
			       "an exception");
	    Tests.threadFail();
	    e.printStackTrace();
	}

    }
}
