/* MyInterruptible - This class implements the Interruptible interface
**		     and is used by the InterruptibleThread class in the
**		     AsynchronouslyInterruptedExceptionTests
**		     to provide it's interruptible logic. A shared
**		     semaphore is passed into the constructor so that
**		     the master test can wait for and be notified by the
**		     interruptible logic.  The testcase passed in is used
**		     to perform the appropriate logic for the subtest being
**		     run.
*/
import javax.realtime.*;

public class MyInterruptible extends RealtimeThread implements Interruptible
{
    private Semaphore semaphore;
    private int testCase;
    private boolean disabled;

    public MyInterruptible(Semaphore shared_semaphore,int tcase) 
    {

	System.out.println("MyInterruptible: MyInterruptible(Semaphore, int)");

	semaphore = shared_semaphore;
	testCase = tcase;
	disabled = false;
    }
	
    public void run(AsynchronouslyInterruptedException e) throws 
    AsynchronouslyInterruptedException {
	int i;
	boolean x;

	System.out.println("MyInterruptible: run()");

	switch (testCase) {
	case 3: // disable() test
	    disabled = e.disable();
	    if (disabled != true) {
		System.out.println("MyInterruptible: e.disable returned "+
				   "unexpected false value in Subtest 3 of "+
				   "AsynchronouslyInterrputedExceptionTest");
		Tests.threadFail();
				// Wake up master to continue with testing
		semaphore.sNotify();  
	    } else {
		semaphore.sNotify();
		for (i=0;i<100000;i++); 
		disabled = false;
		e.enable();		

		for (i=0; i<1000000;i++);

		System.out.println("MyInterruptible: interrupt never "+
				   "received in Subtest 3 of "+
				   "AsynchronouslyInterruptedExceptionTest");
		Tests.threadFail();
	    }
	    break;

	case 6:  // enable() test
	    e.disable();
	    x = e.enable();
	    if (x != true) {
		System.out.println("MyInterruptible: e.enable() returned "+
				   "unexpected false value from Subtest 6 of "+
				   "AsynchronouslyInterruptedExceptionTest");
		Tests.threadFail();
	    }
	    semaphore.sNotify();
	    for (i=0; i<1000000; i++);
	
	    System.out.println("MyInterruptible: interrupt never received in "+
			       "Subtest 6 of "+
			       "AsynchronouslyInterruptedExceptionTest");
	    Tests.threadFail();
			 
	case 7:  // fire() test
	    semaphore.sNotify();	

	    for (i=0;i<1000000;i++);

	    System.out.println("MyInterruptible: interrupt never received in "+
			       "Subtest 7 of "+
			       "AsynchronouslyInterruptedExceptionTest");
	    Tests.threadFail();

	    break;

		
	case 10:  // isEnabled() test
	    semaphore.sNotify();
	    for (i=0;i<1000000;i++);

	    System.out.println("MyInterruptible: interrupt never received in "+
			       "Subtest 10 of "+
			       "AsynchronouslyInterruptedExceptionTest");
	    Tests.threadFail();

	    break;

	case 11:  // isEnabled() test
	    e.disable();
	    semaphore.sNotify();
	    for (i=0;i<1000000;i++);
	    e.enable();
	    System.out.println("MyInterruptible: interrupt never received in "+
			       "Subtest 11 of "+
			       "AsynchronouslyInterruptedExceptionTest");
	    Tests.threadFail();

	    break;

	case 12: // happened() test
	    if (e.happened(true) != false) {
		System.out.println("MyInterruptible: e.happened() "+
				   "incorrectly recognizes this as the "+
				   "current exception");
		Tests.threadFail();
	    }
	    semaphore.sNotify();
	    for (i=0;i<1000000;i++);
	    break;
			     		
	default:
	    System.out.println("MyInterruptible: bad testcase passed in from "+
			       "AsynchronouslyInterruptedExceptionTest");
	    Tests.threadFail();
	}

    }
	
    public void interruptAction(AsynchronouslyInterruptedException e)
    {
	switch (testCase) {
	case 3:
	    if (disabled == true)
		System.out.println("MyInterruptible: MyInterruptible "+
				   "processed interrupt even though "+
				   "interrupt was disabled in Subtest 3 of "+
				   "AsynchronouslyInterruptedExceptionTest");
	    Tests.threadFail();
	    break;

	case 12:
	    if (e.happened(true) == false) {
		System.out.println("MyInterruptible: MyInterruptible "+
				   "processed interrupt even though the "+
				   "state was changed to non-pending via "+
				   "happened() method in Subtest 12 of "+
				   "AsynchronouslyInterruptedExceptionTest");
		Tests.threadFail();
	    }
	    break;

	}

	System.out.println("**** MyInterruptible's run method was excepted "+
			   "in Subtest " + testCase);
    }
	
}
