//package javax.realtime.calltests;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2000  All Rights Reserved
 */


/*				PeriodicParametersTest
 
Subtest 1:
	"public PeriodicParameters(HighResolutionTime start, RelativeTime 
	period, RelativeTime cost, RelativeTime deadline, AsyncEventHandler 
	overrunHandler, AsyncEventHandler missHandler)" where all given 
	parameters are null

Subtest 2:
	"public PeriodicParameters(HighResolutionTime start, RelativeTime 
	period, RelativeTime cost, RelativeTime deadline, AsyncEventHandler 
	overrunHandler, AsyncEventHandler missHandler)" where given parameters
	have value

Subtest 3:
	"public RelativeTime getPeriod()"

Subtest 4:
	"public HighResolutionTime getStart()"

Subtest 5:
	"public void setPeriod(RelativeTime period)"

Subtest 6:
	"public void setStart(HighResolutionTime start)"
*/

import javax.realtime.*;

public class PeriodicParametersTest 
{

    public static void run() 
    {
	PeriodicParameters pp = null;
	Object o = null;

	/* Subtest 1:
	** Constructor "public PeriodicParameters(HighResolutionTime start, 
	** RelativeTime period, RelativeTime cost, RelativeTime deadline, 
	** AsyncEventHandler overrunHandler, AsyncEventHandler missHandler)" 
	** where all given parameters are null
	*/
	Tests.newTest("PeriodicParametersTest");
	Tests.increment();
	try {
	    System.out.println("PeriodicParametersTest: PeriodicParameters("+
			       "null,null,null,null,null,null)");
	    HighResolutionTime start = null;
	    RelativeTime period = null;
	    RelativeTime cost = null;
	    RelativeTime deadline = null;
	    AsyncEventHandler overrunHandler = null;
	    AsyncEventHandler missHandler = null;
	    pp = new PeriodicParameters(start, period, cost, deadline, 
					overrunHandler, missHandler);
	    if( !(pp instanceof PeriodicParameters && pp instanceof 
		  ReleaseParameters) )
		throw new Exception("Return object is not instanceof "+
				    "PeriodicParameters nor "+
				    "ReleaseParameters");
	} catch (Exception e) {
	    System.out.println("PeriodicParametersTest: PeriodicParameters("+
			       "null,null,null,null,null,null) failed");
	    Tests.fail("new PeriodicParameters(null,null,null,null,null,null)",
		       e);
	}
		
	/* Subtest 2:
	** Constructor "public PeriodicParameters(HighResolutionTime start, 
	** RelativeTime period, RelativeTime cost, RelativeTime deadline, 
	** AsyncEventHandler overrunHandler, AsyncEventHandler missHandler)" 
	** where given parameters have value
	*/
	Tests.increment();
	RelativeTime start = new RelativeTime(1024L, 100);
	RelativeTime period = new RelativeTime(1000L, 0);
	RelativeTime cost = new RelativeTime(900L, 0);
	RelativeTime deadline = new RelativeTime(1000L, 0);
	AEventHandler overrunHandler = new AEventHandler();
	AEventHandler missHandler = new AEventHandler();
	try {
	    System.out.println("PeriodicParametersTest: PeriodicParameters("+
			       "HighResolutionTime,RelativeTime,RelativeTime,"+
			       "RelativeTime,null,null)");
	    pp = new PeriodicParameters(start, period, cost, deadline, 
					overrunHandler, missHandler);
	    if( !(pp instanceof PeriodicParameters && pp instanceof 
		  ReleaseParameters) )
		throw new Exception("Return object is not instanceof "+
				    "PeriodicParameters nor "+
				    "ReleaseParameters");
	} catch (Exception e) {
	    System.out.println("PeriodicParametersTest: PeriodicParameters("+
			       "HighResolutionTime,RelativeTime,RelativeTime,"+
			       "RelativeTime,null,null) failed");
	    Tests.fail("new PeriodicParameters(start,period,cost,deadline,"+
		       "null,null)",e);
	}

	/* Subtest 3:
	** Method "public RelativeTime getPeriod()"
	*/
	Tests.increment();
	try {
	    System.out.println("PeriodicParametersTest: getPeriod()");
	    RelativeTime rt = pp.getPeriod();
	    if (rt.equals(period)==false)
		{
		    throw new Exception("Period was not set correctly");
		}
	} catch (Exception e) {
	    System.out.println("PeriodicParametersTest: getPeriod() failed");
	    Tests.fail("pp.getPeriod()",e);
	}

	/* Subtest 4:
	** Method "public HighResolutionTime getStart()"
	** Note: In this test, I know that the return type should be
	** also an instanceof RelativeTime because that is what I used
	** in the constructor of Subtest 2:
	*/
	Tests.increment();
	try {
	    System.out.println("PeriodicParametersTest: getStart()");
	    o = pp.getStart();
	    if( !(o instanceof HighResolutionTime && o instanceof 
		  RelativeTime) )
		throw new Exception("Return object is not instanceof "+
				    "HighResolution nor RelativeTime");
	    RelativeTime rt = (RelativeTime)o;	
	    if (rt.equals(start)==false)
		{
		    throw new Exception("Start time was not set correctly");
		}
	
	} catch (Exception e) {
	    System.out.println("PeriodicParametersTest: getStart() failed");
	    Tests.fail("pp.getStart()",e);
	}

	/*
	** Subtest 5:
	** Method "public void setPeriod(RelativeTime period)"
	*/
	Tests.increment();
	try {
	    System.out.println("PeriodicParametersTest: setPeriod(Relative"+
			       "Time)");
	    RelativeTime rperiod = new RelativeTime(2000L,100);
	    pp.setPeriod(rperiod);
	    RelativeTime rt = pp.getPeriod();
	    if (rt.equals(rperiod)==false) {
		throw new Exception("Period was not set correctly");
	    }
	    
	} catch (Exception e) {
	    System.out.println("PeriodicParametersTest: setPeriod(Relative"+
			       "Time) failed");
	    Tests.fail("pp.setPeriod(rperiod)",e);
	}

	/* Subtest 6:
	** Method "public void setStart(HighResolutionTime start)"
	*/

	Tests.increment();
	try {
	    System.out.println("PeriodicParametersTest: setStart(High"+
			       "ResolutionTime)");
	    AbsoluteTime astart = new AbsoluteTime(new java.util.Date());
	    astart.add(10223L,100);
	    pp.setStart(astart);
	    HighResolutionTime st = pp.getStart();
	    if (st.equals(astart)==false)
		throw new Exception("Start was not set correctly");
	} catch (Exception e) {
	    System.out.println("PeriodicParametersTest: setStart(High"+
			       "ResolutionTime) failed");
	    Tests.fail("pp.setStart(astart)",e);
	}

	Tests.printSubTestReportTotals("PeriodicParametersTest");
    }
}
