//package javax.realtime.calltests;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2000  All Rights Reserved
 */


/*				RawMemoryFloatAccessTest
 
Subtest 1
	"public static RawMemoryFloatAccess createFloatAccess(java.lang.Object
	type, long size)

Subtest 2
	"public static RawMemoryFloatAccess createFloatAccess(java.lang.Object
	type, long base, long size)"

Subtest 3
	"public void setDouble( long offset, double value)" and
	"public double getDouble( long offset)"

Subtest 4
	"public void setDoubles( long offset, double[] doubles, int low, int 
	number)" and
	"public void getDoubles( long offset, double[] doubles, int low, int 
	number))"

Subtest 5
	"public void setFloat( long offset, float value)" and
	"public float getFloat( long offset)"

Subtest 6
	"public void setFloats( long offset, float[] floats, int low, int 
	number)" and
	"public void getFloats( long offset, float[] floats, int low, int 
	number))"

*/

import javax.realtime.*;

public class RawMemoryFloatAccessTest 
{
    public static void run() 
    {
	RawMemoryFloatAccess rmfa = null;
	Object o = null;

	// constructors protected


	Tests.newTest("RawMemoryFloatAccessTest");

	/* Subtest 1
	** Method "public static RawMemoryFloatAccess createFloatAccess(
	** java.lang.Object type, long size)
	*/
	Tests.increment();
		
	try {
	    System.out.println("RawMemoryFloatAccessTest: createFloatAccess("+
			       "Object,long)");
	    long size = 1000;
	    o = RawMemoryFloatAccess.createFloatAccess(new Object(), size);
	    if( !(o instanceof RawMemoryFloatAccess) )
		throw new Exception("Return object is not instanceof "+
				    "RawMemoryFloatAccess");
	} catch (Exception e) {
	    System.out.println("RawMemoryFloatAccessTest: createFloatAccess("+
			       "Object,long) failed");
	    Tests.fail("RawMemoryFloatAccess.createFloatAccess()",e);
	}

	/* Subtest 2
	** Method "public static RawMemoryFloatAccess createFloatAccess(
	** java.lang.Object type, long base, long size)"
	*/
	Tests.increment();
	try {
	    System.out.println("RawMemoryFloatAccessTest: createFloatAccess("+
			       "Object,long,long)");
	    long base = 0;
	    long size = 1000;
	    o = RawMemoryFloatAccess.createFloatAccess(new Object(), base, 
						       size); 
	    if( !(o instanceof RawMemoryFloatAccess) )
		throw new Exception("Return object is not instanceof "+
				    "RawMemoryFloatAccess");
	} catch (Exception e) {
	    System.out.println("RawMemoryFloatAccessTest: createFloatAccess("+
			       "Object,long,long) failed");
	    Tests.fail("RawMemoryFloatAccess.createFloatAccess()",e);
	}

	/* Subtest 3
	** Methods "public void setDouble( long offset, double value)" and
	**         "public double getDouble( long offset)"
	*/
	Tests.increment();
	try {
	    System.out.println("RawMemoryFloatAccessTest: setDouble(long,"+
			       "double), getDouble(long)");
	    long base = 0;
	    long size = 1000;
	    rmfa = RawMemoryFloatAccess.createFloatAccess(new Object(), base, 
							  size);
	    rmfa.setDouble(0L,(double)112.5);
	    if(rmfa.getDouble(0L) != 112.5)
		throw new Exception("Double not set properly");
	} catch (Exception e) {
	    System.out.println("RawMemoryFloatAccessTest: setDouble(long,"+
			       "double), getDouble(long) failed");
	    Tests.fail("rmfa.setDouble() and rmfa.getDouble()",e);
	}
		
	/* Subtest 4
	** Methods "public void setDoubles( long offset, double[] doubles, 
	** int low, int number)" and
	**         "public void getDoubles( long offset, double[] doubles, 
	** int low, int number))"
	*/
	Tests.increment();
	try {
	    System.out.println("RawMemoryFloatAccessTest: setDoubles(long,"+
			       "double[],int,int), getDoubles(long,double[],"+
			       "int,int)");
	    double[] doubles = new double[4];
	    doubles[0] = 1;
	    doubles[1] = 2;
	    rmfa.setDoubles(5L, doubles, 0, 2);
	  rmfa.getDoubles(5L, doubles, 2, 2);
	  
	  if (!(doubles[0]==doubles[2] && doubles[1]==doubles[3]))
	      throw new Exception("Doubles were not set properly");
	} catch (Exception e) {
	    System.out.println("RawMemoryFloatAccessTest: setDoubles(long,"+
			       "double[],int,int), getDoubles(long,double[],"+
			       "int,int) failed");
	    Tests.fail("rmfa.setDoubles() and rmfa.getDoubles()",e);
	}
	/* Subtest 5
	** Methods "public void setFloat( long offset, float value)" and
	**         "public float getFloat( long offset)"
	*/
	Tests.increment();
	try {
	    System.out.println("RawMemoryFloatAccessTest: setFloat(long,"+
			       "float), getFloat(long)");
	    rmfa.setFloat(0L,(float)124.5);
	    if(rmfa.getFloat(0L) != 124.5)
		throw new Exception("Float not set properly");
	} catch (Exception e) {
	    System.out.println("RawMemoryFloatAccessTest: setFloat(long,"+
			       "float), getFloat(long) failed");
	    Tests.fail("rmfa.setFloat() and rmfa.getFloat()",e);
	}
	/* Subtest 6
	** Methods "public void setFloats( long offset, float[] floats, int 
	** low, int number)" and
	**         "public void getFloats( long offset, float[] floats, int 
	** low, int number))"
	*/
	Tests.increment();
	try {
	    System.out.println("RawMemoryFloatAccessTest: setFloats(long,"+
			       "float[],int,int), getFloats(long,float[],int,"+
			       "int)");
	    float[] floats = new float[4];
	    floats[0] = 100;
	    floats[1] = 200;
	    rmfa.setFloats(5L, floats, 0, 2);
	    rmfa.getFloats(5L, floats, 2, 2);
	    
	  if (!(floats[0]==floats[2] && floats[1]==floats[3]))
	      throw new Exception("Floats were not set properly");
	} catch (Exception e) {
	    System.out.println("RawMemoryFloatAccessTest: setFloats(long,"+
			       "float[],int,int), getFloats(long,float[],int,"+
			       "int) failed");
	    Tests.fail("rmfa.setFloats() and rmfa.getFloats()",e);
	}
	Tests.printSubTestReportTotals("RawMemoryFloatAccessTest");
    }
}
