/* Tests - this is the main logic that calls every RTSJ API test class.
*/

// package javax.realtime.calltests;

/*
 * Licensed Materials - Property of IBM,
 * (c) Copyright IBM Corp. 2000  All Rights Reserved
 */

import java.util.Vector;
import javax.realtime.*;
import com.timesys.*;


public class Tests 
{

    /* Report option SHOWTESTS - 
    **	true - reports every subtest as it runs
    **	false - only reports failed subtests (preferred)
    */
    private static final boolean SHOWTESTS = false;
	
    private static int threadFailureCount;
    private static int subcount;  //counts individual subtests for one test
    private static int subfailed; //counts individual failed subtests for one 
                                  // test
    private static int count;     //counts all individual subtests 
    private static int failed;    //counts all individual subtests that fail

    private static TestReportFile report;  //Report file

    /*
    ** Runs all tests reporting failures and total counts
    */

    public static void main(java.lang.String[] args) 
    {

	System.setOut(System.err);

	System.out.println("Tests: Starting the testing procedure");
	count = 0;
	failed = 0;
	threadFailureCount=0;

	report = new TestReportFile();         //Create report file

	report.println("Running tests...");
	report.println("");

	/* FIXME - reenable these tests

	System.out.println("Tests: RealtimeThreadTest");
	RealtimeThreadTest.run();

	System.out.println("Tests: NoHeapRealtimeThreadTest");
	NoHeapRealtimeThreadTest.run(); 

	System.out.println("Tests: SchedulerTest");
	SchedulerTest.run();

	System.out.println("Tests: PrioritySchedulerTest");
	PrioritySchedulerTest.run();

	System.out.println("Tests: PriorityParametersTest");
	PriorityParametersTest.run(); 

	System.out.println("Tests: ImportanceParametersTest");
	ImportanceParametersTest.run();

	System.out.println("Tests: PeriodicParametersTest");
	PeriodicParametersTest.run();      

	System.out.println("Tests: AperiodicParametersTest");
	AperiodicParametersTest.run();

	System.out.println("Tests: SporadicParametersTest");
	SporadicParametersTest.run(); 

	System.out.println("Tests: ProcessingGroupParametersTest");
	ProcessingGroupParametersTest.run(); 

	System.out.println("Tests: HeapMemoryTest");
	HeapMemoryTest.run(); 

	System.out.println("Tests: ImmortalMemoryTest");
	ImmortalMemoryTest.run(); 

	System.out.println("Tests: VTMemoryTest");
	VTMemoryTest.run();

	System.out.println("Tests: LTMemoryTest");
	LTMemoryTest.run();

	System.out.println("Tests: PhysicalMemoryFactoryTest");
	PhysicalMemoryFactoryTest.run(); 

	System.out.println("Tests: ImmortalPhysicalMemoryTest");
	ImmortalPhysicalMemoryTest.run(); 

	System.out.println("Tests: ScopedPhysicalMemoryTest");
	ScopedPhysicalMemoryTest.run(); 

	System.out.println("Tests: RawMemoryAccessTest");
	RawMemoryAccessTest.run(); 

	System.out.println("Tests: RawMemoryFloatAccessTest");
	RawMemoryFloatAccessTest.run(); 

	System.out.println("Tests: MemoryParametersTest");
	MemoryParametersTest.run(); 

	System.out.println("Tests: MarkAndSweepCollectorTest");
	MarkAndSweepCollectorTest.run(); 

	System.out.println("Tests: PriorityCeilingEmulationTest");
	PriorityCeilingEmulationTest.run(); 

	System.out.println("Tests: PriorityInheritanceTest");
	PriorityInheritanceTest.run();

	//	System.out.println("Tests: WaitFreeDequeueTest");
	//	WaitFreeDequeueTest.run();// causes stack trace
*/
	System.out.println("Tests: WaitFreeReadQueueTest");
	WaitFreeReadQueueTest.run();

	//	System.out.println("Tests: WaitFreeWriteQueueTest");
	//	WaitFreeWriteQueueTest.run();

	System.out.println("Tests: AbsoluteTimeTest");
	AbsoluteTimeTest.run();

	System.out.println("Tests: RelativeTimeTest");
	RelativeTimeTest.run();  

	System.out.println("Tests: RationalTimeTest");
	RationalTimeTest.run();

	System.out.println("Tests: ClockTest");
	ClockTest.run();

	System.out.println("Tests: OneShotTimerTest");
	OneShotTimerTest.run();	

	System.out.println("Tests: PeriodicTimerTest");
	PeriodicTimerTest.run();

	System.out.println("Tests: AsyncEventTest");
	AsyncEventTest.run(); 

	System.out.println("Tests: AsyncEventHandlerTest");
	AsyncEventHandlerTest.run();

	System.out.println("Tests: BoundAsyncEventHandlerTest");
	BoundAsyncEventHandlerTest.run();

	System.out.println("Tests: AsynchronouslyInterruptedExceptionTest");
	AsynchronouslyInterruptedExceptionTest.run(); 

	System.out.println("Tests: TimedTest");
	TimedTest.run(); 

	System.out.println("Tests: POSIXSignalHandlerTest");
	POSIXSignalHandlerTest.run(); 

	System.out.println("Tests: RealtimeSecurityTest");
	RealtimeSecurityTest.run();

	//	System.out.println("Tests: RealtimeSystemTest");
	//	RealtimeSystemTest.run();

	System.out.println("Tests: IllegalAssignmentErrorTest");
	IllegalAssignmentErrorTest.run();

	System.out.println("Tests: MemoryAccessErrorTest");
	MemoryAccessErrorTest.run();

	System.out.println("Tests: MemoryScopeExceptionTest");
	MemoryScopeExceptionTest.run(); 

	System.out.println("Tests: OffsetOutOfBoundsExceptionTest");
	OffsetOutOfBoundsExceptionTest.run(); 

	System.out.println("Tests: ResourceLimitErrorTest");
	ResourceLimitErrorTest.run();

	System.out.println("Tests: SizeOutOfBoundsExceptionTest");
	SizeOutOfBoundsExceptionTest.run(); 

	System.out.println("Tests: ThrowBoundaryErrorTest");
	ThrowBoundaryErrorTest.run(); 

	System.out.println("Tests: UnsupportedPhysicalMemoryExceptionTest");
	UnsupportedPhysicalMemoryExceptionTest.run(); 

	System.out.println("Tests: Done running tests");

	printReportTotals(); 

	for (int i=0;i<100000000;i++);

	try {
	    int pid = Dispatcher.getPID();
	    System.out.println("");
	    System.out.println(" ***********************************");
	    System.out.println(" **** RTSJ API TESTING COMPLETE ****");
	    System.out.println(" **** Please type: kill -9 " + pid + " ****");
	} catch (Exception e) {
	    System.out.println(" **** Cannot get PID!           ****");
	}
	System.out.println(" ***********************************");
	System.out.println("");
    }

    /*
    ** Global Methods for tracking tests
    */
	
    /*
    ** newTest - marks beginning of new Test 
    */
    static void newTest( String arg )       
    {
	report.println("");
	report.println(arg + ":");      // Prints Test heading
	subcount=0;                     // initializes subtest count
	subfailed=0;			// initializes failed subtests
    }

    /*
    ** increment() - increments count of subtests within a test 
    */
    static void increment()                
    {
	count++;
	subcount++;
	if(SHOWTESTS)                  // Print what test is running
	    report.println("  Running test: " + subcount);
    }

    /*
    ** fail() - reports a failed subtest
    */
    static void fail(String arg)  
    {
	report.println("  " + arg + " failed at Subtest " + subcount);

	failed++;
	subfailed++;
    }

    /*
    ** fail() - reports a failed subtest with exception
    */
    static void fail(String arg, Exception e)      
    {
	report.println("  " + arg + " failed at Subtest " + subcount);
	report.println("    " + e.toString());

	failed++;
	subfailed++;
    }
	
    /*
    ** threadFail() - reports a failed test within a separate thread
    */	
    static void threadFail()
    {
	threadFailureCount++;
    }

    /*
    ** comment() - prints a comment into the report
    */
    static void comment(String arg)
    {
	report.println("  COMMENT: " + arg);
    }

    /* 
    ** printSubTestReporTotals - called at the end of a test to print
    **			     all counts.
    */
    static void printSubTestReportTotals(String arg) 
    {
	report.println("");
	report.println("  " + arg + " TOTALS:");
	report.println("  " + subcount + " tests RAN");
	report.println("  " + subfailed + " tests FAILED");
    }

    /*
    ** printReportTotals - called at the end of all tests to print
    **		       the total counts.
    */
    private static void printReportTotals() 
    {
	report.println("");
	report.println("OVERALL TEST TOTALS:");
	report.println("  " + count + " tests RUN");
	report.println("  " + failed + " tests FAILED");
	report.println("  " + threadFailureCount + " test thread(s) FAILED");
	if (threadFailureCount > 0)
	    report.println("    (see messages beginning with **** "+
			   "THREAD FAILURE:  in the standard output for more "+
			   "information)" );
    }

}
