import javax.realtime.*;
import java.util.*;

public class MemData {

    int type, id, initial, maximum, num_threads, num_objects;

    int use_type;

    MemoryArea mem_area;

    Vector thread_data, object_data;

    Vector objects;

    sanity_checker san_chk;

    public MemData(int i_d, sanity_checker sc) {
	id=i_d;
	num_threads=0;
	num_objects=0;
	thread_data=new Vector();
	object_data=new Vector();
	objects=new Vector();
	initial=0;
	maximum=0;
	san_chk=sc;
    }

    public void setMemoryArea(MemoryArea ma) {
	mem_area=ma;
    }

    public void setUseType(int t) { 
	if(num_threads != 0) {
	    System.out.println("***** ERROR MemData : usage count > 0, yet "+
			       "MemData use_type was set to "+t);
	    san_chk.destroyAll();
	}
	use_type=t; 
    }

    public void release(int id) {

	System.out.println("Inside MemData release, num_obj= "+num_objects);
	for(int q=0;q<num_objects;q++) System.out.println(q+" "+objects.elementAt(q));

	for(int x=0; x < num_objects; x++) {

	    System.out.println("MemData : Object # "+x);
	    ((ObjectData)objects.elementAt(x)).dump();


	    if(((ObjectData)objects.elementAt(x)).getId() == id) {
		objects.remove(x);
		
		if(mem_area instanceof MyHeapMemory) {
		    ((MyHeapMemory)mem_area).remove(id);
		} else if(mem_area instanceof MyImmortalMemory) {
		    ((MyImmortalMemory)mem_area).remove(id);
		} else if(mem_area instanceof MyLTMemory) {
		    ((MyLTMemory)mem_area).remove(id);
		} else if(mem_area instanceof MyVTMemory) {
		    ((MyVTMemory)mem_area).remove(id);
		} else {
		    System.out.println("MemData : ERROR - mem_area of "+
				       "unknown type");
		}
	    }
	}
    }

    public void setType(int t) { type=t; }
    public void setInitial(int init) { initial=init; }
    public void setMaximum(int max) { maximum=max; }

    public int getId() { return id; }

    public MemoryArea getMemArea() { 
	return mem_area;
    }

    public void dump(int caller) {

	System.out.println("-----Begin MemData Dump-----");

	System.out.print("type : ");

	switch(type) {

	case MemoryFactory.HEAP_MEMORY :
	    System.out.print("HEAP");
	    break;
	case MemoryFactory.IMMORTAL_MEMORY :
	    System.out.print("IMMORTAL");
	    break;
	case MemoryFactory.VT_MEMORY :
	    System.out.print("VT");
	    break;
	case MemoryFactory.LT_MEMORY :
	    System.out.print("LT");
	    break;

	}

	System.out.println(" id : "+id+" initial : "+initial+
			   " maximum : "+maximum+" num objs : "+num_objects);

	if(caller != sanity_checker.THREAD) {
	    System.out.println("Threads");
	    for(int x=0; x < num_threads; x++) {
		((ThreadData)thread_data.elementAt(x)).dump(
					    sanity_checker.MEMORY);
	    }
	}

	if(caller != sanity_checker.OBJECT) {
	    System.out.println("Objects");
	    for(int x=0; x < num_objects; x++) {
		((ObjectData)object_data.elementAt(x)).dump(
					    sanity_checker.MEMORY);
	    }
	}
	System.out.println("-----End MemData Dump-----");

    }

    public void dump() { dump(-1); }

    public void addThreadData(ThreadData td) {
	thread_data.addElement(td);
	num_threads++;
    }

    public void removeThreadData(ThreadData td) {
	for(int x=0; x < num_threads; x++) {
	    if(td.getId()==((ThreadData)thread_data.elementAt(x)).getId()) {
		thread_data.remove(x);
		num_threads--;
	    }
	}
    }

    public void addObjectData(ObjectData od) {
	object_data.addElement(od);
	num_objects++;
    }

    public void remObjectData(ObjectData od) {
	for(int x=0; x < num_objects; x++) {
	    if(od.getId()==((ObjectData)object_data.elementAt(x)).getId()) {
		object_data.remove(x);
		num_objects--;
	    }
	}
    }
    
    public MemoryArea getMemoryArea() { return mem_area; }

}



