import java.util.*;
import javax.realtime.*;

public class sanity_checker {

    // This is to describe the memory that is given to a thread
    // as is applicable to the spec. changes
    final static public int STATIC=0;
    final static public int NESTED=1;

    final static public int THREAD=1;
    final static public int OBJECT=2;
    final static public int MEMORY=3;

    // lock insures a stepwise execution of the threads. While this will
    // definitely kill any interleaving testing that would normally be done,
    // (FIXME is this ok?) it will also make this project doable by one 
    // person
    static int lock;

    int num_threads, num_mem_areas, num_objects;

    Vector thr_data, mem_area_data, object_data;

    static int id=0;

    ThreadFactory tf;
    MemoryFactory mf;
    ObjectFactory of;
 
    ThreadData just_removed;

    public sanity_checker() {

	System.out.println("*** san_chk initializing");

	lock=0;
	num_threads=0;
	num_mem_areas=0;
	thr_data=new Vector();
	mem_area_data=new Vector();
	num_objects=0;
	object_data=new Vector();
    }

    /*
      This is my way of making this entire class synchronized. Basically,
      whenever a thread wants to execute, the first thing that it's going
      to do is call lock(), execute to completion, and then call unlock()
    */
    public synchronized void lock(Thread thr) {

	System.out.println("*** san_chk lock() entered by ");
	callDump(thr);

	// To keep all of the threads from interleaving and causing a bunch 
	// of headaches. This will reduce the usefullness of the tester
	// slightly, but if necessary there can be certain parts of each
	// test that are locked, as opposed to the whole shebang and also
	// it is possible to make methods synchronized
	while(lock == 1) {
	    thr.yield();
	    try {
		Thread.sleep(1000);
	    } catch (Throwable t) {
		System.out.println(t);
	    }
	}

	lock=1;

	System.out.println("*** san_chk locked by ");
	callDump(thr);
	
	// This is to introduce a delay of one second in between all thread
	// executions so that a human can look at the screen and see what's
	// going on
	try {
	    Thread.sleep(1000);
	} catch (Throwable t) {
	    System.out.println(t);
	}
	
	System.out.println("\n\n\n*****STARTING A NEW CYCLE*****%%%%%\n\n\n");

    }

    public void unlock(Thread thr) {
	check_sanity();
	System.out.println("*** san_chk unlocked by ");
	callDump(thr);
	lock=0;
    }

    // FIXME - these need to be done, and are the most important part of
    // the project.
    void check_threads() {}
    void check_memory_areas() {}

    void check_objects() {

	// This will remove any object that has called it's selfDestruct() 
	// method so that references to it will not be given to any threads
	// any more
	for(int x=0; x < num_objects; x++) {
	    ObjectData od=((ObjectData)object_data.elementAt(x));
	    if(od.getIsAssignable() == 0) {
		if(od.getReferenceCount()==0) {

		    detach(od.getMemData().getMemoryArea(), od.getId());

		    System.out.println("*** san_chk is removing the object");
		    od.dump();
		    od.releaseObject();
		    object_data.remove(x);
		    num_objects--;
		}
	    }
	}

    }
    
    // This will dump out all of the relevant information after each thread
    // has gone through one "cycle" of execution
    void data_dump() {

	System.out.println("**********Data Dump**********");

	System.out.println("\nThreads --- ("+num_threads+")\n");
	for(int x=0; x<num_threads; x++) {
            System.out.print("T#"+x+"::");
            ((ThreadData)thr_data.elementAt(x)).dump();
        }

	System.out.println("\nMemory Areas --- ("+num_mem_areas+")\n");
	for(int x=0; x<num_mem_areas; x++) {
            System.out.print("M#"+x+"::");
            ((MemData)mem_area_data.elementAt(x)).dump();
        }

	System.out.println("\nObject Data --- ("+num_objects+")\n");
	for(int x=0; x<num_objects; x++) {
            System.out.print("O#"+x+"::");
            ((ObjectData)object_data.elementAt(x)).dump();
        }

	System.out.println("**********Done Dump**********");

    }

    public void check_sanity() {
	check_threads();
	check_memory_areas();
	check_objects();
	data_dump();
    }

    public int getNumObjects() {
	return num_objects;
    }

    public int getNumThreads() {
	return num_threads;
    }

    public int getNumMemAreas() {
	return num_mem_areas;
    }

    public int uniqueId() { id++; return id; }

    public void addThreadData(ThreadData td) {
	thr_data.addElement(td);
	num_threads++;
    }

    public void addMemData(MemData md) {
	mem_area_data.addElement(md);
	num_mem_areas++;
    }

    public void addObjectData(ObjectData od) {
	object_data.addElement(od);
	num_objects++;
    }

    public MemData getMemData(int id) {
	for(int x=0; x<num_mem_areas; x++) {
	    if(((MemData)mem_area_data.elementAt(x)).getId()==id) {
		return ((MemData)mem_area_data.elementAt(x));
	    }
	}
	return null;
    }

    public MemData getMemDataIndex(int index) {
	return ((MemData)mem_area_data.elementAt(index));
    }

    // This is what is used by a function that wants to get a random object
    public ObjectData getObjectIndex(int index) {
	return ((ObjectData)object_data.elementAt(index));
    }

    // attach will be the catch phrase for associating two *Datas
    public void attach(Thread t, MemoryArea m, int md_type) {

	ThreadData td=convert(t);
	MemData md=convert(m);
	
	md.setUseType(md_type);
	md.addThreadData(td);

	td.addMemData(md);

	System.out.println("*** san_chk Just attached a thread and a memory "+
			   "area");
	td.dump();
	md.dump();

    }

    public void attach(Thread t, int mo) {

	ThreadData td=convert(t);
	ObjectData od=convert(mo);

	td.addObjectData(od);
	od.addThreadData(td);

	System.out.println("*** san_chk Just attached a thread and an object");
	td.dump();
	od.dump();

    }

    public void attach(MemoryArea ma, int mo) {
	
	MemData md=convert(ma);
	ObjectData od=convert(mo);

	md.addObjectData(od);
        od.setMemData(md);

    }

    ThreadData convert(Thread t) {
	for(int x=0; x<num_threads; x++) {
	    if(((ThreadData)thr_data.elementAt(x)).getThread() == t) {
		return ((ThreadData)thr_data.elementAt(x));
	    }
	}
	if(just_removed != null) {
	    if(just_removed.getThread()==t) {
		System.out.println("san_chk returning just_removed");
		return just_removed;
	    }
	}

	System.out.println("***** sanity_checker convert(thread) : thread "+
			   "not found");
	return null;
    }

    ObjectData convert(int o) {
	for(int x=0; x<num_objects; x++) {
	    if(((ObjectData)object_data.elementAt(x)).getId() == o) {
		return ((ObjectData)object_data.elementAt(x));
	    }
	}
	System.out.println("***** sanity_checker convert(int) : object "+
			   "not found");
	return null;
    }

    MemData convert(MemoryArea ma) {

	for(int x=0; x<num_mem_areas; x++) {
	    if(((MemData)mem_area_data.elementAt(x)).getMemoryArea() == ma) {
		return(((MemData)mem_area_data.elementAt(x)));
	    }
	}

	return null;

    }

    public void detach(MemoryArea ma, int obj) {
	
	MemData md=convert(ma);
	ObjectData od=convert(obj);

	md.remObjectData(od);

    }

    public void detach(Thread t, int o) {
	ThreadData td=convert(t);
	ObjectData od=convert(o);

	td.removeObjectData(od);
	od.removeThreadData(td);

	System.out.println("*** san_chk just detached a thread and an object");
	td.dump();
	od.dump();
    }

    public void destroyAll() {

	System.out.println("***** sanity_checker destroyAll() setting quit "+
			   "for all threads in the system");

	for(int x=0; x<num_threads; x++) {

	    System.out.print("***** sanity_checker killing T # "+x+" of "+
			     num_threads);

	    ThreadData td=((ThreadData)thr_data.elementAt(x));

	    System.out.println(" Id # "+td.getId());

	    switch (td.getType()) {
		
	    case ThreadFactory.USER_THREAD :
		((MyThread)td.getThread()).setQuit();
		break;
            case ThreadFactory.RT_THREAD :
		((MyRealtimeThread)td.getThread()).setQuit();
		break;
	    case ThreadFactory.NHRT_THREAD :
		((MyNoHeapRealtimeThread)td.getThread()).setQuit();
		break;

	    }
	}
	
	tf.setQuit();
	mf.setQuit();
	of.setQuit();

	System.out.println("***** sanity_checker destroyAll() has set quit "+
			   "for all threads in the system");

    }

    // FIXME not done
    public void pushThreadData(Thread t) {}
    
    // FIXME also not done
    public ThreadData popThreadData(Thread t) {
	ThreadData td = new ThreadData(-1);
	return td;
    }

    public void removeThread(Thread t) {

	ThreadData td=convert(t);
	System.out.println("*** san_chk is removing a thread");
	td.dump();
	for(int x=0; x < num_threads; x++) {
	    td=((ThreadData)thr_data.elementAt(x));
	    if(td.getThread()==t) {
		thr_data.remove(x);
		num_threads--;
		just_removed=td;
	        // FIXME - does the memory stuff also have to be updated here?
		return;
	    }
	}
    }

    public void callDump(Thread t) {

	if(t instanceof MemoryFactory) System.out.println("*** san_chk "+
				       "callDump()==MemoryFactory");
	else if(t instanceof ThreadFactory) System.out.println("*** san_chk "+
					    "callDump()==ThreadFactory");
	else if(t instanceof ObjectFactory) System.out.println("*** san_chl "+
   				            "callDump()==ObjectFactory");
	else if(t instanceof mem_tester) System.out.println("*** san_chk "+
				         "callDump()==mem_tester");
        else if(t instanceof MyThread) {
	    System.out.println("*** san_chk callDump for MyThread");
	    convert(t).dump();
	} else if(t instanceof MyRealtimeThread) {
	    System.out.println("*** san_chk callDump for MyRealtimeThread");
	    convert(t).dump();
	} else if(t instanceof MyNoHeapRealtimeThread) {
	    System.out.println("*** san_chk callDump for "+
			       "MyNoHeapRealtimeThread");
	    convert(t).dump();
	}

    }

}

