/*
 * -----------------------------------------------------------------
 *  ansi_java_md.c :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)ansi_java_md.c	1.33 00/11/11
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <setjmp.h>

#include "javavm/export/jni.h"
#include "porting/ansi_c/java.h"

/* %comment dlong001 */
#include "javavm/include/utils.h"
#include "javavm/include/common_exceptions.h"

/* *** RTSJ *** */
#include <semaphore.h>
/* *** JSTR *** */

static jmp_buf jmp_env;
static int in_main = 0;
static int exitCode = 0;

static void
safeExit(int status)
{
    if (in_main) {
	exitCode = status;
	longjmp(jmp_env, 1);
    } else {
	fprintf(stderr, "Unexpected VM exit(%d)\n", status);
    }
}

int
ansiJavaMain(int argc, const char** argv)
{
    JavaVM*         jvm = 0;
    JNIEnv*         env = 0;
    JavaVMInitArgs  args;
    int i;

    args.version = JNI_VERSION_1_2;
    args.ignoreUnrecognized = JNI_FALSE;
    /* Allocate JVM arguments and put argv in there.
       These are parsed up in Java by the CVM class.

       argv contains the incoming options to the VM, like -Xms20m, as
       well as the program name and its arguments. We're going to try
       to do as much command line parsing as possible in Java. Here we
       just parcel up argv so that the arguments we pass into
       JNI_CreateJavaVM are spec-compliant. Namely, we prepend the
       program name and its arguments by "-Xcvm". These are recognized
       by CVM.parseCommandLineOptions and used to build up the list of
       arguments for the main() method. Note that it is also valid to
       call JNI_CreateJavaVM without the special -Xcvm arguments (as a
       user embedding the VM in another application would) and do all
       command line parsing and constructing of arguments for main()
       in C.

       %comment kbr001

       %comment kbr002
    */       
#ifdef ANSI_USE_EXIT_HOOK
#define EXTRA_ARGS 3
#else
#define EXTRA_ARGS 2
#endif
#define ARGV_OFFSET 1
#define OPT_INDEX(i) ((i) + EXTRA_ARGS - ARGV_OFFSET)
    args.nOptions = EXTRA_ARGS + argc - 1;
    args.options =
	(JavaVMOption*) malloc(args.nOptions * sizeof(JavaVMOption));

    args.options[0].optionString = "vfprintf";
    args.options[0].extraInfo = (void *)vfprintf;
    args.options[1].optionString = "abort";
    args.options[1].extraInfo = (void *)abort;
#ifdef ANSI_USE_EXIT_HOOK
    args.options[2].optionString = "exit";
    args.options[2].extraInfo = (void *)exit;
#endif

    for (i = ARGV_OFFSET; i < argc; i++) {
	if (argv[i][0] != '-') {
	    /* Prepend "-Xcvm" to the rest of the command line options */
	    while (i < argc) {
		char* tmpStr;
		/* Make room for "-Xcvm" and NULL */
		tmpStr = malloc(sizeof(char) *
				    (strlen(argv[i]) + 6));
		strcpy(tmpStr, "-Xcvm");
		strcat(tmpStr, argv[i]);
		args.options[OPT_INDEX(i)].optionString = tmpStr;
		args.options[OPT_INDEX(i)].extraInfo = NULL;
		i++;
	    }
	    break;
	} else {
	    if (strcmp(argv[i], "-XsafeExit") == 0) {
		args.options[OPT_INDEX(i)].optionString = (char*) argv[i];
		args.options[OPT_INDEX(i)].extraInfo = (void *)safeExit;
	    } else {
		args.options[OPT_INDEX(i)].optionString = (char*) argv[i];
		args.options[OPT_INDEX(i)].extraInfo = NULL;
	    }
	}
    }

    in_main = 0;

    /* *** RTSJ *** */
    // Added by Pratik & Kunal.
    {
        extern sem_t vmWait;
        if(sem_init(&vmWait, 0, 0) != 0)
        {
            printf("Unable to initialise vmWait semaphore. Aborting!\n");
            return -1;
        }
    }
    /* *** JSTR *** */

    if (JNI_CreateJavaVM(&jvm, (void **)&env, &args) == JNI_OK &&
	args.version == JNI_VERSION_1_2)
    {
	jclass			cvmClass;
	jmethodID		getParseStatusID;
	jmethodID		getMainClassNameID;
	jmethodID		getMainArgumentsID;
	jobject			mainClassName;
	const char*		mainClassChars;
	jobject			mainArgs;
	volatile jclass		mainClass = NULL;
	volatile jmethodID	mainMethodID;
	jint            	parseStatus;

    /* *** RTSJ *** */
    // JNI_CreateJavaVM completed successfully.
    // Post the semaphore.
    {
        extern sem_t vmWait;
        if(sem_post(&vmWait) != 0)
        {
            printf("Unable to post semaphore vmWait. Aborting.\n");
            return -1;
        }
    }
    /* *** JSTR *** */
    
	/*
	 * First of all, free the memory that we allocated. The options
	 * have already been parsed and copied by JNI_CreateJavaVM
	 */
	for (i = ARGV_OFFSET; i < argc; i++) {
	    if (argv[i][0] != '-') {
		/* Prepend "-Xcvm" to the rest of the command line options */
		while (i < argc) {
		    CVMassert(args.options[OPT_INDEX(i)].optionString !=
			      NULL);
		    CVMassert(args.options[OPT_INDEX(i)].extraInfo == NULL);

		    free(args.options[OPT_INDEX(i)].optionString);
		    args.options[OPT_INDEX(i)].optionString = NULL;
		    i++;
		}
		break;
	    }
	}
	free(args.options);

	/* this is the first place to put the CoreExtension loader
	 */
    // Following intialisation code commented out by Pratik on 
    // 16th April 2001. Initialisation is now being done in JNI_CreateJavaVM
#if 0
	{
	  jclass coreExtLoaderClass;
	  jmethodID bootMethodID;
	  extern int JVMCeiInitialized;

	  printf("Loading Core Extensions Loader\n");

	  coreExtLoaderClass = (*env)->FindClass(env,"com/timesys/CoreExtension/Loader");
	  if (coreExtLoaderClass == NULL)
	    {
	      printf("Could not find Loader class\n");
	      goto uncaughtException;
	    }

	  bootMethodID = (*env)->GetStaticMethodID(env, coreExtLoaderClass, "boot", "()V");

	  if (bootMethodID == NULL)
	    {
	      printf("Could not find the boot method for Loader");
	      goto uncaughtException;
	    }

	  (*env)->CallStaticVoidMethod(env, coreExtLoaderClass, bootMethodID);
	  if ((*env)->ExceptionOccurred(env))
	    goto uncaughtException;

	  printf("CoreExtensions initialized!!! \n");
	  JVMCeiInitialized = 1;
	}
#endif

	args.options = NULL; /* This has all been processed by
				JNI_CreateJavaVM */
	cvmClass = CVMcbJavaInstance(CVMsystemClass(sun_misc_CVM));
	getParseStatusID =
	    (*env)->GetStaticMethodID(env, cvmClass,
				      "getParseStatus",
				      "()I");
	getMainClassNameID =
	    (*env)->GetStaticMethodID(env, cvmClass,
				      "getMainClassName",
				      "()Ljava/lang/String;");
	getMainArgumentsID =
	    (*env)->GetStaticMethodID(env, cvmClass,
				      "getMainArguments",
				      "()[Ljava/lang/String;");

	parseStatus =
	    (*env)->CallStaticIntMethod(env, cvmClass, getParseStatusID);
	if ((*env)->ExceptionCheck(env)) {
	    goto uncaughtException;
	}
	if (parseStatus == sun_misc_CVM_ARG_PARSE_EXITVM) {
	    goto exitingVM;
	} else {
	    /* The error case must have been filtered out */
	    CVMassert(parseStatus == sun_misc_CVM_ARG_PARSE_OK);
	}
	mainClassName = (*env)->CallStaticObjectMethod(env, cvmClass,
						       getMainClassNameID);
	if ((*env)->ExceptionCheck(env) || (mainClassName == NULL)) {
	    goto uncaughtException;
	}
	mainArgs = (*env)->CallStaticObjectMethod(env, cvmClass,
						  getMainArgumentsID);
	if (mainArgs == NULL) {
	    goto uncaughtException;
	}
	
	mainClassChars = (*env)->GetStringUTFChars(env, mainClassName, NULL);
	if ((*env)->ExceptionCheck(env) || (mainClassChars == NULL)) {
	    (*env)->DeleteLocalRef(env, mainClassName);
	    goto uncaughtException;
	}
	
	(*env)->DeleteLocalRef(env, mainClassName);
	mainClass = (*env)->FindClass(env, mainClassChars);
	(*env)->ReleaseStringUTFChars(env, mainClassName, mainClassChars);

	if (mainClass == NULL) { /* exception already thrown */
	    goto uncaughtException;
	}

	mainMethodID =
	    (*env)->GetStaticMethodID(env, mainClass,
				      "main", "([Ljava/lang/String;)V");
	if (mainMethodID == NULL) {
	    goto uncaughtException;
	}

	/*
	 * The vm spec says that the main method must be public, although
	 * the JDK does not check for this.
	 */
	if (!CVMmbIs(mainMethodID, PUBLIC)) {
	    CVMthrowIllegalAccessException(CVMjniEnv2ExecEnv(env),
					   "%M is not public", mainMethodID);
	    goto uncaughtException;
	}

	if (setjmp(jmp_env) == 0) {
	    in_main = 1;
	    (*env)->CallStaticVoidMethod(env, mainClass, mainMethodID,
					 mainArgs);
	} else {
	    return exitCode;
	}

    uncaughtException:
	/* Print stack traces of exceptions which propagate out of
	   main(). Should we call ThreadGroup.uncaughtException()
	   here? JDK1.2 doesn't. */
	{
	    jthrowable exceptionObj = (*env)->ExceptionOccurred(env);
	    if (exceptionObj != NULL) {
		(*env)->ExceptionDescribe(env);
		(*env)->DeleteLocalRef(env, exceptionObj);
	    }
	}

	(*env)->DeleteLocalRef(env, mainClass);

    exitingVM:
	if ((*jvm)->DetachCurrentThread(jvm) != 0) {
	    fprintf(stderr, "Could not detach main thread.\n");
	}

	(*jvm)->DestroyJavaVM(jvm);
    } else {
	fprintf(stderr, "Could not create JVM.\n");
    }
    
    /* *** RTSJ *** */
    {
        extern sem_t vmWait;
        sem_destroy(&vmWait);
    }
    /* *** JSTR *** */
    
    return 0;
}
