/*
 * @(#)posix_io_md.c	1.16 00/11/11
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/porting/io.h"
#include "javavm/include/porting/int.h"
#include "javavm/include/porting/doubleword.h"
#include "javavm/include/interpreter.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include <error.h>

/* NOTE: the implementations here definitely do not support the
   concept of "interruptible I/O". They are also not thread-safe. Any
   necessary locks should be implemented up in Java. */

/* %comment dlong001 */
extern int errno;
CVMInt32
POSIXioGetLastErrorString(char *buf, CVMInt32 len)
{
   if (errno == 0) {
	return 0;
    } else {
	const char *s = strerror(errno);
	int n = strlen(s);
	if (n >= len) n = len - 1;
	strncpy(buf, s, n);
	buf[n] = '\0';
	return n;
    }
}

/*
 * Return the last error string, without any buffers.
 */
char *
POSIXioReturnLastErrorString()
{
    return strerror(errno);
}

#ifndef POSIX_HAVE_CUSTOM_NATIVE_PATH
char *
POSIXioNativePath(char *path)
{
    return path;
}
#endif

CVMInt32
POSIXioFileType(const char *path)
{
    int ret;
    struct stat buf;

    if ((ret = stat(path, &buf)) == 0) {
      mode_t mode = buf.st_mode & S_IFMT;
      if (mode == S_IFREG) return CVM_IO_FILETYPE_REGULAR;
      if (mode == S_IFDIR) return CVM_IO_FILETYPE_DIRECTORY;
      return CVM_IO_FILETYPE_OTHER;
    }
    return ret;
}


#ifndef POSIX_HAVE_64_BIT_IO
static int
fileMode(int fd, int *mode)
{
    int ret;
    struct stat buf;

    ret = fstat(fd, &buf);
    (*mode) = buf.st_mode;
    return ret;
}

CVMInt32
POSIXioOpen(const char *name, CVMInt32 openMode,
	  CVMInt32 filePerm)
{
    CVMInt32 fd = open(name, openMode, filePerm);
    if (fd >= 0) {
	int mode;
	int statres = fileMode(fd, &mode);
	if (statres == -1) {
	    close(fd);
	    return -1;
	}
	if (S_ISDIR(mode)) {
	    close(fd);
	    return -1;
	}
    }
    return fd;
}
#endif

CVMInt32
POSIXioClose(CVMInt32 fd)
{
    return close(fd);
}

#if !defined(POSIX_HAVE_64_BIT_IO) && !defined(POSIX_HAVE_CUSTOM_LSEEK)
CVMInt64
POSIXioSeek(CVMInt32 fd, CVMInt64 offset, CVMInt32 whence)
{
    return lseek(fd, offset, whence);
}
#endif

#ifndef POSIX_HAVE_64_BIT_IO
CVMInt32
POSIXioSetLength(CVMInt32 fd, CVMInt64 length)
{
    /* %comment dlong002 */
    CVMInt32 length32;
    CVMInt64 length64;
    length32 = CVMlong2Int(length);
    length64 = CVMint2Long(length32);

    if (CVMlongNe(length64, length)) {
	/* 32-bit overflow */
	return -1;
    }
#ifdef POSIX_HAVE_FTRUNCATE
    return ftruncate(fd, length32);
#elif defined(POSIX_HAVE_FIOTRUNC)
    /* %comment dlong004 */
    return ioctl(fd, FIOTRUNC, length32);
#else
    return -1;
#endif
}
#endif

CVMInt32
POSIXioSync(CVMInt32 fd)
{
#ifdef POSIX_HAVE_FSYNC
    return fsync(fd);
#elif defined(POSIX_HAVE_FIOFLUSH)
    /* %comment dlong004 */
    return ioctl(fd, FIOFLUSH, 0);
#else
    return -1;
#endif
}

#if !defined(POSIX_HAVE_64_BIT_IO) && !defined(POSIX_HAVE_CUSTOM_IO_AVAILABLE)
CVMInt32
POSIXioAvailable(CVMInt32 fd, CVMInt64 *bytes)
{
    CVMInt64 cur, end;
    int mode;

    if (fileMode(fd, &mode) >= 0) {
        if (S_ISCHR(mode) || S_ISFIFO(mode) || S_ISSOCK(mode)) {
	    int n;
            if (ioctl(fd, FIONREAD, &n) >= 0) {
		/* %comment dlong003 */
		*bytes = CVMint2Long(n);
                return 1;
            }
        }
    }
    {
	CVMInt32 cur32, end32;
	if ((cur32 = lseek(fd, 0L, SEEK_CUR)) == -1) {
	    return 0;
	} else if ((end32 = lseek(fd, 0L, SEEK_END)) == -1) {
	    return 0;
	} else if (lseek(fd, cur32, SEEK_SET) == -1) {
	    return 0;
	}
	/* %comment dlong003 */
	cur = CVMint2Long(cur32);
	end = CVMint2Long(end32);
    }
    *bytes = CVMlongSub(end, cur);
    return 1;
}
#endif

size_t
POSIXioRead(CVMInt32 fd, void *buf, CVMUint32 nBytes)
{
    int n;
    CVMExecEnv* ee = CVMgetEE();
      
//    if(CVMpendingAIException(ee) && CVMATCdirty(ee))
//	    return -1;
    
    ee->isIOBlocked = CVM_TRUE;    
   
read_again:	    if((n =  read(fd, buf, nBytes))<0){
			  //  printf("Dirty flag value:: %d\n",CVMATCdirty(ee));
			    if(errno == EINTR && CVMATCdirty(ee)<=0){
					CVMATCsetDirty(ee,0);
					goto read_again;
			    }
			   /* else{
				JNIEnv *env = CVMexecEnv2JniEnv(ee);
			    	jclass clazz = (*env)->FindClass(env,"java/lang/IOException");	
				(*env)->ThrowNew(env,clazz,"Blocked IO Operation Interrupted");
			    }*/
		    }

    ee->isIOBlocked = CVM_FALSE;
    
    return n;
}

size_t
POSIXioWrite(CVMInt32 fd, const void *buf,
	   CVMUint32 nBytes)
{
    int n;
    CVMExecEnv* ee = CVMgetEE();
      
    
//    if(CVMpendingAIException(ee) && CVMATCdirty(ee))
//	    return -1;
   
    ee->isIOBlocked = CVM_TRUE;   

write_again:	    if((n =  write(fd, buf, nBytes))<0){
			  //  printf("Dirty flag value:: %d\n",CVMATCdirty(ee));
			    if(errno == EINTR && CVMATCdirty(ee)<0){
                                        CVMATCsetDirty(ee,0);
                                        goto write_again;
                            }				    
		    }

		    
    ee->isIOBlocked = CVM_FALSE;
    
   return n;
   
}

#ifndef POSIX_HAVE_64_BIT_IO
CVMInt32
POSIXioFileSizeFD(CVMInt32 fd, CVMInt64 *size)
{
    int ret;
    struct stat buf;

    ret = fstat(fd, &buf);
    /* %comment dlong003 */
    *size = CVMint2Long(buf.st_size);
    return ret;
}
#endif
