/*
 * -----------------------------------------------------------------
 * posix_threads_md.c
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */



/*
 * @(#)posix_threads_md.c	1.15 00/11/11
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/porting/threads.h"
#include "javavm/include/porting/ansi/stdlib.h"
#include "porting/posix/threads.h"
#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <signal.h>

struct args {
    void (*func)(void *);
    void *arg;
};

static void *
start_func(void *a)
{
    struct args *args = (struct args *)a;
    void (*func)(void *) = args->func;
    void *arg = args->arg;
    free(a);
    (*func)(arg);
    return 0;
}
CVMInt32
POSIXthreadCreate(CVMSize stackSize, CVMInt16 priority,
    void (*func)(void *), void *arg)
{
  return POSIXthreadCreateEx(stackSize, priority, -1, func, arg);
}

/* Dio - extended version of thread create to include the policy */

CVMInt32
POSIXthreadCreateEx(CVMSize stackSize, CVMInt16 priority, CVMInt16 policy,
    void (*func)(void *), void *arg)
{
    struct args *args;
    pthread_t pthreadID;
    struct sched_param param;
    pthread_attr_t attr;
    int result;

    args = (struct args *)malloc(sizeof *args);
    if (args == NULL) {
	return 0;
    }
    args->func = func;
    args->arg = arg;
    pthread_attr_init(&attr);
    if (stackSize != 0) {
	errno = pthread_attr_setstacksize(&attr, stackSize);
	if (errno != 0) {
	    perror("pthread_attr_setstacksize");
	}
    }
    pthread_attr_getschedparam(&attr, &param);
    /* %comment dlong001 */
    param.sched_priority = priority;
    pthread_attr_setschedparam(&attr, &param);

    // Dio the thread should inherit the priority from the parent
    // until it can reassign it's own priority later on (RTThreadAdaptor.c)

    pthread_attr_setinheritsched(&attr, PTHREAD_INHERIT_SCHED);

    /* Don't leave thread/stack around after exit for join */
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    //    printf ("ISHIN - pthread is created \n");

    result = pthread_create(&pthreadID, &attr, start_func, args);
    pthread_attr_destroy(&attr);




    if (result != 0) {
	perror("pthread_create");
	free(args);
	return 0;
    } else {
	return 1;
    }
}

static pthread_key_t key;

void
POSIXthreadInitStaticState()
{
    pthread_key_create(&key, NULL);
}

void
POSIXthreadDestroyStaticState()
{
    pthread_key_delete(key);
}

static void
POSIXthreadSetSelf(const CVMThreadID *self)
{
    pthread_setspecific(key, self);
}

CVMThreadID *
POSIXthreadGetSelf()
{
    return (CVMThreadID *)pthread_getspecific(key);
}

void
POSIXthreadAttach(CVMThreadID *self)
{
    POSIX_COOKIE(self) = pthread_self();
    POSIXthreadSetSelf(self);
}

void
POSIXthreadDetach(CVMThreadID *self)
{
    POSIXthreadSetSelf(NULL);
}

void
POSIXthreadSetPriority(CVMThreadID *t, CVMInt32 priority)
{
    struct sched_param param;
    int policy;
    pthread_getschedparam(POSIX_COOKIE(t), &policy, &param);
    param.sched_priority = priority;
    pthread_setschedparam(POSIX_COOKIE(t), policy, &param);
}

void
POSIXthreadGetPriority(CVMThreadID *t, CVMInt32 *priority)
{
    struct sched_param param;
    int policy;
    pthread_getschedparam(POSIX_COOKIE(t), &policy, &param);
    *priority = param.sched_priority;
}

void
POSIXthreadSendSig(CVMThreadID *t)
{
   if(pthread_kill(POSIX_COOKIE(t),SIGUSR2)<0){
	   
//	   printf("Error sending signal to thread %x",POSIX_COOKIE(t));
   }

}


