/*
 * @(#)Version.java	1.4 00/05/19
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

package sun.misc;
import java.io.PrintStream;

public class Version {

    private static String java_version;

    private static final String java_runtime_name =
	"Java(TM) 2, Micro Edition";

    private static final String java_runtime_version =
	"1.0";

    static {
	init();
    }

    public static void init() {
	System.setProperty("java.runtime.version", java_runtime_version);
	System.setProperty("java.runtime.name", java_runtime_name);
	java_version = System.getProperty("java.version");
    }

    /**
     * In case you were wondering this method is called by java -version.
     * Sad that it prints to stderr; would be nicer if default printed on
     * stdout.
     */
    public static void print(boolean longVersion) {
	print(System.err, longVersion);
    }

    /**
     * Give a stream, it will print version info on it.
     */
    public static void print(PrintStream ps, boolean longVersion) {
	String java_vm_version =
	    (String)System.getProperty("java.vm.version");
	if (longVersion) {
	    /* First line: platform version. */
	    ps.println("java version \"" + java_version + "\"");
	    
	    /* Second line: runtime version (ie, libraries). */
	    ps.println(java_runtime_name + " (build " +
		       java_vm_version + ")");

	    /* Third line: JVM information. */
	    String java_vm_name    =
		(String)System.getProperty("java.vm.name");
	    String java_vm_info    =
		(String)System.getProperty("java.vm.info");
	    ps.println(java_vm_name + " (build " + java_vm_version + ", " +
		       java_vm_info + ")");
	} else {
	    /* Second line: runtime version (ie, libraries). */
	    ps.println("java full version \"" + java_vm_version + "\"");

	}
    }

}

// Help Emacs a little because this file doesn't end in .java.
//
// Local Variables: ***
// mode: java ***
// End: ***
