/*
 * @(#)defs.h	1.63 00/09/06
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifndef _INCLUDED_DEFS_H
#define _INCLUDED_DEFS_H

#include "javavm/include/porting/defs.h"
#include "javavm/include/porting/vm-defs.h"

/*
 * Basic type definitions
 */

typedef char      CVMUtf8;

#define CVMoffsetof(T, f)    (CVMUint32)(&(((T *)0)->f))

/*
 * Object format
 */
CVM_STRUCT_TYPEDEF(CVMObjectHeader);
CVM_UNION_TYPEDEF (CVMVariousBits);
CVM_UNION_TYPEDEF (CVMJavaVal8);
CVM_UNION_TYPEDEF (CVMJavaVal16);
CVM_UNION_TYPEDEF (CVMJavaVal32);
CVM_UNION_TYPEDEF (CVMJavaVal64);
CVM_STRUCT_TYPEDEF(CVMjava_lang_Object);
CVM_STRUCT_TYPEDEF(CVMSyncVector);

typedef CVMjava_lang_Object CVMObject;  /* Short-hand for CVMjava_lang_Object */

/*
 * Array types
 */
CVM_STRUCT_TYPEDEF(CVMArrayOfByte);	/* CVMArrayOfByte    */
CVM_STRUCT_TYPEDEF(CVMArrayOfShort);	/* CVMArrayOfShort   */
CVM_STRUCT_TYPEDEF(CVMArrayOfChar);	/* CVMArrayOfChar    */
CVM_STRUCT_TYPEDEF(CVMArrayOfBoolean);	/* CVMArrayOfBoolean */
CVM_STRUCT_TYPEDEF(CVMArrayOfInt);	/* CVMArrayOfInt     */
CVM_STRUCT_TYPEDEF(CVMArrayOfRef);	/* CVMArrayOfRef     */
CVM_STRUCT_TYPEDEF(CVMArrayOfFloat);	/* CVMArrayOfFloat   */
CVM_STRUCT_TYPEDEF(CVMArrayOfLong);	/* CVMArrayOfLong    */
CVM_STRUCT_TYPEDEF(CVMArrayOfDouble);	/* CVMArrayOfDouble  */
CVM_STRUCT_TYPEDEF(CVMArrayOfAnyType);	/* CVMArrayOfAnyType */

/*
 * ICell types
 */
CVM_STRUCT_TYPEDEF(CVMjava_lang_ObjectICell);	/* CVMjava_lang_ObjectICell */
CVM_STRUCT_TYPEDEF(CVMObjectICell);		/* ObjectICell         */

CVM_STRUCT_TYPEDEF(CVMArrayOfByteICell);	/* CVMArrayOfByteICell    */
CVM_STRUCT_TYPEDEF(CVMArrayOfShortICell);	/* CVMArrayOfShortICell   */
CVM_STRUCT_TYPEDEF(CVMArrayOfCharICell);	/* CVMArrayOfCharICell    */
CVM_STRUCT_TYPEDEF(CVMArrayOfBooleanICell);	/* CVMArrayOfBooleanICell */
CVM_STRUCT_TYPEDEF(CVMArrayOfIntICell);		/* CVMArrayOfIntICell     */
CVM_STRUCT_TYPEDEF(CVMArrayOfRefICell);		/* CVMArrayOfRefICell     */
CVM_STRUCT_TYPEDEF(CVMArrayOfFloatICell);	/* CVMArrayOfFloatICell   */
CVM_STRUCT_TYPEDEF(CVMArrayOfLongICell);	/* CVMArrayOfLongICell    */
CVM_STRUCT_TYPEDEF(CVMArrayOfDoubleICell);	/* CVMArrayOfDoubleICell  */
CVM_STRUCT_TYPEDEF(CVMArrayOfAnyTypeICell);	/* CVMArrayOfAnyTypeICell */

/*
 * Class data structures
 */
CVM_STRUCT_TYPEDEF(CVMClassBlock);
CVM_UNION_TYPEDEF (CVMGCBitMap);
CVM_STRUCT_TYPEDEF(CVMBigGCBitMap);
CVM_STRUCT_TYPEDEF(CVMArrayInfo);
CVM_STRUCT_TYPEDEF(CVMInterfaces);
CVM_STRUCT_TYPEDEF(CVMInterfaceTable);
CVM_STRUCT_TYPEDEF(CVMMethodBlock);
CVM_STRUCT_TYPEDEF(CVMMethodRange);
CVM_STRUCT_TYPEDEF(CVMMethodArray);
CVM_STRUCT_TYPEDEF(CVMCheckedExceptions);
CVM_STRUCT_TYPEDEF(CVMJavaMethodDescriptor);
CVM_STRUCT_TYPEDEF(CVMExceptionHandler);
CVM_STRUCT_TYPEDEF(CVMLineNumberEntry);
CVM_STRUCT_TYPEDEF(CVMLocalVariableEntry);
CVM_STRUCT_TYPEDEF(CVMFieldBlock);
CVM_STRUCT_TYPEDEF(CVMFieldRange);
CVM_STRUCT_TYPEDEF(CVMFieldArray);
CVM_STRUCT_TYPEDEF(CVMInnerClassInfo);
CVM_STRUCT_TYPEDEF(CVMInnerClassesInfo);
CVM_UNION_TYPEDEF(CVMConstantPool);
CVM_STRUCT_TYPEDEF(CVMTransitionConstantPool);
CVM_UNION_TYPEDEF (CVMConstantPoolEntry);
CVM_STRUCT_TYPEDEF (CVMStackMapEntry);
CVM_STRUCT_TYPEDEF(CVMStackMaps);

/*
 * Classes which the VM uses directly
 */

typedef CVMObjectICell CVMThreadICell;		/* java.lang.Thread */
typedef CVMObjectICell CVMThrowableICell;	/* java.lang.Throwable */
typedef CVMObjectICell CVMStringICell;		/* java.lang.String */
typedef CVMObjectICell CVMClassICell;		/* java.lang.Class */
typedef CVMObjectICell CVMClassLoaderICell;	/* java.lang.ClassLoader */

/*
 * Interpreter data structures
 */
CVM_STRUCT_TYPEDEF(CVMExecEnv);

/*
 * Synchronization-related types.
 */

typedef CVMBool CVMTryLockFunc     (CVMExecEnv*     ee,
				    CVMObject*      obj);
typedef CVMBool CVMLockFunc        (CVMExecEnv*     ee,
				    CVMObjectICell* indirectObj);

typedef CVMBool CVMNotifyFunc      (CVMExecEnv*     ee,
				    CVMObjectICell* indirectObj);
typedef CVMBool CVMNotifyAllFunc   (CVMExecEnv*     ee,
				    CVMObjectICell* indirectObj);
typedef CVMBool CVMWaitFunc        (CVMExecEnv*     ee,
				    CVMObjectICell* indirectObj,
				    CVMJavaLong     millis);

CVM_STRUCT_TYPEDEF(CVMReentrantMutex);
CVM_STRUCT_TYPEDEF(CVMSysMutex);
CVM_STRUCT_TYPEDEF(CVMSysMonitor);
CVM_STRUCT_TYPEDEF(CVMNamedSysMonitor);
CVM_STRUCT_TYPEDEF(CVMObjMonitor);
CVM_STRUCT_TYPEDEF(CVMOwnedMonitor);

/*
 * Types for consistent states
 */

CVM_STRUCT_TYPEDEF(CVMCState);
CVM_STRUCT_TYPEDEF(CVMTCState);


/*
 * Class loading data structures
 */
CVM_STRUCT_TYPEDEF(CVMLoaderCacheEntry);
#ifdef CVM_CLASSLOADING
CVM_STRUCT_TYPEDEF(CVMLoaderConstraint);
CVM_STRUCT_TYPEDEF(CVMSeenClass);
CVM_STRUCT_TYPEDEF(CVMClassPathEntry);
#endif

/*
 * Stack data structures
 */
CVM_UNION_TYPEDEF (CVMStackVal32);
CVM_STRUCT_TYPEDEF(CVMStack);
CVM_STRUCT_TYPEDEF(CVMStackChunk);
CVM_STRUCT_TYPEDEF(CVMFrame);
CVM_STRUCT_TYPEDEF(CVMLocalRootsFrame);
CVM_STRUCT_TYPEDEF(CVMFreelistFrame);
CVM_STRUCT_TYPEDEF(CVMInterpreterFrame);
CVM_STRUCT_TYPEDEF(CVMJavaFrame);
CVM_STRUCT_TYPEDEF(CVMTransitionFrame);
CVM_STRUCT_TYPEDEF(CVMStackWalkContext);

/*
 * VM global state
 */
CVM_STRUCT_TYPEDEF(CVMGlobalState);
CVM_STRUCT_TYPEDEF(CVMGCGlobalState);

/*
 * VM initialization options
 */
CVM_STRUCT_TYPEDEF(CVMOptions);

/*
 * GC-related types
 */
CVM_STRUCT_TYPEDEF(CVMGCOptions);

/*
 * A 'ref callback' called on each discovered root
 */
typedef void (*CVMRefCallbackFunc)(CVMObject** refAddr, void* data);

/*
 * A predicate to test liveness of a given reference.
 */
typedef CVMBool (*CVMRefLivenessQueryFunc)(CVMObject** refAddr, void* data);

#ifdef CVM_DEBUG
/*
 * A per-object callback function, to be called during heap dumps
 */
typedef void (*CVMObjectCallbackFunc)(CVMObject* obj, CVMClassBlock* cb, 
				      CVMUint32  objSize, void* data);
#endif

/*
 * Scan all GC references in a frame. Each frame carries a pointer
 * to one of these.
 */
typedef void CVMFrameGCScannerFunc(CVMExecEnv* ee,
				   CVMFrame* thisFrame,
				   CVMStackChunk* thisChunk,
				   CVMRefCallbackFunc refCallback,
				   void* data,
				   CVMGCOptions* gcOpts);

/*
 * JNI types
 */

struct JNINativeInterface;

typedef const struct JNINativeInterface * JNIEnv;

#endif /* _INCLUDED_DEFS_H */
