/*
 * @(#)doubleword.h	1.17 00/05/19
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifndef _INCLUDED_PORTING_DOUBLEWORD_H
#define _INCLUDED_PORTING_DOUBLEWORD_H

#include "javavm/include/porting/defs.h"
#include "javavm/include/porting/vm-defs.h"

/*
 * 64-bit value access
 *
 * Convert 64-bit Java longs between native representation and JVM
 * two-word representation:
 */

/*
 * Read a 64-bit Java long in two-word "JVM" representation from "location"
 * and return the result in native representation.  There is no restriction
 * on the JVM respresentation, except that a value written using
 * CVMlong2Jvm() must return the same value when CVMjvm2Long() is called
 * (The operation is reversible.)
 */

CVMJavaLong CVMjvm2Long(const CVMUint32 location[2]);

/*
 * Read a 64-bit Java long in native representation from "val"
 * and write the result in two-word JVM representation to "location"
 */

void CVMlong2Jvm(CVMUint32 location[2], CVMJavaLong val);

/*
 * Convert 64-bit Java doubles between native representation and JVM
 * two-word representation:
 */

/*
 * Read a 64-bit Java double in two-word representation from "location"
 * and return the result in native representation.  There is no restriction
 * on the JVM respresentation, except that a value written using
 * CVMdouble2Jvm() must return the same value when CVMjvm2Dobule()
 * is called (the operation is reversible.)
 */

CVMJavaDouble CVMjvm2Double(const CVMUint32 location[2]);

/*
 * Read a 64-bit Java double in native representation from "val"
 * and write the result in two-word JVM representation to "location"
 */

void CVMdouble2Jvm(CVMUint32 location[2], CVMJavaDouble val);

/*
 * Copy two typeless 32-bit words from one location to another.
 * This is semantically equivalent to:
 * 
 * to[0] = from[0];
 * to[1] = from[1];
 *
 * but this interface is provided for those platforms that could
 * optimize this into a single 64-bit transfer.
 */

void CVMmemCopy64(CVMUint32 to[2], const CVMUint32 from[2]);

/*
 * Java long manipulation.
 *
 * The result argument is, once again, an lvalue.
 *
 * Conversions:
 */

/*
 * Convert long bits to double bits, according to Double.longBitsToDouble()
 */

extern CVMJavaDouble CVMlongBits2Double(CVMJavaLong val);

/*
 * Convert double bits to long bits, according to Double.doubleToRawLongBits()
 */

extern CVMJavaLong CVMdouble2LongBits(CVMJavaDouble val);

/*
 * Convert long to int, according to "l2i" bytecode semantics
 */
extern CVMJavaInt CVMlong2Int(CVMJavaLong val);

/*
 * Convert long to float, according to "l2f" bytecode semantics
 */
extern CVMJavaFloat CVMlong2Float(CVMJavaLong val);

/*
 * Convert long to double, according to "l2d" bytecode semantics
 */
extern CVMJavaDouble CVMlong2Double(CVMJavaLong val);

/*
 * Convert long to void *
 */
extern void * CVMlong2VoidPtr(CVMJavaLong val);

/*
 * Convert void * to long
 */
extern CVMJavaLong CVMvoidPtr2Long(void * val);

/*
 * Convert int to long, according to "i2l" bytecode semantics
 */
extern CVMJavaLong CVMint2Long(CVMJavaInt val);

/*
 * Convert double to long, according to "d2l" bytecode semantics
 */
extern CVMJavaLong CVMdouble2Long(CVMJavaDouble val);

/*
 * Arithmetic:
 *
 * The functions below follow the semantics of the
 * ladd, land, ldiv, lmul, lor, lxor, and lrem bytecodes,
 * respectively.
 */

CVMJavaLong CVMlongAdd(CVMJavaLong op1, CVMJavaLong op2);
CVMJavaLong CVMlongAnd(CVMJavaLong op1, CVMJavaLong op2);
CVMJavaLong CVMlongDiv(CVMJavaLong op1, CVMJavaLong op2);
CVMJavaLong CVMlongMul(CVMJavaLong op1, CVMJavaLong op2);
CVMJavaLong CVMlongOr (CVMJavaLong op1, CVMJavaLong op2);
CVMJavaLong CVMlongSub(CVMJavaLong op1, CVMJavaLong op2);
CVMJavaLong CVMlongXor(CVMJavaLong op1, CVMJavaLong op2);
CVMJavaLong CVMlongRem(CVMJavaLong op1, CVMJavaLong op2);

/*
 * Shift:
 *
 * The functions below follow the semantics of the
 * lushr, lshl, and lshr bytecodes, respectively.
 */

CVMJavaLong CVMlongUshr(CVMJavaLong op1, CVMJavaInt op2);
CVMJavaLong CVMlongShl (CVMJavaLong op1, CVMJavaInt op2);
CVMJavaLong CVMlongShr (CVMJavaLong op1, CVMJavaInt op2);

/*
 * Unary:
 *
 * Return the negation of "op" (-op), according to
 * the semantics of the lneg bytecode.
 */

CVMJavaLong CVMlongNeg(CVMJavaLong op);

/*
 * Return the complement of "op" (~op)
 */

CVMJavaLong CVMlongNot(CVMJavaLong op);

/*
 * Comparisons (returning an CVMBool: CVM_TRUE or CVM_FALSE)
 *
 * To 0:
 */

CVMInt32 CVMlongLtz(CVMJavaLong op);     /* op <= 0 */
CVMInt32 CVMlongGez(CVMJavaLong op);     /* op >= 0 */
CVMInt32 CVMlongEqz(CVMJavaLong op);     /* op == 0 */

/*
 * Between operands:
 */

CVMInt32 CVMlongEq(CVMJavaLong op1, CVMJavaLong op2);    /* op1 == op2 */
CVMInt32 CVMlongNe(CVMJavaLong op1, CVMJavaLong op2);    /* op1 != op2 */
CVMInt32 CVMlongGe(CVMJavaLong op1, CVMJavaLong op2);    /* op1 >= op2 */
CVMInt32 CVMlongLe(CVMJavaLong op1, CVMJavaLong op2);    /* op1 <= op2 */
CVMInt32 CVMlongLt(CVMJavaLong op1, CVMJavaLong op2);    /* op1 <  op2 */
CVMInt32 CVMlongGt(CVMJavaLong op1, CVMJavaLong op2);    /* op1 >  op2 */

/*
 * Comparisons (returning an CVMJavaInt value: 0, 1, or -1)
 *
 * Between operands:
 *
 * Compare "op1" and "op2" according to the semantics of the
 * "lcmp" bytecode.
 */

CVMInt32 CVMlongCompare(CVMJavaLong op1, CVMJavaLong op2);

/*
 * Constants (each a CVMJavaLong that can be passed to any one of the
 * operations above)
 *
 * The functions below follow the semantics of the
 * lconst_0 and lconst_1 bytecodes, respectively.
 */

CVMJavaLong CVMlongConstZero();
CVMJavaLong CVMlongConstOne();

/*
 ******************************************
 * Java double floating-point manipulation.
 ******************************************
 *
 * The result argument is, once again, an lvalue.
 *
 * Conversions:
 */

/*
 * Convert double to int, according to "d2i" bytecode semantics
 */

CVMJavaInt CVMdouble2Int(CVMJavaDouble val);

/*
 * Convert double to float, according to "d2f" bytecode semantics
 */

CVMJavaFloat CVMdouble2Float(CVMJavaDouble val);

/*
 * Convert int to double, according to "i2d" bytecode semantics
 */

CVMJavaDouble CVMint2Double(CVMJavaInt val);

/*
 * Arithmetic:
 *
 * The functions below follow the semantics of the
 * dadd, dsub, ddiv, dmul, and drem bytecodes, respectively.
 */

CVMJavaDouble CVMdoubleAdd(CVMJavaDouble op1, CVMJavaDouble op2);
CVMJavaDouble CVMdoubleSub(CVMJavaDouble op1, CVMJavaDouble op2);
CVMJavaDouble CVMdoubleDiv(CVMJavaDouble op1, CVMJavaDouble op2);
CVMJavaDouble CVMdoubleMul(CVMJavaDouble op1, CVMJavaDouble op2);
CVMJavaDouble CVMdoubleRem(CVMJavaDouble op1, CVMJavaDouble op2);

/*
 * Unary:
 *
 * Return the negation of "op" (-op), according to
 * the semantics of the dneg bytecode.
 */

CVMJavaDouble CVMdoubleNeg(CVMJavaDouble op);

/*
 * Comparisons (returning an CVMInt32 value: 0, 1, or -1)
 *
 * Between operands:
 *
 * Compare "op1" and "op2" according to the semantics of the
 * "dcmpl" (direction is -1) or "dcmpg" (direction is 1) bytecodes.
 */

CVMInt32 CVMdoubleCompare(CVMJavaDouble op1, CVMJavaDouble op2,
			  CVMInt32 direction);

/*
 * Constants (each a CVMJavaDouble that can be passed to any one of the
 * operations above)
 *
 * The functions below follow the semantics of the
 * dconst_0 and dconst_1 bytecodes, respectively.
 */

CVMJavaDouble CVMdoubleConstZero();
CVMJavaDouble CVMdoubleConstOne();

#include CVM_HDR_DOUBLEWORD_H

#endif /* _INCLUDED_PORTING_DOUBLEWORD_H */
