/*
 * @(#)float.h	1.12 00/05/19
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifndef _INCLUDED_PORTING_FLOAT_H
#define _INCLUDED_PORTING_FLOAT_H

#include "javavm/include/porting/defs.h"
#include "javavm/include/porting/vm-defs.h"

/*
 * Java floating-point float value manipulation.
 *
 * The result argument is, once again, an lvalue.
 *
 * Arithmetic:
 *
 * The functions below follow the semantics of the
 * fadd, fsub, fmul, fdiv, and frem bytecodes,
 * respectively.
 */

CVMJavaFloat CVMfloatAdd(CVMJavaFloat op1, CVMJavaFloat op2);
CVMJavaFloat CVMfloatSub(CVMJavaFloat op1, CVMJavaFloat op2);
CVMJavaFloat CVMfloatMul(CVMJavaFloat op1, CVMJavaFloat op2);
CVMJavaFloat CVMfloatDiv(CVMJavaFloat op1, CVMJavaFloat op2);
CVMJavaFloat CVMfloatRem(CVMJavaFloat op1, CVMJavaFloat op2);

/*
 * Unary:
 *
 * Return the negation of "op" (-op), according to
 * the semantics of the fneg bytecode.
 */

CVMJavaFloat CVMfloatNeg(CVMJavaFloat op);

/*
 * Comparisons (returning an int value: 0, 1, or -1)
 *
 * Between operands:
 *
 * Compare "op1" and "op2" according to the semantics of the
 * "fcmpl" (direction is -1) or "fcmpg" (direction is 1) bytecodes.
 */

CVMInt32 CVMfloatCompare(CVMJavaFloat op1, CVMJavaFloat op2,
			 CVMInt32 direction);

/*
 * Conversion:
 */

/* 
 * Convert float to int, according to "f2i" bytecode semantics
 */

CVMJavaInt CVMfloat2Int(CVMJavaFloat op);

/*
 * Convert float to double, according to "f2d" bytecode semantics
 */

CVMJavaDouble CVMfloat2Double(CVMJavaFloat op);

/*
 * Convert float to long, according to "f2l" bytecode semantics
 */
extern CVMJavaLong CVMfloat2Long(CVMJavaFloat op);

#include CVM_HDR_FLOAT_H

#endif /* _INCLUDED_PORTING_FLOAT_H */
