/*
 * @(#)linker.h	1.8 00/10/25
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifndef _INCLUDED_PORTING_LINKER_H
#define _INCLUDED_PORTING_LINKER_H

#include "javavm/include/porting/defs.h"
#include "javavm/include/porting/vm-defs.h"

/*
 * Dynamic linking support.
 *
 * These routines look very similar to the dynamic linking facilities
 * on Solaris and Win32. However, on VxWorks there is currently no
 * notion of a local symbol table (submitted as an RFE; TSR# 143425.)
 * Therefore symbol lookup is slow on VxWorks as we have to iterate
 * through the entire system symbol table every time. It is required
 * by the Java libraries that sit above these routines that a notion
 * of local symbol namespaces be maintained.
 *
 * On VxWorks these routines require three libraries to be built into
 * your project if they are not already: loadLib, moduleLib, and
 * symLib.
 */

/*
 * Dynamically link in a shared object. This takes a platform-
 * dependent, absolute pathname to the DSO. Calling code is
 * responsible for constructing this name. (We have yet to determine
 * how this will be done both for user libraries, which have the
 * properties java.library.path and java.path.separator available, and
 * system or "JVM helper" libraries like the JVMDI/JDWP backend, which
 * must be loaded before the VM is initialized.) Returns an abstract
 * "handle" to the shared object which must be used for later symbol
 * queries, or NULL if an error occurred. (sysLoadLibrary in the JDK)
 *
 * If absolutePathName is NULL, a handle to the "main program" is
 * returned.
 */
void *CVMdynlinkOpen(const void *absolutePathName);

/*
 * Find a function or data pointer in an already-opened shared object.
 * Takes the dynamic shared object handle (see CVMdlopen, above) and a
 * platform-dependent symbol as argument (typically, but not always, a
 * char *). On systems which support it this can be a global symbol
 * table lookup; otherwise an internal list of shared objects will
 * have to be searched for the symbol. (sysFindLibraryEntry in the
 * JDK). CVMdynlinkSym() is also responsible for adding any necessary
 * "decorations" to the symbol name before doing the lookup.
 */

void *CVMdynlinkSym(void *dsoHandle, const void *name);

/*
 * Close a dynamic shared object. This should probably have the
 * semantics that its symbols are unloaded, but VxWorks, at least,
 * doesn't currently appear to have a module unloading facility. Takes
 * the dynamic shared object handle from CVMdlopen(), above.
 * (sysUnloadLibrary in the JDK)
 */

void CVMdynlinkClose(void *dsoHandle);

/*
 * This is a Java-specific function which creates a string for the
 * dynamic lib open call by adding the appropriate pre and extensions
 * to a filename and the path. HOLDER is the destination buffer;
 * HOLDERLEN its size in bytes (will this be satisfactory for
 * WinCE?). PNAME is the directory prefix or NULL; FNAME is the "base
 * name" of the library. For example, CVMdynlinkBuildLibName(buf,
 * bufLen, "foo", "bar") might result in a library name of
 * "foo/libbar.so". All buffers are of a platform-specific data type
 * but are typically char *'s. Returns CVM_FALSE if the HOLDER buffer
 * is not big enough.
 */

CVMBool CVMdynlinkBuildLibName(void *holder, int holderlen,
			       void *pname, void *fname);

#include CVM_HDR_LINKER_H

#endif  /* #defined _INCLUDED_PORTING_LINKER_H */
