/*
 * -----------------------------------------------------------------
 * preloader.h : This file contains, references to romized classes.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)preloader.h	1.38 00/11/13
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*
 * This file defines the interface to the ROMized classes.
 */

#ifndef _INCLUDED_PRELOADER_H
#define _INCLUDED_PRELOADER_H

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/interpreter.h"

/*
 * Class lookup by name. Only used from debuggers.
 */
#ifdef CVM_DEBUG
extern const CVMClassBlock*
CVMpreloaderLookup(const char* className);
#endif

/*
 * Returns the cb of the romized class with the specified CVMClassTypeID.
 * This excludes VM-defined primitive types, such as "int" (though if a
 * user managed to define an "int" class, that would be fine).
 */
extern CVMClassBlock*
CVMpreloaderLookupFromType(CVMClassTypeID classType);

/*
 * Only for lookup of primitive VM-defined types, by type name.
 */
extern CVMClassBlock*
CVMpreloaderLookupPrimitiveClassFromType(CVMClassTypeID classType);

/*
 * Preloader initialization and destruction routines
 */
extern void
CVMpreloaderInit();

extern void
CVMpreloaderInitializeStringInterning(CVMExecEnv* ee);

extern void
CVMpreloaderDestroy();

/*
 * Avoid generating all stackmaps in advance of need. Instead, rewrite
 * code as required so that stackmap generation can be done lazily.
 */
extern CVMBool
CVMpreloaderDisambiguateAllMethods();

/*
 * Iterate over all preloaded classes,
 * and call 'callback' on each class.
 */
extern void
CVMpreloaderIterateAllClasses(CVMExecEnv* ee, 
			  CVMClassCallbackFunc callback,
			  void* data);

/*
 * Provide convenient handles to commonly used classes.
 */
#ifndef IN_ROMJAVA
#define CVM_CLASSBLOCK_DECL(claz) extern const CVMClassBlock claz##_Classblock

/*
 * Well known VM classes
 */
CVM_CLASSBLOCK_DECL(sun_misc_CVM);
CVM_CLASSBLOCK_DECL(sun_misc_Launcher);
CVM_CLASSBLOCK_DECL(sun_misc_SystemThread);
CVM_CLASSBLOCK_DECL(java_lang_Object);
CVM_CLASSBLOCK_DECL(java_lang_Class);
CVM_CLASSBLOCK_DECL(java_lang_ClassLoader);
CVM_CLASSBLOCK_DECL(java_lang_ClassLoader_NativeLibrary);
CVM_CLASSBLOCK_DECL(java_lang_Shutdown);
CVM_CLASSBLOCK_DECL(java_lang_String);
CVM_CLASSBLOCK_DECL(java_lang_Thread);
CVM_CLASSBLOCK_DECL(java_lang_ThreadGroup);
CVM_CLASSBLOCK_DECL(java_lang_Throwable);
CVM_CLASSBLOCK_DECL(java_lang_Exception);
CVM_CLASSBLOCK_DECL(java_lang_Error);
CVM_CLASSBLOCK_DECL(java_lang_ThreadDeath);
CVM_CLASSBLOCK_DECL(java_lang_Cloneable);
CVM_CLASSBLOCK_DECL(java_lang_System);
CVM_CLASSBLOCK_DECL(java_io_Serializable);
CVM_CLASSBLOCK_DECL(java_lang_reflect_Field);
CVM_CLASSBLOCK_DECL(java_lang_reflect_Method);
CVM_CLASSBLOCK_DECL(java_lang_reflect_Constructor);
CVM_CLASSBLOCK_DECL(java_lang_ref_SoftReference); 
CVM_CLASSBLOCK_DECL(java_lang_ref_WeakReference);
CVM_CLASSBLOCK_DECL(java_lang_ref_PhantomReference);
CVM_CLASSBLOCK_DECL(java_lang_ref_FinalReference);
CVM_CLASSBLOCK_DECL(java_lang_ref_Finalizer);
CVM_CLASSBLOCK_DECL(java_lang_ref_Reference);

/* *** RTSJ *** */
/* Added by Pratik. */

// Added to get finalization of realtime objects to work. 
// (deprecated as of 18th March 2001. I should remove this.)
CVM_CLASSBLOCK_DECL(javax_realtime_Finalizer);
// Added to get memory constraints on RealtimeThreads to work.
CVM_CLASSBLOCK_DECL(javax_realtime_RealtimeThread);  

/* *** JSTR *** */

/*
 * Exception classes
 */

#ifdef CVM_CLASSLOADING
CVM_CLASSBLOCK_DECL(java_lang_ClassCircularityError);
CVM_CLASSBLOCK_DECL(java_lang_ClassFormatError);
CVM_CLASSBLOCK_DECL(java_lang_IllegalAccessError);
CVM_CLASSBLOCK_DECL(java_lang_InstantiationError);
CVM_CLASSBLOCK_DECL(java_lang_LinkageError);
CVM_CLASSBLOCK_DECL(java_lang_UnsupportedClassVersionError);
CVM_CLASSBLOCK_DECL(java_lang_VerifyError);
#endif

#ifdef CVM_DYNAMIC_LINKING
CVM_CLASSBLOCK_DECL(java_lang_UnsatisfiedLinkError);
#endif

#ifdef CVM_REFLECT
CVM_CLASSBLOCK_DECL(java_lang_NegativeArraySizeException);
CVM_CLASSBLOCK_DECL(java_lang_NoSuchFieldException);
CVM_CLASSBLOCK_DECL(java_lang_NoSuchMethodException);
#endif

#if defined(CVM_CLASSLOADING) || defined(CVM_REFLECT)
CVM_CLASSBLOCK_DECL(java_lang_IncompatibleClassChangeError);
#endif

CVM_CLASSBLOCK_DECL(java_lang_AbstractMethodError);
CVM_CLASSBLOCK_DECL(java_lang_ArithmeticException);
CVM_CLASSBLOCK_DECL(java_lang_ArrayIndexOutOfBoundsException);
CVM_CLASSBLOCK_DECL(java_lang_ArrayStoreException);
CVM_CLASSBLOCK_DECL(java_lang_ClassCastException);
CVM_CLASSBLOCK_DECL(java_lang_ClassNotFoundException);
CVM_CLASSBLOCK_DECL(java_lang_CloneNotSupportedException);
CVM_CLASSBLOCK_DECL(java_lang_IllegalAccessException);
CVM_CLASSBLOCK_DECL(java_lang_IllegalArgumentException);
CVM_CLASSBLOCK_DECL(java_lang_IllegalMonitorStateException);
CVM_CLASSBLOCK_DECL(java_lang_InstantiationException);
CVM_CLASSBLOCK_DECL(java_lang_InternalError);
CVM_CLASSBLOCK_DECL(java_lang_InterruptedException);
CVM_CLASSBLOCK_DECL(java_lang_NoClassDefFoundError);
CVM_CLASSBLOCK_DECL(java_lang_NoSuchFieldError);
CVM_CLASSBLOCK_DECL(java_lang_NoSuchMethodError);
CVM_CLASSBLOCK_DECL(java_lang_NullPointerException);
CVM_CLASSBLOCK_DECL(java_lang_OutOfMemoryError);
CVM_CLASSBLOCK_DECL(java_lang_StackOverflowError);
CVM_CLASSBLOCK_DECL(java_lang_StringIndexOutOfBoundsException);
CVM_CLASSBLOCK_DECL(java_lang_UnsupportedOperationException);

CVM_CLASSBLOCK_DECL(java_io_InvalidClassException);
CVM_CLASSBLOCK_DECL(java_io_IOException);

CVM_CLASSBLOCK_DECL(java_lang_reflect_Method_ArgumentException);
CVM_CLASSBLOCK_DECL(java_lang_reflect_Method_AccessException);
CVM_CLASSBLOCK_DECL(java_lang_reflect_Constructor_ArgumentException);
CVM_CLASSBLOCK_DECL(java_lang_reflect_Constructor_AccessException);

/* *** RTSJ *** */ 
CVM_CLASSBLOCK_DECL(javax_realtime_IllegalAssignmentError);  // Added by Pratik to get interpreter to throw the right error.
CVM_CLASSBLOCK_DECL(javax_realtime_ResourceLimitError);
CVM_CLASSBLOCK_DECL(javax_realtime_InaccessibleAreaException);
CVM_CLASSBLOCK_DECL(javax_realtime_MemoryAccessError);
/* *** JSTR *** */

/*
 * Arrays of basic types
 */
CVM_CLASSBLOCK_DECL(manufacturedArrayOfBoolean);
CVM_CLASSBLOCK_DECL(manufacturedArrayOfChar);
CVM_CLASSBLOCK_DECL(manufacturedArrayOfFloat);
CVM_CLASSBLOCK_DECL(manufacturedArrayOfDouble);
CVM_CLASSBLOCK_DECL(manufacturedArrayOfByte);
CVM_CLASSBLOCK_DECL(manufacturedArrayOfShort);
CVM_CLASSBLOCK_DECL(manufacturedArrayOfInt);
CVM_CLASSBLOCK_DECL(manufacturedArrayOfLong);
CVM_CLASSBLOCK_DECL(manufacturedArrayOfObject);

/*
 * Primitive type classes.
 */
CVM_CLASSBLOCK_DECL(primitiveClass_boolean);
CVM_CLASSBLOCK_DECL(primitiveClass_char);
CVM_CLASSBLOCK_DECL(primitiveClass_float);
CVM_CLASSBLOCK_DECL(primitiveClass_double);
CVM_CLASSBLOCK_DECL(primitiveClass_byte);
CVM_CLASSBLOCK_DECL(primitiveClass_short);
CVM_CLASSBLOCK_DECL(primitiveClass_int);
CVM_CLASSBLOCK_DECL(primitiveClass_long);
CVM_CLASSBLOCK_DECL(primitiveClass_void);

/*
 * Wrapper object classes.
 */
CVM_CLASSBLOCK_DECL(java_lang_Boolean);
CVM_CLASSBLOCK_DECL(java_lang_Character);
CVM_CLASSBLOCK_DECL(java_lang_Float);
CVM_CLASSBLOCK_DECL(java_lang_Double);
CVM_CLASSBLOCK_DECL(java_lang_Byte);
CVM_CLASSBLOCK_DECL(java_lang_Short);
CVM_CLASSBLOCK_DECL(java_lang_Integer);
CVM_CLASSBLOCK_DECL(java_lang_Long);

#undef CVM_CLASSBLOCK_DECL

extern CVMInt32 CVM_staticData[];

#define CVMpreloaderGetRefStatics()      (&CVM_staticData[1])
#define CVMpreloaderNumberOfRefStatics() (CVM_staticData[0])

/* The address of the CVMClassBlock for java_lang_Object, for instance,
 * is CVMsystemClass(java_lang_Object)
 * It is an R-value only, of course, since the underlying struct is const.
 */
#define CVMsystemClass(cname) ((CVMClassBlock*)(&cname##_Classblock))

/* The address of the CVMClassBlock for an array of basic type 'basictypename.
 * For instance CVMsystemArrayClassOf(Int).
 *
 * It is an R-value only, of course, since the underlying struct is const.
 */
#define CVMsystemArrayClassOf(basictypename) \
    ((CVMClassBlock*)&manufacturedArrayOf##basictypename##_Classblock)

/* The address of the CVMClassBlock for a primitive type.
 * For instance CVMprimitiveClass(void).
 *
 * It is an R-value only, of course, since the underlying struct is const.
 */
#define CVMprimitiveClass(basictypename) \
    ((CVMClassBlock*)&primitiveClass_##basictypename##_Classblock)

#endif

#endif /* _INCLUDED_PRELOADER_H */
