/*
 * @(#)verify.h	1.6 00/08/07
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/export/jni.h"
#include "javavm/export/jvm.h"


/*
 * Support for a VM-independent class format checker.
 */ 
typedef struct {
    CVMUint32 code;    /* byte code */
    CVMUint32 excs;    /* exceptions */
    CVMUint32 etab;    /* catch table */
    CVMUint32 lnum;    /* line number */
    CVMUint32 lvar;    /* local vars */
} CVMmethod_size_info;

typedef struct {
    CVMUint32 constants;       /* constant pool */
    CVMUint32 reducedConstants;/* constant pool without trailing utf8s */
    CVMUint32 fields;
    CVMUint32 methods;
    CVMUint32 javamethods;     /* # of methods implemented in java */
    CVMUint32 interfaces;
    CVMUint32 staticFields;    /* number of words of static fields */
    CVMUint32 staticRefFields; /* number of words of static ref fields */
    CVMUint32 innerclasses;    /* # of records in InnerClasses attr */

    method_size_info clinit;   /* memory used in clinit */
    method_size_info main;     /* used everywhere else */
} CVMclass_size_info;

/* 
 * Byte code verifier.
 *
 * Returns JNI_FALSE if verification fails. A detailed error message
 * will be places in msg_buf, whose length is specified by buf_len.
 */
jboolean 
VerifyClass(CVMExecEnv *ee, 
	    CVMClassBlock *cb,
	    char * msg_buf, 
	    jint buf_len);

/* 
 * This performs class format checks and fills in size 
 * information for the class file and returns:
 *  
 *   0: good
 *  -1: out of memory
 *  -2: bad format
 *  -3: unsupported version
 *  -4: bad class name
 */
jint
CVMverifyClassFormat(const char *class_name,
		     const unsigned char *data,
		     unsigned int data_size,
		     class_size_info *size,
		     char *message_buffer,
		     jint buffer_length,
		     jboolean measure_only,
		     jboolean check_relaxed); 
