/*
 * @(#)DependenceArc.java	1.3 00/05/19
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

package dependenceAnalyzer;

/*
 * A DependenceArc connects two DependenceNode structures.
 * The "from" node has a dependence on the "to" node.
 * The "type" of dependence is not defined by us here, but depends
 * on the type of analysis being done. See the DependenceAnalyzer
 * classes. An alternative design would be to use subclassing of
 * DependenceArc to indicate dependence type. Let me know.
 *
 * Generally, we will know either the "from" or the "to" from 
 * context (as we're traversing a dependsOn or dependedOn enumeration)
 * and will only need to look at the other.
 */
public class DependenceArc {
    public DependenceNode from(){ return fromNode; }
    public DependenceNode to()  { return toNode; }
    public int		  type(){ return arcType; }

    public DependenceArc( DependenceNode isFrom, DependenceNode isTo, int dependenceType ){
	fromNode = isFrom;
	toNode   = isTo;
	arcType  = dependenceType;
    }

    public boolean equals( Object o ){
	try {
	    DependenceArc other = (DependenceArc)o;
	    return ( (this.fromNode==other.fromNode) && (this.toNode==other.toNode) && (this.arcType==other.arcType) );
	} catch (Throwable t ){
	    return false;
	}
    }

    // state is all private
    // object is immutable.
    private DependenceNode fromNode;
    private DependenceNode toNode;
    private int		   arcType;
}
