/*
 * @(#)MemberDependenceNode.java	1.7 00/11/09
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

package dependenceAnalyzer;
import util.Set;
import java.util.*;
import util.EmptyEnumeration;

public class MemberDependenceNode extends DependenceNode{

    /*
     * Flag values.
     */
    public static final int EXCLUDED		= 1;
    public static final int REQUIRED		= 2;

    public static final int STATIC		= 4;
    public static final int PRIVATE		= 8;
    public static final int METHOD		=16;
    public static final int FIELD		=32;
    public static final int INIT		=64;
    public static final int CLINIT		=128;
    public static final int OVERRIDDEN		=256;
    public static final int IMPUTED		=512;
    public static final int NATIVE		=1024;

    // a mask
    public static final int NO_OVERRIDING	= STATIC|PRIVATE|FIELD|INIT|CLINIT;

    /*
     * Constructor.
     */
    public MemberDependenceNode( Object nm, int flagval ){
	super( nm );
	flags = flagval;
    }

    /*
     * Public methods.
     */
    public
    Enumeration overrides(){ 
	if ( memberOverrides == null ) return EmptyEnumeration.instance;
	return memberOverrides.elements();
    }

    public
    Enumeration overriddenBy(){ 
	if ( memberOverriddenBy == null ) return EmptyEnumeration.instance;
	return memberOverriddenBy.elements();
    }

    /*
     * These are all package-visible parts of the implementation.
     */
    Set		memberOverriddenBy;
    Set		memberOverrides;

    public void dump( java.io.PrintStream o ){
	for ( int j = 0; j < 16; j++ ){ 
	    switch ( flags & ( 1<<j ) ){
	    case EXCLUDED: o.print("Excluded "); break;
	    case REQUIRED: o.print("Required "); break;
	    case STATIC:   o.print("static "); break;
	    case PRIVATE:  o.print("private "); break;
	    case METHOD:   o.print("method "); break;
	    case FIELD:    o.print("field "); break;
	    case INIT:	   o.print("constructor "); break;
	    case CLINIT:   o.print("static initializer "); break;
	    case OVERRIDDEN:o.print("overridden "); break;
	    case IMPUTED:  o.print("imputed "); break;
	    case NATIVE:   o.print("native "); break;
	    }
	}
	o.println( nodeName.toString() );
	if ( memberOverriddenBy != null ){
	    o.println("	is overridden "+memberOverriddenBy.size()+" times");
	}
    }

}
