/*
 * @(#)JDKHeader.java	1.8 00/05/19
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

package runtime;
import Const;
import jcc.Util;
import util.*;
import components.*;
import vm.*;
import java.io.PrintStream;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;

public class JDKHeader extends HeaderDump{
    boolean		doPackPragma;
    // users might want to change this, though it is very unlikely.
    public String	packPragma = "#pragma pack(4)";
    public String	unpackPragma = "#pragma pack()";

    public JDKHeader( boolean dopack ){
	super( '_' );
	doPackPragma = dopack;
    }

    // this data is really per-header.
    // it is guarded by synchronous method access
    Hashtable	classesNamed;

    void generateForward( String classname ){
	String target = strsub( classname, CDelim );
	if ( classesNamed.containsKey( target ) ) return; // already under control
	o.println("struct H"+ target +";" );
	classesNamed.put( target, target );
    }

    class ExternGenerator extends util.SignatureIterator {
	ExternGenerator( String s ){ super( s ); }

	public void do_array( int depth, int start, int end ){
	    if ( sig.charAt(start) == Const.SIGC_CLASS ){
		generateForward( className( start, end ) );
	    }
	}

	public void do_object( int start, int end ){
	    generateForward( className( start, end ) );
	}
    }

    private void
    generateForwardRef( String sig ) throws DataFormatException {
	// primitive types generate no forward references,
	// so don't ever bother with them.
	if ( !ExternGenerator.isPrimitiveType( sig ) ){
	    (new ExternGenerator( sig )).iterate();
	}
	
    }

    class ParamTypeGenerator extends util.SignatureIterator {
	String	returnType;
	Vector  paramTypes = new Vector();

	ParamTypeGenerator( String s ){ super( s ); }

	private void put( String typestring ){
	    if ( isReturnType ) returnType = typestring;
	    else paramTypes.addElement( typestring );
	}

	public void do_boolean(){ put("/*boolean*/ long "); }
	public void do_byte(){    put( "/*byte*/ char "); }
	public void do_short(){   put( "short " ); }
	public void do_char(){    put( "unicode " ); }
	public void do_int(){     put( "long " ); }
	public void do_long(){    put( "int64_t " ); }
	public void do_float(){   put( "float " ); }
	public void do_double(){  put( "double " ); }
	public void do_void(){    put( "void " ); }

	public void do_array( int depth, int start, int end ){
	    if ( depth > 1 ){
		put("struct HArrayOfArray *");
	    }else switch (sig.charAt(start)){
		case Const.SIGC_BYTE:	put( "struct HArrayOfByte *" );
					break;
		case Const.SIGC_CHAR:	put( "struct HArrayOfChar *" );
					break;
		case Const.SIGC_SHORT:	put( "struct HArrayOfShort *" );
					break;
		case Const.SIGC_BOOLEAN:
		case Const.SIGC_INT:	put( "struct HArrayOfInt *" );
					break;
		case Const.SIGC_LONG:	put( "struct HArrayOfLong *" );
					break;
		case Const.SIGC_FLOAT:	put( "struct HArrayOfFloat *" );
					break;
		case Const.SIGC_DOUBLE:	put( "struct HArrayOfDouble *" );
					break;
		case Const.SIGC_ARRAY:	put( "struct HArrayOfArray *" );
					break;
		case Const.SIGC_CLASS:	if ( className(start,end).equals("java/lang/String") )
					    put( "struct HArrayOfString *" );
					else
					    put( "struct HArrayOfObject *" );
					break;
		default:		put( "void *" );
					break;
	    }
	}

	public void do_object( int start, int end ){
	    put( "struct H"+Util.convertToClassName(className(start,end)) +" *");
	}

    }



    boolean generateForwardRefs( ClassMemberInfo fields[] ){
	try {
	    for (int i=0; i < fields.length; i++ ){
		ClassMemberInfo cmi = fields[i];
		if ( !cmi.isStaticMember()){
		    generateForwardRef( cmi.type.string );
		}
	    }
	} catch ( Exception e ){
	    e.printStackTrace();
	    return false;
	}
	return true;
    }

    private boolean
    generateNatives( MethodInfo methods[], boolean cplusplusguard ){
	if ( methods == null ) return true;
	boolean anyMethods = false;
	String namePrefix = " "+className+CDelim;
	for ( int i =0; i < methods.length; i++ ){
	    MethodInfo m = methods[i];
	    if ( (m.access&Const.ACC_NATIVE) == 0 ) continue;
	    if ( cplusplusguard && ! anyMethods ){
		o.println("#ifdef __cplusplus\nextern \"C\"{\n#endif");
		anyMethods = true;
	    }
	    String methodsig = m.type.string;
	    ParamTypeGenerator ptg;
	    try {
		generateForwardRef( methodsig );
		ptg = new ParamTypeGenerator( methodsig );
		ptg.iterate();
	    } catch ( Exception e ){
		e.printStackTrace();
		return false;
	    }
	    o.print("extern ");
	    o.print( ptg.returnType);
	    o.print( namePrefix );
	    o.print( m.name.string );
	    o.print("(");
	    Enumeration tlist = ptg.paramTypes.elements();
	    // implicit first parameter of "this"
	    o.print("struct H"+className+" *");
	    while (tlist.hasMoreElements()){
		o.print(", ");
		o.print( tlist.nextElement() );
	    }
	    o.println(");");
	}
	if ( cplusplusguard && anyMethods ){
	    o.println("#ifdef __cplusplus\n}\n#endif");
	}
	return true;
    }

    String CFieldType( String sig ){
	switch ( sig.charAt(0) ){
	case Const.SIGC_ARRAY:
	    switch ( sig.charAt(1) ){
	    case Const.SIGC_BYTE:	return "struct HArrayOfByte *";
	    case Const.SIGC_CHAR:	return "struct HArrayOfChar *";
	    case Const.SIGC_BOOLEAN:	return "struct HArrayOfInt *";
	    case Const.SIGC_SHORT:	return "struct HArrayOfShort *";
	    case Const.SIGC_INT:	return "struct HArrayOfInt *";
	    case Const.SIGC_LONG:	return "struct HArrayOfLong *";
	    case Const.SIGC_FLOAT:	return "struct HArrayOfFloat *";
	    case Const.SIGC_DOUBLE:	return "struct HArrayOfDouble *";
	    case Const.SIGC_ARRAY:	return "struct HArrayOfArray *";
	    case Const.SIGC_CLASS:	return "struct HArrayOfObject *";
	    default:		return "void *";
	    }
	case Const.SIGC_BOOLEAN:return "/*boolean*/ int32_t ";
	case Const.SIGC_BYTE:	return "/*byte*/ int32_t ";
	case Const.SIGC_SHORT:	return "/*short*/ int32_t ";
	case Const.SIGC_CHAR:	return "/*char*/ int32_t ";
	case Const.SIGC_INT:	return "int32_t ";
	case Const.SIGC_LONG:	return "int64_t ";
	case Const.SIGC_FLOAT:	return "float ";
	case Const.SIGC_DOUBLE:	return "double ";
	case Const.SIGC_CLASS:	return "struct H"+Util.convertToClassName(sig.substring(1,sig.length()-1)) +" *";
	default:		return "void *";
	}
    }

    int generateInstanceField( FieldInfo finfo, int curoff ){
	int thisOff = finfo.instanceOffset;
	if ( thisOff != curoff ){
	    // generate padding. This will never happen.
	    // negative padding will, of course, cause compiler error.
	    o.println("	char	pad"+thisOff+"["+(4*(thisOff-curoff))+"];");
	}
	o.println("\t"+CFieldType( finfo.type.string)+finfo.name.string+";");
	return thisOff+ finfo.nSlots;
    }

    void generateStruct( FieldConstant refFieldtable[] ){
	if ( doPackPragma )
	    o.println( packPragma );
	o.println("typedef struct Class"+className+" {" );
	if ( (refFieldtable == null)||(refFieldtable.length==0) ){
	    o.println("	char	PAD; /* ANSI C requires at least one field */");
	}else{
	    int n = refFieldtable.length;
	    int curoff = 0;
	    for ( int i = 0; i < n; i++ ){
		curoff = generateInstanceField( refFieldtable[i].find(), curoff );
	    }
	}
	o.println("} Class"+className+";" );
	o.println("HandleTo("+className+");");
	if ( doPackPragma )
	    o.println( unpackPragma );
    }

    private void prolog(){
	o.println("/* DO NOT EDIT THIS FILE - it is machine generated */\n#include <native.h>");
	o.println("/* Header for class "+className+" */\n");
	o.println("#ifndef _Included_"+className);
	o.println("#define _Included_"+className);
    }
    private void epilog(){
	o.println("#endif");
    }

    synchronized public boolean
    dumpHeader( ClassInfo c, PrintStream outfile ){
	o = outfile;
	className = Util.convertToClassName( c.className );
	if ( classesNamed == null )
	    classesNamed = new Hashtable();
	else
	    classesNamed.clear();
	prolog();
	generateForwardRefs( c.fields );
	if ( !c.className.equals("java/lang/Object" ) && !c.className.equals("java/lang/Class")){
	    generateStruct( c.refFieldtable );
	}
	generateConsts( className, c.fields );
	generateNatives( c.methods, true );
	epilog();
	return true;
    }

    synchronized public boolean
    dumpExternals( ClassInfo c, PrintStream outfile ){
	o = outfile;
	className = Util.convertToClassName( c.className );
	generateNatives( c.methods, false );
	return true;
    }

}
