/*
 * @(#)CVMClassFactory.java	1.11 00/07/17
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

package vm;
import Const;
import components.ClassInfo;
import components.ConstantPool;
import runtime.CVMWriter;
import util.Localizer;
import java.util.Comparator;
import java.util.Hashtable;

public class CVMClassFactory implements VMClassFactory, Comparator {
    
    static String primitiveTypeName[] = {
	"void", "boolean", "char", "byte",
	"short", "int", "long", "float",
	"double"
    };

    /*
     * STATIC INITIALIZER:
     * MAKE SURE THE PRIMITIVE TYPES GET SMALL TYPEID NUMBERS
     * DO THIS BEFORE CONSTRUCTION OF ANY CVMClass!
     */
    static{
	for ( int i = 0 ; i < primitiveTypeName.length; i++ ){
	    CVMDataType.lookupClassname( primitiveTypeName[i] );
	}
    }
    
    public ClassClass newVMClass( ClassInfo c ){
	return new CVMClass( c );
    }

    private static boolean
    setType( String name, int value ){
	ClassInfo ci = ClassInfo.lookupClass( name );
	ClassClass cc;
	if ( (ci == null) || ( ( cc = ci.vmClass ) == null) ){
	    return false;
	}
	((CVMClass)cc).typeCode = value;
	return true;


    }

    private static void
    setAllClassID(){
	// null had better work here!
	// we'll find out soon enough.
	ClassClass allClasses[] = ClassClass.getClassVector( null );
	int n = allClasses.length;
	for ( int i = 0; i < n; i++ ){
	    CVMClass cvmc = (CVMClass) allClasses[i];
	    cvmc.classId = CVMDataType.lookupClassname(CVMWriter.getUTF(cvmc.ci.thisClass.name));
	}
    }

    public void
    setTypes(){
	setType( "void", Const.T_VOID );
	setType( "boolean", Const.T_BOOLEAN );
	setType( "char", Const.T_CHAR );
	setType( "byte", Const.T_BYTE );
	setType( "short", Const.T_SHORT );
	setType( "int", Const.T_INT );
	setType( "long", Const.T_LONG );
	setType( "float", Const.T_FLOAT );
	setType( "double", Const.T_DOUBLE );
	setAllClassID();
    }

    public ConstantPool 
    makeResolvable(ConstantPool cp, Hashtable missingClasses ) {
	//
	// For the CVM, this is actually pretty trivial stuff...
        //
        System.err.println(Localizer.getString("classclass.warning_it_has_an_impure_shared_constant_pool"));
	CVMClass.makeResolvable(cp.getConstants(), missingClasses);
	return cp;
    }

    /*
     * Compare two CVMClass objects by classid
     */
    public int
    compare( Object o1, Object o2 ){
	// just do the casts.
	// unacceptable comparison results in
	// classcast exception.
	CVMClass c1 = (CVMClass)o1;
	CVMClass c2 = (CVMClass)o2;
        int classId1 = c1.classId;
        int classId2 = c2.classId;

        // NOTE: This comparison method is used in the sorting of CVM Class
        // instances in CVM_ROMClasses[].  The Class instances of the Class
        // and Array types are sorted in incremental order of typeid values.
        // The only exception is that deep array types (i.e. array types with
        // depth >= 7) will be sorted based only on the basetype field of
        // their typeids (i.e. the depth field is ignored ... hence the
        // masking operations below).  This makes it possible to access Class
        // instances for Class and Deep array types in CVM_ROMClasses[] by
        // indexing.
        if ((classId1 & CVMTypeCode.CVMtypeArrayMask) == CVMTypeCode.CVMtypeBigArray) {
            classId1 &= CVMTypeCode.CVMtypeBasetypeMask;
        }
        if ((classId2 & CVMTypeCode.CVMtypeArrayMask) == CVMTypeCode.CVMtypeBigArray) {
            classId2 &= CVMTypeCode.CVMtypeBasetypeMask;
        }
        return classId1-classId2; // should turn into -1, 0, 1. oops.
    }
}
