/*
 * @(#)gen_markcompact.c	1.19 00/11/11
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*
 * This file includes the implementation of a copying semispace generation.
 * This generation can act as a young or an old generation.
 */

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/directmem.h"

/*
 * This file is generated from the GC choice given at build time.
 */
#include "generated/javavm/include/gc_config.h"

#include "javavm/include/gc_common.h"
#include "javavm/include/gc/gc_impl.h"

#include "javavm/include/gc/generational/generational.h"

#include "javavm/include/gc/generational/gen_markcompact.h"

#include "javavm/include/porting/system.h"

/*
 * Range checks
 */
static CVMBool
CVMgenMarkCompactInSpace(CVMGenSpace* space, CVMObject* ref)
{
    CVMUint32* heapPtr = (CVMUint32*)ref;
    return ((heapPtr >= space->allocBase) &&
	    (heapPtr <  space->allocTop));
}

static CVMBool 
CVMgenMarkCompactInGeneration(CVMGeneration* gen, CVMObject* ref)
{
    CVMGenMarkCompactGeneration* thisGen = (CVMGenMarkCompactGeneration*)gen;
    return CVMgenMarkCompactInSpace(thisGen->theSpace, ref);
}

static CVMBool 
CVMgenMarkCompactInThisGen(CVMGenMarkCompactGeneration* thisGen, 
			   CVMObject* ref)
{
    return CVMgenMarkCompactInSpace(thisGen->theSpace, ref);
}

/*
 * Allocate a MarkCompact for the current semispaces generation
 */
static CVMGenSpace*
CVMgenMarkCompactAllocSpace(CVMUint32* heapSpace, CVMUint32 numBytes)
{
    CVMUint32 allocSize = sizeof(CVMGenSpace);
    CVMGenSpace* space = (CVMGenSpace*)calloc(1, allocSize);
    CVMUint32 numWords = numBytes / 4;
    if (space != NULL) {
	space->allocTop  = &heapSpace[numWords];
	space->allocPtr  = &heapSpace[0];
	space->allocBase = &heapSpace[0];
    }
    return space;
}

/*
 * Make sure the allocation pointers of the mark-compact generation
 * are reflected at the "superclass" level for the shared GC code.
 */
static void
CVMgenMarkCompactMakeCurrent(CVMGenMarkCompactGeneration* thisGen,
			     CVMGenSpace* thisSpace)
{
    CVMGeneration* super = &thisGen->gen;
    super->allocPtr  = thisSpace->allocPtr;
    super->allocBase = thisSpace->allocBase;
    super->allocTop  = thisSpace->allocTop;
    super->allocMark = thisSpace->allocPtr;
}

/* Forward declaration */
static CVMBool
CVMgenMarkCompactCollect(CVMGeneration* gen,
			 CVMExecEnv*    ee, 
			 CVMUint32      numBytes, /* collection goal */
			 CVMGCOptions*  gcOpts);

/*
 * Scan objects in contiguous range, without doing any special handling.
 */
static void
scanObjectsInRangePartial(CVMGenMarkCompactGeneration* thisGen,
			  CVMExecEnv* ee, CVMGCOptions* gcOpts,
			  CVMUint32* base, CVMUint32* top,
			  CVMRefCallbackFunc callback, void* callbackData)
{
    CVMUint32* curr = base;
    CVMtraceGcCollect(("GC[MC,%s]: "
		       "Scanning object range (partial) [%x,%x)\n",
		       gcOpts->fullGC ? "full" : "partial",
		       base, top));
    while (curr < top) {
	CVMObject* currObj    = (CVMObject*)curr;
	CVMClassBlock* currCb = CVMobjectGetClass(currObj);
	CVMUint32  objSize    = CVMobjectSizeGivenClass(currObj, currCb);
	CVMobjectWalkRefs(ee, gcOpts, currObj, currCb, {
	    if (*refPtr != 0) {
		(*callback)(refPtr, callbackData);
	    }
	});
	/* iterate */
	curr += objSize / 4;
    }
    CVMassert(curr == top); /* This had better be exact */
}

static void
CVMgenMarkCompactUpdateTableForObjectSlot(CVMObject** refPtr,
					void* data)
{
    CVMGenMarkCompactGeneration* thisGen = (CVMGenMarkCompactGeneration*)data;
    CVMObject* ref = *refPtr;
    if ((ref != NULL) &&
	!CVMgenMarkCompactInThisGen(thisGen, ref) &&
	!CVMobjectIsInROM(ref)) {
	*CARD_TABLE_SLOT_ADDRESS_FOR(refPtr) = CARD_DIRTY_BYTE;
    }
}

/*
 * Scan objects in contiguous range, and do all special handling as well.
 */
static void
scanObjectsInRange(CVMExecEnv* ee, CVMGCOptions* gcOpts,
		   CVMUint32* base, CVMUint32* top,
		   CVMRefCallbackFunc callback, void* callbackData)
{
    CVMUint32* curr = base;
    CVMtraceGcCollect(("GC[MC,%s]: "
		       "Scanning object range (full) [%x,%x)\n",
		       gcOpts->fullGC ? "full" : "partial",
		       base, top));
    while (curr < top) {
	CVMObject* currObj = (CVMObject*)curr;
	CVMClassBlock* currCb = CVMobjectGetClass(currObj);
	CVMUint32  objSize    = CVMobjectSizeGivenClass(currObj, currCb);
	CVMobjectWalkRefsWithSpecialHandling(ee, gcOpts, currObj, currCb, {
	    if (*refPtr != 0) {
		(*callback)(refPtr, callbackData);
	    }
	}, callback, callbackData);
	/* iterate */
	curr += objSize / 4;
    }
    CVMassert(curr == top); /* This had better be exact */
}

/*
 * Scan objects in contiguous range, and do all special handling as well.
 * Ignore unmarked objects.
 */
static void
scanObjectsInRangeSkipUnmarked(CVMGenMarkCompactGeneration* thisGen,
			       CVMExecEnv* ee, CVMGCOptions* gcOpts,
			       CVMUint32* base, CVMUint32* top,
			       CVMRefCallbackFunc callback, void* callbackData)
{
    CVMUint32* curr = base;
    CVMtraceGcCollect(("GC[MC,%s]: "
		       "Scanning object range (skip unmarked) [%x,%x)\n",
		       gcOpts->fullGC ? "full" : "partial",
		       base, top));
    while (curr < top) {
	CVMObject* currObj    = (CVMObject*)curr;
	CVMUint32  classWord  = CVMobjectGetClassWord(currObj);
	CVMClassBlock* currCb = CVMobjectGetClassFromClassWord(classWord);
	CVMUint32  objSize    = CVMobjectSizeGivenClass(currObj, currCb);
	if (CVMobjectMarkedOnClassWord(classWord)) {
	    CVMobjectWalkRefsWithSpecialHandling(ee, gcOpts, currObj,
						 classWord, {
		if (*refPtr != 0) {
		    (*callback)(refPtr, callbackData);
		}
	    }, callback, callbackData);
	}
	/* iterate */
	curr += objSize / 4;
    }
    CVMassert(curr == top); /* This had better be exact */
}

/*
 * Re-build the old-to-young pointer tables after an old generation is
 * done. 
 */
static void
CVMgenMarkCompactRebuildBarrierTable(CVMGenMarkCompactGeneration* thisGen,
				   CVMExecEnv* ee,
				   CVMGCOptions* gcOpts,
				   CVMUint32* startRange,
				   CVMUint32* endRange)
{
    CVMtraceGcCollect(("GC[MC,%d,%s]: "
		       "Rebuilding barrier table for range [%x,%x)\n",
		       thisGen->gen.generationNo,
		       gcOpts->fullGC ? "full" : "partial",
		       startRange, endRange));
    /*
     * No need to look at classes here. They are scanned separately in
     * CVMgenScanAllRoots(), so no need to make a record of them in the
     * barrier records.
     */
    scanObjectsInRangePartial(thisGen, ee, gcOpts, startRange, endRange,
			      CVMgenMarkCompactUpdateTableForObjectSlot,
			      thisGen);
    CVMgenBarrierObjectHeadersUpdate((CVMGeneration*)thisGen,
				     ee, gcOpts, startRange, endRange);
}

static void       
CVMgenMarkCompactScanYoungerToOlderPointers(CVMGeneration* gen,
					  CVMExecEnv* ee,
					  CVMGCOptions* gcOpts,
					  CVMRefCallbackFunc callback,
					  void* callbackData)
{
    CVMassert(CVM_FALSE); /* this generation can only be old */
}

static void       
CVMgenMarkCompactScanOlderToYoungerPointers(CVMGeneration* gen,
					  CVMExecEnv* ee,
					  CVMGCOptions* gcOpts,
					  CVMRefCallbackFunc callback,
					  void* callbackData)
{
    CVMtraceGcCollect(("GC[MC,%d,%s]: "
		       "Scanning older to younger ptrs\n",
		       gen->generationNo,
		       gcOpts->fullGC ? "full" : "partial"));
    /*
     * To get all older to younger pointers, traverse the records in
     * the write barrier tables.
     */
    CVMgenBarrierPointersTraverse(gen, ee, gcOpts, callback, callbackData);
}


/*
 * Start and end GC for this generation. 
 */
static void
CVMgenMarkCompactStartGC(CVMGeneration* gen,
		       CVMExecEnv* ee, CVMGCOptions* gcOpts)
{
    CVMtraceGcCollect(("GC[MC,%d,%s]: Starting GC\n",
		       gen->generationNo,
		       gcOpts->fullGC ? "full" : "partial"));
}

static void
CVMgenMarkCompactEndGC(CVMGeneration* gen,
		     CVMExecEnv* ee, CVMGCOptions* gcOpts)
{
    CVMtraceGcCollect(("GC[MC,%d,%s]: Ending GC\n",
		       gen->generationNo,
		       gcOpts->fullGC ? "full" : "partial"));
}

/*
 * Return remaining amount of memory
 */
static CVMUint32
CVMgenMarkcompactFreeMemory(CVMGeneration* gen, CVMExecEnv* ee)
{
    return (CVMUint32)((CVMUint8*)gen->allocTop - (CVMUint8*)gen->allocPtr);
}

/*
 * Return total amount of memory
 */
static CVMUint32
CVMgenMarkcompactTotalMemory(CVMGeneration* gen, CVMExecEnv* ee)
{
    return (CVMUint32)((CVMUint8*)gen->allocTop - (CVMUint8*)gen->allocBase);
}

/* Iterate over promoted pointers in a generation. A young space
   collection accummulates promoted objects in an older
   generation. These objects need to be scanned for pointers into
   the young generation.
   
   At the start of a GC, the current allocation pointer of each
   generation is marked. Therefore the promoted objects to scan are those
   starting from 'allocMark' to the current 'allocPtr'.
  */
static void
CVMgenMarkCompactScanPromotedPointers(CVMGeneration* gen,
				      CVMExecEnv* ee,
				      CVMGCOptions* gcOpts,
				      CVMRefCallbackFunc callback,
				      void* callbackData)
{
    CVMUint32* startRange = gen->allocMark;
    CVMUint32* endRange   = gen->allocPtr;

    CVMtraceGcCollect(("GC[MC,%s]: "
		       "Scanning promoted ptrs [%x,%x)\n",
		       gcOpts->fullGC ? "full" : "partial",
		       startRange, endRange));
    /* loop, until promoted object scans result in no more promotions */
    do {
	scanObjectsInRange(ee, gcOpts, startRange, endRange,
			   callback, callbackData);
	startRange = endRange;
	/* gen->allocPtr may have been incremented */
	endRange = gen->allocPtr; 
    } while (startRange < endRange);

    CVMassert(endRange == gen->allocPtr);
    /*
     * Remember these pointers for further young space collections.
     * Don't forget to clear class marks again, since we need to be
     * able to traverse already scanned classes again, in order to
     * build barrier tables.
     */
    CVMgenMarkCompactRebuildBarrierTable((CVMGenMarkCompactGeneration*)gen,
					 ee, gcOpts,
					 gen->allocMark, gen->allocPtr);
    gen->allocMark = gen->allocPtr; /* Advance mark for the next batch
				       of promotions */
    /*
     * Make sure that gen->allocPtr and gen->fromSpace->allocPtr are
     * in agreement (i.e. in the same space).
     */
    CVMassert(CVMgenMarkCompactInThisGen((CVMGenMarkCompactGeneration*)gen,
					 (CVMObject*)gen->allocPtr) ||
	      (gen->allocPtr == gen->allocTop));
    /*
     * This is the observable fromSpace pointer. We can probably
     * get away with not committing this pointer.
     */
    ((CVMGenMarkCompactGeneration*)gen)->theSpace->allocPtr = gen->allocPtr;
}

/*
 * Copy numBytes from 'source' to 'dest'. Assume word copy, and assume
 * that the copy regions are disjoint.  
 */
static void
CVMgenMarkCompactCopyDisjointWords(CVMUint32* dest, CVMUint32* source, 
				   CVMUint32 numBytes)
{
    CVMUint32*       d = dest;
    const CVMUint32* s = source;
    CVMUint32        n = numBytes / 4; /* numWords to copy */

    do {
	*d++ = *s++;
    } while (--n != 0);
}

/*
 * Copy numBytes _down_ from 'source' to 'dest'. Assume word copy.
 */
static void
CVMgenMarkCompactCopyDown(CVMUint32* dest, CVMUint32* source,
			  CVMUint32 numBytes)
{
    if (dest == source) {
	return;
    } else {
	CVMUint32*       d = dest;
	const CVMUint32* s = source;
	CVMUint32        n = numBytes / 4; /* numWords to copy */
	
	CVMassert(d < s); /* Make sure we are copying _down_ */
	do {
	    *d++ = *s++;
	} while (--n != 0);
    }
}


/* Promote 'objectToPromote' into the current generation. Returns
   the new address of the object, or NULL if the promotion failed. */
static CVMObject*
CVMgenMarkCompactPromoteInto(CVMGeneration* gen,
			     CVMObject* objectToPromote,
			     CVMUint32  objectSize)
{
    CVMObject* dest;
    CVMgenContiguousSpaceAllocate(gen, objectSize, dest);
    if (dest != NULL) {
	CVMgenMarkCompactCopyDisjointWords((CVMUint32*)dest,
					 (CVMUint32*)objectToPromote,
					 objectSize);
    }
    CVMtraceGcScan(("GC[MC,%d]: "
		    "Promoted object %x (len %d, class %C), to %x\n",
		    gen->generationNo,
		    objectToPromote, objectSize,
		    CVMobjectGetClass(objectToPromote),
		    dest));
    return dest;
}

/*
 * Allocate a MarkCompact generation.
 */
CVMGeneration*
CVMgenMarkCompactAlloc(CVMUint32* space, CVMUint32 totalNumBytes)
{
    CVMGenMarkCompactGeneration* thisGen  = (CVMGenMarkCompactGeneration*)
	calloc(sizeof(CVMGenMarkCompactGeneration), 1);
    CVMUint32 numBytes = totalNumBytes; /* Respect the no. of bytes passed */
    CVMGenSpace* theSpace;

    if (thisGen == NULL) {
	return NULL;
    }
    theSpace = CVMgenMarkCompactAllocSpace(space, numBytes);
    
    if (theSpace == NULL) {
	free(thisGen);
	return NULL;
    }

    /*
     * Initialize thisGen. 
     */
    thisGen->theSpace = theSpace;

    /*
     * Start the world on theSpace
     */
    CVMgenMarkCompactMakeCurrent(thisGen, theSpace);

    /* 
     * And finally, set the function pointers for this generation
     */
    thisGen->gen.collect = CVMgenMarkCompactCollect;
    thisGen->gen.scanOlderToYoungerPointers =
	CVMgenMarkCompactScanOlderToYoungerPointers;
    thisGen->gen.scanYoungerToOlderPointers =
	CVMgenMarkCompactScanYoungerToOlderPointers;
    thisGen->gen.promoteInto = 
	CVMgenMarkCompactPromoteInto;
    thisGen->gen.scanPromotedPointers =
	CVMgenMarkCompactScanPromotedPointers;
    thisGen->gen.inGeneration = 
	CVMgenMarkCompactInGeneration;
    thisGen->gen.startGC = CVMgenMarkCompactStartGC;
    thisGen->gen.endGC = CVMgenMarkCompactEndGC;
    thisGen->gen.freeMemory = CVMgenMarkcompactFreeMemory;
    thisGen->gen.totalMemory = CVMgenMarkcompactTotalMemory;

    CVMdebugPrintf(("GC[MC]: Initialized mark-compact gen "
		  "for generational GC\n"));
    CVMdebugPrintf(("\tSize of the space in bytes=%d\n"
		  "\tLimits of generation = [0x%x,0x%x)\n",
		  numBytes, theSpace->allocBase, theSpace->allocTop));
    return (CVMGeneration*)thisGen;
}

/*
 * Free all the memory associated with the current mark-compact generation
 */
void
CVMgenMarkCompactFree(CVMGenMarkCompactGeneration* thisGen)
{
    free(thisGen->theSpace);

    thisGen->theSpace  = NULL;

    free(thisGen);
 }

static CVMObject*
CVMgenMarkCompactGetForwardingPtr(CVMGenMarkCompactGeneration* thisGen,
				  CVMObject* ref)
{
    CVMUint32  forwardingAddress = CVMobjectVariousWord(ref);
    CVMassert(CVMgenMarkCompactInThisGen(thisGen, ref));
    CVMassert(CVMobjectMarked(ref));
    return (CVMObject*)forwardingAddress;
}

#ifdef MAX_STACK_DEPTHS
static CVMUint32 idxMax = 0;
static CVMUint32 preservedIdxMax = 0;
#endif

static void
addToTodo(CVMGenMarkCompactGeneration* thisGen, CVMObject* ref)
{
    thisGen->todoList[thisGen->todoIdx] = ref;
    thisGen->todoIdx++;
#ifdef MAX_STACK_DEPTHS
    if (thisGen->todoIdx >= idxMax) {
	idxMax = thisGen->todoIdx;
	CVMconsolePrintf("**** todo: max depth=%d\n", idxMax);
    }
#endif /* MAX_STACK_DEPTHS */
    if (thisGen->todoIdx >= thisGen->todoSize) {
	CVMpanic("**** OVERFLOW OF TODO STACK!!! ****"
		 " in mark-compact GC!!! ****");
    }
}

/*
 * Restore all preserved header words.
 */
static void
restorePreservedHeaderWords(CVMGenMarkCompactGeneration* thisGen)
{
    CVMInt32 idx;
    CVMtraceGcScan(("GC[MC,%d]: Restoring %d preserved headers\n",
		    thisGen->gen.generationNo,
		    thisGen->preservedIdx));
    while ((idx = --thisGen->preservedIdx) >= 0) {
	CVMMCPreservedItem* item = &thisGen->preservedItems[idx];
	CVMObject* movedRef = (CVMObject*)(item->movedRef);
	/*
	 * Make sure that the object we are looking at now has the default
	 * various word set up for now
	 */
	CVMassert(CVMobjectTrivialClassWord(CVMobjectVariousWord(movedRef)));
	/*
	 * And also, that we did not evacuate this word for nothing
	 */
	CVMassert(!CVMobjectTrivialClassWord(item->originalWord));
	CVMobjectVariousWord(movedRef) = item->originalWord;
	CVMtraceGcScan(("GC[MC,%d]: "
			"Restoring header %x for obj 0x%x at i=%d\n",
			thisGen->gen.generationNo, item->originalWord,
			movedRef, idx));
    }
}

/*
 * Preserve a header word for an object in its new address.
 */
static void
preserveHeaderWord(CVMGenMarkCompactGeneration* thisGen,
		   CVMUint32* movedRef, CVMUint32 originalWord)
{
    CVMMCPreservedItem* item = &thisGen->preservedItems[thisGen->preservedIdx];
    item->movedRef     = movedRef;
    item->originalWord = originalWord;
    CVMtraceGcScan(("GC[MC,%d]: Preserving header %x for obj 0x%x at i=%d\n",
		    thisGen->gen.generationNo, originalWord, movedRef,
		    thisGen->preservedIdx));
    thisGen->preservedIdx++;
#ifdef MAX_STACK_DEPTHS
    if (thisGen->preservedIdx >= preservedIdxMax) {
	preservedIdxMax = thisGen->preservedIdx;
	CVMconsolePrintf("**** preserved: max depth=%d\n", preservedIdxMax);
    }
#endif /* MAX_STACK_DEPTHS */
    if (thisGen->preservedIdx >= thisGen->preservedSize) {
	CVMpanic("**** OVERFLOW OF PRESERVED HEADER WORDS" 
		 " in mark-compact GC!!! ****");
    }
}

/* Sweep the heap, compute the compacted addresses, write them into the
   original object headers, and return the new allocPtr of this space. */
static CVMUint32*
sweep(CVMGenMarkCompactGeneration* thisGen, CVMUint32* base, CVMUint32* top)
{
    CVMUint32* forwardingAddress = base;
    CVMUint32* curr = base;
    CVMtraceGcCollect(("GC[MC,%d]: Sweeping object range [%x,%x)\n",
		       thisGen->gen.generationNo, base, top));
    while (curr < top) {
	CVMObject* currObj    = (CVMObject*)curr;
	CVMUint32  classWord  = CVMobjectGetClassWord(currObj);
	CVMClassBlock* currCb = CVMobjectGetClassFromClassWord(classWord);
	CVMUint32  objSize    = CVMobjectSizeGivenClass(currObj, currCb);
	if (CVMobjectMarkedOnClassWord(classWord)) {
	    volatile CVMUint32* headerAddr   = &CVMobjectVariousWord(currObj);
	    CVMUint32  originalWord = *headerAddr;
	    CVMtraceGcScan(("GC[MC,%d]: obj 0x%x -> 0x%x\n",
			    thisGen->gen.generationNo, curr,
			    forwardingAddress));
	    if (!CVMobjectTrivialClassWord(originalWord)) {
		/* Preserve the old header word with the new address
                   of object, but only if it is non-trivial. */
		preserveHeaderWord(thisGen, forwardingAddress, originalWord);
	    }
	    /* Set forwardingAddress in header */
	    *headerAddr = (CVMUint32)forwardingAddress;
	    /* Pretend to have copied this live object! */
	    forwardingAddress += objSize / 4;
	}
	/* iterate */
	curr += objSize / 4;
    }
    CVMassert(curr == top); /* This had better be exact */
    CVMtraceGcCollect(("GC[MC,%d]: Swept object range [%x,%x) -> 0x%x\n",
		       thisGen->gen.generationNo, base, top,
		       forwardingAddress));
    return forwardingAddress;
}

/* Finally we can move objects, reset marks, and restore original
   header words. */
static void 
compact(CVMGenMarkCompactGeneration* thisGen, CVMUint32* base, CVMUint32* top)
{
    CVMUint32* curr = base;
    CVMtraceGcCollect(("GC[MC,%d]: Compacting object range [%x,%x)\n",
		       thisGen->gen.generationNo, base, top));
    while (curr < top) {
	CVMObject*     currObj   = (CVMObject*)curr;
	CVMUint32      classWord = CVMobjectGetClassWord(currObj);
	CVMClassBlock* currCb    = CVMobjectGetClassFromClassWord(classWord);
	CVMUint32      objSize   = CVMobjectSizeGivenClass(currObj, currCb);

	if (CVMobjectMarkedOnClassWord(classWord)) {
	    CVMUint32* destAddr = (CVMUint32*)
		CVMgenMarkCompactGetForwardingPtr(thisGen, currObj);
	    CVMobjectClearMarkedOnClassWord(classWord);
#ifdef CVM_DEBUG
	    /* For debugging purposes, make sure the deleted mark is
	       reflected in the original copy of the object as well. */
	    CVMobjectSetClassWord(currObj, classWord);
#endif
	    CVMgenMarkCompactCopyDown(destAddr, curr, objSize);
	    /*
	     * First assume that all objects had the default various
	     * word. We'll patch the rest shortly.
	     * 
	     * Note that this clears the GC bits portion. No worries,
	     * we won't need that anymore in this generation.
	     */
	    CVMobjectVariousWord((CVMObject*)destAddr) =
		CVM_OBJECT_DEFAULT_VARIOUS_WORD;
	    /* Write the class word whose mark has just been cleared */
	    CVMobjectSetClassWord((CVMObject*)destAddr, classWord);
	}
	/* iterate */
	curr += objSize / 4;
    }
    /* Restore the "non-trivial" old header words
       with the new address of object */
    restorePreservedHeaderWords(thisGen);
    CVMassert(curr == top); /* This had better be exact */
}

/*
 * Gray an object known to be in the old MarkCompact generation.
 */
static void
CVMgenMarkCompactGrayObject(CVMGenMarkCompactGeneration* thisGen,
			  CVMObject* ref)
{
    /*
     * ROM objects should have been filtered out by the time we get here
     */
    CVMassert(!CVMobjectIsInROM(ref));

    /*
     * Also, all objects that don't belong to this generation need to have
     * been filtered out by now.
     */
    CVMassert(CVMgenMarkCompactInThisGen(thisGen, ref));

    /*
     * If this object has not been encountered yet, queue it up.
     */
    if (!CVMobjectMarked(ref)) {
	CVMobjectSetMarked(ref);
	/* queue it up */
	CVMtraceGcCollect(("GC[MC,%d]: Graying object %x\n",
			   thisGen->gen.generationNo, ref));
	addToTodo(thisGen, ref);
    }
}

/*
 * Gray a slot if it points into this generation.
 */
static void
CVMgenMarkCompactFilteredGrayObject(CVMGenMarkCompactGeneration* thisGen,
				    CVMObject* ref)
{
    /*
     * Gray only if the slot points to this generation's from space. 
     *
     * Also handles the case of root slots that have been already
     * scanned and grayed, and therefore contain a pointer into
     * to-space -- those are ignored.
     */
    if (CVMgenMarkCompactInThisGen(thisGen, ref)) {
	CVMgenMarkCompactGrayObject(thisGen, ref);
    }
}

static void
CVMgenMarkCompactFilteredGrayCallback(CVMObject** refPtr, void* data)
{
    CVMGenMarkCompactGeneration* thisGen = (CVMGenMarkCompactGeneration*)data;
    CVMObject* ref = *refPtr;
    CVMgenMarkCompactFilteredGrayObject(thisGen, ref);
}

static void
CVMgenMarkCompactFilteredUpdateRoot(CVMObject** refPtr, void* data)
{
    CVMGenMarkCompactGeneration* thisGen = (CVMGenMarkCompactGeneration*)data;
    CVMObject* ref = *refPtr;
    if (!CVMgenMarkCompactInThisGen(thisGen, ref)) {
	return;
    }
    /* The forwarding address is valid only for already marked
       objects. If the object is not marked, then this slot has already
       been 'seen'. */
    CVMassert(CVMobjectMarked(ref));
    *refPtr = CVMgenMarkCompactGetForwardingPtr(thisGen, ref);
}

typedef struct CVMGenMarkCompactTransitiveScanData {
    CVMExecEnv* ee;
    CVMGCOptions* gcOpts;
    CVMGenMarkCompactGeneration* thisGen;
} CVMGenMarkCompactTransitiveScanData;

static void
CVMgenMarkCompactBlackenObject(CVMGenMarkCompactTransitiveScanData* tsd,
			       CVMObject* ref, CVMClassBlock* refCb)
{
    CVMGenMarkCompactGeneration* thisGen = tsd->thisGen;
    CVMExecEnv* ee = tsd->ee;
    CVMGCOptions* gcOpts = tsd->gcOpts;
    CVMassert(ref != 0);

    CVMtraceGcCollect(("GC[MC,%d,full]: Blacken object %x\n",
		       thisGen->gen.generationNo,
		       ref));
    /*
     * We'd better have grayed this object already.
     */
    CVMassert(CVMobjectMarked(ref));

    /*
     * Queue up all non-null object references. Handle the class as well.
     */
    CVMobjectWalkRefsWithSpecialHandling(ee, gcOpts, ref, refCb, {
	if (*refPtr != 0) {
	    CVMgenMarkCompactFilteredGrayObject(thisGen, *refPtr);
	}
    }, CVMgenMarkCompactFilteredGrayCallback, thisGen);
}

static void
CVMgenMarkCompactFollowRoots(CVMGenMarkCompactGeneration* thisGen,
			     CVMGenMarkCompactTransitiveScanData* tsd)
{
    CVMtraceGcCollect(("GC[MC,%d]: Following roots\n",
		       thisGen->gen.generationNo));
    while (thisGen->todoIdx > 0) {
	CVMObject* ref;
	CVMClassBlock* refCB;
	--thisGen->todoIdx;
	ref = thisGen->todoList[thisGen->todoIdx];
	refCB = CVMobjectGetClass(ref);
	/*
	 * Blackening will queue up more objects
	 */
	CVMgenMarkCompactBlackenObject(tsd, ref, refCB);
    }
}

static void
CVMgenMarkCompactScanTransitively(CVMObject** refPtr, void* data)
{
    CVMGenMarkCompactTransitiveScanData* tsd =
	(CVMGenMarkCompactTransitiveScanData*)data;
    CVMGenMarkCompactGeneration* thisGen = tsd->thisGen;
    CVMObject* ref;

    CVMassert(refPtr != 0);
    ref = *refPtr;
    CVMassert(ref != 0);

    CVMgenMarkCompactFilteredGrayObject(thisGen, ref);

    /*
     * Now scan all its children as well
     */
    CVMassert(tsd->gcOpts->fullGC);
    CVMgenMarkCompactFollowRoots(thisGen, tsd);
}


/*
 * Test whether a given reference is live or dying. If the reference
 * does not point to the current generation, answer TRUE. The generation
 * that does contain the pointer will give the right answer.
 */
static CVMBool
CVMgenMarkCompactRefIsLive(CVMObject** refPtr, void* data)
{
    CVMGenMarkCompactGeneration* thisGen = (CVMGenMarkCompactGeneration*)data;
    CVMObject* ref;

    CVMassert(refPtr != NULL);

    ref = *refPtr;
    CVMassert(ref != NULL);

    /*
     * ROM objects are always live
     */
    if (CVMobjectIsInROM(ref)) {
	return CVM_TRUE;
    }
    /*
     * If this reference is not in this generation, just assume
     * it's live. 
     */
    if (!CVMgenMarkCompactInThisGen(thisGen, ref)) {
	return CVM_TRUE;
    }
    /* Did somebody else already scan or forward this object? It's live then */
    return CVMobjectMarked(ref);
}

static CVMBool
CVMgenMarkCompactCollect(CVMGeneration* gen,
			 CVMExecEnv*    ee, 
			 CVMUint32      numBytes, /* collection goal */
			 CVMGCOptions*  gcOpts)
{
    CVMGenMarkCompactTransitiveScanData tsd;
    CVMGenMarkCompactGeneration* thisGen = (CVMGenMarkCompactGeneration*)gen;
    CVMGenSpace* space;
    CVMBool success;
    CVMGenSpace* extraSpace;
    CVMUint32* newAllocPtr;

    CVMassert(gcOpts->fullGC);
    CVMtraceGcStartStop(("GC[MC,%d,full]: Collecting\n",
			 gen->generationNo));

    extraSpace = gen->prevGen->getExtraSpace(gen->prevGen);

    thisGen->todoList = (CVMObject**)extraSpace->allocBase;
    thisGen->todoIdx  = 0;
    thisGen->todoSize = extraSpace->allocTop - extraSpace->allocBase;
	
    /*
     * Prepare to do transitive scans 
     */
    tsd.ee = ee;
    tsd.gcOpts = gcOpts;
    tsd.thisGen = thisGen;

    /* The mark phase */
    gcOpts->discoverWeakReferences = CVM_TRUE;

    /*
     * Scan all roots that point to this generation. The root callback is
     * transitive, so 'children' are aggressively processed.
     */
    CVMgenScanAllRoots((CVMGeneration*)thisGen,
		       ee, gcOpts, CVMgenMarkCompactScanTransitively, &tsd);
    /*
     * Don't discover any more weak references.
     */
    gcOpts->discoverWeakReferences = CVM_FALSE;

    /*
     * At this point, we know which objects are live and which are not.
     * Do the special objects processing.
     */
    CVMgcProcessSpecialWithLivenessInfo(ee, gcOpts,
					CVMgenMarkCompactRefIsLive, thisGen,
					CVMgenMarkCompactScanTransitively,
					&tsd);

    /* Prepare the data structure that preserves original second header
       words. We are done with the TODO stack, so we can re-use extraSpace. */
    thisGen->preservedItems = (CVMMCPreservedItem*)extraSpace->allocBase;
    thisGen->preservedIdx   = 0;
    thisGen->preservedSize  = (CVMMCPreservedItem*)extraSpace->allocTop - 
	(CVMMCPreservedItem*)extraSpace->allocBase;
	
    /* The sweep phase. This phase computes the new addresses of each
       object and writes it in the second header word. Evacuated
       original header words will be kept in the 'preservedItems'
       list. */
    newAllocPtr = sweep(thisGen, gen->allocBase, gen->allocPtr);
    CVMassert(newAllocPtr <= gen->allocPtr);

    /* At this point, the new addresses of each object are written in
       the headers of the old. Update all pointers to refer to the new
       copies. */

    /* Update the roots */
    CVMgenScanAllRoots((CVMGeneration*)thisGen, ee, gcOpts,
		       CVMgenMarkCompactFilteredUpdateRoot, thisGen);
    CVMgcScanSpecial(ee, gcOpts,
		     CVMgenMarkCompactFilteredUpdateRoot, thisGen);
    /* And update all interior pointers */
    scanObjectsInRangeSkipUnmarked(thisGen,
				   ee, gcOpts, gen->allocBase, gen->allocPtr,
				   CVMgenMarkCompactFilteredUpdateRoot,
				   thisGen);
    
    /* Finally we can move objects, reset marks, and restore original
       header words. */
    compact(thisGen, gen->allocBase, gen->allocPtr);

    gen->allocPtr = newAllocPtr;

    /*
     * Now since we are not in the young generation, we have to
     * rebuild the barrier structures.  
     */
    space = thisGen->theSpace;
    space->allocPtr = newAllocPtr;
    CVMassert(space->allocPtr == gen->allocPtr);
    CVMgenMarkCompactMakeCurrent(thisGen, space);

    CVMgenClearBarrierTable();
    CVMgenMarkCompactRebuildBarrierTable(thisGen, ee, gcOpts,
					 space->allocBase,
					 space->allocPtr);
    success = (numBytes / 4 <= space->allocTop - space->allocPtr);
    CVMtraceGcStartStop(("GC[MC,%d,full]: Done, success for %d bytes %s\n",
			 gen->generationNo,
			 numBytes,
			 success ? "TRUE" : "FALSE"));

    /*
     * Can we allocate in this space after this GC?
     */
    return success;
}
