/*
 * -----------------------------------------------------------------
 *  globals.c :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)globals.c	1.109 00/10/25
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/sync.h"
#include "javavm/include/globals.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/preloader.h"
#include "javavm/include/globalroots.h"
#include "javavm/include/directmem.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/stackmaps.h"
#include "javavm/include/utils.h"

#include "native/java/util/zip/zip_util.h"

#include "javavm/include/typeid.h"
#include "javavm/include/string.h"

#include "javavm/include/gc/gc_impl.h"

#include "javavm/include/clib.h"

#ifdef CVM_JVMDI
#include "javavm/include/jvmdi_jni.h"
#endif

#include "native/com/timesys/CoreExtension/rtmonitoradapter.h"

CVMGlobalState CVMglobals;

/*
 * Capacities for the class table of dynamcially loaded classes.
 */
#define CVM_MIN_CLASSTABLE_STACKCHUNK_SIZE 100
#define CVM_MAX_CLASSTABLE_STACK_SIZE      \
    100 * CVM_INITIAL_CLASSTABLE_STACK_SIZE
#define CVM_INITIAL_CLASSTABLE_STACK_SIZE   \
    CVM_MIN_CLASSTABLE_STACKCHUNK_SIZE

/* 
 * Capacities for the global roots stack. It holds both regular global
 * roots and jni global refs.
 */
#define CVM_MIN_GLOBALROOTS_STACKCHUNK_SIZE 500
#define CVM_MAX_GLOBALROOTS_STACK_SIZE      \
    100 * CVM_MIN_GLOBALROOTS_STACKCHUNK_SIZE
#define CVM_INITIAL_GLOBALROOTS_STACK_SIZE   \
    CVM_MIN_GLOBALROOTS_STACKCHUNK_SIZE

/*
 * Capacities for the JNI weak global refs stack.
 */
#define CVM_MIN_WEAK_GLOBALROOTS_STACKCHUNK_SIZE \
    CVM_MIN_GLOBALROOTS_STACKCHUNK_SIZE
#define CVM_MAX_WEAK_GLOBALROOTS_STACK_SIZE      \
    CVM_MAX_GLOBALROOTS_STACK_SIZE
#define CVM_INITIAL_WEAK_GLOBALROOTS_STACK_SIZE   \
    CVM_INITIAL_GLOBALROOTS_STACK_SIZE

/*
 * Capacities for the Class global roots stack.
 */
#define CVM_MIN_CLASS_GLOBALROOTS_STACKCHUNK_SIZE 500
#define CVM_MAX_CLASS_GLOBALROOTS_STACK_SIZE      \
    100 * CVM_MIN_CLASS_GLOBALROOTS_STACKCHUNK_SIZE
#define CVM_INITIAL_CLASS_GLOBALROOTS_STACK_SIZE   \
    CVM_MIN_CLASS_GLOBALROOTS_STACKCHUNK_SIZE

/*
 * Capacities for the ClassLoader global roots stack.
 */
#define CVM_MIN_CLASSLOADER_GLOBALROOTS_STACKCHUNK_SIZE 30
#define CVM_MAX_CLASSLOADER_GLOBALROOTS_STACK_SIZE      \
    10 * CVM_MIN_CLASS_GLOBALROOTS_STACKCHUNK_SIZE
#define CVM_INITIAL_CLASSLOADER_GLOBALROOTS_STACK_SIZE   \
    CVM_MIN_CLASS_GLOBALROOTS_STACKCHUNK_SIZE

CVMBool CVMinitVMGlobalState(CVMGlobalState *gs, CVMOptions *options)
{
    CVMExecEnv *ee = &gs->mainEE;
    /*
     * First zero everything out
     */
    memset((void*)gs, 0, sizeof(CVMGlobalState));

    /*
     * Now initialize the global state
     */
    CVMtraceMisc(("Initializing global state\n"));

    CVMpreloaderInit();

    gs->threadList = NULL;
    gs->threadIDCount = 0;
#ifdef CVM_DEBUG

#endif


#ifdef CVM_CLASSLOADING
    CVMsysMutexInit(&gs->nullClassLoaderLock, "NULL classloader lock", 1);
#endif
    CVMsysMutexInit(&gs->heapLock, "heap lock", 3);
    CVMsysMutexInit(&gs->threadLock, "thread-list lock", 5);
    CVMsysMutexInit(&gs->classTableLock,
		    "class table lock", 7);
#ifdef CVM_CLASSLOADING
    CVMsysMutexInit(&gs->constantpoolLock, "constant pool lock", 9);
    CVMsysMutexInit(&gs->codeLock, "code lock", 11);
#endif
    CVMsysMutexInit(&gs->loaderCacheLock, "loader cache lock", 13);
    /* NOTE: This is not a leaf mutex. It must remain below the global
       roots lock since a couple of the JVMDI routines allocate and
       deallocate global roots while holding the debugger lock. In
       addition, others of them perform indirect memory accesses while
       holding the debugger lock, so if we ever had a lock associated
       with those accesses it would need to have a higher rank (less
       importance) than this one. */
#ifdef CVM_JVMDI
    CVMsysMutexInit(&gs->debuggerLock, "debugger lock", 15);
#endif

    /* Are these all "leaf" mutexes? */
    /* NOTE: doesn't appear so (see CVMlocksForThreadSuspendAcquire,
       interpreter.c) */
    CVMsysMutexInit(&gs->globalRootsLock, "global roots lock", 20);
    CVMsysMutexInit(&gs->weakGlobalRootsLock, "weak global roots lock", 21);
    CVMsysMutexInit(&gs->typeidLock, "typeid lock", 22);

    /* Dio: set the syncLock to PI so we can have a predictable
     * behavior on inflation
     */
 
    {
      extern void RTMonAdaptorStaticFilterInitPIMutex(CVMSysMutex *p, char *s, int num);
      
      RTMonAdaptorStaticFilterInitPIMutex(&gs->syncLock, "fast sync lock",24);
    }

    /*
    CVMsysMutexInit(&gs->syncLock, "fast sync lock", 24);
    */

    CVMsysMutexInit(&gs->internLock, "intern table lock", 26);

    CVMcondvarInit(&gs->threadCountCV, &gs->threadLock.rmutex.mutex);

    /* NOTE: This is theoretically where the "checked" version of the
       JNI would be installed if we had it in our source tree.  Here,
       if JVMDI is present, we replace some functions in the JNI
       interface with the JVMDI's instrumented versions, which will
       call back into the debugger if native code uses the JNI to set
       or get a field. */

    /* NOTE: this should come BEFORE the first call to CVMinitExecEnv.
       This mutates the global JNI vector which later gets pointed to
       by the various execution environments. */
#ifdef CVM_JVMDI
    gs->jvmdiDebuggingEnabled = options->debugging;
    if (gs->jvmdiDebuggingEnabled) {
	CVMjvmdiInstrumentJNINativeInterface();
    }
#endif    

    gs->userThreadCount = 1;
    gs->threadCount = 1;

    if (!CVMinitExecEnv(ee, CVM_TRUE)) {
	return CVM_FALSE;
    }

    if (!CVMinitGCRootStack(ee, &gs->globalRoots,
		       CVM_INITIAL_GLOBALROOTS_STACK_SIZE,
		       CVM_MAX_GLOBALROOTS_STACK_SIZE, 
		       CVM_MIN_GLOBALROOTS_STACKCHUNK_SIZE,
		       CVMglobalrootFrameScanner)) {
	return CVM_FALSE;
    }

    if (!CVMinitGCRootStack(ee, &gs->weakGlobalRoots,
		       CVM_INITIAL_WEAK_GLOBALROOTS_STACK_SIZE,
		       CVM_MAX_WEAK_GLOBALROOTS_STACK_SIZE, 
		       CVM_MIN_WEAK_GLOBALROOTS_STACKCHUNK_SIZE,
		       CVMglobalrootFrameScanner)) {
	return CVM_FALSE;
    }

    if (!CVMinitGCRootStack(ee, &gs->classGlobalRoots,
		       CVM_INITIAL_CLASS_GLOBALROOTS_STACK_SIZE,
		       CVM_MAX_CLASS_GLOBALROOTS_STACK_SIZE, 
		       CVM_MIN_CLASS_GLOBALROOTS_STACKCHUNK_SIZE,
		       CVMglobalrootFrameScanner)) {
	return CVM_FALSE;
    }

    if (!CVMinitGCRootStack(ee, &gs->classLoaderGlobalRoots,
		       CVM_INITIAL_CLASSLOADER_GLOBALROOTS_STACK_SIZE,
		       CVM_MAX_CLASSLOADER_GLOBALROOTS_STACK_SIZE, 
		       CVM_MIN_CLASSLOADER_GLOBALROOTS_STACKCHUNK_SIZE,
		       CVMglobalrootFrameScanner)) {
	return CVM_FALSE;
    }

    if (!CVMinitStack(ee, &gs->classTable,
		 CVM_INITIAL_CLASSTABLE_STACK_SIZE,
		 CVM_MAX_CLASSTABLE_STACK_SIZE, 
		 CVM_MIN_CLASSTABLE_STACKCHUNK_SIZE,
		 CVMfreelistFrameCapacity, 
		 NULL)) {
	return CVM_FALSE;
    }

    CVMfreelistFrameInit((CVMFreelistFrame*)gs->classTable.currentFrame);

#if CVM_MICROLOCK_TYPE == CVM_MICROLOCK_DEFAULT
    {
	int i;

	for (i = 0; i < CVM_NUM_SYS_MICROLOCKS; ++i) {
	    CVMmutexInit(&gs->sysMicroLock[i]);
	}
	for (i = 0; i < CVM_NUM_OBJ_MICROLOCKS; ++i) {
	    CVMmutexInit(&gs->objMicroLock[i]);
	}
    }
#endif

    {
	int i;
	int rank = 30;

	for (i = 0; i < CVM_NUM_CONSISTENT_STATES; ++i) {
	    CVMcsInit(CVM_CSTATE(i), CVMcstateNames[i], rank);
	    ++rank;
	}
    }

    /*
     * Initialize the target global state
     */
    CVMinitVMTargetGlobalState(&gs->target);

    /*
     * Initialize the particular GC global state
     */
    CVMgcimplInitGlobalState(&gs->gc);

    /*
     * JNI JavaVM state
     */
    CVMinitJNIJavaVM(&gs->javaVM);

    /*
     * TypeId system, which also registers well-known types.
     */
    CVMtypeidInit(ee);

    /*
     * Intern table, which just has a global variable to set.
     */
    CVMinternInit();
    
    /*
     * Classes system
     */
    CVMclassModuleInit(ee);
#ifdef CVM_CLASSLOADING
    gs->classVerificationLevel = options->classVerificationLevel;
#endif

    /*
     * Now that both typeid and class system are initialized,
     * register pre-loaded packages.
     */
    CVMtypeidRegisterPreloadedPackages();

    /*
     * Stackmap computer
     */
    CVMstackmapComputerInit();

    /*
     * Random number generator
     */
    CVMrandomInit();

    /*
     * Weak reference handling initialization
     */
    CVMweakrefInit();

#ifdef CVM_REFLECT
    /* lookup mb for java.lang.reflect.Constructor.newInstance() */
    {
	CVMMethodTypeID methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
            ee, "newInstance", "([Ljava/lang/Object;)Ljava/lang/Object;");
	const CVMClassBlock* cb =
	    CVMsystemClass(java_lang_reflect_Constructor);
	gs->java_lang_reflect_Constructor_newInstance = 
	    CVMclassGetNonstaticMethodBlock(cb, methodTypeID);
	CVMassert(gs->java_lang_reflect_Constructor_newInstance != NULL);
    }

    /* lookup mb for java.lang.reflect.Method.invoke() */
    {
	CVMMethodTypeID methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
            ee, "invoke",
	    "(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;");
	const CVMClassBlock* cb =
	    CVMsystemClass(java_lang_reflect_Method);
	gs->java_lang_reflect_Method_invoke = 
	    CVMclassGetNonstaticMethodBlock(cb, methodTypeID);
	CVMassert(gs->java_lang_reflect_Method_invoke != NULL);
    }
#endif /* CVM_REFLECT */

    /* lookup mb for java.lang.ref.Finalizer.register */
    {
	CVMMethodTypeID methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
            ee, "register", "(Ljava/lang/Object;)V");
	const CVMClassBlock* cb = CVMsystemClass(java_lang_ref_Finalizer);
	gs->java_lang_ref_Finalizer_register = 
	    CVMclassGetStaticMethodBlock(cb, methodTypeID);
	CVMassert(gs->java_lang_ref_Finalizer_register != NULL);
    }

    /* lookup mb for javax.realtime.Finalizer.register 
	 * -Pratik
     * (deprecated. I should remove this)
	 */
    /*{
	CVMMethodTypeID methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
            ee, "register", "(Ljavax/realtime/MemoryArea;Ljava/lang/Object;)V");
	const CVMClassBlock* cb = CVMsystemClass(javax_realtime_Finalizer);
	gs->javax_realtime_Finalizer_register = 
	    CVMclassGetStaticMethodBlock(cb, methodTypeID);
	CVMassert(gs->javax_realtime_Finalizer_register != NULL);
    }*/
    
    /* lookup mb for javax.realtime.Realtime.allowAllocation
	 * -Pratik
	 */
    {
	CVMMethodTypeID methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
            ee, "allowAllocation", "(Ljavax/realtime/MemoryArea;I)Z");
	const CVMClassBlock* cb = CVMsystemClass(javax_realtime_RealtimeThread);
	gs->javax_realtime_RealtimeThread_allowAllocation = 
	    CVMclassGetNonstaticMethodBlock(cb, methodTypeID);
	CVMassert(gs->javax_realtime_RealtimeThread_allowAllocation != NULL);
    }
    
	/* 
     * Lookup fb for java.lang.ref.Reference.lock. GC has to know about
     * this lock to handle weak references correctly.
     */
    {
	CVMFieldTypeID lockTID = CVMtypeidLookupFieldIDFromNameAndSig(
            ee, "lock", "Ljava/lang/ref/Reference$Lock;");
	CVMFieldTypeID pendingTID = CVMtypeidLookupFieldIDFromNameAndSig(
            ee, "pending", "Ljava/lang/ref/Reference;");
	const CVMClassBlock* cb = CVMsystemClass(java_lang_ref_Reference);
	CVMFieldBlock* lockFB = 
	    CVMclassGetStaticFieldBlock(cb, lockTID);
	CVMFieldBlock* pendingFB = 
	    CVMclassGetStaticFieldBlock(cb, pendingTID);
	CVMassert(lockFB != NULL);
	CVMassert(pendingFB != NULL);
	gs->java_lang_ref_Reference_lock = &CVMfbStaticField(lockFB);
	gs->java_lang_ref_Reference_pending = &CVMfbStaticField(pendingFB);
    }

#ifdef CVM_DEBUG_STACKTRACES
    /* lookup mb for java.lang.Throwable.fillInStackTrace() */
    {
	CVMMethodTypeID methodTypeID = CVMtypeidLookupMethodIDFromNameAndSig(
            ee, "fillInStackTrace", "()Ljava/lang/Throwable;");
	const CVMClassBlock* cb = CVMsystemClass(java_lang_Throwable);
	gs->java_lang_Throwable_fillInStackTrace = 
	    CVMclassGetNonstaticMethodBlock(cb, methodTypeID);
	CVMassert(gs->java_lang_Throwable_fillInStackTrace != NULL);
    }

#ifdef CVM_DEBUG
    /* 
     * Lookup fb for System.out. This is useful for debugging sessions
     * if you have a Throwable object that you want to pass to
     * CVMprintStackTrace, since it requires that you pass it an object
     * that implements println().
     */
    {
	CVMFieldTypeID fieldTypeID = CVMtypeidLookupFieldIDFromNameAndSig(
            ee, "out", "Ljava/io/PrintStream;");
	const CVMClassBlock* cb = CVMsystemClass(java_lang_System);
	gs->java_lang_System_out = 
	    CVMclassGetStaticFieldBlock(cb, fieldTypeID);
	CVMassert(gs->java_lang_System_out != NULL);
    }
#endif /* CVM_DEBUG */
#endif /* CVM_DEBUG_STACKTRACES */

#ifdef CVM_DEBUG
    /*
     * Turn off all debugging flags by default.
     */

    if (options->traceFlagsStr != NULL) {
	gs->debugFlags = strtol(options->traceFlagsStr, NULL, 0);
    } else {
	gs->debugFlags = 0;
    }
#endif /* CVM_DEBUG */

    if (!CVMgcInitHeap(options)) {
	CVMconsolePrintf("Cannot start VM "
			 "(out of memory while initializing)\n");
	return CVM_FALSE;
    }

    /*
     * Some desperation exception objects we preallocate.
     */
    gs->preallocatedOutOfMemoryError = CVMID_getGlobalRoot(ee);
    gs->preallocatedStackOverflowError = CVMID_getGlobalRoot(ee);
    if (gs->preallocatedOutOfMemoryError == NULL ||
	gs->preallocatedStackOverflowError == NULL) {
	return CVM_FALSE;
    }
    CVMID_allocNewInstance(ee, CVMsystemClass(java_lang_OutOfMemoryError),
			   gs->preallocatedOutOfMemoryError);
    CVMID_allocNewInstance(ee, CVMsystemClass(java_lang_StackOverflowError),
			   gs->preallocatedStackOverflowError);
    if (CVMID_icellIsNull(gs->preallocatedOutOfMemoryError) ||
	CVMID_icellIsNull(gs->preallocatedStackOverflowError)) {
	return CVM_FALSE;
    }

    /*
     * Default sizes
     */
    if (options->nativeStackSizeStr != NULL) {
	CVMInt32 nativeStackSize = CVMatoml(options->nativeStackSizeStr);
	if (nativeStackSize == -1) {
	    CVMconsolePrintf("Warning: Could not parse "
			     "stack size string \"%s\", setting default\n",
			     options->nativeStackSizeStr);
	    /* gs->nativeStackSize remains 0 */
	} else if (nativeStackSize < CVM_THREAD_MIN_C_STACK_SIZE) {
	    CVMconsolePrintf("Stack size %d below minimum stack size (%d)\n",
		nativeStackSize, CVM_THREAD_MIN_C_STACK_SIZE);
	    return CVM_FALSE;
	} else {
	    gs->nativeStackSize = nativeStackSize;
	}
    }

    /*
     * Hooks
     */
    gs->vfprintfHook = options->vfprintfHook;
    gs->exitHook = (void (*)(int))options->exitHook;
    gs->abortHook = (void (*)())options->abortHook;
    gs->safeExitHook = (void (*)(int))options->safeExitHook;

#ifdef CVM_HAVE_PROCESS_MODEL
    gs->fullShutdown = options->fullShutdownFlag;
#else
    gs->fullShutdown = CVM_TRUE;
#endif

    gs->exit_procs = NULL;

#ifdef CVM_JVMDI
    /* jvmdi global variables initialization */
    CVMjvmdiStaticsInit(&gs->jvmdiStatics);
#endif

    return CVM_TRUE;
}

void CVMdestroyVMGlobalState(CVMExecEnv *ee, CVMGlobalState *gs)
{
    /* These should be roughly in the reverse of initialization order */

    /* 
     * Destroy dynamically loaded classes.
     */
    CVMclassModuleDestroy( ee );

    if (gs->preallocatedOutOfMemoryError != NULL) {
	CVMID_freeGlobalRoot(ee, gs->preallocatedStackOverflowError);
    }
    if (gs->preallocatedStackOverflowError != NULL) {
	CVMID_freeGlobalRoot(ee, gs->preallocatedOutOfMemoryError);
    }

    CVMeeSyncDestroyGlobal(ee);

    CVMgcDestroyHeap();

    CVMcondvarDestroy(&gs->threadCountCV);

    CVMsysMutexDestroy(&gs->heapLock);
    CVMsysMutexDestroy(&gs->loaderCacheLock);
    CVMsysMutexDestroy(&gs->syncLock);
    CVMsysMutexDestroy(&gs->threadLock);
    CVMsysMutexDestroy(&gs->globalRootsLock);
    CVMsysMutexDestroy(&gs->weakGlobalRootsLock);
    CVMsysMutexDestroy(&gs->classTableLock);
    CVMsysMutexDestroy(&gs->typeidLock);
    CVMsysMutexDestroy(&gs->internLock);
#ifdef CVM_CLASSLOADING
    CVMsysMutexDestroy(&gs->constantpoolLock);
    CVMsysMutexDestroy(&gs->codeLock);
    CVMsysMutexDestroy(&gs->nullClassLoaderLock);
#endif
#ifdef CVM_JVMDI
    CVMsysMutexDestroy(&gs->debuggerLock);
#endif
    CVMdestroyGCRootStack(&gs->globalRoots);
    CVMdestroyGCRootStack(&gs->weakGlobalRoots);
    CVMdestroyGCRootStack(&gs->classGlobalRoots);
    CVMdestroyGCRootStack(&gs->classLoaderGlobalRoots);
    CVMdestroyStack(&gs->classTable);

    {
        int i;

        for (i = 0; i < CVM_NUM_CONSISTENT_STATES; ++i) {
            CVMcsDestroy(CVM_CSTATE(i));
        }
    }

#if CVM_MICROLOCK_TYPE == CVM_MICROLOCK_DEFAULT
    {
	int i;

	for (i = 0; i < CVM_NUM_SYS_MICROLOCKS; ++i) {
	    CVMmutexDestroy(&gs->sysMicroLock[i]);
	}
	for (i = 0; i < CVM_NUM_OBJ_MICROLOCKS; ++i) {
	    CVMmutexDestroy(&gs->objMicroLock[i]);
	}
    }
#endif

    CVMdestroyVMTargetGlobalState(&gs->target);

    CVMgcimplDestroyGlobalState(&gs->gc);

    CVMdestroyJNIJavaVM(&gs->javaVM);

    CVMdestroyExecEnv(ee);

    /*
     * Destroy class related data structures
     * Including those for preloaded classes.
     */
    CVMpreloaderDestroy();
    CVMtypeidDestroy();
    CVMinternDestroy();


    /*
     * Destroy stackmap computer related data structures
     */
    CVMstackmapComputerDestroy();

    /*
     * Destroy zip_util related data structures
     */
    CVMzutilDestroyZip();

#ifdef CVM_JVMDI
    /* Free jvmdi global variables */
    CVMjvmdiStaticsDestroy(&gs->jvmdiStatics);
#endif
}
