/*
 * @(#)CVMStringTable.java	1.4 00/10/31
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

package runtime;
import java.util.Enumeration;
import Const;
import components.*;
import vm.*;

public class CVMStringTable extends vm.StringTable {

    CVMStringIntern internTable = new CVMStringIntern();

    final boolean writeStringData( String dataName, String charArrayClassBlockName, CCodeWriter out ){
	int n = arrangeStringData();

	if ( n == 0 ) return false; // nothing to do here.

	char v[] = new char[n];
	data.getChars( 0, n, v, 0 );

	//
	// first typedef the structure we're about to create
	out.print("STATIC const struct { CVMObjectHeader hdr;\n    CVMJavaInt	length;\n    CVMJavaChar stringData["+n+"]; } ");
	out.print( dataName );
	out.println( " = {" );
	//
	// fill in the header information
	// hdr.various = {{0}} ??
	int hashCode = 0;
	out.println("	CVM_ROM_OBJECT_HDR_INIT0("+charArrayClassBlockName+
	    ","+hashCode+"),");
	out.println("	"+n+", {");
	//
	// finally write the char array.
	int mod = 0;
	for (int i = 0; i < n; i++) { 
	    if (mod == 0) 
	        out.write('\t');
	    out.printHexInt(v[i]);
	    out.write(',');
	    if (++mod == 12) { 
	        out.write('\n');
		mod = 0;
	    }
	} 
	if (mod != 0)
	    out.write('\n');
	out.print(" } };\n\n");

	return true;

    }

    private final static int maxCom = 20;
    private static byte cbuf[] = new byte[maxCom + 3];
    private void commentary( String s, CCodeWriter out ){
	int i;
	int m = Math.min( maxCom, s.length() );
	for ( i = 0; i < m; i++ ){
	    char c = s.charAt(i);
	    if ( ' ' <= c && c <= '~' ) {
		if (c == '*') {
		    /*
		     * Prevent embedded comments by escaping * with \
		     */
		    cbuf[i++] = (byte)'\\';
		}
		cbuf[i] = (byte) c;
	    } else {
		cbuf[i] = (byte)'?';
	    }
	}
	out.write( out.commentLeader, 0, out.commentLeader.length );
	out.write( cbuf, 0, m );
	if ( m < s.length() ){
	    out.write('.');
	    out.write('.');
	    out.write('.');
	}
	out.write( out.commentTrailer, 0, out.commentTrailer.length );
    }

    public int writeStrings(CCodeWriter out, String tableName ) {
	String dataName = tableName+"_data";

	ClassInfo charArrayClass = ClassInfo.lookupClass("[C");
	if ( charArrayClass == null ){
	    System.err.println("String writer cannot find array-of-char class");
	    return 0;
	}

	if ( ! writeStringData( dataName, ((CVMClass)(charArrayClass.vmClass)).getNativeName()+"_Classblock" , out ) )
	    return 0;

	out.println("\n#define CVM_INIT_JAVA_STRING(hashCode,offset,count)\\\n"+
	    "    CVM_ROM_STRING_INIT({(CVMArrayOfChar*)&"+dataName+"},"+
	    "offset,count,hashCode)");

	int offset;
	out.println("STATIC const struct java_lang_String " + tableName + "[] = {");
	offset = 0;

	// unfortunately, order matters.
	// unfortunately, the order that matters is not the order
	// available throught the Enumeration.
	// Thus we sort first, output second.
	StringConstant stringTable[] = new StringConstant[ internedStringCount() ];

	for (Enumeration e = allStrings(); e.hasMoreElements(); offset++) { 
	    StringConstant s = (StringConstant)e.nextElement();
	    stringTable[ s.unicodeIndex ] = s;
	}
	for ( int i = 0; i < offset; i++ ){
	    StringConstant s = stringTable[i];
	    String string = s.str.string;
	    int hashCode = 0;
	    out.print("\tCVM_INIT_JAVA_STRING(");
	    out.printHexInt(hashCode);
	    out.write(',');
	    out.printHexInt(s.unicodeOffset);
	    out.write(',');
	    out.printHexInt(string.length());
	    out.print("),");
	    commentary(string, out);
	    internTable.internStringConstant( s );
	}
	out.println("};\n\n");

	return offset;
    }

    public void writeStringInternTable( CCodeWriter out, String tableName, String stringArrayName ){
	internTable.writeInternTable( out, tableName, stringArrayName );
    }
}



class CVMInternSegment {
    int 	     capacity = 2003 /* or use 943 = 41*23*/;
    int		     content = 0;
    boolean	     isFull = false;
    StringConstant   data[] = new StringConstant[capacity];
    CVMInternSegment    next;

    private int maxLoad = (capacity*65) / 100;/* formerly 85% */

    boolean fullEnough(){
	if ( ++content > maxLoad ){
	    isFull = true;
	    if ( next == null ){
		// as we expect...
		next = new CVMInternSegment();
	    }
	}
	return isFull;
    }

    void writeDataArray(
	String dataGenerationName,
	String stringArrayName,
	String stringArrayClassBlockName,
	CCodeWriter out )
    {
	//
	// first typedef the array we're about to create
	out.print("STATIC struct { CVMObjectHeader hdr;\n    CVMJavaInt	length;\n    CVMObjectICell data["+capacity+"]; } ");
	out.print( dataGenerationName );
	out.println( " = {" );
	//
	// fill in the header information
	// hdr.various = {{1}} ??
	int hashCode = 0;
	out.println("	CVM_ROM_OBJECT_HDR_INIT0("+stringArrayClassBlockName+
	    ","+hashCode+"),");
	out.println("	"+capacity+", {");
	//
	// finally write the String reference array.
	for (int i = 0; i < capacity; i++) { 
	    StringConstant sc = data[i];
	    if ( sc == null ){
		out.println("    {0},");
	    } else {
		out.println("    {(CVMObject*)&"+stringArrayName+"["+sc.unicodeIndex+"]},");
	    }
	}
	out.println("}};");
    }

    int writeHeader(
	boolean isStatic,
	String thisname,
	String linkname,
	String dataTargetName,
	CCodeWriter out )
    {
	// now write out the CVMInternSegment struct.
	if( isStatic ){
	    out.print("STATIC const struct java_lang_CVMInternSegment ");
	} else {
	    out.print("const struct java_lang_CVMInternSegment ");
	}
	out.print(thisname);
	// hdr.various = {{0x60}} ??
	int hashCode = 0;
	out.println(" = { CVM_ROM_OBJECT_HDR_INIT(java_lang_CVMInternSegment,"
	    +hashCode+"),\n");
	out.print(capacity);
	out.print(", ");
	out.print(content);
	out.print(", ");
	out.print(isFull?1:0);
	out.print(",\n    {(CVMArrayOfRef*)&");
	out.print(dataTargetName);
	out.print("}, {");
	out.print(linkname);
	out.println("}, "+maxLoad+" };");

	return capacity;
    }

}

class CVMStringIntern{

    CVMInternSegment internTable = new CVMInternSegment();
    CVMInternSegment endTable    = internTable;

    StringConstant internStringConstant( StringConstant val ){
	int h = val.str.string.hashCode();

	CVMInternSegment curseg = internTable;
	do{
	    int h1 = (h&0x7fffffff) % curseg.capacity;
	    int h2 = (h&15)+1;
	    int i = h1;
	    StringConstant candidate;
	    while ( (candidate=curseg.data[i]) != null ){
		if ( val.equals( candidate ) ){
		    return candidate;
		}
		i += h2;
		if ( i >= curseg.capacity ){
		    i -= curseg.capacity;
		}
	    }
	    if ( curseg == endTable ){
		// not found in this table.
		// insert at current location.
		curseg.data[i] = val;
		if ( curseg.fullEnough()){
		    endTable = curseg.next; // which got set as a side effect...
		}
		return val;
	    }
	    curseg = curseg.next;
	}while( true );
    }

    public void writeInternTable(
	CCodeWriter out,
	String tableName,
	String stringArrayName )
    {
	ClassInfo stringArrayClass = ClassInfo.lookupClass("[Ljava/lang/String;");
	if ( stringArrayClass == null ){
	    System.err.println("String writer cannot find array-of-String class");
	    return;
	}
	String stringArrayClassname = ((CVMClass)(stringArrayClass.vmClass)).getNativeName()+"_Classblock";
	// the StringIntern class has no EMV equivalent -- it becomes
	// part of java.lang.String. Only the InternSegment exists.
	// Go write them out.
	int segno = 0;
	String thisTargetName = tableName;
	for ( CVMInternSegment curseg = internTable; curseg != null; curseg = curseg.next ) {
	    /*
	     * Each segment may have two names: the name where it
	     * gets generated, and the name where it will end up.
	     * If they are different, then some copying has to get done
	     * at startup.
	     *
	     * A segment can be readonly only if no thread will ever
	     * lock on it. Locking is only done at insertion. No insertion
	     * can be done if curseg.isFull. Thus if curseg.isFull, then
	     * this can be readonly.
	     */
	    boolean readonlyTarget = curseg.isFull;
	    String thisGenerationName;
	    String dataTargetName = thisTargetName+"_data";
	    String dataGenerationName;
	    if ( readonlyTarget ) {
		thisGenerationName = thisTargetName;
		dataGenerationName = dataTargetName;
	    } else {
		thisGenerationName = thisTargetName+"Master";
		dataGenerationName = dataTargetName+"Master";
	    }
	    String nextTargetName;
	    String nextLink;
	    curseg.writeDataArray( dataGenerationName, stringArrayName, stringArrayClassname, out );

	    if ( curseg.next == null ){
		nextTargetName = nextLink = "0";
	    } else {
		nextTargetName = tableName+segno;
		nextLink = "(CVMObject*)&"+nextTargetName;
		if ( curseg.next.isFull ){
		    out.print("const ");
		}
		out.println("struct java_lang_CVMInternSegment "+nextTargetName+";");
	    }
	    if ( !readonlyTarget ){
		out.println("struct java_lang_CVMInternSegment "+thisTargetName+";");
		out.println("char "+dataTargetName+"[ sizeof "+dataGenerationName+" ];");
	    }

	    curseg.writeHeader( (segno!=0)&!readonlyTarget,
		thisGenerationName, nextLink, 
		dataTargetName, out );
	    if ( !readonlyTarget ){
		// it will have to be moved.
		CVMInitInfo.addInfo("&"+thisGenerationName, "&"+thisTargetName,
		    "sizeof(struct java_lang_CVMInternSegment)" );
		CVMInitInfo.addInfo("&"+dataGenerationName, "&"+dataTargetName,
		    "sizeof("+dataGenerationName+")" );
	    }
	    thisTargetName = nextTargetName;
	    segno += 1;
	}
    }
}
