/*
 * @(#)threads_md.c	1.22 00/11/04
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#define VXWORKS_PRIVATE
#include "javavm/include/porting/threads.h"
#include <vxWorks.h>
#include <taskLib.h>
#include <taskVarLib.h>
#include <fppLib.h>
#include <stdio.h>
#include <assert.h>


/* %comment: mam052 */
#define JAVAPRIO2VXPRIO(p)	(100 - (p))
#define VXPRIO2JAVAPRIO(p)	(100 - (p))

static CVMThreadID *mySelf = NULL;

static void
fpuInit()
{
    FP_CONTEXT fpContext;
    fppSave(&fpContext);
    /* disable floating-point traps */
#if defined(__i386__)
    /* set the fp control word precision to 53-bit for both single and double precision */
    fpContext.fpcr = 0xffff027f; 
    fppRestore(&fpContext);
#elif defined(__sparc__)
    fpContext.fsr &= ~0x0f800000;
    fppRestore(&fpContext);
    fppSave(&fpContext);
#endif
#if 1
    {
	int options = -1;
	int status = taskOptionsGet(0, &options);
	assert(status == OK);
	assert(options & VX_FP_TASK);
    }
#endif
}

static int
start_func(void (*func)(void *), void *arg)
{

    taskVarAdd(0, (int *)&mySelf);

    (*func)(arg);
    return 0;
}

int javaThreadStackSize = 0x10000;

CVMInt32
CVMthreadCreate(CVMSize stackSize, CVMInt16 priority,
    void (*func)(void *), void *arg)
{
    int task_options = VX_FP_TASK;
    int taskID;
    int vxPrio;

    vxPrio = JAVAPRIO2VXPRIO(priority);
    if (stackSize == 0) {
	stackSize = javaThreadStackSize;
    }
    taskID = taskSpawn(NULL, vxPrio, task_options, stackSize,
                            (FUNCPTR)start_func, (int)func,
                            (int)arg,0,0,0,0,0,0,0,0);

    if (taskID == ERROR) {
#ifdef CVM_DEBUG
	perror("taskSpawn");
#endif
	return 0;
    } else {
	return 1;
    }
}

void
CVMthreadYield(void)
{
    taskDelay(0);
}

void
CVMthreadSuspend(CVMThreadID *t)
{
    taskSuspend(t->taskID);
}

void
CVMthreadResume(CVMThreadID *t)
{
    taskResume(t->taskID);
}

void
CVMthreadSetPriority(CVMThreadID *t, CVMInt32 priority)
{
    taskPrioritySet(t->taskID, JAVAPRIO2VXPRIO(priority));
}

CVMThreadID *
CVMthreadSelf()
{
    return mySelf;
}

void
CVMthreadAttach(CVMThreadID *self)
{
    self->taskID = taskIdSelf();

    fpuInit();
    vxworksSyncInit(self);

    mySelf = self;

    {
        TASK_DESC desc;
        taskInfoGet(0, &desc);
        self->stackTop = desc.td_pStackLimit;
    }

}

void
CVMthreadDetach(CVMThreadID *self)
{
    mySelf = NULL;
    vxworksSyncDestroy(self);
}

CVMBool
CVMthreadStackCheck(CVMThreadID *self, CVMUint32 redZone)
{
    return (char *)self->stackTop + redZone < (char *)&self;
}


void
CVMthreadInterruptWait(CVMThreadID *thread)
{
    taskLock();
    thread->interrupted = CVM_TRUE;
    if (thread->in_wait) {
        vxworksSyncInterruptWait(thread);
    }
    taskUnlock();
}

CVMBool
CVMthreadIsInterrupted(CVMThreadID *thread, CVMBool clearInterrupted)
{
    CVMBool wasInterrupted;
    taskLock();
    wasInterrupted = thread->interrupted;
    if (clearInterrupted) {
        thread->interrupted = CVM_FALSE;
    }
    taskUnlock();
    return wasInterrupted;
}

/*
 * Private to porting layer implementation
 */

void
threadInitStaticState()
{
}

void
threadDestroyStaticState()
{
}
