/*
 * @(#)java_props_md.c	1.11 00/11/09
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include <locale.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <ioLib.h>
#include <kernelLib.h>

#include "java_props.h"

static const char *valueOrDefault(const char *v, const char *def)
{
    return v != NULL ? v : def;
}

/* This function gets called very early, before VM_CALLS are setup.
 * Do not use any of the VM_CALLS entries!!!
 */
void
CVMgetJavaProperties(java_props_t *sprops)
{
    const char *v; /* tmp var */

    if (sprops->user_dir) {
        return;
    }

    /* tmp dir */
    sprops->tmp_dir = valueOrDefault(getenv("TMP"), "/tmp");

    /* Printing properties */
#ifdef SOLARIS
    sprops->printerJob = "sun.awt.motif.PSPrinterJob";
#else
    sprops->printerJob = NULL;
#endif

    /* Java 2D properties */
#ifdef SOLARIS
    sprops->graphics_env = "sun.awt.X11GraphicsEnvironment";
#else
    sprops->graphics_env = NULL;
#endif
    sprops->awt_toolkit = NULL;

    v = getenv("JAVA_FONTS");
    /*
    sprops->font_dir = v ? v :
    "/usr/openwin/lib/X11/fonts/Type1:/usr/openwin/lib/X11/fonts/TrueType";
    */

    /* If JAVA_FONTS is not set, the system-dependent fontpath will
     * be evaluated later
     */
    sprops->font_dir = v ? v : "";

    /* os properties */
    {
	sprops->os_name = "VxWorks";
	sprops->os_version = kernelVersion();

#undef I80X86
#undef SIMSPARCSOLARIS
#define STR_I80X86		"x86"
#define STR_SIMSPARCSOLARIS	"sparc"
#define OS_ARCH0(x)	STR_##x
#define OS_ARCH(x)	OS_ARCH0(x)
#ifdef CPU_FAMILY
	sprops->os_arch = OS_ARCH(CPU_FAMILY);
#else
	sprops->os_arch = "Unknown";
#endif
    }

    sprops->language = "en";
    sprops->encoding = "8859_1";
    
#if _BYTE_ORDER == _LITTLE_ENDIAN
    sprops->unicode_encoding = "UnicodeLittle";
#else
    sprops->unicode_encoding = "UnicodeBig";
#endif

    /* user properties */
    {
	sprops->user_name = valueOrDefault(getenv("USER"), "nobody");
	sprops->user_home = valueOrDefault(getenv("HOME"), "/");
    }

    /* User TIMEZONE */
    sprops->timezone = getenv("JAVA_TIMEZONE");

    /* Current directory */
    {
        char buf[MAX_FILENAME_LENGTH];
	char *cwd;

        getcwd(buf, sizeof(buf));

	cwd = malloc(strlen(buf) + 1);
	strcpy(cwd, buf);
	sprops->user_dir = cwd;
    }

    sprops->file_separator = "/";
    sprops->path_separator = ";";
    sprops->line_separator = "\n";

    /*
     * Now for some path information
     * 
     * For the defaults, assume that we are 
     * in <cvmroot>/build/<platform>/bin
     */

    /* System classes */
    sprops->sysclasspath = J2ME_CLASSESJAR;      

    sprops->ext_dirs     = "../extclasses.zip";   /* Extension classes */
    sprops->dll_dir      = "../lib";              /* Where to find libraries */
    sprops->library_path = "../lib";              /* Where to find libraries */
    sprops->java_home    = "..";                  /* Installation directory */
}
