/*
 * @(#)ConnectionBase.java	1.11 00/11/17
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */


package com.sun.cdc.io;

import java.io.*;
import javax.microedition.io.*;

/**
 * Base class for Connection protocols.
 *
 * @author  Nik Shaylor
 * @version 1.1 2/3/2000
 */
abstract public class ConnectionBase extends GeneralBase implements Connection, ConnectionBaseInterface {

    /**
     * Open a connection to a target.
     *
     * @param string           The URL for the connection
     * @param mode             The access mode
     * @param timeouts         A flag to indicate that the called wants timeout exceptions
     *
     * @exception IllegalArgumentException If a parameter is invalid.
     * @exception ConnectionNotFoundException If the connection cannot be found.
     * @exception IOException  If some other kind of I/O error occurs.
     */
    abstract public void open(String name, int mode, boolean timeouts) throws IOException;

    /**
     * Open a connection to a target.
     *
     * @param string           The URL for the connection
     * @param mode             The access mode
     * @param timeouts         A flag to indicate that the called wants timeout exceptions
     * @return                 A new Connection object
     *
     * @exception IllegalArgumentException If a parameter is invalid.
     * @exception ConnectionNotFoundException If the connection cannot be found.
     * @exception IOException  If some other kind of I/O error occurs.
     */
    public Connection openPrim(String name, int mode, boolean timeouts) throws IOException {
        open(name, mode, timeouts);
        return this;
    }

    /**
     * Open and return an input stream for a connection.
     *
     * @return                 An input stream
     * @exception IOException  If an I/O error occurs
     */
    public InputStream openInputStream() throws IOException {
        throw new RuntimeException("No openInputStream");
    }

    /**
     * Open and return an output stream for a connection.
     *
     * @return                 An input stream
     * @exception IOException  If an I/O error occurs
     */
    public OutputStream openOutputStream() throws IOException {
        throw new RuntimeException("No openOutputStream");
    }

    /**
     * Open and return a data input stream for a connection.
     *
     * @return                 An input stream
     * @exception IOException  If an I/O error occurs
     */
    public DataInputStream openDataInputStream() throws IOException {
        InputStream is = openInputStream();
        if(is instanceof DataInputStream) {
            return (DataInputStream)is;
        } else {
            return new DataInputStream(is);
        }
    }

    /**
     * Open and return a data output stream for a connection.
     *
     * @return                 An input stream
     * @exception IOException  If an I/O error occurs
     */
    public DataOutputStream openDataOutputStream() throws IOException {
        OutputStream os = openOutputStream();
        if(os instanceof DataOutputStream) {
            return (DataOutputStream)os;
        } else {
            return new DataOutputStream(os);
        }
    }

}
