/*
 * @(#)string_impl.h	1.3 00/05/22
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifndef INCLUDED_STRING_IMPL_H
#define INCLUDED_STRING_IMPL_H

/*
 * This is the struct declaration for the strings that
 * the ROMizer instantiates.
 */
struct java_lang_String {
	CVMObjectHeader		hdr;
	CVMArrayOfCharICell	value;
	CVMJavaInt		offset;
	CVMJavaInt		count;
};

#define CVM_ROM_STRING_INIT(value, offset, count, hashCode)	\
    {CVM_ROM_OBJECT_HDR_INIT(java_lang_String,hashCode),	\
	value, offset, count}

/*
 * String.intern internal data structure.
 * These are on a linked list, of which the first
 * is (or are) read-only.
 * The header is fixed, but the data following is of
 * variable size, of which "capacity" is the key.
 *
 * capacity is the total number of slots in this segment of
 *   the intern table (currently these are all the same, but need not
 *   be).
 *
 * load is the current number of occupied slots.
 *
 * maxLoad is:
 *  the maximum number of slots we want occupied: currently
 *	about 65% of the capacity for reasonably re-hash
 *	behavour: AND ALSO....
 *   an odd number for pre-loaded segments, or
 *   an even number for dynamically allocated segments
 *   (as a flag for CVMinternDestroy).
 * a refCount value of 255 is an unused entry
 * a refCount value of 254 is "sticky"
 * a refCount value of 253 is a deleted entry
 *
 * 
 * What these really look like in use is:
 * struct CVMInternSegment {
 * 	struct CVMInternSegment ** nextp;
 * 	CVMUint32	capacity;
 * 	CVMUint32	load;
 * 	CVMUint32	maxLoad;
 * 	CVMStringICell	data[capacity];
 * 	CVMUint8	refCount[capacity];
 *	struct CVMInternSegment * next;
 * } CVMInternSegment;
 * except that read-only segments, of course, have a next cell
 * separate from the data.
 *
 * Access to the variable-offset part is through macros.
 */

#define CVM_INTERN_SEGMENT_HEADER \
	struct CVMInternSegment ** nextp; \
	CVMUint32	capacity; \
	CVMUint32	load; \
	CVMUint32	maxLoad;

typedef struct CVMInternSegment {
    CVM_INTERN_SEGMENT_HEADER
    CVMStringICell	data[1];
} CVMInternSegment;

#define CVMInternRefCount(hp) ( (CVMUint8*)(&((hp)->data[(hp)->capacity])) )
#define CVMInternNext(hp) ( *(hp->nextp) )

#define CVMInternUnused		255
#define CVMInternSticky		254
#define CVMInternDeleted	253

#ifndef IN_ROMJAVA
extern const struct CVMInternSegment CVMInternTable;
#endif

#endif
