/*
 * -----------------------------------------------------------------
 * CVM.c :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)CVM.c	1.29 00/11/01
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*
 * All methods in this file use the CNI native method interface. See
 * cjp07/13/99 for details.
 */

/* ROBBINS SEZ: added to make rempote exceptions work */
#define CVM_REMOTE_EXCEPTIONS_SUPPORTED

#include "javavm/include/interpreter.h"
#include "javavm/include/directmem.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/utils.h"
#include "javavm/include/common_exceptions.h"
#include "javavm/include/gc_common.h"
#include "javavm/export/jvm.h"
#include "generated/offsets/java_lang_Thread.h"

CNIResultCode
CNIsun_misc_CVM_checkDebugFlags(CVMExecEnv* ee, CVMStackVal32 *arguments,
				CVMMethodBlock **p_mb)
{
#ifdef CVM_DEBUG
    CVMJavaInt flags = arguments[0].j.i;
    arguments[0].j.i = CVMcheckDebugFlags(flags);
#else
    arguments[0].j.i = 0;
#endif
    return CNI_SINGLE;
}

CNIResultCode
CNIsun_misc_CVM_setDebugFlags(CVMExecEnv* ee, CVMStackVal32 *arguments,
			      CVMMethodBlock **p_mb)
{
#ifdef CVM_DEBUG
    CVMJavaInt flags = arguments[0].j.i;
    arguments[0].j.i = CVMsetDebugFlags(flags);
#else
    arguments[0].j.i = 0;
#endif
    return CNI_SINGLE;
}

CNIResultCode
CNIsun_misc_CVM_clearDebugFlags(CVMExecEnv* ee, CVMStackVal32 *arguments,
				CVMMethodBlock **p_mb)
{
#ifdef CVM_DEBUG
    CVMJavaInt flags = arguments[0].j.i;
    arguments[0].j.i = CVMclearDebugFlags(flags);
#else
    arguments[0].j.i = 0;
#endif
    return CNI_SINGLE;
}

CNIResultCode
CNIsun_misc_CVM_restoreDebugFlags(CVMExecEnv* ee, CVMStackVal32 *arguments,
				  CVMMethodBlock **p_mb)
{
#ifdef CVM_DEBUG
    CVMJavaInt flags = arguments[0].j.i;
    CVMJavaInt oldvalue = arguments[1].j.i;
    arguments[0].j.i = CVMrestoreDebugFlags(flags, oldvalue);
#else
    arguments[0].j.i = 0;
#endif
    return CNI_SINGLE;
}

/*
 * executeClinit is responsible for getting the <clinit> method of the
 * specified class executed. It could do this by just using JNI to
 * invoke the <clinit> method, but this causes undesireable C recursion
 * in the interpreter. Instead we just store the mb of the <clinit>
 * method in *p_mb, and return CVM_NEW_MB to the interpreter. This
 * signals the interpreter to invoke the method stored in *p_mb.
 */
CNIResultCode
CNIsun_misc_CVM_executeClinit(CVMExecEnv* ee, CVMStackVal32 *arguments,
			      CVMMethodBlock **p_mb)
{
    CVMClassBlock* cb = CVMgcUnsafeClassRef2ClassBlock(ee, &arguments[0].j.r);
    CVMMethodBlock* clinitmb;
    CVMD_gcSafeExec(ee, {
	clinitmb = CVMclassGetStaticMethodBlock(cb, CVMglobals.clinitTid);
    });
    CVMtraceClinit(("[Initializing %C]\n", cb));
    if (clinitmb != NULL) {
	CVMtraceClinit(("[Running static initializer for %C]\n", cb));
	/* Return the new mb */
	*p_mb = clinitmb;
	return CNI_NEW_MB;
    } else {
	return CNI_VOID;
    }
}

/*
 * If the class is dynamically loaded and has a <clinit> method, then
 * free up the memory allocated for the <clinit> method. Note the mb,
 * stays around, but all the code and other data located in the jmd
 * is freed up.
 *
 * (Formerly, the entire body of this was brackeketed by the
 * ifdef CVM_CLASSLOADING block. But now that we want to
 * free the stackmaps even for preloaded classes, we need to
 * go through the motions anyway.)
 */

CNIResultCode
CNIsun_misc_CVM_freeClinit(CVMExecEnv* ee, CVMStackVal32 *arguments,
			   CVMMethodBlock **p_mb)
{
    CVMClassBlock* cb = CVMgcUnsafeClassRef2ClassBlock(ee, &arguments[0].j.r);
    CVMMethodBlock* clinitmb;
    CVMJavaMethodDescriptor* jmd;

    CVMD_gcSafeExec(ee, {
	clinitmb = CVMclassGetStaticMethodBlock(cb, CVMglobals.clinitTid);
	if (clinitmb != NULL) {
	    if (CVMmbIsJava(clinitmb)) {
		jmd = CVMmbJmd(clinitmb);
		if ( *CVMjmdStackMaps(jmd) != NULL ){
		    free(*CVMjmdStackMaps(jmd));
		    *CVMjmdStackMaps(jmd) = NULL;
		}
		if ( (!CVMcbIsInROM(cb))
		   || (CVMjmdFlags(jmd)&CVM_JMD_DID_REWRITE) ) {
		    free(jmd);
		    CVMmbJmd(clinitmb) = NULL;
		}
		if ( !CVMcbIsInROM(cb) ){
		    CVMclassFreeLocalVariableTableFieldIDs(ee, clinitmb);
		}
	    }
	}
    });
    return CNI_VOID;
}

#ifdef FOR_EXAMPLE
CNIResultCode
CNIsun_misc_CVM_arraycopy(CVMExecEnv* ee, CVMStackVal32 *arguments,
			  CVMMethodBlock **p_mb)
{
    CVMObjectICell *src = &arguments[0].j.r;
    CVMJavaInt src_position = arguments[1].j.i;
    CVMObjectICell *dst = &arguments[2].j.r;
    CVMJavaInt dst_position = arguments[3].j.i;
    CVMJavaInt length = arguments[4].j.i;

    /* For now, until we have an optimized, unsafe version of arraycopy */
    CVMD_gcSafeExec(ee, {
	JNIEnv *env = CVMexecEnv2JniEnv(ee);
	JVM_ArrayCopy(env, NULL, src, src_position, dst, dst_position, length);
    });

    return CNI_VOID;
}
#endif

CNIResultCode
CNIsun_misc_CVM_disableRemoteExceptions(CVMExecEnv* ee,
					CVMStackVal32 *arguments,
					CVMMethodBlock **p_mb)
{
    CVMdisableRemoteExceptions(ee);
    return CNI_VOID;
}

CNIResultCode
CNIsun_misc_CVM_enableRemoteExceptions(CVMExecEnv* ee,
				       CVMStackVal32 *arguments,
				       CVMMethodBlock **p_mb)
{
    CVMenableRemoteExceptions(ee);
    return CNI_VOID;
}

CNIResultCode
CNIsun_misc_CVM_throwRemoteException(CVMExecEnv* ee,
				     CVMStackVal32 *arguments,
				     CVMMethodBlock **p_mb)
{
#ifdef CVM_REMOTE_EXCEPTIONS_SUPPORTED
    CVMObjectICell *threadICell = &arguments[0].j.r;
    CVMObjectICell *exceptionICell = &arguments[1].j.r;
    CVMJavaLong eetop;
    CVMExecEnv *targetEE;

    CVMD_fieldReadLong(CVMID_icellDirect(ee, threadICell),
		       CVMoffsetOfjava_lang_Thread_eetop,
		       eetop);
    targetEE = (CVMExecEnv *)CVMlong2VoidPtr(eetop);

    CVMgcUnsafeThrowRemoteException(ee, targetEE, 
				    CVMID_icellDirect(ee, exceptionICell));

    /* %comment: rt037 */
#else
    CVMassert(CVM_FALSE);
#endif
    return CNI_VOID;
}

CNIResultCode
CNIsun_misc_CVM_throwLocalException(CVMExecEnv* ee,
				    CVMStackVal32 *arguments,
				    CVMMethodBlock **p_mb)
{
    CVMObjectICell *exceptionICell = &arguments[0].j.r;

    CVMgcUnsafeThrowLocalException(ee, CVMID_icellDirect(ee, exceptionICell));

    return CNI_EXCEPTION;
}

CNIResultCode
CNIsun_misc_CVM_gcDumpHeapSimple(CVMExecEnv* ee,
				 CVMStackVal32 *arguments,
				 CVMMethodBlock **p_mb)
{
#ifdef CVM_DEBUG
    CVMD_gcSafeExec(ee, {
	CVMgcDumpHeapSimple();
    });
#endif

    return CNI_VOID;
}

CNIResultCode
CNIsun_misc_CVM_gcDumpHeapVerbose(CVMExecEnv* ee,
				  CVMStackVal32 *arguments,
				  CVMMethodBlock **p_mb)
{
#ifdef CVM_DEBUG
    CVMD_gcSafeExec(ee, {
	CVMgcDumpHeapVerbose();
    });
#endif

    return CNI_VOID;
}

CNIResultCode
CNIsun_misc_CVM_gcDumpHeapStats(CVMExecEnv* ee,
				CVMStackVal32 *arguments,
				CVMMethodBlock **p_mb)
{
#ifdef CVM_DEBUG
    CVMD_gcSafeExec(ee, {
	CVMgcDumpHeapStats();
    });
#endif

    return CNI_VOID;
}
