/*
 * @(#)classverify.c	1.14 00/09/21
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifdef CVM_CLASSLOADING

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/common_exceptions.h"
#include "javavm/include/typeid.h"
#include "javavm/include/globals.h"
#include "javavm/include/verify.h"
#include "javavm/include/directmem.h"
#include "javavm/export/jvm.h"
#include "javavm/export/jni.h"

CVMBool
CVMclassVerify(CVMExecEnv* ee, CVMClassBlock* cb)
{
#ifdef CVM_TRUSTED_CLASSLOADERS	
    CVMcbSetRuntimeFlag(cb, VERIFIED);
    return CVM_TRUE;
#else
    if (CVMcbCheckRuntimeFlag(cb, VERIFIED)) {
	return CVM_TRUE;
    }
    CVMtraceClassLoading(("CL: Verifying class %C.\n", cb));

    /* Do some quick integrity tests on interfaces. */
    if (CVMcbIs(cb, INTERFACE)) { 
	int i;
	if ((CVMcbSuperclass(cb) == NULL) ||
	    (CVMcbSuperclass(cb) != CVMsystemClass(java_lang_Object))) { 
	        CVMthrowVerifyError(ee, "Interface %C has bad superclass", cb);
		goto failed;
	}
	for (i = 0; i < CVMcbMethodCount(cb); i++) {
	    CVMMethodBlock* mb = CVMcbMethodSlot(cb, i);
	    if (CVMmbIs(mb, STATIC)) {
		if (!CVMtypeidIsStaticInitializer(CVMmbNameAndTypeID(mb))) { 
		    /* Only internal interface methods can be static */
		    CVMthrowVerifyError(
                        ee, "Illegal static method %M in interface %C",
			mb, cb);
		    goto failed;
		}
	    }
	}
    } else if (CVMcbSuperclass(cb) != NULL) { 
	/*EMPTY*/
        /* Ensuring that final classes are not subclassed, and that
	 * final methods are not overridden.
	 * DEFERRED into PrepareMethods
	 */
    } else if (cb != CVMsystemClass(java_lang_Object)) {
	CVMthrowVerifyError(ee, "Class %C does not have superclass", cb);
	goto failed;
    }
	
    {
        CVMBool       result;
        char          message[256];
	message[0] = 0;
	result = 
	    VerifyClass(ee, cb, message, sizeof(message));
	if (!result) {
	    if (!CVMexceptionOccurred(ee)) {
		CVMthrowVerifyError(ee, "%s", message);
	    }
	    goto failed;
	}
    }
    CVMcbSetRuntimeFlag(cb, ee, VERIFIED);
    CVMtraceClassLoading(("CL: Done verifying class %C.\n", cb));
    return CVM_TRUE;

 failed:
    CVMtraceClassLoading(("CL: Failed verifying class %C.\n", cb));
    return CVM_FALSE;
#endif /* CVM_TRUSTED_CLASSLOADERS */
}

#endif /* CVM_CLASSLOADING */
