/*
 * @(#)gen_semispace.c	1.39 00/10/17
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*
 * This file includes the implementation of a copying semispace generation.
 * This generation can act as a young or an old generation.
 */

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/directmem.h"

/*
 * This file is generated from the GC choice given at build time.
 */
#include "generated/javavm/include/gc_config.h"

#include "javavm/include/gc_common.h"
#include "javavm/include/gc/gc_impl.h"

#include "javavm/include/gc/generational/generational.h"

#include "javavm/include/gc/generational/gen_semispace.h"

/*
 * Range checks
 */
static CVMBool
CVMgenSemispaceInSpace(CVMGenSpace* space, CVMObject* ref)
{
    CVMUint32* heapPtr = (CVMUint32*)ref;
    return ((heapPtr >= space->allocBase) &&
	    (heapPtr <  space->allocTop));
}

static CVMBool 
CVMgenSemispaceInGeneration(CVMGeneration* gen,
			    CVMObject* ref)
{
    CVMGenSemispaceGeneration* thisGen = (CVMGenSemispaceGeneration*)gen;
    return CVMgenSemispaceInSpace(thisGen->fromSpace, ref) ||
	CVMgenSemispaceInSpace(thisGen->toSpace, ref);
}

static CVMBool 
CVMgenSemispaceInOld(CVMGenSemispaceGeneration* thisGen, CVMObject* ref)
{
    return CVMgenSemispaceInSpace(thisGen->fromSpace, ref);
}

/*
 * Allocate a semispace for the current semispaces generation
 */
static CVMGenSpace*
CVMgenOneSemispaceAlloc(CVMUint32* heapSpace, CVMUint32 numBytes)
{
    CVMUint32 allocSize = sizeof(CVMGenSpace);
    CVMGenSpace* space = (CVMGenSpace*)calloc(1, allocSize);
    CVMUint32 numWords = numBytes / 4;
    if (space != NULL) {
	space->allocTop  = &heapSpace[numWords];
	space->allocPtr  = &heapSpace[0];
	space->allocBase = &heapSpace[0];
    }
    return space;
}

/*
 * Make a semispace the current semispace of the generation.
 */
static void
CVMgenSemispaceMakeCurrent(CVMGenSemispaceGeneration* thisGen,
			   CVMGenSpace* thisSpace)
{
    CVMGeneration* super = &thisGen->gen;
    super->allocPtr  = thisSpace->allocPtr;
    super->allocBase = thisSpace->allocBase;
    super->allocTop  = thisSpace->allocTop;
    super->allocMark = thisSpace->allocPtr;
}

/* Forward declaration */
static CVMBool
CVMgenSemispaceCollect(CVMGeneration* gen,
		       CVMExecEnv*    ee, 
		       CVMUint32      numBytes, /* collection goal */
		       CVMGCOptions*  gcOpts);

/*
 * Scan objects in contiguous range, without doing any special handling.
 */
static void
scanObjectsInRangePartial(CVMExecEnv* ee, CVMGCOptions* gcOpts,
			  CVMUint32* base, CVMUint32* top,
			  CVMRefCallbackFunc callback, void* callbackData)
{
    CVMUint32* curr = base;
    CVMtraceGcCollect(("GC[SS,%s]: "
		       "Scanning object range (partial) [%x,%x)\n",
		       gcOpts->fullGC ? "full" : "partial",
		       base, top));
    while (curr < top) {
	CVMObject* currObj    = (CVMObject*)curr;
	CVMClassBlock* currCb = CVMobjectGetClass(currObj);
	CVMUint32  objSize    = CVMobjectSizeGivenClass(currObj, currCb);
	CVMobjectWalkRefs(ee, gcOpts, currObj, currCb, {
	    if (*refPtr != 0) {
		(*callback)(refPtr, callbackData);
	    }
	});
	/* iterate */
	curr += objSize / 4;
    }
    CVMassert(curr == top); /* This had better be exact */
}

static void
CVMgenSemispaceUpdateTableForObjectSlot(CVMObject** refPtr,
					void* data)
{
    CVMGenSemispaceGeneration* thisGen = (CVMGenSemispaceGeneration*)data;
    CVMObject* ref = *refPtr;
    if ((ref != NULL) &&
	!CVMgenSemispaceInOld(thisGen, ref) &&
	!CVMobjectIsInROM(ref)) {
	CVMassert(!CVMgenSemispaceInGeneration((CVMGeneration*)thisGen, ref));
	*CARD_TABLE_SLOT_ADDRESS_FOR(refPtr) = CARD_DIRTY_BYTE;
    }
}

static CVMBool
CVMgenSemispaceIsYoung(CVMGenSemispaceGeneration* thisGen)
{
    return thisGen->gen.generationNo == 0;
}

/*
 * Scan objects in contiguous range, and do all special handling as well.
 */
static void
scanObjectsInRangeFull(CVMExecEnv* ee, CVMGCOptions* gcOpts,
		       CVMUint32* base, CVMUint32* top,
		       CVMRefCallbackFunc callback, void* callbackData)
{
    CVMUint32* curr = base;
    CVMtraceGcCollect(("GC[SS,%s]: "
		       "Scanning object range (full) [%x,%x)\n",
		       gcOpts->fullGC ? "full" : "partial",
		       base, top));
    while (curr < top) {
	CVMObject* currObj = (CVMObject*)curr;
	CVMClassBlock* currCb = CVMobjectGetClass(currObj);
	CVMUint32  objSize    = CVMobjectSizeGivenClass(currObj, currCb);
	CVMobjectWalkRefsWithSpecialHandling(ee, gcOpts,
					     currObj,
					     currCb, {
	    if (*refPtr != 0) {
		(*callback)(refPtr, callbackData);
	    }
	}, callback, callbackData);
	/* iterate */
	curr += objSize / 4;
    }
    CVMassert(curr == top); /* This had better be exact */
}

/*
 * Re-build the old-to-young pointer tables after an old generation is
 * done. 
 */
static void
CVMgenSemispaceRebuildBarrierTable(CVMGenSemispaceGeneration* thisGen,
				   CVMExecEnv* ee,
				   CVMGCOptions* gcOpts,
				   CVMUint32* startRange,
				   CVMUint32* endRange)
{
    CVMassert(!CVMgenSemispaceIsYoung(thisGen)); /* Do this only for old */
    CVMtraceGcCollect(("GC[SS,%d,%s]: "
		       "Rebuilding barrier table for range [%x,%x)\n",
		       thisGen->gen.generationNo,
		       gcOpts->fullGC ? "full" : "partial",
		       startRange, endRange));
    /*
     * No need to look at classes here. They are scanned separately in
     * CVMgenScanAllRoots().
     */
    scanObjectsInRangePartial(ee, gcOpts, startRange, endRange,
			      CVMgenSemispaceUpdateTableForObjectSlot,
			      thisGen);
    CVMgenBarrierObjectHeadersUpdate((CVMGeneration*)thisGen,
				     ee, gcOpts, startRange, endRange);
}

static void       
CVMgenSemispaceScanYoungerToOlderPointers(CVMGeneration* gen,
					  CVMExecEnv* ee,
					  CVMGCOptions* gcOpts,
					  CVMRefCallbackFunc callback,
					  void* callbackData)
{
    /*
     * Iterate over the current space objects, and call 'callback' on
     * each reference slot found in there. 'callback' is passed in by
     * an older generation to get young-to-old pointers.  
     *
     * This generation has just been GC'ed, so the space we need to iterate
     * on is 'fromSpace', which has all remaining live pointers to older
     * generations.
     */
    CVMGenSemispaceGeneration* thisGen = (CVMGenSemispaceGeneration*)gen;
    CVMGenSpace* currentSpace = thisGen->fromSpace;
    CVMUint32* startRange     = currentSpace->allocBase;
    CVMUint32* endRange       = currentSpace->allocPtr;
    CVMtraceGcCollect(("GC[SS,%d,%s]: "
		       "Scanning younger to older ptrs [%x,%x)\n",
		       thisGen->gen.generationNo,
		       gcOpts->fullGC ? "full" : "partial",
		       startRange, endRange));
    /*
     * Classes are handled as part of the root scan. Don't worry
     * about them here.
     */
    scanObjectsInRangePartial(ee, gcOpts, startRange, endRange,
			      callback, callbackData);
}

static void       
CVMgenSemispaceScanOlderToYoungerPointers(CVMGeneration* gen,
					  CVMExecEnv* ee,
					  CVMGCOptions* gcOpts,
					  CVMRefCallbackFunc callback,
					  void* callbackData)
{
    CVMtraceGcCollect(("GC[SS,%d,%s]: "
		       "Scanning older to younger ptrs\n",
		       gen->generationNo,
		       gcOpts->fullGC ? "full" : "partial"));
    /*
     * To get all older to younger pointers, traverse the records in
     * the write barrier tables.
     */
    CVMgenBarrierPointersTraverse(gen, ee, gcOpts, callback, callbackData);
}


/*
 * Start and end GC for this generation. 
 */
static void
CVMgenSemispaceStartGC(CVMGeneration* gen,
		       CVMExecEnv* ee, CVMGCOptions* gcOpts)
{
    CVMtraceGcCollect(("GC[SS,%d,%s]: Starting GC\n",
		       gen->generationNo,
		       gcOpts->fullGC ? "full" : "partial"));
}

static void
CVMgenSemispaceEndGC(CVMGeneration* gen,
		     CVMExecEnv* ee, CVMGCOptions* gcOpts)
{
    CVMtraceGcCollect(("GC[SS,%d,%s]: Ending GC\n",
		       gen->generationNo,
		       gcOpts->fullGC ? "full" : "partial"));
}

/*
 * Return remaining amount of memory
 */
static CVMUint32
CVMgenSemispaceFreeMemory(CVMGeneration* gen, CVMExecEnv* ee)
{
    return (CVMUint32)((CVMUint8*)gen->allocTop - (CVMUint8*)gen->allocPtr);
}

/*
 * Return total amount of memory
 */
static CVMUint32
CVMgenSemispaceTotalMemory(CVMGeneration* gen, CVMExecEnv* ee)
{
    return (CVMUint32)((CVMUint8*)gen->allocTop - (CVMUint8*)gen->allocBase);
}

/* Iterate over promoted pointers in a generation. A young space
   collection accummulates promoted objects in an older
   generation. These objects need to be scanned for pointers into
   the young generation.
   
   At the start of a GC, the current allocation pointer of each
   generation is marked. Therefore the promoted objects to scan are those
   starting from 'allocMark' to the current 'allocPtr'.
  */
static void
CVMgenSemispaceScanPromotedPointers(CVMGeneration* gen,
				    CVMExecEnv* ee,
				    CVMGCOptions* gcOpts,
				    CVMRefCallbackFunc callback,
				    void* callbackData)
{
    CVMUint32* startRange = gen->allocMark;
    CVMUint32* endRange   = gen->allocPtr;

    CVMtraceGcCollect(("GC[SS,%s]: "
		       "Scanning promoted ptrs [%x,%x)\n",
		       gcOpts->fullGC ? "full" : "partial",
		       startRange, endRange));
    /* loop, until promoted object scans result in no more promotions */
    /*
     * First clear class marks, since we want to be able to iterate over
     * classes scanned before promotion.
     */
    do {
	if (gcOpts->fullGC) {
	    scanObjectsInRangeFull(ee, gcOpts, startRange, endRange,
				   callback, callbackData);
	} else {
	    scanObjectsInRangePartial(ee, gcOpts, startRange, endRange,
				      callback, callbackData);
	}
	startRange = endRange;
	/* gen->allocPtr may have been incremented */
	endRange = gen->allocPtr; 
    } while (startRange < endRange);

    CVMassert(endRange == gen->allocPtr);
    /*
     * Remember these pointers for further young space collections.
     * Don't forget to clear class marks again, since we need to be
     * able to traverse already scanned classes again, in order to
     * build barrier tables.
     */
    CVMgenSemispaceRebuildBarrierTable((CVMGenSemispaceGeneration*)gen,
				       ee, gcOpts,
				       gen->allocMark, gen->allocPtr);
    gen->allocMark = gen->allocPtr; /* Advance mark for the next batch
				       of promotions */
    /*
     * Make sure that gen->allocPtr and gen->fromSpace->allocPtr are
     * in agreement (i.e. in the same space).
     */
    CVMassert(CVMgenSemispaceInOld((CVMGenSemispaceGeneration*)gen,
				   (CVMObject*)gen->allocPtr) ||
	      (gen->allocPtr == gen->allocTop));
    /*
     * This is the observable fromSpace pointer. We can probably
     * get away with not committing this pointer.
     */
    ((CVMGenSemispaceGeneration*)gen)->fromSpace->allocPtr = gen->allocPtr;
}

/*
 * Copy numBytes from 'source' to 'dest'. Assume word copy, and assume
 * that the copy regions are disjoint.  
 */
static void
CVMgenSemispaceCopyDisjointWords(CVMUint32* dest, CVMUint32* source, 
				 CVMUint32 numBytes)
{
    CVMUint8* sourceEnd = (CVMUint8*)source + numBytes;
    while (source < (CVMUint32*)sourceEnd) {
	*dest++ = *source++;
    }
}

/* Promote 'objectToPromote' into the current generation. Returns
   the new address of the object, or NULL if the promotion failed. */
static CVMObject*
CVMgenSemispacePromoteInto(CVMGeneration* gen,
			   CVMObject* objectToPromote,
			   CVMUint32  objectSize)
{
    CVMObject* dest;
    CVMgenContiguousSpaceAllocate(gen, objectSize, dest);
    if (dest != NULL) {
	CVMgenSemispaceCopyDisjointWords((CVMUint32*)dest,
					 (CVMUint32*)objectToPromote,
					 objectSize);
    }
    CVMtraceGcScan(("GC[SS,%d]: "
		    "Promoted object %x (len %d, class %C), to %x\n",
		    gen->generationNo,
		    objectToPromote, objectSize,
		    CVMobjectGetClass(objectToPromote),
		    dest));
    return dest;
}

static CVMGenSpace*
CVMgenSemispaceGetExtraSpace(CVMGeneration* gen)
{
    CVMGenSemispaceGeneration* thisGen = (CVMGenSemispaceGeneration*)gen;
    /*
     * Make sure it is unused
     */
    CVMassert(gen->allocBase != thisGen->toSpace->allocBase);
    return thisGen->toSpace;
}

/*
 * Allocate a semispaces generation. This includes the actual semispaces
 * as well as all the supporting data structures.
 */
CVMGeneration*
CVMgenSemispaceAlloc(CVMUint32* space, CVMUint32 totalNumBytes)
{
    CVMGenSemispaceGeneration* thisGen  = (CVMGenSemispaceGeneration*)
	calloc(sizeof(CVMGenSemispaceGeneration), 1);
    CVMUint32 numBytes = totalNumBytes / 2; /* Two equal-sized semispaces */
    CVMGenSpace* fromSpace;
    CVMGenSpace* toSpace;

    if (thisGen == NULL) {
	return NULL;
    }
    fromSpace     = CVMgenOneSemispaceAlloc(space, numBytes);
    toSpace       = CVMgenOneSemispaceAlloc(space + numBytes / 4, numBytes);
    
    if ((fromSpace == NULL) || (toSpace == NULL)) {
	free(fromSpace);
	free(toSpace);
	return NULL;
    }

    /*
     * Initialize thisGen. 
     */
    thisGen->fromSpace = fromSpace;
    thisGen->toSpace   = toSpace;

    /*
     * Start the world on fromSpace
     */
    CVMgenSemispaceMakeCurrent(thisGen, fromSpace);

    /* 
     * And finally, set the function pointers for this generation
     */
    thisGen->gen.collect = CVMgenSemispaceCollect;
    thisGen->gen.scanOlderToYoungerPointers =
	CVMgenSemispaceScanOlderToYoungerPointers;
    thisGen->gen.scanYoungerToOlderPointers =
	CVMgenSemispaceScanYoungerToOlderPointers;
    thisGen->gen.promoteInto = 
	CVMgenSemispacePromoteInto;
    thisGen->gen.scanPromotedPointers =
	CVMgenSemispaceScanPromotedPointers;
    thisGen->gen.inGeneration = 
	CVMgenSemispaceInGeneration;
    thisGen->gen.getExtraSpace = 
	CVMgenSemispaceGetExtraSpace;
    thisGen->gen.startGC = CVMgenSemispaceStartGC;
    thisGen->gen.endGC = CVMgenSemispaceEndGC;
    thisGen->gen.freeMemory = CVMgenSemispaceFreeMemory;
    thisGen->gen.totalMemory = CVMgenSemispaceTotalMemory;
    
    CVMdebugPrintf(("GC[SS]: Initialized semi-space gen for generational GC\n"));
    CVMdebugPrintf(("\tSize of *each* semispace in bytes=%d\n"
		  "\tLimits of generation = [0x%x,0x%x)\n" 
		  "\tFirst semispace      = [0x%x,0x%x)\n" 
		  "\tSecond semispace     = [0x%x,0x%x)\n",
		  numBytes, fromSpace->allocBase, toSpace->allocTop,
		  fromSpace->allocBase, fromSpace->allocTop,
		  toSpace->allocBase, toSpace->allocTop));
    return (CVMGeneration*)thisGen;
}

/*
 * Free all the memory associated with the current semispaces generation
 */
void
CVMgenSemispaceFree(CVMGenSemispaceGeneration* thisGen)
{
    free(thisGen->fromSpace);
    free(thisGen->toSpace);

    thisGen->fromSpace = NULL;
    thisGen->toSpace   = NULL;

    free(thisGen);
}

static void
CVMgenSemispaceSetupCopying(CVMGenSemispaceGeneration* thisGen)
{
    CVMUint32 spaceSize =
	thisGen->toSpace->allocTop - thisGen->toSpace->allocBase;
    /* Always copy from 'fromSpace' to 'toSpace' */
    thisGen->copyBase = thisGen->toSpace->allocBase;
    thisGen->copyTop  = thisGen->copyBase;
    thisGen->copyThreshold = thisGen->toSpace->allocTop - spaceSize / 4;
}

/*
 * GC done. Flip the semispaces.
 */
static void
CVMgenSemispaceFlip(CVMGenSemispaceGeneration* thisGen)
{
    CVMGenSpace* newCurrent = thisGen->toSpace;
    thisGen->toSpace = thisGen->fromSpace;
    thisGen->fromSpace = newCurrent;

    newCurrent->allocPtr = thisGen->copyTop;
    CVMgenSemispaceMakeCurrent(thisGen, newCurrent);
}

/*
 * Copy object, leave forwarding pointer behind
 */
static CVMObject*
CVMgenSemispaceForwardOrPromoteObject(CVMGenSemispaceGeneration* thisGen,
				      CVMObject* ref, CVMUint32 classWord)
{
    CVMClassBlock*  objCb      = CVMobjectGetClassFromClassWord(classWord);
    CVMUint32       objSize    = CVMobjectSizeGivenClass(ref, objCb);
    CVMUint32*      copyTop    = thisGen->copyTop;

    CVMObject* ret;
    CVMBool copyIntoToSpace;

    /* 
     * If this is the old generation, copy into to-space unconditionally.
     *
     * If it is not the old generation, try promotions as well.  Copy
     * into to-space if object age below the desired threshold. If
     * above the threshold, try promotion. If promotion fails, copy
     * into to-space anyway.  
     */
    if (!CVMgenSemispaceIsYoung(thisGen)) {
	ret = NULL;
	copyIntoToSpace = CVM_TRUE;
    } else if (CVMobjGcBitsPlusPlusCompare(ref,
					   CVM_GEN_PROMOTION_THRESHOLD)) {
	ret = NULL;
	copyIntoToSpace = CVM_TRUE;
    } else {
	/* Try promotion */
	ret = thisGen->gen.nextGen->promoteInto(thisGen->gen.nextGen,
						ref, objSize);
	if (ret == NULL) {
	    copyIntoToSpace = CVM_TRUE;
	} else {
	    copyIntoToSpace = CVM_FALSE;
	}
    }
    if (copyIntoToSpace) {
	CVMUint32*      newCopyTop = copyTop + objSize / 4;
	CVMtraceGcCollect(("GC[SS,%d]: "
			   "Forwarding object %x (len %d, class %C), to %x\n",
			   thisGen->gen.generationNo,
			   ref, objSize, objCb, thisGen->copyTop));
	CVMgenSemispaceCopyDisjointWords(copyTop, (CVMUint32*)ref, objSize);
	thisGen->copyTop = newCopyTop;
	ret = (CVMObject*)copyTop;
    }
    CVMassert(ret != NULL);
    /*
     * Set forwarding pointer
     */
    {
	CVMUint32 newClassWord = (CVMUint32)ret;
	/* Mark as forwarded */
	CVMobjectSetMarkedOnClassWord(newClassWord);
	/* ... and forward to the new copy */
	CVMobjectSetClassWord(ref, newClassWord);
    }
    
    return ret;
}

/*
 * Gray an object known to be in the old semispace
 */
static void
CVMgenSemispaceGrayObject(CVMGenSemispaceGeneration* thisGen,
			  CVMObject** refPtr)
{
    CVMObject* ref = *refPtr;
    CVMUint32  classWord = CVMobjectGetClassWord(ref);

    CVMtraceGcCollect(("GC[SS,%d]: Graying object %x\n",
		       thisGen->gen.generationNo,
		       ref));
    /*
     * ROM objects should have been filtered out by the time we get here
     */
    CVMassert(!CVMobjectIsInROM(ref));

    /*
     * Also, all objects that don't belong to this generation need to have
     * been filtered out by now.
     */
    CVMassert(CVMgenSemispaceInOld(thisGen, ref));

    /*
     * If the object is in old space it might already be
     * forwarded. Check that. If it has been forwarded, set the slot
     * to refer to the new copy. If it has not been forwarded, copy
     * the object, and set the forwarding pointer.
     */
    if (CVMobjectMarkedOnClassWord(classWord)) {
	/* Get the forwarding pointer */
	*refPtr = (CVMObject*)CVMobjectClearMarkedOnClassWord(classWord);
	CVMtraceGcCollect(("GC[SS,%d]: Object already forwarded: %x -> %x\n",
			   thisGen->gen.generationNo,
			   ref, *refPtr));
    } else {
	*refPtr = CVMgenSemispaceForwardOrPromoteObject(thisGen, ref,
							classWord);
    }
}

/*
 * Gray a slot if it points into this generation.
 */
static void
CVMgenSemispaceFilteredGrayObject(CVMGenSemispaceGeneration* thisGen,
				  CVMObject** refPtr,
				  CVMObject*  ref)
{
    /*
     * Gray only if the slot points to this generation's from space. 
     *
     * Make sure we don't scan roots twice!
     */
    CVMassert(CVMobjectIsInROM(ref) ||
	      !CVMgenSemispaceInGeneration((CVMGeneration*)thisGen, ref) ||
	      CVMgenSemispaceInOld(thisGen, ref));
    if (CVMgenSemispaceInOld(thisGen, ref)) {
	CVMgenSemispaceGrayObject(thisGen, refPtr);
    }
}

/* 
 * A version of the above where multiple scans of the same slot are allowed.
 * Used for CVMgenSemispaceScanTransitively() only.
 */
static void
CVMgenSemispaceFilteredGrayObjectWithMS(CVMGenSemispaceGeneration* thisGen,
					CVMObject** refPtr,
					CVMObject*  ref)
{
    /*
     * Gray only if the slot points to this generation's from space. 
     */
    if (CVMgenSemispaceInOld(thisGen, ref)) {
	CVMgenSemispaceGrayObject(thisGen, refPtr);
    }
}

/*
 * Scan a non-NULL reference into this generation. Make sure to
 * handle the case that the slot has already been scanned -- the
 * root scan should allow for such a case.
 */
static void
CVMgenSemispaceFilteredHandleRoot(CVMObject** refPtr, void* data)
{
    CVMGenSemispaceGeneration* thisGen = (CVMGenSemispaceGeneration*)data;
    CVMObject* ref = *refPtr;
    CVMassert(refPtr != 0);
    CVMassert(ref != 0);

    CVMtraceGcCollect(("GC[SS,%d]: Handling external reference: [%x] = %x\n",
		       thisGen->gen.generationNo,
		       refPtr, ref));
    CVMgenSemispaceFilteredGrayObject(thisGen, refPtr, ref);
}

static void
CVMgenSemispaceBlackenObjectPartial(CVMGenSemispaceGeneration* thisGen,
				    CVMExecEnv* ee, 
				    CVMGCOptions* gcOpts, CVMObject* ref,
				    CVMClassBlock* refCb)
{
    CVMassert(ref != 0);

    CVMtraceGcCollect(("GC[SS,%d,partial]: Blacken object %x\n",
		       thisGen->gen.generationNo,
		       ref));
    /*
     * Queue up all non-null object references. Handle the class as well.
     */
    CVMobjectWalkRefs(ee, gcOpts, ref, refCb, {
	if (*refPtr != 0) {
	    CVMgenSemispaceFilteredGrayObject(thisGen, refPtr, *refPtr);
	}
    });
}

static void
CVMgenSemispaceBlackenObjectFull(CVMGenSemispaceGeneration* thisGen,
				 CVMExecEnv* ee, 
				 CVMGCOptions* gcOpts, CVMObject* ref,
				 CVMClassBlock* refCb)
{
    CVMassert(ref != 0);

    CVMtraceGcCollect(("GC[SS,%d,full]: Blacken object %x\n",
		       thisGen->gen.generationNo,
		       ref));
    /*
     * Queue up all non-null object references. Handle the class as well.
     */
    CVMobjectWalkRefsWithSpecialHandling(ee, gcOpts, ref, refCb, {
	if (*refPtr != 0) {
	    CVMgenSemispaceFilteredGrayObject(thisGen, refPtr, *refPtr);
	}
    }, CVMgenSemispaceFilteredHandleRoot, thisGen);
}

/*
 * A function to blacken an object. Makes distinction between
 * "full" and "partial" blackening easier.
 */
typedef void (*CVMGenBlackenFunction)(CVMGenSemispaceGeneration* thisGen,
				      CVMExecEnv* ee, 
				      CVMGCOptions* gcOpts, CVMObject* ref,
				      CVMClassBlock* refCb);

static void
CVMgenSemispaceFollowRootsWithBlackener(CVMGenSemispaceGeneration* thisGen,
					CVMExecEnv* ee, CVMGCOptions* gcOpts,
					CVMGenBlackenFunction blackener)
{
    CVMUint32* copyBase = thisGen->copyBase;
    CVMBool    isYoung = CVMgenSemispaceIsYoung(thisGen);
    CVMtraceGcCollect(("GC[SS,%d,%s]: Following roots, "
		       "copyBase=%x, copyTop=%x\n",
		       thisGen->gen.generationNo,
		       (blackener == CVMgenSemispaceBlackenObjectPartial) ?
		           "partial" : "full",
		       thisGen->copyBase, thisGen->copyTop));
    do {
	/*
	 * First traverse the gray objects and blacken them
	 */
	while (copyBase < thisGen->copyTop) {
	    CVMObject* obj = (CVMObject*)copyBase;
	    CVMClassBlock* objCb = CVMobjectGetClass(obj);
	    CVMUint32 objSize = CVMobjectSizeGivenClass(obj, objCb);
	    if (blackener == CVMgenSemispaceBlackenObjectPartial) {
		CVMgenSemispaceBlackenObjectPartial(thisGen, ee, gcOpts, 
						    obj, objCb);
	    } else {
		CVMgenSemispaceBlackenObjectFull(thisGen, ee, gcOpts, 
						 obj, objCb);
	    }
	    copyBase += objSize / 4;
	}
	if (isYoung) {
	    CVMGeneration* nextGen = thisGen->gen.nextGen;
	    if (nextGen->allocPtr > nextGen->allocMark) {
		/* For a young generation, the scan above may have
		   promoted objects. Ask the next generation to scan
		   those objects for pointers into the current */
		nextGen->scanPromotedPointers(nextGen,
		    ee, gcOpts, CVMgenSemispaceFilteredHandleRoot, thisGen);
		/* And of course, scanning the promoted pointers gives
		   us more gray objects to traverse and or more more
		   promoted pointers to scan. Do it. */
	    }
	}
    } while (copyBase < thisGen->copyTop);

    thisGen->copyBase = copyBase;
    CVMassert(thisGen->copyBase == thisGen->copyTop);
}

static void
CVMgenSemispaceFollowRootsFull(CVMGenSemispaceGeneration* thisGen,
			   CVMExecEnv* ee, CVMGCOptions* gcOpts)
{
    CVMgenSemispaceFollowRootsWithBlackener(thisGen, ee, gcOpts,
        CVMgenSemispaceBlackenObjectFull);
}

static void
CVMgenSemispaceFollowRootsPartial(CVMGenSemispaceGeneration* thisGen,
			   CVMExecEnv* ee, CVMGCOptions* gcOpts)
{
    CVMgenSemispaceFollowRootsWithBlackener(thisGen, ee, gcOpts,
        CVMgenSemispaceBlackenObjectPartial);
}

typedef struct CVMGenSemispaceTransitiveScanData {
    CVMExecEnv* ee;
    CVMGCOptions* gcOpts;
    CVMGenSemispaceGeneration* thisGen;
} CVMGenSemispaceTransitiveScanData;

static void
CVMgenSemispaceScanTransitively(CVMObject** refPtr, void* data)
{
    CVMGenSemispaceTransitiveScanData* tsd =
	(CVMGenSemispaceTransitiveScanData*)data;
    CVMGenSemispaceGeneration* thisGen = tsd->thisGen;
    CVMObject* ref;

    CVMassert(refPtr != 0);
    ref = *refPtr;
    CVMassert(ref != 0);

    CVMgenSemispaceFilteredGrayObjectWithMS(thisGen, refPtr, ref);

    /*
     * Now scan all its children as well
     */
    CVMassert(tsd->gcOpts->fullGC);
    CVMgenSemispaceFollowRootsFull(thisGen, tsd->ee, tsd->gcOpts);
}


/*
 * Test whether a given reference is live or dying. If the reference
 * does not point to the current generation, answer TRUE. The generation
 * that does contain the pointer will give the right answer.
 */
static CVMBool
CVMgenSemispaceRefIsLive(CVMObject** refPtr, void* data)
{
    CVMGenSemispaceGeneration* thisGen = (CVMGenSemispaceGeneration*)data;
    CVMObject* ref;

    CVMassert(refPtr != NULL);

    ref = *refPtr;
    CVMassert(ref != NULL);

    /*
     * ROM objects are always live
     */
    if (CVMobjectIsInROM(ref)) {
	return CVM_TRUE;
    }
    /*
     * If this reference is not in this generation, just assume
     * it's live. 
     */
    if (!CVMgenSemispaceInGeneration((CVMGeneration*)thisGen, ref)) {
	return CVM_TRUE;
    }
    return !CVMgenSemispaceInOld(thisGen, ref) || CVMobjectMarked(ref);
}

static CVMBool
CVMgenSemispaceCollectPartial(CVMGeneration* gen,
			      CVMExecEnv*    ee, 
			      CVMUint32      numBytes, /* collection goal */
			      CVMGCOptions*  gcOpts
			      )
{
    CVMGenSemispaceGeneration* thisGen = (CVMGenSemispaceGeneration*)gen;
    CVMGenSpace* space;
    CVMBool success;

    CVMassert(!gcOpts->fullGC);
    CVMtraceGcStartStop(("GC[SS,%d,partial]: Collecting from 0x%p to 0x%p\n",
			 thisGen->gen.generationNo,
			 thisGen->fromSpace->allocBase,
			 thisGen->toSpace->allocBase));

    CVMgenSemispaceSetupCopying(thisGen);

    /*
     * Scan roots that point to this generation.
     */
    CVMgenScanAllRoots((CVMGeneration*)thisGen,
		       ee, gcOpts, CVMgenSemispaceFilteredHandleRoot, thisGen);

    /*
     * Now start from the roots, and copy or promote all live data 
     */
    CVMgenSemispaceFollowRootsPartial(thisGen, ee, gcOpts);

    /*
     * All relevant live data has been copied to toSpace. Flip the spaces.
     */
    CVMgenSemispaceFlip(thisGen);

    /*
     * Now if we are not the young generation, we have to rebuild the
     * barrier structures.
     */
    space = thisGen->fromSpace;
    if (!CVMgenSemispaceIsYoung(thisGen)) {
	CVMgenClearBarrierTable();
	CVMgenSemispaceRebuildBarrierTable(thisGen, ee, gcOpts,
					   space->allocBase,
					   space->allocPtr);
    }
    success = (numBytes / 4 <= space->allocTop - space->allocPtr);
    CVMtraceGcStartStop(("GC[SS,%d,partial]: Done, success for %d bytes %s\n",
			 thisGen->gen.generationNo,
			 numBytes,
			 success ? "TRUE" : "FALSE"));

    /*
     * Have we processed everything?
     */
    CVMassert((gen->nextGen == NULL) ||
	      (gen->nextGen->allocMark == gen->nextGen->allocPtr));

    /*
     * Can we allocate in this space after this GC?
     */
    return success;
}

static CVMBool
CVMgenSemispaceCollectFull(CVMGeneration* gen,
			   CVMExecEnv*    ee, 
			   CVMUint32      numBytes, /* collection goal */
			   CVMGCOptions*  gcOpts
			   )
{
    CVMGenSemispaceTransitiveScanData tsd;
    CVMGenSemispaceGeneration* thisGen = (CVMGenSemispaceGeneration*)gen;
    CVMGenSpace* space;
    CVMBool success;

    CVMassert(gcOpts->fullGC);
    gcOpts->discoverWeakReferences = CVM_TRUE;

    CVMtraceGcStartStop(("GC[SS,%d,full]: Collecting from 0x%p to 0x%p\n",
			 thisGen->gen.generationNo,
			 thisGen->fromSpace->allocBase, 
			 thisGen->toSpace->allocBase));

    CVMgenSemispaceSetupCopying(thisGen);

    /*
     * Scan all roots that point to this generation.
     */
    CVMgenScanAllRoots((CVMGeneration*)thisGen,
		       ee, gcOpts, CVMgenSemispaceFilteredHandleRoot, thisGen);

    /*
     * Now start from the roots, and copy or promote all live data 
     */
    CVMgenSemispaceFollowRootsFull(thisGen, ee, gcOpts);

    /*
     * Don't discover any more weak references.
     */
    gcOpts->discoverWeakReferences = CVM_FALSE;

    /*
     * At this point, we know which objects are live and which are not.
     * Do the special objects processing.
     */
    tsd.ee = ee;
    tsd.gcOpts = gcOpts;
    tsd.thisGen = thisGen;

    CVMgcProcessSpecialWithLivenessInfo(ee, gcOpts,
					CVMgenSemispaceRefIsLive, thisGen,
					CVMgenSemispaceScanTransitively, &tsd);

    /*
     * All relevant live data has been copied to toSpace. Flip the spaces.
     */
    CVMgenSemispaceFlip(thisGen);

    /*
     * Now if we are not the young generation, we have to rebuild the
     * barrier structures.
     */
    space = thisGen->fromSpace;
    if (!CVMgenSemispaceIsYoung(thisGen)) {
	CVMgenClearBarrierTable();
	CVMgenSemispaceRebuildBarrierTable(thisGen, ee, gcOpts,
					   space->allocBase,
					   space->allocPtr);
    }
    success = (numBytes / 4 <= space->allocTop - space->allocPtr);
    CVMtraceGcStartStop(("GC[SS,%d,full]: Done, success for %d bytes %s\n",
			 thisGen->gen.generationNo,
			 numBytes,
			 success ? "TRUE" : "FALSE"));

    /*
     * Have we processed everything?
     */
    CVMassert((gen->nextGen == NULL) ||
	      (gen->nextGen->allocMark == gen->nextGen->allocPtr));

    /*
     * Can we allocate in this space after this GC?
     */
    return success;
}

static CVMBool
CVMgenSemispaceCollect(CVMGeneration* gen,
		       CVMExecEnv*    ee, 
		       CVMUint32      numBytes, /* collection goal */
		       CVMGCOptions*  gcOpts)
{
    if (gcOpts->fullGC) {
	return CVMgenSemispaceCollectFull(gen, ee, numBytes, gcOpts);
    } else {
	return CVMgenSemispaceCollectPartial(gen, ee, numBytes, gcOpts);
    }
}


