/*
 * @(#)globalroots.c	1.21 00/05/19
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/stacks.h"
#include "javavm/include/globals.h"
#include "javavm/include/globalroots.h"

/*
 * Allocate  a global root from the specified stack. Returns NULL
 * on failure. No exception ever thrown.
 */
static CVMObjectICell*
CVMIDprivate_getGlobalRootNoLock(CVMExecEnv* ee, CVMStack* grStack)
{
    CVMStackVal32* result = NULL;
    CVMGlobalRootsFrame* singletonFrame =
	(CVMGlobalRootsFrame*)grStack->currentFrame;

    CVMframeAlloc(ee, grStack, singletonFrame, result);
    
    return &result->ref;
}

static CVMObjectICell*
CVMIDprivate_getGlobalRoot(CVMExecEnv* ee, CVMStack* grStack,
			   CVMSysMutex* lock)
{
    CVMObjectICell* result;
    CVMsysMutexLock(ee, lock);
    result = CVMIDprivate_getGlobalRootNoLock(ee, grStack);
    CVMsysMutexUnlock(ee, lock);
    
    return result;
}

CVMObjectICell*
CVMID_getGlobalRoot(CVMExecEnv* ee)
{
    return CVMIDprivate_getGlobalRoot(ee, &CVMglobals.globalRoots,
				      &CVMglobals.globalRootsLock);
}

CVMObjectICell*
CVMID_getClassGlobalRoot(CVMExecEnv* ee)
{
    return CVMIDprivate_getGlobalRoot(ee, &CVMglobals.classGlobalRoots,
				      &CVMglobals.classTableLock);
}

CVMObjectICell*
CVMID_getClassLoaderGlobalRoot(CVMExecEnv* ee)
{
    return CVMIDprivate_getGlobalRoot(ee,
				      &CVMglobals.classLoaderGlobalRoots,
				      &CVMglobals.classTableLock);
}

CVMObjectICell*
CVMID_getWeakGlobalRoot(CVMExecEnv* ee)
{
    return CVMIDprivate_getGlobalRoot(ee, &CVMglobals.weakGlobalRoots,
				      &CVMglobals.weakGlobalRootsLock);
}

static void
CVMIDprivate_freeGlobalRootNoLock(CVMExecEnv* ee, CVMObjectICell* glRoot,
				  CVMStack* grStack)
{
    CVMGlobalRootsFrame* singletonFrame =
	(CVMGlobalRootsFrame*)grStack->currentFrame;
    CVMStackVal32* refToFree = (CVMStackVal32*)glRoot;

    CVMframeFree(singletonFrame, refToFree);
}

static void
CVMIDprivate_freeGlobalRoot(CVMExecEnv* ee, CVMObjectICell* glRoot,
			    CVMStack* grStack, CVMSysMutex* lock)
{
    CVMsysMutexLock(ee, lock);
    CVMIDprivate_freeGlobalRootNoLock(ee, glRoot, grStack);
    CVMsysMutexUnlock(ee, lock);
}

void
CVMID_freeClassGlobalRoot(CVMExecEnv* ee, CVMObjectICell* glRoot)
{
    CVMIDprivate_freeGlobalRoot(ee, glRoot,
				&CVMglobals.classGlobalRoots,
				&CVMglobals.classTableLock);
}

void
CVMID_freeClassLoaderGlobalRoot(CVMExecEnv* ee, CVMObjectICell* glRoot)
{
    CVMIDprivate_freeGlobalRoot(ee, glRoot,
				&CVMglobals.classLoaderGlobalRoots,
				&CVMglobals.classTableLock);
}

void
CVMID_freeGlobalRoot(CVMExecEnv* ee, CVMObjectICell* glRoot)
{
    CVMIDprivate_freeGlobalRoot(ee, glRoot, &CVMglobals.globalRoots,
				&CVMglobals.globalRootsLock);
}

void
CVMID_freeWeakGlobalRoot(CVMExecEnv* ee, CVMObjectICell* glRoot)
{
    CVMIDprivate_freeGlobalRoot(ee, glRoot, &CVMglobals.weakGlobalRoots,
				&CVMglobals.weakGlobalRootsLock);
}

void
CVMglobalrootFrameScanner(CVMExecEnv* ee,
			  CVMFrame* grFrame, CVMStackChunk* grChunk,
			  CVMRefCallbackFunc callback, void* data,
			  CVMGCOptions* opts)
{
    CVMwalkRefsInGCRootFrame(grFrame,
			     ((CVMFreelistFrame*)grFrame)->vals,
			     grChunk, callback, data, CVM_TRUE);
}
