/*
 * -----------------------------------------------------------------
 *  interpreter.c :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)interpreter.c	1.288 00/11/16
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/* ROBBINS SEZ: added to make rempote exceptions work */
#define CVM_REMOTE_EXCEPTIONS_SUPPORTED

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/sync.h"
#include "javavm/include/globals.h"
#include "javavm/include/stacks.h"
#include "javavm/include/stackwalk.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/jni_impl.h"
#include "javavm/include/localroots.h"
#include "javavm/include/globalroots.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/gc_common.h"
#include "javavm/include/signature.h"
#include "javavm/include/preloader.h"
#include "javavm/include/basictypes.h"
#include "javavm/include/stackmaps.h"
#include "javavm/include/bcattr.h"
#include "javavm/include/common_exceptions.h"
#include "generated/javavm/include/opcodes.h"

#include "generated/offsets/java_lang_Throwable.h"
#ifdef CVM_CLASSLOADING
#include "generated/offsets/java_lang_ClassLoader.h"
#endif

#include "javavm/include/clib.h"
#include "javavm/include/porting/threads.h"
#include "javavm/include/porting/int.h"
#include "javavm/include/porting/system.h"

#ifdef CVM_CLASSLOADING
#include "javavm/include/porting/linker.h"
#endif

#ifdef CVM_REFLECT
#include "javavm/include/reflect.h"
#endif

#ifdef CVM_JVMDI
#include "javavm/include/jvmdi_impl.h"
#endif

/* For jtHeapMemoryArea object */
#include "jvmcei.h"
#include "javavm/include/mem_registry.h"
/* - Pratik */

/* Capacities for the interpreter stack. */
#define CVM_MIN_INTERPRETER_STACKCHUNK_SIZE 1024
#define CVM_MAX_INTERPRETER_STACK_SIZE      \
    100 * CVM_MIN_INTERPRETER_STACKCHUNK_SIZE
#define CVM_INITIAL_INTERPRETER_STACK_SIZE   \
    CVM_MIN_INTERPRETER_STACKCHUNK_SIZE

/* Capacities for the thread's local roots stack. */
#define CVM_MIN_LOCALROOTS_STACKCHUNK_SIZE 20
#define CVM_MAX_LOCALROOTS_STACK_SIZE      \
    3 * CVM_MIN_LOCALROOTS_STACKCHUNK_SIZE
#define CVM_INITIAL_LOCALROOTS_STACK_SIZE   \
    CVM_MIN_LOCALROOTS_STACKCHUNK_SIZE

static CVMBool
CVMarrayIsAssignable(CVMExecEnv* ee, 
		     CVMClassBlock* srcArrayCb,
		     CVMClassBlock* dstArrayCb);


#if defined(CVM_DEBUG) || defined(CVM_DEBUG_DUMPSTACK) || defined(CVM_DEBUG_STACKTRACES)
static char*
CVMaddstr(const char *s, char* buf, char* limit, char term)
{
    char c;
    while ((c = *s) && c != term && buf < limit) {
	*buf++ = c;
	s++;
    }
    return buf;
}
#endif /* CVM_DEBUG */


#if defined(CVM_DEBUG) || defined(CVM_DEBUG_DUMPSTACK)
static char*
CVMadddec(CVMInt32 n, char* buf, char* limit)
{ 
    char dec[11];
    int i = sizeof(dec) - 1;
    dec[i] = 0;

    if (n < 0) {
        n = -n;
        if (buf < limit)
            *buf++ = '-';
    }

    while (n != 0) {
        i--;
        dec[i] = (char)((n % 10) + '0');
        n = n / 10;
    }
    return CVMaddstr(&dec[i], buf, limit, 0);
}

#endif /* CVM_DEBUG */

#ifdef CVM_DEBUG_CLASSINFO
CVMInt32
CVMpc2lineno(CVMMethodBlock *mb, CVMUint16 pc_offset)
{
    CVMUint32 length;

    CVMassert(!CVMmbIs(mb, NATIVE));

    length = CVMmbLineNumberTableLength(mb);
    if (length > 0) {
	CVMLineNumberEntry *ln = CVMmbLineNumberTable(mb);
	CVMUint16 l = 0;
	CVMUint16 u = length;
	if (pc_offset < ln[l].startpc) { 
	    return -1;
	}
	if (pc_offset >= ln[u - 1].startpc) {
	    return ln[u - 1].lineNumber;
	}
	while (l < u) {
	    int m = (l + u) >> 1;
	    if (pc_offset < ln[m].startpc) {
	        u = m;
	    } else if (pc_offset >= ln[m + 1].startpc) {
	        l = m;
	    } else {
	        /* ln[m].startpc <= pc_offset < ln[m + 1].startpc */ 
	        return ln[m].lineNumber;
	    }
	}
	CVMassert(CVM_FALSE);
    }
    return -1;
}
#endif

#ifdef CVM_DEBUG_STACKTRACES
CVMMethodBlock* 
CVMpc2method(CVMClassBlock* cb, CVMUint8* pc)
{
    CVMMethodBlock* mb = NULL;
    int idx = 0;
    while (idx < CVMcbMethodCount(cb)) {
	mb = CVMcbMethodSlot(cb, idx);
	if (CVMmbIs(mb, NATIVE)) {
	    if (pc == CVMmbNativeCode(mb)) {
		return mb;
	    }
	} else if (!CVMmbIs(mb, ABSTRACT)) {
	    /* We need to check for NULL in case the mb is for a freed
	     * up <clinit> method. */
	    if (CVMmbJmd(mb) != NULL) {
		CVMUint8* codeStart = (CVMUint8*)CVMmbJavaCode(mb);
		if (pc >= codeStart && pc < codeStart + CVMmbCodeLength(mb)) {
		    return mb;
		}
	    }
	}
	idx++;
    }
    CVMassert(CVM_FALSE);
    return mb;  /* avoid a crash by returing last mb */
}
#endif

/*
 * Convert the current pc to a string in the format:
 *    <classname>.<methodname>(Native Method)
 *        or
 *    <classname>.<methodname>(<sourcefilename>:<linenumber>)
 *
 * pc can be specified by either passing in the frame (in which case mb
 * and pc should be NULL) or the <mb,pc> pair, in which case frame
 * should be NULL.
 */
#if defined(CVM_DEBUG) || defined(CVM_DEBUG_DUMPSTACK) || defined(CVM_DEBUG_STACKTRACES)
void
CVMpc2string(CVMFrame* frame, CVMMethodBlock* mb, CVMUint8* pc,
	     char *buf, char* limit)
{
    CVMClassBlock*  cb;

    CVMassert(buf < limit);
    CVMassert(mb != 0 || (frame != NULL && frame->mb != NULL));

    if (mb == NULL)
	mb = frame->mb;
    cb = CVMmbClassBlock(mb);

    --limit;	/* Save room for terminating '\0' */
    buf += CVMformatString(buf, limit - buf, "%C.%M", cb, mb);
    if (CVMmbIs(mb, NATIVE)) {
	buf = CVMaddstr("(Native Method)", buf, limit, 0);
    } else if (frame != NULL && CVMframeIsTransition(frame)) {
	buf = CVMaddstr("(Transition Method)", buf, limit, 0);
    } else {
#ifdef CVM_DEBUG_CLASSINFO
	const char* fn = CVMcbSourceFileName(cb);
	if (pc == NULL) {
	    pc = CVMframePc(frame);
	}
	if (fn != NULL) {
	    CVMInt32    lineno;

	    /* Just want the short name */
	    fn = strrchr(fn, '/');
	    if (fn) {
		/* skip past last separator */
		fn += 1;
	    } else {
		fn = CVMcbSourceFileName(cb);
	    }
		
	    buf = CVMaddstr("(", buf, limit, 0);
	    buf = CVMaddstr(fn, buf, limit, 0);
	    lineno = CVMpc2lineno(mb, pc - CVMmbJavaCode(mb));
	    if (lineno >= 0) {
		buf = CVMaddstr(":", buf, limit, 0);
		buf = CVMadddec(lineno, buf, limit);
	    }
	    buf = CVMaddstr(")", buf, limit, 0);
	} else {
	    buf = CVMaddstr("(Unknown Source)", buf, limit, 0);
	}
#else
	buf = CVMaddstr("(Unknown Source)", buf, limit, 0);
#endif /* CVM_DEBUG_CLASSINFO */
    }
    *buf = 0;
    return;
}
#endif /* CVM_DEBUG */

CVMBool 
CVMinitExecEnv(CVMExecEnv* ee, CVMBool userThread)
{
    CVMBool result;

    /*
     * Unless we are prepared to zero out every field by hand, then
     * we have to do this...
     */
    memset(ee, 0, sizeof *ee);

    CVMCstackBufferInit(ee);	/* Unset EE-buffer flag */

    ee->userThread = userThread;

    /* Need to initialize consistent states before anything else
     * or some debugging asserts may get upset.
     */
    {
	int i;

	for (i = 0; i < CVM_NUM_CONSISTENT_STATES; ++i) {
	    CVMtcsInit(CVM_TCSTATE(ee, i));
	}
    }

#ifdef CVM_DEBUG
    /* Also make sure this EE doesn't think we hold any locks. */
    ee->sysLocks = NULL;
#endif

    /* The first frame has to be a JNI frame so we can allocate
     * jni local refs from it.
     */
    result = CVMinitStack(ee, &ee->interpreterStack, 
			  CVM_INITIAL_INTERPRETER_STACK_SIZE,
			  CVM_MAX_INTERPRETER_STACK_SIZE, 
			  CVM_MIN_INTERPRETER_STACKCHUNK_SIZE,
			  CVMJNIFrameCapacity,
			  CVMjniFrameScanner);
    if (!result) {
	return CVM_FALSE; /* exception NOT thrown */
    }

    CVMjniFrameInit(CVMgetJNIFrame(ee->interpreterStack.currentFrame));


    result = CVMinitStack(ee, &ee->localRootsStack, 
			  CVM_INITIAL_LOCALROOTS_STACK_SIZE,
			  CVM_MAX_LOCALROOTS_STACK_SIZE, 
			  CVM_MIN_LOCALROOTS_STACKCHUNK_SIZE,
			  CVMbaseFrameCapacity,
			  CVMlocalrootFrameScanner);
    if (!result) {
	return CVM_FALSE; /* exception NOT thrown */
    }

    CVMinitJNIEnv(&ee->jniEnv);

    /* Allocate a JNI local ref for the ee's exception holders
     */
    CVMlocalExceptionICell(ee)    = CVMjniCreateLocalRef(ee);	/* null ref */
#ifdef CVM_REMOTE_EXCEPTIONS_SUPPORTED
    CVMremoteExceptionICell(ee)   = CVMjniCreateLocalRef(ee);	/* null ref */
#endif
    CVMcurrentExceptionICell(ee)  = CVMjniCreateLocalRef(ee);	/* null ref */
    CVMmiscICell(ee)              = CVMjniCreateLocalRef(ee);	/* null ref */
    CVMfinalizerRegisterICell(ee) = CVMjniCreateLocalRef(ee);	/* null ref */
    CVMmemoryAreaICell(ee)        = CVMjniCreateLocalRef(ee);	/* null ref (Pratik) */
    ee->perThreadStack = NULL;

    /*This  frame supposed to be the first frame of the context -Kalpesh*/
    ee->interpreterStack.currentFrame->ATCflags.val.depth=0;
    /* temporary -Kalpesh */
    ee->isIOBlocked = CVM_FALSE;
    ee->dirtyAie = 0;
  //  ee->pid = 0;

    
   /* ROBBINS SEZ: 
     * init the ATC-deferred data
     */
    CVMpendingAIExceptionICell(ee) = CVMjniCreateLocalRef(ee); /* null ref */
    CVMclearPendingAIExceptionFlag(ee);

    /*
     * Clear flags and objects (which should already be null)
     */
    CVMclearLocalException(ee);
    CVMclearRemoteException(ee);
    
    CVMcurrentThreadICell(ee) = CVMjniCreateLocalRef(ee);

    /* attach thread */
    CVMthreadAttach(CVMexecEnv2threadID(ee));
    ee->threadState = CVM_THREAD_RUNNING;
    /* Added by Kunal - make isNoHeapRealtimeThread = CVM_FALSE*/
    /* changed following line to use macro. - Pratik */
    CVMunsetNoHeapRealtimeThread(ee);
    
    /* Disable constraints checking. It will be anabled only for
     * RealtimeThreads
     * -Pratik
     */
    CVMdisableCheckForConstraints(ee);
    
    /*
     * Link it in.
     */

    CVMsysMutexLock(ee, &CVMglobals.threadLock);
    {
	CVMExecEnv **firstPtr = &CVMglobals.threadList;
	ee->prevEEPtr = firstPtr;
	ee->nextEE = *firstPtr;
	if (*firstPtr != 0) {
	    (*firstPtr)->prevEEPtr = &ee->nextEE;
	}
	*firstPtr= ee;
#ifdef CVM_DEBUG
	ee->threadID = CVMglobals.threadIDCount;
	++CVMglobals.threadIDCount;
#endif
    }
    
    /* Initialise memoryArea object */
    //First check if filters are loaded
    {
      extern intFuncPtr JVMCeiFilters[];
      extern int JVMCeiInitialized;

      if(JVMCeiInitialized && JVMCeiFilters[JVMCEF_INIT_EXECENV] != NULL)
	  {
	    JVMCeiFilters[JVMCEF_INIT_EXECENV]((void*) ee);
	  }
    }
   
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);

    CVMeeSyncInit(ee);
   
	return CVM_TRUE;
}

void 
CVMdestroyExecEnv(CVMExecEnv* ee)
{
#ifdef CVM_DEBUG
    CVMassert(ee->sysLocks == NULL);
#endif

    CVMeeSyncDestroy(ee);

    /*
     * Unlink it.  After this point, GC will no longer
     * scan this thread.
     */
    CVMsysMutexLock(ee, &CVMglobals.threadLock);
    *ee->prevEEPtr = ee->nextEE;
    if (ee->nextEE != 0) {
	ee->nextEE->prevEEPtr = ee->prevEEPtr;
    }
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);

    {
	int i;

	for (i = 0; i < CVM_NUM_CONSISTENT_STATES; ++i) {
	    CVMtcsDestroy(CVM_TCSTATE(ee, i));
	}
    }

    {
	/* %comment dlong001 */
#ifdef CVM_DEBUG
	CVMlocalExceptionICell(ee) = NULL;
	CVMremoteExceptionICell(ee) = NULL;
	CVMcurrentExceptionICell(ee) = NULL;
	CVMmiscICell(ee) = NULL;
	CVMfinalizerRegisterICell(ee) = NULL;
	CVMcurrentThreadICell(ee) = NULL;
	CVMmemoryAreaICell(ee) = NULL;   // - Pratik.
#endif
    }

	if(ee->perThreadStack != NULL)
	{
	  extern intFuncPtr JVMCeiFilters[];
	  JVMCeiFilters[JVMCEF_DESTROY_EXEC_ENV](ee);
	}
    
	CVMdestroyStack(&ee->interpreterStack);
    CVMdestroyStack(&ee->localRootsStack);

    CVMdestroyJNIEnv(&ee->jniEnv);

    CVMCstackFreeBuffer(ee); /* free the EE buffer used by C stack IO operations */

    /* detach thread */
    CVMthreadDetach(CVMexecEnv2threadID(ee));

    CVMsysMutexLock(ee, &CVMglobals.threadLock);
    CVMassert(CVMglobals.threadCount > 0);
    --CVMglobals.threadCount;
    if (ee->userThread) {
	CVMassert(CVMglobals.userThreadCount > 0);
	--CVMglobals.userThreadCount;
    }
    CVMcondvarNotify(&CVMglobals.threadCountCV);
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);

    /*
     * The thread is no longer allowed to access VM data structures
     * after this point.
     */

#ifdef CVM_DEBUG
    CVMassert(ee->sysLocks == NULL);
#endif
}

CVMExecEnv *
CVMgetEE()
{
    CVMThreadID *self = CVMthreadSelf();
    if (self == NULL) {
	return NULL;
    } else {
	return CVMthreadID2ExecEnv(self);
    }
}

/*
 * Scan reference-typed arguments by looking at signature.
 */
static void
CVMinterpreterIterateRefArguments(CVMMethodBlock* mb, CVMStackVal32* list,
				  CVMRefCallbackFunc callback, void* data)
{
    CVMMethodTypeID  tid = CVMmbNameAndTypeID(mb);
    CVMterseSigIterator terseSig;
    char       arg;

    /* 
     * The terse signature is all we need (and even less than that)
     */
    CVMtypeidGetTerseSignatureIterator( tid, &terseSig );
    if (!CVMmbIs(mb, STATIC)) {
	CVMObject** refPtr = (CVMObject**)list;
	CVMassert(*refPtr != 0);
	callback(refPtr, data);
	list++;
    }	
    while((arg = CVM_TERSE_ITER_NEXT(terseSig)) != CVM_TYPEID_ENDFUNC) {
	if (arg == CVM_TYPEID_OBJ) {
	    CVMObject** refPtr = (CVMObject**)list;
	    if (*refPtr != 0) {
		callback(refPtr, data);
	    }
	}
	list++;
    }
}

void
CVMtransitionFrameScanner(CVMExecEnv* ee,
			  CVMFrame* frame, CVMStackChunk* chunk,
			  CVMRefCallbackFunc callback, void* data,
			  CVMGCOptions* opts)
{
    CVMassert(CVMframeIsTransition(frame));

    CVMtraceGcScan(("Scanning frame for %C.%M ",
		    CVMmbClassBlock(frame->mb), frame->mb));

    /*
     * If the transition frame was marked as one that returns a
     * ref and we are at the opc_exitinterpreter instruction,
     * then the result is our responsiblity and not that of the
     * java method that pushed the result. There are three possible
     * states the frame can be in:
     */
    if 	(CVMframePc(frame) != NULL &&
	  *CVMframePc(frame) == opc_exittransition) {
	/* 
	 * (1) we've caught the small window between the callee
	 * having completed the return, but the caller hasn't
	 * fetched the result and popped the transition frame yet.
	 * Scan the result if it's a reference.
	 * 
	 * This is covering the checkpoint at opc_?return (handle_return:).
	 */
	if (CVMtypeidGetReturnType(CVMmbNameAndTypeID(frame->mb)) ==
	    CVM_TYPEID_OBJ)
	{
	    CVMObject** refPtr = (CVMObject**)&frame->topOfStack[-1];
	    CVMtraceGcScan(("(transition, ref result)\n"));
	    if (*refPtr != 0) {
		(*callback)(refPtr, data);
	    }
	} else {
	    CVMtraceGcScan(("(transition, no ref result)\n"));
	}
    } else if (frame->topOfStack == 
	       (CVMStackVal32*)CVMframeOpstack(frame, Transition)) {
	/* 
	 * (2) We've caught the big window between the callee having taken
	 * responsibility for the arguments, and the callee having completed
	 * the return, so there is nothing to scan. This only happens when 
	 * the transition frame was used to invoke a java method from jni.
	 * When a transtion frame is used to invoke a jni method from jni,
	 * the arguments are always scannable.
	 */
	CVMtraceGcScan(("(transition, no scan needed)\n"));
	return;
    } else {
	/* 
	 * (3) If the invoke opcode in the transition frame had to
	 * expand the stack, you would hit this one. You could also
	 * hit it if after having pushed a transition frame and added
	 * a large number of arguments, and then. Lastly, there is a
	 * small gcsafe section in CVMjniInvoke() that occurs after
	 * the transition frame is pushed and before we become
	 * gcUnsafe to push the arguments which will also cause you
	 * to end up here. Scan all the arguments.
	 */
	CVMStackVal32* argsStart = 
	    (CVMStackVal32*)CVMframeOpstack(frame, Transition);
	CVMtraceGcScan(("(transition, scanning arguments)\n"));
	CVMinterpreterIterateRefArguments(frame->mb, argsStart,
					  callback, data);
    }
}

/*
 * Find the innermost exception handler for a PC, for stackmap purposes only.
 */
static CVMUint8*
CVMfindInnermostHandlerFor(CVMJavaMethodDescriptor* jmd, CVMUint8* pc)
{
    CVMExceptionHandler *eh = CVMjmdExceptionTable(jmd);
    CVMExceptionHandler *ehEnd = eh + CVMjmdExceptionTableLength(jmd);
    CVMUint8* codeBegin = CVMjmdCode(jmd);
    CVMUint16 relPC = pc - codeBegin;
	
    /* Check each exception handler to see if the pc is in its range */
    for (; eh < ehEnd; eh++) { 
	if (eh->startpc <= relPC && relPC < eh->endpc) { 
	    return codeBegin + eh->handlerpc;
	}
    }
    return 0;
}

/*
 * Get the stackmap entry for a frame. If (*missingStackMapOK), a NULL entry
 * is normal.
 */
static CVMStackMapEntry*
getStackmapEntry(CVMFrame* frame, CVMJavaMethodDescriptor* jmd,
		 CVMStackMaps* stackmaps, CVMBool* missingStackmapOK)
{
    CVMStackMapEntry* smEntry;
    /*
     * The stackmaps have been created. Now let's check for the entry.
     */
    smEntry = CVMstackmapGetEntryForPC(stackmaps,
				       CVMframePc(frame) - CVMjmdCode(jmd));
    if (smEntry == NULL) {
	CVMUint8* handlerPC;
	if (frame->topOfStack != (CVMStackVal32*)CVMframeOpstack(frame, Java)){
	    *missingStackmapOK = CVM_FALSE;
	    /* Not throwing an exception; therefore no entry */
	    return NULL;
	} else {
	    handlerPC = CVMfindInnermostHandlerFor(jmd, CVMframePc(frame));
	    if (handlerPC == NULL) {
		/* Not handled here -- OK to be missing stackmap then, since
		   we are going to be abandoning this frame anyway */
		*missingStackmapOK = CVM_TRUE;
		return NULL;
	    } else {
		/* The handler is here. Now we must make sure that we
		   find a stackmap for 'handlerPC'. */
		*missingStackmapOK = CVM_FALSE;
	    }
	    smEntry = CVMstackmapGetEntryForPC(stackmaps, 
					       handlerPC - CVMjmdCode(jmd));
	}
    }
    return smEntry;
}

/*
 * Ensure space for and perform stackmap allocations for a given frame.
 * This is called at the onset of GC _before_ CVMgcimplDoGC() is called.
 * If this returns an error, we don't even call CVMgcimplDoGC(), so it
 * does not have to deal with the error. This can be thought of as a
 * pre-verification pass to ensure that the GC stackwalk will never have
 * to encounter a stackmap creation failure.
 *
 * CVM_TRUE  if stackmaps successfully created
 * CVM_FALSE on error
 */
CVMBool
CVMjavaFrameEnsureStackmaps(CVMExecEnv* ee, CVMFrame* frame)
{
    CVMJavaMethodDescriptor* jmd;
    CVMStackMaps*    stackmaps;
    CVMUint32        pcOffset;

    CVMassert(CVMframeIsJava(frame));

    jmd = CVMmbJmd(frame->mb);

    pcOffset = CVMframePc(frame) - CVMjmdCode(jmd);

    CVMtraceGcScan(("Ensuring stackmaps generation in frame for %C.%M (maxLocals=%d, locals=0x%x, stack=0x%x, pc=%d[0x%x])\n",
		    CVMmbClassBlock(frame->mb), frame->mb,
		    CVMmbMaxLocals(frame->mb),
		    CVMframeLocals(frame),
		    CVMframeOpstack(frame, Java),
		    pcOffset,
		    CVMjmdCode(jmd)));

    stackmaps = *CVMjmdStackMaps(jmd);
    if (stackmaps == NULL) {
	/* Compute stackmaps, do not consider conditional GC points */
	if (CVMstackmapCompute(frame->mb, CVM_FALSE)) {
	    stackmaps = *CVMjmdStackMaps(jmd);
	}
    } 

    if (stackmaps == NULL) {
	return CVM_FALSE; /* Could not create stackmaps */
    } else {
	CVMBool missingStackmapOK;
	CVMStackMapEntry* smEntry;
	CVMOpcode opcode;

	smEntry = getStackmapEntry(frame, jmd, stackmaps, &missingStackmapOK);
	if (smEntry != NULL) {
	    return CVM_TRUE; /* Found an entry; all is fine */
	}
	if (missingStackmapOK) {
	    /* It is OK to not have an entry only if the frame
	       was throwing an exception and it is not handled in this
	       frame. */
	    return CVM_TRUE;
	}
	opcode = *CVMframePc(frame);
	if (!CVMbcAttr(opcode, COND_GCPOINT)) {
	    /* This can't happen! */
	    CVMassert(CVM_FALSE);
	    return CVM_FALSE;
	} else {
	    /* Now, if there is no entry, and we are at a
	       conditional GC point, re-create stackmaps, and
	       re-store them */
	    CVMStackMaps* oldStackmaps = *CVMjmdStackMaps(jmd);
	    *CVMjmdStackMaps(jmd) = NULL;
	    free(oldStackmaps);
	    
	    CVMstackmapCompute(frame->mb, CVM_TRUE);
	    stackmaps = *CVMjmdStackMaps(jmd);
	    if (stackmaps == NULL) {
		/* The new stackmaps could not be created: fail */
		return CVM_FALSE;
	    }
	    smEntry = getStackmapEntry(frame, jmd, 
				       stackmaps, &missingStackmapOK);
	    CVMassert(smEntry != NULL);
	    return CVM_TRUE;
	}
    }
}

/*
 * Scan a JavaFrame found on the interpreter stack.
 */
void
CVMjavaFrameScanner(CVMExecEnv* ee,
		    CVMFrame* frame, CVMStackChunk* chunk,
		    CVMRefCallbackFunc callback, void* data,
		    CVMGCOptions* gcOpts)
{
    CVMJavaMethodDescriptor* jmd;

    CVMStackVal32*   scanPtr;
    CVMObject**      receiverObjPtr;
    CVMStackMaps*    stackmaps;
    CVMStackMapEntry* smEntry;
    CVMUint16*       smChunk;
    CVMUint16        bitmap;
    CVMUint32        bitNo, i;
    CVMUint32        noSlotsToScan;
    CVMUint32        pcOffset;
    CVMBool          missingStackmapOK;

    CVMassert(CVMframeIsJava(frame));

    jmd = CVMmbJmd(frame->mb);

    CVMassert(CVMframePc(frame) >= CVMjmdCode(jmd));

    pcOffset = CVMframePc(frame) - CVMjmdCode(jmd);

    CVMassert(pcOffset < CVMjmdCodeLength(jmd));

    CVMtraceGcScan(("Scanning frame for %C.%M (maxLocals=%d, locals=0x%x, stack=0x%x, pc=%d[0x%x])\n",
		    CVMmbClassBlock(frame->mb), frame->mb,
		    CVMmbMaxLocals(frame->mb),
		    CVMframeLocals(frame),
		    CVMframeOpstack(frame, Java),
		    pcOffset,
		    CVMjmdCode(jmd)));

    stackmaps = *CVMjmdStackMaps(jmd);

    /* A previous pass ensures that the stackmaps are indeed generated. */
    CVMassert(stackmaps != NULL);

    smEntry = getStackmapEntry(frame, jmd, stackmaps, &missingStackmapOK);
    CVMassert((smEntry != NULL) || missingStackmapOK);
    if (smEntry == NULL) {
	goto done;
    }

    /*
     * Get ready to read the stackmap data
     */
    smChunk = &smEntry->state[0];
    bitmap = *smChunk;

    bitmap >>= 1; /* Skip flag */
    bitNo  = 1;

    /*
     * Scan locals
     */
    scanPtr = (CVMStackVal32*)CVMframeLocals(frame);
    noSlotsToScan = CVMjmdMaxLocals(jmd);

    for (i = 0; i < noSlotsToScan; i++) {
	if ((bitmap & 1) != 0) {
	    CVMObject** slot = (CVMObject**)scanPtr;
	    if (*slot != 0) {
		callback(slot, data);
	    }
	}
	scanPtr++;
	bitmap >>= 1;
	bitNo++;
	if (bitNo == 16) {
	    bitNo = 0;
	    smChunk++;
	    bitmap = *smChunk;
	}
    }

    scanPtr = (CVMStackVal32*)CVMframeOpstack(frame, Java);
    noSlotsToScan = frame->topOfStack - scanPtr;

    /*
     * The stackmaps for the variables and the operand stack are
     * consecutive. Keep bitmap, bitNo, smChunk as is.
     */
    for (i = 0; i < noSlotsToScan; i++) {
	if ((bitmap & 1) != 0) {
	    CVMObject** slot = (CVMObject**)scanPtr;
	    if (*slot != 0) {
		callback(slot, data);
	    }
	}
	scanPtr++;
	bitmap >>= 1;
	bitNo++;
	if (bitNo == 16) {
	    bitNo = 0;
	    smChunk++;
	    bitmap = *smChunk;
	}
    }

    /*
     * The class of the method executing. Do this only if not all classes
     * are roots.
     */
    if (!gcOpts->allClassesAreRoots) {
	CVMClassBlock* cb = CVMmbClassBlock(frame->mb);
	CVMscanClassIfNeeded(ee, cb, callback, data);
    }
    
done:

    /*
     * And last but not least, javaframe->receiverObj
     */
    receiverObjPtr = (CVMObject**)&CVMframeReceiverObj(frame);
    if (*receiverObjPtr != 0) {
	callback(receiverObjPtr, data);
    }
}

/*
 * Scan a free-list frame.
 */
void
CVMfreelistFrameScanner(CVMExecEnv* ee,
			CVMFrame* frame, CVMStackChunk* chunk,
		        CVMRefCallbackFunc callback, void* data,
			CVMGCOptions* opts)
{
    CVMassert(CVMframeIsFreelist(frame));
    /*
     * Walk the refs in this free-listing root frame.
     */
    CVMwalkRefsInGCRootFrame(frame,
			     ((CVMFreelistFrame*)frame)->vals,
			     chunk, callback, data, CVM_TRUE);
}

/*
 * Return the pc of the exception handler for the specified exception.
 */
/* NOTE: this is needed by JVMDI, so was made nonstatic */
CVMUint8*
CVMgcSafeFindPCForException(CVMExecEnv* ee, CVMFrame* frame, 
			    CVMClassBlock* exceptionClass, CVMUint8* pc)
{
    CVMMethodBlock* mb = frame->mb;

    CVMassert(CVMD_isgcSafe(ee));
    CVMassert(mb != NULL);

    {
	/* Search current class for catch frame, based on pc value,
	   matching specified type. */
	CVMJavaMethodDescriptor* jmd = CVMmbJmd(mb);
	CVMExceptionHandler *eh = CVMjmdExceptionTable(jmd);
	CVMExceptionHandler *ehEnd = eh + CVMjmdExceptionTableLength(jmd);
	CVMUint16 pcTarget = 
	    pc - CVMjmdCode(jmd); /* offset from start of code */
	CVMConstantPool* cp = CVMframeCp(frame);
	
	/* Check each exception handler to see if the pc is in its range */
	for (; eh < ehEnd; eh++) { 
	    if (eh->startpc <= pcTarget && pcTarget < eh->endpc) { 
		/* The pc is in the handlers range. Now we need to make
		 * sure it handles the specified exception.
		 */
		CVMUint16 catchtype = eh->catchtype;
		CVMClassBlock* catchClass;

		/* Special note put in for finally.  Always succeed. */
		if (catchtype == 0)
		    goto found;

		/* See if the class of the exception that the handler catches
		 * is the object's class or one of its superclasses.
		 */
		
#ifdef CVM_CLASSLOADING
		/* Make sure the class of the catchtype is resolved */
		if (!CVMcpResolveEntryFromClass(ee, CVMmbClassBlock(mb),
						cp, catchtype)) {
			return 0;
		}
#endif
		catchClass = CVMcpGetCb(cp, catchtype);

		/* check the objects class and each superclass for a match */
		{
		    CVMClassBlock* cb;
		    for (cb = exceptionClass; cb != 0;
			 cb = CVMcbSuperclass(cb)) {
			if (cb == catchClass)
			    goto found;
		    }
		}
	    }
	}
	/* not found in this execution environment */
	return 0;
    found:
	return CVMjmdCode(jmd) + eh->handlerpc;
    }
}

/*
 * Handle the current outstanding exception in the ee. Return the frame
 * that handles the exception after updating the frame's pc.
 *
 * Arguments:
 *    ee:           ee of the thread that needs an exception handled.
 *    pc:           current pc for thread.
 *    topOfStack:   current topOfStack for thread.
 *    frame:        frame to start searching for a handler.
 *    initialframe: last frame in which a handler should be checked for.
 *    exceptionObj: place in which to return the exception object.
 */
CVMFrame*
CVMgcUnsafeHandleException(CVMExecEnv* ee, CVMFrame* frame,
			   CVMFrame* initialframe)
{
    CVMBool AIExceptionThrown = CVM_FALSE;

    CVMStack*        stack = &ee->interpreterStack;
    CVMClassBlock*   exceptionClass = NULL;

    CVMassert(CVMD_isgcUnsafe(ee));
    // printf("[CVMgcUnsafeHandleException] 1\n");
 new_exception_thrown:

    /* retrieve the exception object */

    /* ROBBINS SEZ: 
     * This is the new exception retrieving code 
     * TAKE NOTE: it gives precedence to local and remote
     * exceptions over AI excepions, this distinction need
     * only be made in a race however,so it dosn't really
     * matter.
     */
    if (CVMlocalExceptionOccurred(ee)) {
	CVMsetCurrentExceptionObj(ee, CVMgetLocalExceptionObj(ee));
	CVMclearLocalException(ee);
    } else if(CVMremoteExceptionOccurred(ee)){
        CVMsetCurrentExceptionObj(ee, CVMgetRemoteExceptionObj(ee));
        CVMclearRemoteException(ee);
    } else {
      /* Must be an AI Exception */
      
      /* we aquire a mutex on the current thread object so that */
      /* posting threads can't change the exception data on us */
      CVMgcUnsafeLockThreadMutex(ee,{},{
	/* failure action */
	CVMthrowOutOfMemoryError(ee,NULL);
	goto new_exception_thrown;
      });

      AIExceptionThrown = CVM_TRUE;

      /* set the exc env's current exception obj which is used
       * for locating the correct catch block.
       */
      CVMsetCurrentExceptionObj(ee, CVMgetPendingAIExceptionObj(ee));      

      /* release the thread mutex */
      CVMgcUnsafeUnlockThreadMutex(ee,{},{
	/* failure action */
	CVMthrowIllegalMonitorStateException(ee, "current thread not owner");
	goto new_exception_thrown;
      });

      /* But that's not all folks, we also have to clear the interrupted status
       * of this thread. As part of a last-minute clarification to the spec, 
       * RealtimeThread's interrupt() method sets the interrupted status, which
       * is cleared at the time of propagation
       */
      CVMthreadIsInterrupted(CVMexecEnv2threadID(ee),CVM_TRUE);

    }
    
    /* get the cb for the exception object */
    exceptionClass = CVMobjectGetClass(CVMgetCurrentExceptionObj(ee));

    
    /* Search the java stack for the frame that handles the
     * exception. We only search the frames that this invocation
     * of CVMunsafeExecuteJavaMethod pushed. Otherwise we just put the
     * exception back in the ee return to the caller .
     */
    while (frame != initialframe) {
	CVMUint8* pc = CVMframePc(frame);


	/* Skip transition frames since they can't catch exceptions. */
	/* ROBBINS SEZ: also if propigation an aie, skip interruptible frames */
	if ((!CVMframeIsTransition(frame)) && 
	    ((AIExceptionThrown==CVM_TRUE?
	      CVMATCPermission(frame)!=NONDEFERRED:CVM_TRUE))){
	    
	  /* If we're here because of an AIE in the first place, we don't bother,
	   * If, the frame is AI suceptable and there is a pending AIE, (this
	   * is the same check that's in CHECK_FOR_REMOTE_AND_AI_EXCEPTIONS in
	   * executejava.c).
	   */
	  
	  if(!AIExceptionThrown){
	    if((CVMATCPermission(frame)==NONDEFERRED) &&
	       (CVMpendingAIException(ee)))
	    {
	      CVMdebugPrintf(("AI EXCEPTION PENDING IN RAISED NON-DEFERED FRAME\n"));
	      /*we go trough similar motions as the CHECK macro*/
	      
	      /* we aquire a mutex on the current thread object so that */
	      /* posting threads can't change the exception data on us */
	      CVMgcUnsafeLockThreadMutex(ee,{},{
		/* failure action */
		CVMthrowOutOfMemoryError(ee,NULL);
		goto new_exception_thrown;
	      });
	      /* we now have a lock on the thread object*/	      

	      CVMsetCurrentExceptionObj(ee, CVMgetPendingAIExceptionObj(ee));      
	      exceptionClass = CVMobjectGetClass(CVMgetCurrentExceptionObj(ee));

	      /* okay, we're done, release the mutex */
	      CVMgcUnsafeUnlockThreadMutex(ee,{},{
		/* failure action */
		CVMthrowIllegalMonitorStateException(ee, "current thread not owner");
		goto new_exception_thrown;
	      });
	    }
	  }
	      
	  CVMD_gcSafeExec(ee, {
	    pc = CVMgcSafeFindPCForException(ee, frame, 
					     exceptionClass, pc);
	  });
	  
	    /* If a pc was returned, then that's the pc of the code
	     * that will handle the exception.
	     */
	  if (pc != 0) {
	    CVMframePc(frame) = pc;
	    CVMtraceExceptions(("<%d> Exception caught:     %P\n",
				ee->threadID, frame));
	    break;
	  }
	}

	/* Looks like this frame doesn't handle the exception.
	 * Pop it and setup state for looking in the next frame.
	 * We also need to unlock the receiverObj if the method
	 * was synchronized. 
	 */
	CVMtraceExceptions(("<%d> Exception not caught: %P\n",
			    ee->threadID, frame));
	if (CVMframeIsJava(frame) && CVMmbIs(frame->mb, SYNCHRONIZED)) {
            CVMObjectICell *objICell = &CVMframeReceiverObj(frame);
            CVMBool success;
            /* %comment lamm003 */
            success = CVMobjectTryUnlock(ee, CVMID_icellDirect(ee, objICell));
            if (!success) {
                /* NOTE: CVMobjectUnlock may become safe. */
                success = CVMobjectUnlock(ee, objICell);
            }
	    if (!success) {
		CVMthrowIllegalMonitorStateException(
                    ee, "current thread not owner");
	    }
	}
#ifdef CVM_JVMDI
	/** The JVMDI spec is unclear about whether the frame
	    pop event is sent just before or just after the
	    frame is popped. The JDK implementation sends the
	    event just before the frame is popped. */
	/** Bugid 4191820 seems to imply that the callback must
	    be done before the frame is popped, and that no
	    local variable slots should have been overwritten
	    by the return value. */
	if (CVMglobals.jvmdiDebuggingEnabled) {
	    CVMD_gcSafeExec(ee, {
		CVMjvmdiNotifyDebuggerOfFramePop(ee);
	    });
	}
#endif
	CVMpopFrame(stack, frame);
	stack->currentFrame = frame;       /* decache new current frame */

	/*
	 * CVMgcSafeFindPCForException() may have thrown an  exception,
	 * so we need forget about the original exception and start 
	 * working on the new one.
	 *
	 * We also need to modify here to handle AIE if special case(not sure) -Kalpesh
	 */
	if (CVMlocalExceptionOccurred(ee)) {
	    goto new_exception_thrown;
	}
    }
	   
    /* If we didn't find a handler for the exception, then we need
     * to rethrow the exception. Whoever called CVMunsafeExecuteJavaMethod
     * will handle it.
     */
    if (frame == initialframe) {
      printf("no frame found, re-throwing exception\n");
	CVMgcUnsafeThrowLocalException(ee, CVMgetCurrentExceptionObj(ee));
	CVMsetCurrentExceptionObj(ee, NULL);
    }

    return frame;
}

/*
 * CVMsignalError - signal an exception. The exception message can
 * be a format string followed by arguments just like printf.
 */

static void
CVMgcSafeSignalError(CVMExecEnv* ee, CVMClassBlock* exceptionCb,
		     const char* emessage);

void
CVMsignalError(CVMExecEnv* ee, CVMClassBlock* exceptionCb,
	       const char *format, ...)
{
    va_list argList;
    va_start(argList, format);
    CVMsignalErrorVaList(ee, exceptionCb, format, argList);
    va_end(argList);
}

void
CVMsignalErrorVaList(CVMExecEnv* ee, CVMClassBlock* exceptionCb, 
		     const char* format, va_list ap)
{
#undef CVM_ERR_STRING_BUF_SIZE
#define CVM_ERR_STRING_BUF_SIZE 256
    char *emessageBuf = NULL; 
    char *emessage = NULL; 
    // printf("[CVMsignalErrorVaList] 1\n");
    /* C stack check */
    if (!CVMCstackCheckSize(ee, CVM_REDZONE_CVMsignalErrorVaList,
                            "CVMsignalErrorVaList", CVM_TRUE)) {
        return;
    }
    // printf("[CVMsignalErrorVaList] 2\n");

    if (format != NULL) {
        emessageBuf = malloc(CVM_ERR_STRING_BUF_SIZE); 
        if (emessageBuf == NULL) 
	    emessage = "<exception message unavailable due to low memory>";
        else { 
	    emessage = emessageBuf;
	    CVMformatStringVaList(emessageBuf, CVM_ERR_STRING_BUF_SIZE, 
	        format, ap);
	}
    }
    // printf("[CVMsignalErrorVaList] 3\n");

    if (CVMD_isgcUnsafe(ee)) {
	CVMD_gcSafeExec(ee, {
	    CVMgcSafeSignalError(ee, exceptionCb, emessage);
	});
    } else {
	CVMgcSafeSignalError(ee, exceptionCb, emessage);
    }
    if (emessageBuf != NULL)
        free(emessageBuf);
    // printf("[CVMsignalErrorVaList] 4\n");

}

static void
CVMgcSafeSignalError(CVMExecEnv* ee, CVMClassBlock* exceptionCb,
		     const char* emessage)
{
    CVMassert(CVMD_isgcSafe(ee));
    if (CVMlocalExceptionOccurred(ee)) 
      {
	printf("[CVMgcSafeSignalError] msg(%s)\n",emessage);
	if (strcmp("Maximum Concurrent Locks Exceeded",emessage) == 0)
	  {
	    printf("[CVMgcSafeSignalError] clearing exception \n");
	    CVMclearLocalException(ee);
	  }
      }

    CVMassert(!CVMlocalExceptionOccurred(ee));


    CVMtraceExceptions(("[<%d> Signaling Error: %C, \"%s\"]\n", 
			ee->threadID, exceptionCb,
			emessage ? emessage : ""));

#ifdef CVM_DEBUG
    if (CVMcheckDebugFlags(CVM_DEBUGFLAG(TRACE_EXCEPTIONS))) {
	CVMdumpStack(&ee->interpreterStack, CVM_FALSE, CVM_FALSE, 100);
    }
#endif

    /*
     * Instantiate and throw the requested excepton object.
     */
    CVMID_localrootBegin(ee) {
	CVMID_localrootDeclare(CVMThrowableICell, exceptionICell);

	/*
	 * Instantiate the exception object. If we can't then 
	 * throw a preallocated object.
	 */
	CVMID_allocNewInstance(ee, exceptionCb, exceptionICell);
	if (CVMID_icellIsNull(exceptionICell)) {
	    CVMdebugPrintf(("Could not allocate exception object \"%C\"; "
			    "throwing a pre-allocated "
			    "OutOfMemoryError object instead\n",
			    exceptionCb));
	    CVMgcSafeThrowLocalException(
                ee, CVMglobals.preallocatedOutOfMemoryError);
	    goto localRootEnd;
	}

	/* 
	 * If a message string was passed in, then allocate a String object
	 * for it and assign it to the Throwable.msg field. We don't care if
	 * CVMnewStringUTF fails. The msg field just ends up being null
	 * in that case.
	 */
	if (emessage) {
	    CVMID_localrootDeclare(CVMStringICell, stringICell);
	    CVMnewStringUTF(ee, stringICell, emessage);
	    CVMID_fieldWriteRef(ee, exceptionICell, 
				CVMoffsetOfjava_lang_Throwable_detailMessage,
				stringICell);
	}

	/* Fill in the backtrace into the exception object. */
	CVMfillInStackTrace(ee, exceptionICell);

	/* Throw the exception object. */
   	if (!CVMlocalExceptionOccurred(ee)) {
	   CVMgcSafeThrowLocalException(ee, exceptionICell);
	}

    localRootEnd:;
    } CVMID_localrootEnd();
}

/*
 * CVMpushTransitionFrame - pushes a transition frame. Called when we
 * are about to make a method call from native code. Transition frames store
 * the method arguments and result and also have the benefit of
 * not requiring the intepreter loop to check for (frame == initialframe)
 * each time a frame is popped. Instead, the opc_exittransition opcode
 * in the transition frame causes the interpreter loop to exit. They also
 * simplify the pushing of the CVMJavaFrame or CVMJNIFrame by leaving it
 * up to the interpreter to handle it in the normal way.
 */
CVMTransitionFrame* CVMpushTransitionFrame(CVMExecEnv* ee, CVMMethodBlock* mb)
{
    CVMFrame* frame = CVMeeGetCurrentFrame(ee);

    CVMassert(CVMD_isgcUnsafe(ee));

    CVMpushFrame(ee, &ee->interpreterStack, frame, frame->topOfStack,
		 /* transition frame size in words plus argument size */
		 CVM_TRANSITIONFRAME_SIZE / sizeof(CVMStackVal32)
		 + CVMmbArgsSize(mb),
		 0,         /* frame offset (no local variables) */
		 CVMtransitionFrameScanner,
		 /* action if pushFrame fails */
		 {
		     CVMdebugPrintf(("CVMsetupTransitionFrame - "
				     "CVMpushFrame failed.\n"));
		     return NULL;  /* exception already thrown */
		 },
		 /* action to take if stack expansion occurred */
		 {});

    /* Initialize the transition frame. */
    CVM_RESET_TRANSITION_TOS(frame->topOfStack, frame);
    ee->interpreterStack.currentFrame = frame;

    /* be default, don't let opc_exittransition increment the pc. */
    CVMframeIncrementPcFlag(frame) = CVM_FALSE;

    /*
     * We must do this while still gcUnsafe or CVMtransitionFrameScanner()
     * will crash if a gc starts up before CVMframePc() gets initialized
     * to its proper value.
     */
    CVMframePc(frame) = NULL;

    /* Setting the frame's mb to the mb of the method we want to invoke.
     * We really don't need an mb for the transition frame,
     * but we do need to know what the arguments are so they can be
     * scanned. The arguments of the method we are going to call serve
     * us well, so we use them by saving away the mb in the frame.
     */
    /* NOTE: The JVMDI now relies on the fact that
       transition frames have a valid methodblock. If you decide to
       remove this line, you must instead surround it with #ifndef
       JVMDI. */
    frame->mb = mb;

    return (CVMTransitionFrame*)frame;    
}

/* Documented in interpreter.h */
CVMClassBlock*
CVMgetCallerClass(CVMExecEnv* ee, int skip)
{
    /* CVMFrame frame_buf; */
    CVMFrame* frame = CVMgetCallerFrame(CVMeeGetCurrentFrame(ee), skip);
    if (frame) {
	CVMassert(frame->mb != NULL);
	return CVMmbClassBlock(frame->mb);
    } else {
	return NULL;
    }
}

/* Documented in interpreter.h */
CVMFrame*
CVMgetCallerFrameSpecial(CVMFrame* frame, int n, CVMBool skipSpecial)
{
    /*
     * %comment kbr001
     *
     * %comment dlong002
     *
     * %comment dlong003
     *
     */
#ifdef CVM_REFLECT
    CVMMethodBlock *reflectInvokeMb =
	CVMglobals.java_lang_reflect_Method_invoke;
    CVMMethodBlock *reflectNewInstanceMb =
	CVMglobals.java_lang_reflect_Constructor_newInstance;
#endif
    CVMMethodBlock *newInstanceMb =
	CVMglobals.java_lang_Class_newInstance;

    while (CVM_TRUE) {
        /* skip transition, and maybe Method.invoke(),
           Constructor.newInstance(), and Class.newInstance()
           frames. */
        while (frame != NULL && (CVMframeIsTransition(frame)
				 || frame->mb == NULL
				 || (skipSpecial &&
				 (frame->mb == newInstanceMb
#ifdef CVM_REFLECT
				 || frame->mb == reflectInvokeMb
				 || frame->mb == reflectNewInstanceMb
#endif
		)))) {
	    frame = CVMframePrev(frame);
	}

	if (frame == NULL) {
	    return NULL;
	}

	if (n-- > 0) {
	    /* skip a real frame */
	    frame = CVMframePrev(frame);
	} else {
	    break;
	}
    }

    CVMassert(frame->mb != NULL);

    return frame;
}

/*
 * Check whether srcCb is assignable to dstCb
 */
CVMBool
CVMisAssignable(CVMExecEnv* ee, CVMClassBlock* srcCb, 
		CVMClassBlock* dstCb)
{
    if (!CVMisArrayClass(srcCb)) {
	/* srcCb is an object class. Check subclass relationship */
        return CVMisSubclassOf(ee, srcCb, dstCb);
    } else {
	if (CVMisArrayClass(dstCb)) {
	    /*
	     * Both srcCb and dstCb are arrays. Check array assignability.
	     */
	    return CVMarrayIsAssignable(ee, srcCb, dstCb);
	} else {
	    /*
	     * If srcCb is an array class and dstCb is not,
	     * then only the following assignments are legal.
	     */
	    return (dstCb == CVMsystemClass(java_lang_Object) || 
		    dstCb == CVMsystemClass(java_lang_Cloneable) || 
		    dstCb == CVMsystemClass(java_io_Serializable));
	}
    }
}

/*
 * CVMgcUnsafeIsInstanceOf - Check if obj is an instance of cb. obj can be an
 * array object. 
 */
CVMBool
CVMgcUnsafeIsInstanceOf(CVMExecEnv* ee, CVMObject* obj, 
			CVMClassBlock* cb)
{
    CVMassert(CVMD_isgcUnsafe(ee));
    /* null can be cast to anything EXCEPT in the interpreter
       (opc_instanceof)! */
    if (obj == 0) {            
        return CVM_TRUE;
    } 
    return CVMisAssignable(ee, CVMobjectGetClass(obj), cb);

}


/*
 * CVMisSubclassOf - Check if subclasscb is an subclass of cb. subclasscb
 * must be a non-array class.
 *
 * -if cb is a class type, subclasscb must be the same class as cb,
 *  or a subclass of cb.
 * -if cb is an interface type, subclasscb must implement interface cb.
 * 
 */
CVMBool
CVMisSubclassOf(CVMExecEnv* ee, CVMClassBlock* subclasscb, 
		CVMClassBlock* cb)
{
    if (subclasscb == cb) {
	return CVM_TRUE;
    }

    if (CVMcbIs(cb, INTERFACE)) {
	/*
	 * CVMimplementsInterface checks the subclasscb's CVMInterfaceTable,
	 * which is a sumary of all the interfaces subclasscb implements.
	 * Therefore, there is no need to iterate over superclasses.
	 */
	return CVMimplementsInterface(ee, subclasscb, cb);
    } else {
	/* cb is a class type */ 
	subclasscb = CVMcbSuperclass(subclasscb);
	while (subclasscb != NULL) {
            if (subclasscb == cb) {
                return CVM_TRUE;
	    }
            subclasscb = CVMcbSuperclass(subclasscb);
        }
	return CVM_FALSE;
    }
}


/* 
 * Return TRUE if srcArrayCb is assignable to dstArrayCb
 */
static CVMBool
CVMarrayIsAssignable(CVMExecEnv* ee,
		     CVMClassBlock* srcArrayCb,
		     CVMClassBlock* dstArrayCb)
{
    int              srcArrayDepth     = CVMarrayDepth(srcArrayCb);
    CVMBasicType     srcArrayBaseType  = CVMarrayBaseType(srcArrayCb);
    CVMClassBlock*   srcArrayBaseClass = CVMarrayBaseCb(srcArrayCb);

    int              dstArrayDepth     = CVMarrayDepth(dstArrayCb);
    CVMBasicType     dstArrayBaseType  = CVMarrayBaseType(dstArrayCb);
    CVMClassBlock*   dstArrayBaseClass = CVMarrayBaseCb(dstArrayCb);

    if (srcArrayDepth > dstArrayDepth) {
	/*
	 * The source array is 'deeper' than the destination array.
	 * The only way they could be assignable is if the destination
	 * array is an array of Object, Cloneable or Serializable.
	 */
        return (dstArrayBaseClass == CVMsystemClass(java_lang_Object) ||
                dstArrayBaseClass == CVMsystemClass(java_lang_Cloneable) ||
                dstArrayBaseClass == CVMsystemClass(java_io_Serializable));
    } else if (srcArrayDepth == dstArrayDepth) {
	/*
	 * Equal depths.
	 * 
	 * If the base types of both arrays are not of the same category
	 * they are definitely not assignable
	 */
	if (dstArrayBaseType != srcArrayBaseType) {
	    return CVM_FALSE;
	}

	/*
	 * Now the classes are assignable iff:
	 * 
	 * both base types are the same non-object basic type
	 * 
	 *   or
	 *
	 * both base types are object types and the source type
	 * is a subclass of the destination type.
	 */
	return ((srcArrayBaseType != CVM_T_CLASS) ||
		(CVMisSubclassOf(ee, srcArrayBaseClass, dstArrayBaseClass)));
    } else {
	/*
	 * The destination array is 'deeper' than the source array,
	 * so they are not assignable.
	 */
        return CVM_FALSE;
    }
}

/* 
 * Return TRUE if cb implements the interface interfacecb.
 */
CVMBool
CVMimplementsInterface(CVMExecEnv* ee, CVMClassBlock* cb, 
		       CVMClassBlock* interfacecb)
{
    int i;

    if (CVMcbCheckRuntimeFlag(cb, LINKED)) {
	/*
	 * If the class has already been linked, then the CVMInterfaceTable
	 * contains all interfaces that the class implements. There is no
	 * need to look at super classes or super interfaces.
	 */
	for ( i = 0; i < CVMcbInterfaceCount(cb); i++) {
	    if (CVMcbInterfacecb(cb, i) == interfacecb) {
		return CVM_TRUE;
	    }
	}
	return CVM_FALSE;
    } else {
	/*
	 * If the class has not been linked, then only the declared interfaces
	 * are in the CVMInterfaceTable. In order to check all of the
	 * class' implemented interfaces, we need to call CVMisSubclassOf()
	 * for each declared interface of the class and for each declared
	 * interface of each superclass.
	 */
	/* C stack checks */
        if (!CVMCstackCheckSize(ee, CVM_REDZONE_CVMimplementsInterface, 
	    "CVMimplementsInterface", CVM_TRUE)) {
	    return CVM_FALSE;
	}

	do {
            CVMClassBlock * subclasscb;
            for (i = 0; i < CVMcbImplementsCount(cb); i++) {
                subclasscb = CVMcbInterfacecb(cb, i);
                if (subclasscb == interfacecb) {
                    return CVM_TRUE;
                }
                else if (CVMimplementsInterface(ee, subclasscb, interfacecb)){
                    return CVM_TRUE;
                }
                else if (CVMlocalExceptionOccurred(ee)) {
                    return CVM_FALSE;
                }
            }
            cb = CVMcbSuperclass(cb);
	} while (cb != NULL);
	return CVM_FALSE;
    }
}

/*
 * Multi-dimensional array creation
 */
void
CVMmultiArrayAlloc(CVMExecEnv*     ee,
		   CVMInt32        nDimensions,
		   CVMInt32*       dimensions,
		   CVMClassBlock*  arrayCb,
		   CVMObjectICell* resultCell)
{
    CVMID_localrootBegin(ee); {
	/*
	 * Temporary array to hold the root of the tree
	 */
	CVMID_localrootDeclare(CVMArrayOfRefICell, rootArray);

	/*
	 * The level of the tree we are constructing right now.
	 */
	CVMInt32 currDepth;

	/*
	 * The siblings at level currDepth - 1, each one an array of length
	 * 'prevNodeWidth', linked by their 0th elements.
	 */
	CVMID_localrootDeclare(CVMArrayOfRefICell, prevNodeSet);
	CVMInt32 prevNodeWidth;

	/*
	 * The node we are currently filling in at level currDepth - 1
	 */
	CVMID_localrootDeclare(CVMArrayOfRefICell, prevOneNode);

	/*
	 * The siblings at level currDepth, each one an array of length
	 * 'currNodeWidth', being created and linked together.  
	 */
	CVMID_localrootDeclare(CVMArrayOfRefICell, currNodeSet);
	CVMInt32 currNodeWidth;

	/*
	 * The node we are currently creating at level currDepth.
	 */
	CVMID_localrootDeclare(CVMArrayOfRefICell, currOneNode);

	/*
	 * The type of the arrays we are constructing at level currDepth.
	 */
	CVMClassBlock* currCb;

	/*
	 * rootArray - the root of the tree. rootArray[0] will hold
	 * the final created array (and keep the whole thing from
	 * falling apart in case GC occurs).
	 */
	CVMID_allocNewArray(ee, 
			    CVM_T_CLASS,
			    (CVMClassBlock*)
			    CVMbasicTypeArrayClassblocks[CVM_T_CLASS],
			    1, 
			    (CVMObjectICell*)rootArray);

	if (CVMID_icellIsNull(rootArray)) {
	    /*
	     * returning 0
	     */
	    CVMID_icellSetNull(resultCell);
	    goto endOfJob;
	}

	CVMID_icellAssign(ee, prevNodeSet, rootArray);
	prevNodeWidth = 1;
	currCb = arrayCb;

	/*
	 * We now have rootArray acting as the previous level to level 0.
	 * It has only one slot, and the allocator has already set it to 0,
	 * so it will appear as a linked list with one node alone.
	 *
	 * Now go through all the dimensions and do the work
	 */
	for (currDepth = 0; currDepth < nDimensions; currDepth++) {
	    CVMBool lastIteration;

	    /*
	     * Each node created at this level will be currNodeWidth wide
	     */
	    currNodeWidth = dimensions[currDepth];
	    CVMID_icellSetNull(currNodeSet);

	    /*
	     * If we are at the last iteration, remember that. In this
	     * case, we will not be linking nodes created at level
	     * currDepth.
	     */
	    lastIteration = (currDepth == nDimensions - 1);

	    /*
	     * Go through each of the previous level's nodes, allocate
	     * and link the current nodes if needed, and fill them in
	     * the slots of the previous level nodes.  
	     */
	    do {
		CVMUint32 idx;
		/*
		 * Keep the node and advance to the next link before
		 * we bash the first element.
		 */
		CVMID_icellAssign(ee, prevOneNode, prevNodeSet);
		CVMID_arrayReadRef(ee, prevNodeSet, 0,
				   (CVMObjectICell*)prevNodeSet);
		for (idx = 0; idx < prevNodeWidth; idx++) {
		    /*
		     * Allocate and link
		     */
		    CVMID_allocNewArray(ee,
					CVMarrayElemTypeCode(currCb),
					currCb,
					currNodeWidth,
					(CVMObjectICell*)currOneNode);
		    if (CVMID_icellIsNull(currOneNode)) {
			/*
			 * returning 0
			 */
			CVMID_icellSetNull(resultCell);
			goto endOfJob;
		    }
		    /*
		     * Link the children only if we are going to 
		     * iterate again.
		     */
		    if (!lastIteration) {
			CVMID_arrayWriteRef(ee, currOneNode, 0,
					    (CVMObjectICell*)currNodeSet);
			CVMID_icellAssign(ee, currNodeSet, currOneNode);
		    }
		    /*
		     * And fill a parent slot
		     */
		    CVMID_arrayWriteRef(ee, prevOneNode, idx,
					(CVMObjectICell*)currOneNode);
		}
	    } while (!CVMID_icellIsNull(prevNodeSet));

	    if (!lastIteration) {
		/*
		 * We've filled level currDepth. On to the next.
		 */
		CVMID_icellAssign(ee, prevNodeSet, currNodeSet);
		prevNodeWidth = currNodeWidth;
		currCb = CVMarrayElementCb(currCb);
		/*
		 * Make sure that we aggressively create all sub array
		 * classes whenever an array class is created.
		 */
		CVMassert(currCb != 0);
	    }
	}
	/*
	 * Now rootArray[0] has what we want. Send it back to the caller.
	 */
	CVMID_arrayReadRef(ee, rootArray, 0, resultCell);

    } 
    /* 
     * We might get here if an error occurred in the above. We've set
     * resultCell to contain 0, so the caller will handle it.  
     */
 endOfJob: 
    CVMID_localrootEnd();
}

/*
 * Functions for enabling and disabling remote exceptions. It's only safe
 * to call these functions if ee is for the current thread.
 */

#ifdef CVM_REMOTE_EXCEPTIONS_SUPPORTED

void
CVMdisableRemoteExceptions(CVMExecEnv* ee)
{
    CVMassert(ee == CVMgetEE());
    ee->remoteExceptionsDisabledCount++;
}

void
CVMenableRemoteExceptions(CVMExecEnv* ee)
{
    CVMassert(ee == CVMgetEE());
    CVMassert(ee->remoteExceptionsDisabledCount > 0);
    ee->remoteExceptionsDisabledCount--;
    return;
}

CVMBool
CVMremoteExceptionsDisabled(CVMExecEnv* ee) {
    CVMassert(ee == CVMgetEE());
    return (ee->remoteExceptionsDisabledCount != 0);
}

#endif

/*
 * Find the length of the variable length instruction at 'iStream'.
 * This is called if CVMopcodeLengths[*iStream] == 0.
 */
CVMUint32
CVMopcodeGetLengthVariable(const CVMUint8* iStream)
{
    CVMUint8  opcode  = *iStream;
    /*
     * Variable length. Intelligence required.
     */
    switch(opcode) {
	case opc_tableswitch: {
	    CVMInt32* lpc  = (CVMInt32*)CVMalignWordUp(iStream + 1);
	    CVMInt32  low  = CVMgetAlignedInt32(&lpc[1]);
	    CVMInt32  high = CVMgetAlignedInt32(&lpc[2]);
	    if ((low > high) || (high - low > 65535)) {
		return -1; /* invalid instruction */
	    }
	    /*
	     * Skip default, low, high, and the (high-low+1) jump offsets
	     */
	    return (CVMUint8*)(&lpc[3 + (high - low + 1)]) - iStream;
	}
        case opc_lookupswitch: {
	    CVMInt32* lpc    = (CVMInt32*)CVMalignWordUp(iStream+1);
	    CVMInt32  npairs = CVMgetAlignedInt32(&lpc[1]);
	    if ((npairs < 0) || (npairs > 65535)) {
		return -1; /* invalid instruction */
	    }
	    /*
	     * Skip default, npairs, and npairs pairs.
	     */
	    return (CVMUint8*)(&lpc[2 + npairs * 2]) - iStream;
	}
        case opc_wide: {
	    switch(iStream[1]) {
	        case opc_ret:
	        case opc_iload: case opc_istore: 
	        case opc_fload: case opc_fstore:
	        case opc_aload: case opc_astore:
	        case opc_lload: case opc_lstore:
	        case opc_dload: case opc_dstore: 
		    return 4;
		case opc_iinc:
		    return 6;
		default: 
		    return -1; /* invalid instruction */
	    }
	}
    default:
	return -1; /* I don't know this bytecode! */
    }
}

/* Documented in interpreter.h */
CVMBool
CVMverifyClassAccess(CVMExecEnv* ee, 
		     CVMClassBlock* currentClass, CVMClassBlock* newClass, 
		     CVMBool resolverAccess)
{
    if (currentClass == NULL ||
	currentClass == newClass ||
	CVMcbIs(newClass, PUBLIC) ||
	CVMisSameClassPackage(ee, currentClass, newClass)) {
        return CVM_TRUE;
    }

#ifdef CVM_CLASSLOADING
    /* 
     * One last chance to pass verification. See similar code in 
     * CVMverifyMemberAccess2 for details.
     */
    if (resolverAccess) {	 	/* (a) */
	CVMBool isTrustedClassLoader;   /* (b) */
	CVMBool isSameClassLoader;      /* (c) */
	CVMBool isSameProtectionDomain; /* (d) */

	isTrustedClassLoader = 
	    CVMisTrustedClassLoader(ee, CVMcbClassLoader(currentClass));
	isSameClassLoader =
	    CVMcbClassLoader(currentClass) == CVMcbClassLoader(newClass);
	CVMID_icellSameObjectNullCheck(ee,
				       CVMcbProtectionDomain(currentClass), 
				       CVMcbProtectionDomain(newClass), 
				       isSameProtectionDomain);

	/* Make sure (b), (c), and (d) hold true. */
	if (!CVM_NEED_VERIFY(!isTrustedClassLoader) 
	    && isSameProtectionDomain
	    && isSameClassLoader)
	{
	    return CVM_TRUE;
	}
    }
#endif
   return CVM_FALSE;
}    

/* Documented in interpreter.h */
CVMBool
CVMverifyMemberAccess2(CVMExecEnv* ee,
		       CVMClassBlock* currentClass,
		       CVMClassBlock* memberClass, 
		       CVMUint32 access, CVMBool resolverAccess,
		       CVMBool protectedRestriction)
{
    /* Since we're going to be doing some private stuff here,
       let's make sure some invariants are met */
    CVMassert(CVM_METHOD_ACC_PUBLIC == CVM_FIELD_ACC_PUBLIC);
    CVMassert(CVM_METHOD_ACC_PROTECTED == CVM_FIELD_ACC_PROTECTED);
    CVMassert(CVM_METHOD_ACC_PRIVATE == CVM_FIELD_ACC_PRIVATE);

#undef IsPublic
#undef IsProtected
#undef IsPrivate
#define IsPublic(x)	((x) & CVM_METHOD_ACC_PUBLIC)
#define IsProtected(x)	((x) & CVM_METHOD_ACC_PROTECTED)
#define IsPrivate(x)	((x) & CVM_METHOD_ACC_PRIVATE)

    if (currentClass == NULL ||
        currentClass == memberClass ||
	IsPublic(access)) {
	return CVM_TRUE;
    }

    /* 
     * Access is ok if the accessor is a subclass of the memberClass
     * and the member is protected.
     */
    if (IsProtected(access) && !protectedRestriction) {
	/* See if currentClass is a subclass of memberClass */
	CVMClassBlock* cb;
	for (cb = CVMcbSuperclass(currentClass); cb != NULL;
	     cb = CVMcbSuperclass(cb)) {
	    if (cb == memberClass) {
		return CVM_TRUE;
	    }
	}
    }
   
    /* 
     * Access is ok if both classes are in the same package and
     * the member is not private.
     */
    if (!IsPrivate(access)) {
	if (CVMisSameClassPackage(ee, currentClass, memberClass)) {
	    return CVM_TRUE;
	}
    }


#ifdef CVM_CLASSLOADING
    /* 
     * Bug #4186924. One last chance to pass verification. Needed for
     * buggy 1.1.x compiler. All of the following must be met:
     *	a. The access is performed by byte code, not through reflection.
     *	b. The class performing the access is loaded by a trusted class
     *	   loader (null loader, extension loader, and system loader).
     *	c. The class performing the access and the class being accessed
     *	   are loaded in the same class loader.
     *	d. The class performing the access and the class being accessed
     *	   have the same protection domain.
     */
    if (resolverAccess) {	 	/* (a) */
	CVMBool isTrustedClassLoader;   /* (b) */
	CVMBool isSameClassLoader;      /* (c) */
	CVMBool isSameProtectionDomain; /* (d) */

	isTrustedClassLoader = 
	    CVMisTrustedClassLoader(ee, CVMcbClassLoader(currentClass));
	isSameClassLoader =
	    (CVMcbClassLoader(currentClass) == CVMcbClassLoader(memberClass));
	CVMID_icellSameObjectNullCheck(ee,
				       CVMcbProtectionDomain(currentClass), 
				       CVMcbProtectionDomain(memberClass), 
				       isSameProtectionDomain);

	/* Make sure (b), (c), and (d) hold true. */
	if (!CVM_NEED_VERIFY(!isTrustedClassLoader) 
	    && isSameProtectionDomain
	    && isSameClassLoader)
	{
	    return CVM_TRUE;
	}
    }
#endif

    return CVM_FALSE;
}

/* Documented in interpreter.h */
CVMBool
CVMverifyMemberAccess(CVMExecEnv* ee,
		      CVMClassBlock* currentClass,
		      CVMClassBlock* memberClass, 
		      int access, CVMBool classloaderOnly)
{
    return CVMverifyMemberAccess2(ee, currentClass, memberClass, access,
				  classloaderOnly, CVM_FALSE);
}

/*
 * Global locking functionality.
 */

void
CVMlocksForGCAcquire(CVMExecEnv* ee)
{
    /* NOTE that we do NOT seize the heap lock. The GC thread has
       already seized this. (We should assert that it's already
       acquired, like the JDK does.) */
    CVMsysMutexLock(ee, &CVMglobals.threadLock);
#ifdef CVM_CLASSLOADING
    CVMsysMutexLock(ee, &CVMglobals.constantpoolLock);
#endif
    CVMsysMutexLock(ee, &CVMglobals.typeidLock);
    CVMsysMutexLock(ee, &CVMglobals.syncLock);
    CVMsysMutexLock(ee, &CVMglobals.internLock);
    /* %comment dlong004 */
}

void
CVMlocksForGCRelease(CVMExecEnv* ee)
{
    CVMsysMutexUnlock(ee, &CVMglobals.internLock);
    CVMsysMutexUnlock(ee, &CVMglobals.syncLock);
    CVMsysMutexUnlock(ee, &CVMglobals.typeidLock);
#ifdef CVM_CLASSLOADING
    CVMsysMutexUnlock(ee, &CVMglobals.constantpoolLock);
#endif
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
    /* %comment dlong004 */
}

void
CVMlocksForThreadSuspendAcquire(CVMExecEnv* ee)
{
    /* NOTE: this used to be called lock_for_debugger in JDK 1.2 */
#ifdef CVM_CLASSLOADING
    CVMsysMutexLock(ee, &CVMglobals.nullClassLoaderLock);
#endif
    CVMsysMutexLock(ee, &CVMglobals.heapLock);
    CVMsysMutexLock(ee, &CVMglobals.threadLock);
    CVMsysMutexLock(ee, &CVMglobals.classTableLock);
#ifdef CVM_CLASSLOADING
    CVMsysMutexLock(ee, &CVMglobals.constantpoolLock);
    CVMsysMutexLock(ee, &CVMglobals.codeLock);
#endif
    CVMsysMutexLock(ee, &CVMglobals.loaderCacheLock);
#ifdef CVM_JVMDI
    CVMsysMutexLock(ee, &CVMglobals.debuggerLock);
#endif
    CVMsysMutexLock(ee, &CVMglobals.globalRootsLock);
    CVMsysMutexLock(ee, &CVMglobals.weakGlobalRootsLock);
    CVMsysMutexLock(ee, &CVMglobals.typeidLock);
    CVMsysMutexLock(ee, &CVMglobals.syncLock);
    /* %comment dlong004 */

    /* %comment dlong005 */
}

void
CVMlocksForThreadSuspendRelease(CVMExecEnv* ee)
{
    /* NOTE: this used to be called lock_for_debugger in JDK 1.2 */

    /* %comment dlong006 */

    /* %comment dlong004 */
    CVMsysMutexUnlock(ee, &CVMglobals.syncLock);
    CVMsysMutexUnlock(ee, &CVMglobals.typeidLock);
    CVMsysMutexUnlock(ee, &CVMglobals.weakGlobalRootsLock);
    CVMsysMutexUnlock(ee, &CVMglobals.globalRootsLock);
#ifdef CVM_JVMDI
    CVMsysMutexUnlock(ee, &CVMglobals.debuggerLock);
#endif
    CVMsysMutexUnlock(ee, &CVMglobals.loaderCacheLock);
#ifdef CVM_CLASSLOADING
    CVMsysMutexUnlock(ee, &CVMglobals.codeLock);
    CVMsysMutexUnlock(ee, &CVMglobals.constantpoolLock);
#endif
    CVMsysMutexUnlock(ee, &CVMglobals.classTableLock);
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
    CVMsysMutexUnlock(ee, &CVMglobals.heapLock);
#ifdef CVM_CLASSLOADING
    CVMsysMutexUnlock(ee, &CVMglobals.nullClassLoaderLock);
#endif
}

void
CVMthreadSuspendConsistentRequest(CVMExecEnv* ee, CVMExecEnv* targetEE)
{
    /*
       %comment dlong007
     */
    /* Roll all threads forward to consistent points. */
    int i;

    for (i = 0; i < CVM_NUM_CONSISTENT_STATES; ++i) {
	CVMcsReachConsistentState(ee, i);
    }
}

void
CVMthreadSuspendConsistentRelease(CVMExecEnv* ee, CVMExecEnv* targetEE)
{
    /* Allow threads to become inconsistent again. */
    /* %comment dlong007 */
    {
	int i;

	for (i = 0; i < CVM_NUM_CONSISTENT_STATES; ++i) {
	    CVMcsResumeConsistentState(ee, i);
	}
    }
}

/*
 * CVMlookupNativeMethodCode - look up the native code for a dynamically
 * loaded JNI method.
 */

#ifdef CVM_CLASSLOADING

static void*
CVMfindBuiltinEntry(const char* name)
{
    if (CVMglobals.cvmDynHandle == NULL) {
	CVMglobals.cvmDynHandle = CVMdynlinkOpen(NULL);
	if (CVMglobals.cvmDynHandle == NULL) {
	    CVMconsolePrintf("Fatal Error: "
			     "Could not open cvm as a shared library.\n");
	    return CVM_FALSE;
	}
    }

    return CVMdynlinkSym(CVMglobals.cvmDynHandle, name);
}

static CVMUint8*
CVMclassLoaderFindNative(CVMExecEnv* ee, CVMClassLoaderICell* loader, 
			 char* nativeMethodName)
{
    CVMUint8* nativeCode = NULL;
    CVMassert(!CVMlocalExceptionOccurred(ee));

    if (loader == NULL) {
	/* This is a NULL class loader so the entry must be built in */
	nativeCode = (CVMUint8*)CVMfindBuiltinEntry(nativeMethodName);
	if (nativeCode != NULL) {
	    return nativeCode;
	}
    }
    {
	/* Ask the ClassLoader to load the native method. */
	JNIEnv*   env = CVMexecEnv2JniEnv(ee);
	CVMUint8* nativeCode = NULL;
	jstring   jname;

	if ((*env)->PushLocalFrame(env, 10) < 0) {
	    CVMclearLocalException(ee);
	    return NULL;
	}
	CVMassert(!CVMlocalExceptionOccurred(ee));

        jname = (*env)->NewStringUTF(env, nativeMethodName);
	if (jname == NULL) {
	    CVMclearLocalException(ee);
	    nativeCode = NULL;
	} else {
	    /* Call java/lang/ClassLoader.findNative() */
	    jlong lresult = (*env)->CallStaticLongMethod(
                env,  
		CVMcbJavaInstance(CVMsystemClass(java_lang_ClassLoader)), 
		CVMglobals.java_lang_ClassLoader_findNative,
		loader, jname);
	    if (CVMlocalExceptionOccurred(ee)) {
		CVMclearLocalException(ee);
		nativeCode = NULL;
	    } else {
		nativeCode = (CVMUint8*)CVMlong2VoidPtr(lresult);
	    }
	}
    
	(*env)->PopLocalFrame(env, 0);
	return nativeCode;
    }
}

CVMBool
CVMlookupNativeMethodCode(CVMExecEnv* ee, CVMMethodBlock* mb)
{
    char*     nativeMethodName;
    CVMUint8* nativeCode;
    CVMClassBlock* cb = CVMmbClassBlock(mb);

    /*
     * Try to lookup the JNI function using the short mangled name.
     */
    nativeMethodName = 
	CVMmangleMethodName(ee, mb, CVM_MangleMethodName_JNI_SHORT);
    if (nativeMethodName == NULL) {
	return CVM_FALSE; /* exception already thrown */
    }
    nativeCode = CVMclassLoaderFindNative(ee, CVMcbClassLoader(cb),
					  nativeMethodName);
    free(nativeMethodName);
    if (nativeCode != NULL) {
	goto success;
    }

    /*
     * Try to lookup the JNI function using the long mangled name.
     */
    nativeMethodName = 
	CVMmangleMethodName(ee, mb, CVM_MangleMethodName_JNI_LONG);
    if (nativeMethodName == NULL) {
	return CVM_FALSE; /* exception already thrown */
    }
    nativeCode = CVMclassLoaderFindNative(ee, CVMcbClassLoader(cb),
					  nativeMethodName);
    free(nativeMethodName);
    if (nativeCode != NULL) {
	goto success;
    }

    CVMthrowUnsatisfiedLinkError(ee, "%C.%M", cb, mb);
    return CVM_FALSE;

 success:
    /*
     * Set the method's native code pointer and new invoker index. This
     * may already have been done by another thread racing with us,
     * but doing it a 2nd time is harmless so no locking is required.
     */
    CVMmbNativeCode(mb) = nativeCode;
    if (CVMmbIs(mb, SYNCHRONIZED)) {
	CVMmbInvokerIdx(mb) = CVM_INVOKE_JNI_SYNC_METHOD;
    } else {
	CVMmbInvokerIdx(mb) = CVM_INVOKE_JNI_METHOD;
    }
    return CVM_TRUE;
}
#endif /* CVM_CLASSLOADING */

/*
 * CVMisTrustedClassLoader - returns TRUE if the classloader is trusted.
 * This basically means that it is either the NULL classloader, the
 * system class loader (aka application class loader), or a parent
 * of the system class loader.
 */

#ifdef CVM_CLASSLOADING

CVMBool
CVMisTrustedClassLoader(CVMExecEnv* ee, CVMClassLoaderICell* loader) {
#ifdef CVM_TRUSTED_CLASSLOADERS	
    return CVM_TRUE;
#else
    /* 
     * The NULL class loader is always trusted.
     */
    if (loader == NULL) {
	return CVM_TRUE;
    }

    /* 
     * If the systemClassLoader is not set and "loader" is not the NULL
     * class loader, then it can't be trusted.
     */
    if (CVMglobals.systemClassLoader == NULL) {
	return CVM_FALSE;
    }

    /*
     * See if "loader" is the systemClassLoader or one of its parents.
     */
    {
	CVMBool result = CVM_FALSE;
	CVMD_gcUnsafeExec(ee, {
	    CVMObject* loaderObj = CVMID_icellDirect(ee, loader);
	    CVMObject* systemClassLoaderObj =
		CVMID_icellDirect(ee, CVMglobals.systemClassLoader);

	    while (systemClassLoaderObj != NULL) {
		if (systemClassLoaderObj == loaderObj) {
		    result = CVM_TRUE;
		    break;
		}
		CVMD_fieldReadRef(systemClassLoaderObj, 
				  CVMoffsetOfjava_lang_ClassLoader_parent,
				  systemClassLoaderObj);
	    }
	});
	
	return result;
    }
#endif /* CVM_TRUSTED_CLASSLOADERS */
}

#endif /* CVM_CLASSLOADING */

/*
 * Shutdown support
 */

/* Some of this borrowed from JDK */

static int
CVMshutdown()
{
    return 0;
}

int
CVMprepareToExit(void)
{
    struct exit_proc *pExit;

    pExit = CVMglobals.exit_procs;
    while (pExit != NULL) {
        struct exit_proc *pNext = pExit->next;
        (*pExit->proc)();
        free(pExit);
        pExit = pNext;
    }
    /* reset the globally static defined exit_procs */
    CVMglobals.exit_procs = NULL;
    return CVMshutdown();
}

void
CVMexit(int status)
{
    CVMprepareToExit();

    if (CVMglobals.exitHook != NULL) {
	void (*fn)(int) = (void (*)())CVMglobals.exitHook;
	(*fn)(status);
	/* exit() shouldn't return, but if it does, fall through */
    }
    CVMhalt(status);
}

CVMBool
CVMsafeExit(CVMExecEnv *ee, CVMInt32 status)
{
    /*
     * If "safeExit" has been enabled, then treat a "safe" System.exit()
     * by the main thread the same as returning from the main method.
     */
    {
	/*
	 * Safe exit is only supported for the main thread.
	 */
	if (ee == &CVMglobals.mainEE) {
	    /*
	     * %comment dlong007
	     */

	    if (CVMglobals.safeExitHook != NULL) {
		void (*fn)(int) = (void (*)())CVMglobals.safeExitHook;
		JavaVM *jvm = &CVMglobals.javaVM.vector;
	
		(*jvm)->DestroyJavaVM(jvm);

		(*fn)(status);
		/* safeExit() shouldn't return */
		CVMassert(CVM_FALSE);
		return CVM_TRUE;
	    }
	}
    }
    return CVM_FALSE;
}

int
CVMatExit(void (*func)(void))
{
    struct exit_proc *pExit = malloc(sizeof(struct exit_proc));
    if (pExit == NULL) {
        return 1;
    }
    pExit->proc = func;
    pExit->next = CVMglobals.exit_procs;
    CVMglobals.exit_procs = pExit;
    return 0;
}

void
CVMabort(void)
{
    static CVMBool entered = CVM_FALSE;
    if (entered)
        return;
    entered = CVM_TRUE;

    if (CVMglobals.abortHook != NULL) {
	void (*fn)() = (void (*)())CVMglobals.abortHook;
        (*fn)();
    } else {
        abort();
    }
}

void
CVMaddThread(CVMExecEnv *ee, CVMBool userThread)
{
    CVMsysMutexLock(ee, &CVMglobals.threadLock);
    if (userThread) {
	++CVMglobals.userThreadCount;
    }
    ++CVMglobals.threadCount;
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
}

void
CVMremoveThread(CVMExecEnv *ee, CVMBool userThread)
{
    CVMsysMutexLock(ee, &CVMglobals.threadLock);
    CVMassert(CVMglobals.threadCount > 0);
    if (userThread) {
	CVMassert(CVMglobals.userThreadCount > 0);
	--CVMglobals.userThreadCount;
    }
    --CVMglobals.threadCount;
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
}

/*
 * Wait for some threads to exit.
 */
static void
CVMwaitForThreads(CVMExecEnv *ee, CVMBool userThreads, CVMUint32 maxThreads)
{
    CVMUint32 *threadCountPtr;

    if (userThreads) {
	threadCountPtr = &CVMglobals.userThreadCount;
    } else {
	threadCountPtr = &CVMglobals.threadCount;
    }

    CVMsysMutexLock(ee, &CVMglobals.threadLock);
    while (*threadCountPtr > maxThreads) {
	CVMBool interrupted;
#ifdef CVM_DEBUG
	if (userThreads) {
	    CVMconsolePrintf("Waiting for %d user thread(s) to exit...\n",
		CVMglobals.userThreadCount - maxThreads);
	} else {
	    CVMconsolePrintf("Waiting for %d thread(s) to exit...\n",
		CVMglobals.threadCount - maxThreads);
	}
#endif
	interrupted = CVMsysMutexWait(ee, &CVMglobals.threadLock,
	    &CVMglobals.threadCountCV, CVMlongConstZero());
#ifdef CVM_DEBUG
	if (interrupted) {
	    CVMconsolePrintf("interrupted!\n");
	}
#else
	(void) interrupted; /* suppress compiler warning */
#endif
    }
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
}

/*
 * Wait for all of the user threads (except this one, the current thread)
 * to exit.
 */
void
CVMwaitForUserThreads(CVMExecEnv *ee)
{
    CVMwaitForThreads(ee, CVM_TRUE, 1);
}

/*
 * Wait for all threads (except this one) to exit.  This would normally
 * be daemon threads only, but it is probably possible for a daemon
 * thread to create a new user thread.
 */
void
CVMwaitForAllThreads(CVMExecEnv *ee)
{
    CVMwaitForThreads(ee, CVM_FALSE, 1);
}

#ifdef CVM_DEBUG_ASSERTS
int
CVMassertHook() {
    return 0;
}
#endif /* CVM_DEBUG_ASSERTS */

/*
 * Execute the following code to release all references to
 * to the main application ClassLoader, which should allow
 * all application classes to unload. This is used for debugging
 * purposes only.
 */


void
CVMunloadApplicationclasses(CVMExecEnv* ee)
{
#if defined(CVM_CLASSLOADING)
    /* Free the global root that points to the "system" ClassLoader. */
    if (CVMglobals.systemClassLoader != NULL) {
	CVMID_freeGlobalRoot(ee, CVMglobals.systemClassLoader);
    }

   {
	/* Set ClassLoader.scl to null */
	CVMFieldTypeID sclTID = CVMtypeidLookupFieldIDFromNameAndSig(
            ee, "scl", "Ljava/lang/ClassLoader;");
	const CVMClassBlock* cb =
	    CVMsystemClass(java_lang_ClassLoader);
	CVMFieldBlock* sclFB = 
	    CVMclassGetStaticFieldBlock(cb, sclTID);
	CVMassert(sclFB != NULL);
	CVMD_gcUnsafeExec(ee, {
	    CVMfbStaticField(sclFB).i = 0;
	});
    }
    
    {
	/* Set Launcher.launcher to null */
	CVMFieldTypeID launcherTID =
	    CVMtypeidLookupFieldIDFromNameAndSig(
                ee, "launcher", "Lsun/misc/Launcher;");
	const CVMClassBlock* cb =
	    CVMsystemClass(sun_misc_Launcher);
	CVMFieldBlock* launcherFB = 
	    CVMclassGetStaticFieldBlock(cb, launcherTID);
	CVMassert(launcherFB != NULL);
	CVMD_gcUnsafeExec(ee, {
	    CVMfbStaticField(launcherFB).i = 0;
	});
    }
    
    /*
     * Set the Thread.contextClassLoader field of each ee's
     * Thread to null, so Thread instances don't keep any
     * ClassLoaders live.
     * 
     * Also, make sure there are no outstanding exceptions thrown. The
     * exception object contains an array of classes in its
     * backtrace, which will prevent classes from unloading.
     */
    {
	CVMJavaVal32 nullRef = {0};
	CVMFieldTypeID cclTID =
	    CVMtypeidLookupFieldIDFromNameAndSig(
	        ee, "contextClassLoader", "Ljava/lang/ClassLoader;");
	const CVMClassBlock* cb =
	    CVMsystemClass(java_lang_Thread);
	CVMFieldBlock* cclFb = 
	    CVMclassGetNonstaticFieldBlock(cb, cclTID);
	CVMassert(cclFb != NULL);
	
	CVMsysMutexLock(ee, &CVMglobals.threadLock);
	CVM_WALK_ALL_THREADS(ee, currEE, {
	    /* clear Thread.contextClassLoader */
	    CVMID_fieldWrite32(ee, CVMcurrentThreadICell(currEE),
			       CVMfbOffset(cclFb), nullRef);
	    /* clear exceptions */
	    CVMclearLocalException(currEE);
	    CVMclearRemoteException(currEE);
	});
	CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
    }
#endif
    
    /* Run the gc to unload the application classes. */
    CVMgcRunGC(ee);

    /*
     * Now there are new finalizable objects enqueued. Get rid of them
     * synchronously.
     */
    {
	jclass      finalizerClass;
	jclass      referenceClass;
	jmethodID   runFinalizationID;
	jmethodID   handleAllEnqueuedID;
	JNIEnv*     env = CVMexecEnv2JniEnv(ee);

	finalizerClass = CVMcbJavaInstance(CVMsystemClass(java_lang_ref_Finalizer));
	referenceClass = CVMcbJavaInstance(CVMsystemClass(java_lang_ref_Reference));
	runFinalizationID = (*env)->GetStaticMethodID(
            env, finalizerClass, "runAllFinalizers", "()V");
	handleAllEnqueuedID = (*env)->GetStaticMethodID(
            env, referenceClass, "handleAllEnqueued", "()V");
	/* Call Reference.handleAllEnqueued() */
	(*env)->CallStaticVoidMethod(env, referenceClass,
				     handleAllEnqueuedID);
	if ((*env)->ExceptionCheck(env)) {
	    (*env)->ExceptionClear(env);
	}
	/* Call Finalizer.runFinalization() */
	(*env)->CallStaticVoidMethod(env, finalizerClass,
				     runFinalizationID);
	if ((*env)->ExceptionCheck(env)) {
	    (*env)->ExceptionClear(env);
	}
    }
}
