/*
 * -----------------------------------------------------------------
 * jni_impl.c : The implementation of JNI fn calls.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)jni_impl.c	1.249 00/11/13
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/stacks.h"
#include "javavm/include/localroots.h"
#include "javavm/include/globalroots.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/typeid.h"
#include "javavm/include/jni_impl.h"
#include "javavm/include/gc_common.h"
#include "javavm/include/preloader.h"
#include "javavm/include/basictypes.h"
#include "javavm/include/common_exceptions.h"
#include "javavm/include/utils.h"
#include "javavm/include/stackwalk.h"
#include "javavm/include/packages.h"
#include "javavm/include/xrun.h"

#include "generated/offsets/java_lang_Class.h"
#include "generated/offsets/java_lang_String.h"
#include "generated/offsets/java_lang_Thread.h"
#include "generated/jni/java_lang_Thread.h"
#include "generated/cni/sun_misc_CVM.h"

#include "javavm/include/porting/threads.h"
#include "javavm/include/porting/system.h"
#include "javavm/include/porting/ansi/stdarg.h"

#include "javavm/export/jni.h"

/* *** RTSJ *** */
#include "javavm/include/mem_registry.h"
#include "javavm/export/jvmcei.h"

// Declare this here to get rid of compiler warniong. 
// I couldn't delare it in jvmcei.h because of compiler
// dependencies.
// - Pratik.
extern void jvmcei_installAllFilters(JNIEnv *env);
/* *** JSTR *** */

#ifdef CVM_CLASSLOADING
#include "native/java/util/zip/zip_util.h"
#include "javavm/include/porting/io.h"
#include "javavm/include/porting/path.h"
#endif

#ifdef CVM_REFLECT
#include "javavm/include/reflect.h"
#endif

#ifdef CVM_JVMDI
#include "javavm/include/jvmdi_impl.h"
#endif

#undef MIN
#define MIN(x,y) ((x) < (y) ? (x) : (y))

/*
 * If gc latency is of concern, you may want to reduce this buffer size.
 * It will affect the maximum amount of time spent gc-unsafe while copying
 * characters from the heap to a local buffer.
 */
#undef  MAX_CHARS_PER_ACCESS
#define MAX_CHARS_PER_ACCESS	100

/* %comment cjp006 */

#define CVMjniNonNullICellPtrFor(jobj) \
    CVMID_NonNullICellPtrFor(jobj)

/*
 * How many JNI refs can be allocated in this frame? The number of slots
 * under the top-of-stack minus the number of slots in use in this frame.
 */
/* %comment cjp007 */
static jint
CVMjniFrameAvailable(CVMStack* stack, CVMJNIFrame* frame)
{
    if (!CVMaddressInStackChunk(frame, stack->currentStackChunk)) {
	CVMdebugPrintf(("WARNING: guessing JNI capacity!"));
	return 100;  /* take a guess */
    } else {
	return  (jint)(((CVMFrame*)(frame))->topOfStack) - (jint)frame -
	    CVMJNIFrameCapacity - frame->inUse;
    }
}


/* private types */

typedef char
JNI_PushArguments_t(JNIEnv *env, CVMterseSigIterator  *terse_signature,
                    CVMFrame *current_frame, void *args);

/* Forward declarations */

static JNI_PushArguments_t CVMjniPushArgumentsVararg;
static JNI_PushArguments_t CVMjniPushArgumentsArray;

static void CVMjniInvoke(JNIEnv *env, jobject obj, jmethodID methodID,
    JNI_PushArguments_t pushArguments, void *args,
    CVMUint16 info, jvalue *retValue);

/* Definitions */


void JNICALL
CVMjniFatalError(JNIEnv *env, const char *msg)
{
    CVMconsolePrintf("JNI FATAL ERROR: %s\n", msg);
#ifdef CVM_DEBUG
    CVMdumpStack(&CVMjniEnv2ExecEnv(env)->interpreterStack,
		 CVM_FALSE, CVM_FALSE, 100);
#endif
    CVMpanic("JNI Fatal error");
}

#undef CVMjniGcUnsafeRef2Class
#define CVMjniGcUnsafeRef2Class(ee, cl) CVMgcUnsafeClassRef2ClassBlock(ee, cl)

#undef CVMjniGcSafeRef2Class
#define CVMjniGcSafeRef2Class(ee, cl) CVMgcSafeClassRef2ClassBlock(ee, cl)



/* 
 * CreateLocalRef is used to create a local ref in the specified frame. It
 * should check the free list first.
 */
CVMObjectICell*
CVMjniCreateLocalRef(CVMExecEnv *ee)
{
    CVMStack*      jniLocalsStack = &ee->interpreterStack;
    CVMFrame*	   currentFrame   = jniLocalsStack->currentFrame;
    CVMStackVal32* result         = NULL;

    CVMassert(CVMframeIsJNI(currentFrame));

    CVMframeAlloc(ee, jniLocalsStack, CVMgetJNIFrame(currentFrame), result);
    if (result != NULL) {
	return &result->ref;
    } else {
	/* The JNI spec says that if an attempt to allocate a localRef
	 * fails, it's a fatal error since the caller really should have
	 * called EnsureLocalCapacity() first. 
	 */
	CVMjniFatalError(CVMexecEnv2JniEnv(ee), 
			 "Out of memory when expanding local ref "
			 "table beyond capacity");
	return NULL;
    }
}

/* Version information */

jint JNICALL
CVMjniGetVersion(JNIEnv *env)
{
    return JNI_VERSION_1_2;
}

/* 
 * NewLocalRef is used to create a local ref in the specified frame. It
 * should check the free list first.
 */
jobject JNICALL
CVMjniNewLocalRef(JNIEnv *env, jobject obj)
{
    CVMExecEnv*     currentEE      = CVMjniEnv2ExecEnv(env);
    CVMObjectICell* resultCell     = NULL;

    CVMassert(CVMframeIsJNI(CVMeeGetCurrentFrame(currentEE)));

    if ((obj == NULL) || CVMID_icellIsNull(obj)) {
	/*
	 * If ref is NULL, or ref points to a cell containing a NULL
	 * (a cleared weak global ref), return NULL.
	 */
	return NULL;
    } else {
	resultCell = CVMjniCreateLocalRef(currentEE);
	if (resultCell == NULL) {
	    return NULL;
	}
	CVMID_icellAssign(currentEE, resultCell, obj);
	return resultCell;
    }
}

/*
 * Is a given ref in any of the chunks of 'frame'?
 */
static CVMBool
CVMjniRefInFrame(CVMStackChunk* chunk, CVMFrame* frame, CVMStackVal32* ref)
{
    CVMJNIFrame*   jniFrame = CVMgetJNIFrame(frame);
    CVMStackVal32* tos      = frame->topOfStack;
    /*
     * Traverse chunks from the last chunk of the frame to the first
     */
    while (!CVMaddressInStackChunk(frame, chunk)) {
	if ((ref < tos) && (ref >= chunk->data)) {
	    return CVM_TRUE;
	}
	chunk = chunk->prev;
	tos = chunk->end_data;
    }
    /* Now we are looking at the first chunk of the frame.
       Do a range check. */
    if ((ref < tos) && (ref >= jniFrame->vals)) {
	return CVM_TRUE;
    } else {
	return CVM_FALSE;
    }
}

/* 
 * DeleteLocalRef is used delete a JNI local ref in the specified frame.
 */
void JNICALL
CVMjniDeleteLocalRef(JNIEnv *env, jobject obj)
{
    CVMExecEnv*    currentEE      = CVMjniEnv2ExecEnv(env);
    CVMStack*      jniLocalsStack = &currentEE->interpreterStack;
    CVMStackVal32* refToFree	  = (CVMStackVal32*)obj;
    CVMFrame*      frame;
    CVMFrame*      calleeFrame;
    CVMStackChunk* chunk;
    CVMStackWalkContext swc;

    frame = jniLocalsStack->currentFrame;
    CVMassert(frame != NULL);
    CVMassert(CVMframeIsJNI(frame));
    if (refToFree == NULL) {
	return; /* Nothing to do */
    }
    /* 
     * Walk the stack through all local roots and JNI frames. These
     * may be PushLocalFrame frames, or the frame of a JNI
     * method. If the local frame belongs to any one of these frames
     * (each of which may span multiple stack chunks), return it to
     * the right free list.  
     */
    CVMstackwalkInit(jniLocalsStack, &swc);
    frame = swc.frame;
    chunk = swc.chunk;
    do {
	if (CVMjniRefInFrame(chunk, frame, refToFree)) {
	    CVMframeFree(CVMgetJNIFrame(frame), refToFree);
	    return;
	}
	calleeFrame = frame;
	CVMstackwalkPrev(&swc);
	frame = swc.frame;
	chunk = swc.chunk;
	CVMassert(frame != NULL);
    } while (CVMframeIsJNI(frame));
    /*
     * Well, we couldn't find where this ref belongs. Maybe it is an argument
     * to this JNI function.
     *
     * At this point:
     *    frame-        the caller of the JNI method
     *    calleeFrame-  the frame of the JNI method
     */
    if (CVMframeIsInterpreter(frame)) {
	/* We are now at the caller of the JNI method. Its top of stack is set
	   to just above the arguments to the JNI method. See if refToFree
	   was one of these arguments. */
	CVMStackVal32* topOfArgs = frame->topOfStack;
	if ((refToFree <  topOfArgs) &&
	    (refToFree >= topOfArgs - CVMmbArgsSize(calleeFrame->mb))) {
	    CVMID_icellSetNull(obj);
	    return;
	}
    }
    /*
     * If we got here, this cannot be a valid ref. Just ignore -- we can't
     * do anything useful with this.
     */
    CVMdebugPrintf(("Attempt to delete invalid local ref -- ignoring!\n"));
}

/*
 * Ensure local refs capacity in frame (in number of words).
 */
jint JNICALL
CVMjniEnsureLocalCapacity(JNIEnv *env, jint capacity)
{
    CVMExecEnv*    ee             = CVMjniEnv2ExecEnv(env);
    CVMStack*      jniLocalsStack = &ee->interpreterStack;
    CVMFrame*	   currentFrame   = jniLocalsStack->currentFrame;
    CVMJNIFrame*   jniFrame;
    CVMUint32      available;

    CVMassert(CVMframeIsJNI(currentFrame));
    jniFrame = CVMgetJNIFrame(currentFrame);
    available = CVMjniFrameAvailable(jniLocalsStack, jniFrame);

    if (capacity <= available) {
	return JNI_OK; /* We have enough to satisfy the capacity request */
    }
    capacity -= available; /* This is what we really need */
    
    if (!CVMensureCapacity(ee, jniLocalsStack, capacity)) {
	/* The JNI spec says to throw an OutOfMemoryError. However,
	 * CVMensureCapacity may have thrown a StackOverflowError
	 * error instead.
	 */
	CVMclearLocalException(ee);
	CVMthrowOutOfMemoryError(ee, NULL);
	return JNI_ENOMEM;
    } else {
	return JNI_OK;
    }
}

/*
 * Push local frame that can accommodate 'capacity' local refs
 */
jint JNICALL
CVMjniPushLocalFrame(JNIEnv *env, jint capacity)
{
    CVMExecEnv*    ee             = CVMjniEnv2ExecEnv(env);
    CVMStack*      jniLocalsStack = &ee->interpreterStack;
    CVMFrame*      currentFrame   = jniLocalsStack->currentFrame;

    /* Make sure we have the room on the stack for the specified bytes. */
    if (!CVMensureCapacity(ee, jniLocalsStack, 
			   capacity + CVMJNIFrameCapacity)) {
	/* The JNI spec says to throw an OutOfMemoryError. However,
	 * CVMensureCapacity may have thrown a StackOverflowError
	 * error instead.
	 */
	CVMclearLocalException(ee);
	CVMthrowOutOfMemoryError(ee, NULL);
	return JNI_ENOMEM;
    }

    /* We know this will succeed, since we've ensured larger capacity above */
    CVMpushGCRootFrame(ee, jniLocalsStack, currentFrame,
		       CVMJNIFrameCapacity, CVMjniFrameScanner);
    CVMjniFrameInit(CVMgetJNIFrame(currentFrame));
    /* Ready to go! */
    return JNI_OK;
}

/*
 * Pop local frame, and return a ref to the old frame
 */
jobject JNICALL
CVMjniPopLocalFrame(JNIEnv *env, jobject resultArg)
{
    CVMExecEnv*    currentEE      = CVMjniEnv2ExecEnv(env);
    CVMStack*      jniLocalsStack = &currentEE->interpreterStack;
    CVMFrame*      currentFrame   = jniLocalsStack->currentFrame;
    CVMJNIFrame*   jniFrame;
    CVMStackVal32* returnVal;
    CVMObjectICell* result;

    CVMassert(CVMframeIsFreelist(currentFrame));
    CVMassert(currentFrame->mb == NULL);

    /* If the resultArg is NULL, we are not responsible for handing
       the object reference to the previous stack frame. This is
       important for making the JVMDI work. See the Java 1.2 JNI spec
       additions. */
    if (resultArg == NULL) {
	CVMpopGCRootFrame(jniLocalsStack, currentFrame);
	/* It is not required that the previous frame be JNI, because
           we're not going to try to get a "local ref" for the result
           argument. */
	return NULL;
    } else {
	result = CVMjniNonNullICellPtrFor(resultArg);
	CVMID_localrootBegin(currentEE); {
	    CVMID_localrootDeclare(CVMObjectICell, temp);

	    /*
	     * Hold the result in a temporary, so it doesn't get lost
	     * when the locals frame is popped
	     */
	    CVMID_icellAssign(currentEE, temp, result);

	    CVMpopGCRootFrame(jniLocalsStack, currentFrame);

	    if (!CVMframeIsJNI(currentFrame)) {
		CVMjniFatalError(env, "can't return reference to "
				 "non-JNI frame");
	    }

	    CVMassert(CVMframeIsJNI(currentFrame));
	    jniFrame = CVMgetJNIFrame(currentFrame);

	    /*
	     * Now get a new slot for the result
	     */
	    CVMframeAlloc(currentEE, jniLocalsStack, jniFrame, returnVal);
	    CVMID_icellAssign(currentEE, &returnVal->ref, temp);
	} CVMID_localrootEnd();
	/*
	 * Don't return ICells containing NULL. 
	 *
	 * %comment: rt011
	 */
	if (CVMID_icellIsNull(&returnVal->ref)) {
	    return NULL;
	} else {
	    return &returnVal->ref;
	}
    }
}

/*
 * Make global ref, and make it point to the same object 'value' points to
 */
jobject JNICALL
CVMjniNewGlobalRef(JNIEnv *env, jobject ref)
{
    if ((ref == NULL) || CVMID_icellIsNull(ref)) {
	/*
	 * If ref is NULL, or ref points to a cell containing a NULL
	 * (a cleared weak global ref), return NULL.  
	 */
	return NULL;
    } else {
	
	/*
	 * JNI Reference checking, if jobject ref is from Scope Memory,
	 * throw an IllegaAssignment Error
	 */
	
	jclass checkClass;
	jobject memref;
	jboolean retVal = JNI_FALSE;
	extern int JVMCeiInitialized;

	if(JVMCeiInitialized) {
		memref = (jobject)CVMJTregistryGetMemoryAreaForObject(CVMjniEnv2ExecEnv(env), 
											(CVMObjectICell *)ref);
        if(memref == NULL)
            retVal = JNI_FALSE;
        else
        {
		    checkClass = (*env)->FindClass(env, "javax/realtime/ScopedMemory");
    		retVal = (*env)->IsInstanceOf(env, memref, checkClass);
        }
	}	

	if(retVal == JNI_FALSE) {
		CVMExecEnv* currentEE = CVMjniEnv2ExecEnv(env);
		CVMObjectICell* cell = CVMID_getGlobalRoot(currentEE);
		if (cell != NULL) {
	    	CVMID_icellAssign(currentEE, cell, ref);
		} else {
	    	CVMthrowOutOfMemoryError(currentEE, NULL);
		}
		return cell;
	}else {
		printf("CVMjniNewGlobalRef(): Illegal Assignment for ScopedMemory\n");
		return NULL;
	}
    }
}

/*
 * Delete global ref
 */
void JNICALL
CVMjniDeleteGlobalRef(JNIEnv *env, jobject ref)
{
    /* Apparently Sheng's book, which contains the revised JNI spec,
       says that deleting a NULL global ref is a no-op */
    /* %comment: rt012 */
    if (ref != NULL) {
	CVMExecEnv* currentEE = CVMjniEnv2ExecEnv(env);
	CVMID_freeGlobalRoot(currentEE, ref);
    }
}

static jweak JNICALL
CVMjniNewWeakGlobalRef(JNIEnv* env, jobject ref)
{
    if ((ref == NULL) || CVMID_icellIsNull(ref)) {
	/*
	 * If ref is NULL, or ref points to a cell containing a NULL
	 * (a cleared weak global ref), return NULL.
	 */
	return NULL;
    } else {
	CVMExecEnv* currentEE = CVMjniEnv2ExecEnv(env);
	CVMObjectICell* cell = CVMID_getWeakGlobalRoot(currentEE);
	if (cell != NULL) {
	    CVMID_icellAssign(currentEE, cell, ref);
	} else {
	    CVMthrowOutOfMemoryError(currentEE, NULL);
	}
	return cell;
    }
}

static void JNICALL
CVMjniDeleteWeakGlobalRef(JNIEnv* env, jweak ref)
{
    /* Apparently Sheng's book, which contains the revised JNI spec,
       says that deleting a NULL global ref is a no-op */
    /* %comment: rt012 */
    if (ref != NULL) {
	CVMExecEnv* currentEE = CVMjniEnv2ExecEnv(env);
	CVMID_freeWeakGlobalRoot(currentEE, ref);
    }
}

/*
 * CVMjniDefineClass - define a class. Used by Classloader.defineClass.
 */
static jclass JNICALL
CVMjniDefineClass(JNIEnv *env, const char *name, jobject loaderRef,
		  const jbyte *buf, jsize bufLen)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    return CVMdefineClass(ee, name, loaderRef, (CVMUint8*)buf, bufLen, NULL);
}

jclass JNICALL
CVMjniFindClass(JNIEnv *env, const char *name)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);

    CVMClassLoaderICell* loader;

    CVMClassBlock* callerCb = CVMgetCallerClass(ee, 0);

#ifdef CVM_CLASSLOADING
    //CVMassert(!CVMlocalExceptionOccurred(ee));
#ifdef CVM_DEBUG
    if(CVMlocalExceptionOccurred(ee))
    {
        printf("Local Exception occured in CVMjniFindClass\n");
        CVMconsolePrintf("Exception was %I\n", CVMlocalExceptionICell(ee));
        CVMassert(CVM_FALSE);
    }
#else
    CVMassert(!CVMlocalExceptionOccurred(ee));
#endif

    if (callerCb != NULL) {
        /* Special handling to make sure JNI_OnLoad and JNI_OnUnload are
	 * executed in the correct class context.
	 */
	if (callerCb == CVMsystemClass(java_lang_ClassLoader_NativeLibrary)) {
	    jobject clazz;
	    CVMMethodBlock* mb = 
		CVMglobals.java_lang_ClassLoader_NativeLibrary_getFromClass;
	    clazz = (*env)->CallStaticObjectMethod(env,
						   CVMcbJavaInstance(callerCb),
						   mb);
	    if (CVMexceptionOccurred(ee)) {
		return NULL;
	    }
	    callerCb = CVMgcSafeClassRef2ClassBlock(ee, clazz);
	}
	loader = CVMcbClassLoader(callerCb);
    } else {
        /* We call ClassLoader.getSystemClassLoader to obtain the system
	 * class loader. We'll cache it in the CVMglobals so we only have
	 * to do this once.
	 */
	if (CVMglobals.systemClassLoader == NULL) {
	    CVMglobals.systemClassLoader = CVMID_getGlobalRoot(ee);
	    if (CVMglobals.systemClassLoader == NULL) {
		CVMthrowOutOfMemoryError(ee, NULL);
		return NULL;
	    }
	}
	if (CVMID_icellIsNull(CVMglobals.systemClassLoader)) {
	    const CVMClassBlock* loaderCb =
		CVMsystemClass(java_lang_ClassLoader);
	    CVMMethodTypeID methodID = CVMtypeidLookupMethodIDFromNameAndSig(
                ee, "getSystemClassLoader", "()Ljava/lang/ClassLoader;");
	    CVMMethodBlock* mb = 
		CVMclassGetStaticMethodBlock(loaderCb, methodID);
	    jobject systemClassLoader;
	    CVMassert(mb != NULL);
	    systemClassLoader = 
		(CVMClassLoaderICell*)(*env)->CallStaticObjectMethod(
                    env, CVMcbJavaInstance(loaderCb), mb);
	    if (CVMexceptionOccurred(ee)) {
		return NULL;
	    }
	    if (systemClassLoader != NULL) {
		CVMID_icellAssign(ee, CVMglobals.systemClassLoader,
				  systemClassLoader);
		CVMjniDeleteLocalRef(env, systemClassLoader);
	    }
	}
	loader = CVMglobals.systemClassLoader;
    }
#else
    loader = NULL;
#endif /* CVM_CLASSLOADING */

    CVMassert(!CVMlocalExceptionOccurred(ee));
    {
	CVMObjectICell* pd = callerCb ? CVMcbProtectionDomain(callerCb) : NULL;
	CVMClassBlock* cb = 
	    CVMclassLookupByNameFromClassLoader(ee, name, CVM_TRUE, 
						loader, pd, CVM_TRUE);
	if (cb != NULL) {
	    return CVMjniNewLocalRef(env, CVMcbJavaInstance(cb));
	}
	return NULL;
    }
}

/*
 * Native Bootstrap
 */

jint JNICALL
CVMjniRegisterNatives(JNIEnv *env, jclass clazz,
		      const JNINativeMethod *methods, jint nMethods)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock *cb = CVMgcSafeClassRef2ClassBlock(ee, clazz);
    CVMBool inROM = CVMcbIsInROM(cb);
    int i;

    for (i = 0; i < nMethods; ++i) {
        struct CVMMethodBlock *mb;
        char *name = methods[i].name;
        char *sig = methods[i].signature;

	mb = CVMclassGetDeclaredMethodBlock(ee, cb, name, sig);
        if (mb == NULL || !CVMmbIs(mb, NATIVE)) {
            CVMthrowNoSuchMethodError(ee, "%s", name);
            return JNI_EINVAL;
	}
    }

    /* %comment: rt013 */
    for (i = 0; i < nMethods; ++i) {
        struct CVMMethodBlock *mb;
        const char *name = methods[i].name;
        const char *sig = methods[i].signature;
        void *nativeProc = methods[i].fnPtr;
	int invoker;

	mb = CVMclassGetDeclaredMethodBlock(ee, cb, name, sig);
	CVMassert(mb != NULL && CVMmbIs(mb, NATIVE));
#ifdef JDK12
	if (verbose_jni) {
	    jio_fprintf(stderr, "[Registering JNI native method %s.%s]\n",
			cbName(mb->fb.clazz), mb->fb.name);
	}
#endif
	invoker = CVMmbIs(mb, SYNCHRONIZED) ? CVM_INVOKE_JNI_SYNC_METHOD
					    : CVM_INVOKE_JNI_METHOD;
	if (inROM) {
	    CVMassert(CVMmbInvokerIdx(mb) == invoker);
	    CVMassert(CVMmbNativeCode(mb) == nativeProc);
	} else {
	    CVMmbInvokerIdx(mb) = invoker;
	    CVMmbNativeCode(mb) = nativeProc;
	}
    }

    return JNI_OK;
}

jint JNICALL
CVMjniUnregisterNatives(JNIEnv *env, jclass clazz)
{
#ifdef CVM_CLASSLOADING
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock *cb = CVMgcSafeClassRef2ClassBlock(ee, clazz);
    CVMBool inROM = CVMcbIsInROM(cb);
    int i;

    if (inROM) {
	return JNI_OK;
    }

    /* %comment: rt013 */
    for (i = 0; i < CVMcbMethodCount(cb); ++i) {
        CVMMethodBlock* mb = CVMcbMethodSlot(cb, i);

	if (CVMmbIs(mb, NATIVE)) {
	    CVMmbInvokerIdx(mb) = CVM_INVOKE_LAZY_JNI_METHOD;
	    CVMmbNativeCode(mb) = NULL;
	}
    }
    return JNI_OK;

#else
    return JNI_ERR;
#endif
}


static jmethodID
CVMjniGetXMethodID(JNIEnv *env,
    jclass clazz, const char *name, const char *sig, jboolean isStatic)
{
    CVMExecEnv *    ee;
    CVMClassBlock*  cb;
    CVMMethodTypeID tid;
    CVMMethodBlock* mb;

    ee  = CVMjniEnv2ExecEnv(env);
    cb  = CVMjniGcSafeRef2Class(ee, clazz);

    if (!CVMclassInit(ee, cb)) {
	return NULL;
    }

    /*
     * We need to use the "new" function, because the "lookup" function
     * is only valid when we know that the type already exists.
     */
    tid = CVMtypeidNewMethodIDFromNameAndSig(ee, name, sig);

    if (tid != CVM_TYPEID_ERROR) {
	/* never search superclasses for constructors */
	mb  = CVMclassGetMethodBlock(cb, tid, isStatic);
	CVMtypeidDisposeMethodID(ee, tid);
    } else {
	mb = NULL;
	CVMclearLocalException(ee);
    }

    if (mb == NULL) {
        CVMthrowNoSuchMethodError(ee, "%s", name);
	return NULL;
    }
    return mb;
}

jmethodID JNICALL
CVMjniGetStaticMethodID(JNIEnv *env,
    jclass clazz, const char *name, const char *sig)
{
    return CVMjniGetXMethodID(env, clazz, name, sig, JNI_TRUE);
}

jmethodID JNICALL
CVMjniGetMethodID(JNIEnv *env,
    jclass clazz, const char *name, const char *sig)
{
    return CVMjniGetXMethodID(env, clazz, name, sig, JNI_FALSE);
}

static jfieldID
CVMjniGetXFieldID(JNIEnv *env,
    jclass clazz, const char *name, const char *sig, jboolean isStatic)
{
    CVMExecEnv *    ee;
    CVMClassBlock*  cb;
    CVMFieldTypeID       tid;
    CVMFieldBlock*  fb;

    ee  = CVMjniEnv2ExecEnv(env);
    cb  = CVMjniGcSafeRef2Class(ee, clazz);

    if (!CVMclassInit(ee, cb)) {
	return NULL;
    }

    /*
     * We need to use the "new" function, because the "lookup" function
     * is only valid when we know that the type already exists.
     */
    tid = CVMtypeidNewFieldIDFromNameAndSig(ee, name, sig);

    if (tid != CVM_TYPEID_ERROR) {
	fb  = CVMclassGetFieldBlock(cb, tid, isStatic);
	CVMtypeidDisposeFieldID(ee, tid);
    } else {
	fb = NULL;
	CVMclearLocalException(ee);
    }

    if (fb == NULL || (CVMfbIs(fb, STATIC) != isStatic)) {
        CVMthrowNoSuchFieldError(ee, "%s", name);
	return NULL ;
    }
    return fb;
}

jfieldID JNICALL
CVMjniGetStaticFieldID(JNIEnv *env,
    jclass clazz, const char *name, const char *sig)
{
    return CVMjniGetXFieldID(env, clazz, name, sig, JNI_TRUE);
}

jfieldID JNICALL
CVMjniGetFieldID(JNIEnv *env,
    jclass clazz, const char *name, const char *sig)
{
    return CVMjniGetXFieldID(env, clazz, name, sig, JNI_FALSE);
}

/*
 * Make sure we are accessing a real field
 */
#ifdef CVM_DEBUG
static void 
CVMjniSanityCheckFieldAccess(JNIEnv* env, jobject jobj, jfieldID jfb)
{
    CVMClassBlock* objCb;
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);

    /* Simple checks */
    CVMassert(jobj != NULL);
    CVMassert(jfb != NULL);
    CVMassert(!CVMfbIs(jfb, STATIC));
    /* Also make sure that the object class is a subclass of the
       field class */
    CVMID_objectGetClass(ee, jobj, objCb);
    CVMassert(CVMisSubclassOf(ee, objCb, CVMfbClassBlock(jfb)));
}
#else
#define CVMjniSanityCheckFieldAccess(env, jobj, jfb)
#endif

#define CVM_DEFINE_JNI_FIELD_GETTER_AND_SETTER(jType_, elemType_,	\
					       wideType_)		\
jType_ JNICALL								\
CVMjniGet##elemType_##Field(JNIEnv* env, jobject obj, jfieldID fid)	\
{									\
    CVMExecEnv* ee  = CVMjniEnv2ExecEnv(env);				\
    jType_      retVal;							\
									\
    CVMjniSanityCheckFieldAccess(env, obj, fid);			\
    CVMID_fieldRead##wideType_(ee, obj, CVMfbOffset(fid), retVal);     	\
    return retVal;							\
}									\
									\
void JNICALL								\
CVMjniSet##elemType_##Field(JNIEnv* env, jobject obj, 			\
			    jfieldID fid, jType_ rhs)			\
{									\
    CVMExecEnv* ee  = CVMjniEnv2ExecEnv(env);				\
									\
    CVMjniSanityCheckFieldAccess(env, obj, fid);			\
    CVMID_fieldWrite##wideType_(ee, obj, CVMfbOffset(fid), rhs);       	\
}

CVM_DEFINE_JNI_FIELD_GETTER_AND_SETTER(jboolean, Boolean, Int)
CVM_DEFINE_JNI_FIELD_GETTER_AND_SETTER(jchar,    Char,    Int)
CVM_DEFINE_JNI_FIELD_GETTER_AND_SETTER(jshort,   Short,   Int)
CVM_DEFINE_JNI_FIELD_GETTER_AND_SETTER(jfloat,   Float,   Float)
CVM_DEFINE_JNI_FIELD_GETTER_AND_SETTER(jbyte,    Byte,    Int)
CVM_DEFINE_JNI_FIELD_GETTER_AND_SETTER(jint,     Int,     Int)

/*
 * The reference version is slightly different, so don't use above
 * macro.
 */
jobject JNICALL
CVMjniGetObjectField(JNIEnv* env, jobject obj, jfieldID fid)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMObjectICell* resultCell = CVMjniCreateLocalRef(ee);
    if (resultCell == NULL) {
	return NULL;
    }

    CVMjniSanityCheckFieldAccess(env, obj, fid);
    CVMID_fieldReadRef(ee, obj, CVMfbOffset(fid), resultCell);
    /* Don't return ICells containing NULL */
    if (CVMID_icellIsNull(resultCell)) {
	CVMjniDeleteLocalRef(env, resultCell);
	resultCell = NULL;
    }
    return resultCell;
}

void JNICALL
CVMjniSetObjectField(JNIEnv* env, jobject obj, jfieldID fid, jobject rhs)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
	CVMBool ret = CVM_TRUE;
	extern int JVMCeiInitialized;

	if(JVMCeiInitialized) {	
		CVMD_gcUnsafeExec( ee,
			{
                if(rhs != NULL)
    				ret = CVMJTrcAllowReferenceAssignment(ee, CVMID_icellDirect(ee,obj),
				                                      CVMID_icellDirect(ee,rhs));
                else
                    ret = CVM_TRUE;
			});
	}

	if(ret == CVM_FALSE){
		CVMthrowIllegalAssignmentError(ee,"IllegalAssignment Error \n");
		return;
	}
    CVMjniSanityCheckFieldAccess(env, obj, fid);
    CVMID_fieldWriteRef(ee, obj, CVMfbOffset(fid), 
			CVMjniNonNullICellPtrFor(rhs));
}

#define CVM_DEFINE_JNI_64BIT_FIELD_GETTER_AND_SETTER(jType_, elemType_)	\
									\
jType_ JNICALL								\
CVMjniGet##elemType_##Field(JNIEnv* env, jobject obj, jfieldID fid)	\
{									\
    CVMExecEnv* ee  = CVMjniEnv2ExecEnv(env);				\
    jType_ v64;								\
									\
    CVMjniSanityCheckFieldAccess(env, obj, fid);			\
    CVMID_fieldRead##elemType_(ee, obj, CVMfbOffset(fid), v64);		\
    return v64;								\
}									\
									\
void JNICALL								\
CVMjniSet##elemType_##Field(JNIEnv* env, jobject obj, 			\
			    jfieldID fid, jType_ rhs)			\
{									\
    CVMExecEnv* ee  = CVMjniEnv2ExecEnv(env);				\
    CVMjniSanityCheckFieldAccess(env, obj, fid);			\
    CVMID_fieldWrite##elemType_(ee, obj, CVMfbOffset(fid), rhs);       	\
}

CVM_DEFINE_JNI_64BIT_FIELD_GETTER_AND_SETTER(jlong,   Long)
CVM_DEFINE_JNI_64BIT_FIELD_GETTER_AND_SETTER(jdouble, Double)

/* Static fields */

/*
 * Make sure we are accessing a real field
 */
#ifdef CVM_DEBUG
static void
CVMjniSanityCheckStaticFieldAccess(JNIEnv* env, jclass clazz, jfieldID jfb)
{
    CVMClassBlock* cb;
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);

    /* Simple checks */
    CVMassert(clazz != NULL);
    CVMassert(jfb != NULL);
    CVMassert(CVMfbIs(jfb, STATIC));

    /* Also make sure that clazz is the same as or a subclass of the
       field class */
    cb = CVMjniGcSafeRef2Class(ee, clazz);
    CVMassert(CVMisSubclassOf(ee, cb, CVMfbClassBlock(jfb)));
}
#else
#define CVMjniSanityCheckStaticFieldAccess(env, jobj, jfb)
#endif

#define CVM_DEFINE_JNI_STATIC_GETTER_AND_SETTER(jType_, elemType_,	\
					       unionField_)		\
jType_ JNICALL								\
CVMjniGetStatic##elemType_##Field(JNIEnv* env, 				\
                                  jclass clazz, jfieldID fid)		\
{									\
    CVMExecEnv* ee  = CVMjniEnv2ExecEnv(env);				\
    jType_ retVal;							\
									\
    CVMjniSanityCheckStaticFieldAccess(env, clazz, fid);		\
    CVMD_gcUnsafeExec(ee, {						\
        retVal = CVMfbStaticField(fid).##unionField_;			\
    });									\
    return retVal;							\
}									\
									\
void JNICALL								\
CVMjniSetStatic##elemType_##Field(JNIEnv* env, jclass clazz,		\
			          jfieldID fid, jType_ rhs)		\
{									\
    CVMExecEnv* ee  = CVMjniEnv2ExecEnv(env);				\
									\
    CVMjniSanityCheckStaticFieldAccess(env, clazz, fid);		\
    CVMD_gcUnsafeExec(ee, {						\
        CVMfbStaticField(fid).##unionField_ = rhs;			\
    });									\
}

CVM_DEFINE_JNI_STATIC_GETTER_AND_SETTER(jboolean, Boolean, i)
CVM_DEFINE_JNI_STATIC_GETTER_AND_SETTER(jchar,    Char,    i)
CVM_DEFINE_JNI_STATIC_GETTER_AND_SETTER(jshort,   Short,   i)
CVM_DEFINE_JNI_STATIC_GETTER_AND_SETTER(jfloat,   Float,   f)
CVM_DEFINE_JNI_STATIC_GETTER_AND_SETTER(jbyte,    Byte,    i)
CVM_DEFINE_JNI_STATIC_GETTER_AND_SETTER(jint,     Int,     i)

/*
 * The reference versions are slightly different, so don't use above
 * macro.
 */
jobject JNICALL
CVMjniGetStaticObjectField(JNIEnv* env, jclass clazz, jfieldID fid)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMObjectICell* resultCell;


    CVMjniSanityCheckStaticFieldAccess(env, clazz, fid);

    /* Don't return ICells containing NULL */
    if (CVMID_icellIsNull(&CVMfbStaticField(fid).r)) {
	return NULL;
    }

    resultCell = CVMjniCreateLocalRef(ee);
    if (resultCell == NULL) {
	return NULL;
    }

    CVMD_gcUnsafeExec(ee, {
        CVMID_icellAssignDirect(ee,
				resultCell,
				&CVMfbStaticField(fid).r);
    });
    return resultCell;
}

void JNICALL
CVMjniSetStaticObjectField(JNIEnv* env, jclass clazz, jfieldID fid,
			   jobject rhs)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMjniSanityCheckStaticFieldAccess(env, clazz, fid);
    CVMD_gcUnsafeExec(ee, {
        CVMID_icellAssignDirect(ee,
				&CVMfbStaticField(fid).r,
				CVMjniNonNullICellPtrFor(rhs));
    });
}

#define CVM_DEFINE_JNI_64BIT_STATIC_GETTER_AND_SETTER(jType_, 		\
                                                     elemType_,		\
                                                     elemType2_)	\
jType_ JNICALL								\
CVMjniGetStatic##elemType_##Field(JNIEnv* env, 				\
                                  jclass clazz, jfieldID fid)		\
{									\
    CVMExecEnv* ee  = CVMjniEnv2ExecEnv(env);				\
    jType_ v64;								\
									\
    CVMjniSanityCheckStaticFieldAccess(env, clazz, fid);		\
    CVMD_gcUnsafeExec(ee, {						\
	v64 = CVMjvm2##elemType_(&CVMfbStaticField(fid).raw);		\
    });									\
    return v64;								\
}									\
									\
void JNICALL								\
CVMjniSetStatic##elemType_##Field(JNIEnv* env, jclass clazz,		\
			          jfieldID fid, jType_ rhs)		\
{									\
    CVMExecEnv* ee  = CVMjniEnv2ExecEnv(env);				\
    CVMjniSanityCheckStaticFieldAccess(env, clazz, fid);		\
    CVMD_gcUnsafeExec(ee, {						\
	CVM##elemType2_##2Jvm(&CVMfbStaticField(fid).raw, rhs);		\
    });									\
}

CVM_DEFINE_JNI_64BIT_STATIC_GETTER_AND_SETTER(jlong,   Long,   long)
CVM_DEFINE_JNI_64BIT_STATIC_GETTER_AND_SETTER(jdouble, Double, double)

jobject JNICALL
CVMjniAllocObject(JNIEnv *env, jclass clazz)
{
    CVMExecEnv*     ee = CVMjniEnv2ExecEnv(env);
    CVMObjectICell* resultCell;

    resultCell = CVMjniCreateLocalRef(ee);
    if (resultCell == NULL) {
	return NULL;
    }

    CVMD_gcUnsafeExec(ee, {
	CVMClassBlock*  cb = CVMjniGcUnsafeRef2Class(ee, clazz);
	CVMObject*      obj = NULL;

	if (CVMcbIs(cb,INTERFACE) || CVMcbIs(cb,ABSTRACT)) {
	    CVMthrowInstantiationException(ee, "%C", cb);
	} else {
	    obj = CVMgcAllocNewInstance(ee, cb);
	    if (obj == NULL) {
		CVMthrowOutOfMemoryError(ee, NULL);
	    } else {
		CVMID_icellSetDirect(ee, resultCell, obj);
	    }
	}

	/* Free the localref if we didn't allocate an object. */
	if (obj == NULL) {
	    CVMjniDeleteLocalRef(env, resultCell);
	    resultCell=NULL;
	}
    });
    return resultCell;
}

static jobject
CVMjniConstruct(JNIEnv *env, jclass clazz, jmethodID methodID,
	        JNI_PushArguments_t pushArguments, void *args)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    jobject result;
    CVMMethodBlock *mb = methodID;

    result = CVMjniAllocObject(env, clazz);

    if (result == NULL) {
        return NULL;
    }

    if (!CVMtypeidIsConstructor(CVMmbNameAndTypeID(mb))) {
        CVMjniFatalError(env, "a non-constructor passed to NewObject");
    }
    if (CVMmbClassBlock(mb) != CVMjniGcSafeRef2Class(ee, clazz)) {
	CVMjniFatalError(env, "wrong method ID passed to NewObject");
    }

#ifdef JDK12
    CVMjniInvoke(env, result, methodID, pushArguments, args,
		 CVM_INVOKE_VIRTUAL | CVM_TYPEID_VOID, NULL);
#else
    CVMjniInvoke(env, result, methodID, pushArguments, args,
		 CVM_INVOKE_NONVIRTUAL | CVM_TYPEID_VOID, NULL);
#endif

    if (CVMjniExceptionCheck(env)) {
	CVMjniDeleteLocalRef(env, result);
	return NULL;
    }
    return result;
}

jobject JNICALL
CVMjniNewObject(JNIEnv *env, jclass clazz, jmethodID methodID, ...)
{
    jobject result;
    va_list args;
    va_start(args, methodID);
    result = CVMjniConstruct(env, clazz, methodID, CVMjniPushArgumentsVararg,
			     &args);
    va_end(args);
    return result;
}

jobject JNICALL
CVMjniNewObjectA(JNIEnv *env, jclass clazz, jmethodID methodID, jvalue *args)
{
    return CVMjniConstruct(env, clazz, methodID, CVMjniPushArgumentsArray,
			   args);
}

jobject JNICALL
CVMjniNewObjectV(JNIEnv *env, jclass clazz, jmethodID methodID, va_list args)
{
    return CVMjniConstruct(env, clazz, methodID, CVMjniPushArgumentsVararg,
			   &args);
}


jclass JNICALL
CVMjniGetObjectClass(JNIEnv *env, jobject obj)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock *cb;
    CVMID_objectGetClass(ee, obj, cb);
    return CVMjniNewLocalRef(env, CVMcbJavaInstance(cb));
}

jboolean JNICALL
CVMjniIsSameObject(JNIEnv *env, jobject ref1, jobject ref2)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    jboolean res;
    jobject nonnullRef1 = CVMjniNonNullICellPtrFor(ref1);
    jobject nonnullRef2 = CVMjniNonNullICellPtrFor(ref2);
    CVMID_icellSameObject(ee, nonnullRef1, nonnullRef2, res);
    return res;
}

jboolean JNICALL
CVMjniIsInstanceOf(JNIEnv* env, jobject obj, jclass clazz)
{
    if (obj == NULL) {
	return JNI_TRUE;
    } else {
	CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
	CVMBool     retVal;
	CVMD_gcUnsafeExec(ee, {
	    CVMClassBlock* cb = CVMjniGcUnsafeRef2Class(ee, clazz);
	    CVMObject* directObj = CVMID_icellDirect(ee, obj);
	    retVal = CVMgcUnsafeIsInstanceOf(ee, directObj, cb);
	});
	return retVal;
    }
}

/*
 * Array related operations
 *
 * Define CVMjni{Get,Set}<PrimitiveType>ArrayRegion() functions 
 * (16 total)
 */
#define CVM_DEFINE_JNI_ARRAY_REGION_ACCESSORS(arrType, jelemType, nativeType) \
void									   \
CVMjniGet##jelemType##ArrayRegion(JNIEnv *env, arrType array, jsize start, \
				  jsize len, nativeType *buf)		   \
{									   \
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);				   \
    CVMD_gcUnsafeExec(ee, {						   \
	CVMArrayOf##jelemType* directArray =				   \
	    (CVMArrayOf##jelemType*)CVMID_icellDirect(ee, array);	   \
	CVMUint32 arrLen = CVMD_arrayGetLength(directArray);		   \
									   \
	/*								   \
	 * Make sure the java-side array and the passed in buffer    	   \
	 * have the same size elements.					   \
	 */								   \
	CVMassert(CVMarrayElemSize(CVMobjectGetClass(directArray)) == \
		  sizeof(nativeType));					   \
	/* Check for bounds and copy */					   \
	if ((start >= 0) && (len >= 0) && 				   \
	    ((CVMUint32)start + (CVMUint32)len <= arrLen)) {		   \
            CVMD_arrayReadBody##jelemType(buf, directArray, start, len);   \
	} else {							   \
	    CVMthrowArrayIndexOutOfBoundsException(ee, NULL);		   \
	}								   \
    });									   \
}									   \
void									   \
CVMjniSet##jelemType##ArrayRegion(JNIEnv *env, arrType array, jsize start, \
				 jsize len, nativeType *buf)		   \
{									   \
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);				   \
    CVMD_gcUnsafeExec(ee, {						   \
	CVMArrayOf##jelemType* directArray =				   \
	    (CVMArrayOf##jelemType*)CVMID_icellDirect(ee, array);	   \
	CVMUint32 arrLen = CVMD_arrayGetLength(directArray);		   \
									   \
	/*								   \
	 * Make sure the java-side array and the passed in buffer    	   \
	 * have the same size elements.					   \
	 */								   \
	CVMassert(CVMarrayElemSize(CVMobjectGetClass(directArray)) == 	   \
		  sizeof(nativeType));					   \
	/* Check for bounds and copy */					   \
	if ((start >= 0) && (len >= 0) && 				   \
	    ((CVMUint32)start + (CVMUint32)len <= arrLen)) {		   \
            CVMD_arrayWriteBody##jelemType(buf, directArray, start, len);  \
	} else {							   \
	    CVMthrowArrayIndexOutOfBoundsException(ee, NULL);		   \
	}								   \
    });									   \
}									   \

CVM_DEFINE_JNI_ARRAY_REGION_ACCESSORS(jbooleanArray, Boolean, jboolean)
CVM_DEFINE_JNI_ARRAY_REGION_ACCESSORS(jbyteArray,    Byte,    jbyte)
CVM_DEFINE_JNI_ARRAY_REGION_ACCESSORS(jcharArray,    Char,    jchar)
CVM_DEFINE_JNI_ARRAY_REGION_ACCESSORS(jshortArray,   Short,   jshort)
CVM_DEFINE_JNI_ARRAY_REGION_ACCESSORS(jintArray,     Int,     jint)
CVM_DEFINE_JNI_ARRAY_REGION_ACCESSORS(jfloatArray,   Float,   jfloat)
CVM_DEFINE_JNI_ARRAY_REGION_ACCESSORS(jlongArray,    Long,    jlong)
CVM_DEFINE_JNI_ARRAY_REGION_ACCESSORS(jdoubleArray,  Double,  jdouble)

#define CVM_DEFINE_JNI_ARRAY_ELEMENTS_ACCESSORS(arrType, jelemType, nativeType) \
nativeType* JNICALL							   \
CVMjniGet##jelemType##ArrayElements(JNIEnv *env, arrType array,		   \
                                    jboolean *isCopy)			   \
{									   \
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);				   \
    CVMJavaInt  arrLen;							   \
    nativeType* buf;							   \
    CVMArrayOf##jelemType##ICell* arrayCell = 				   \
        (CVMArrayOf##jelemType##ICell*)array;				   \
									   \
    CVMID_arrayGetLength(ee, arrayCell, arrLen);			   \
    buf = (nativeType*)malloc(arrLen * sizeof(nativeType));		   \
    if (buf == NULL) {							   \
	CVMthrowOutOfMemoryError(ee, NULL);    				   \
        return NULL;							   \
    }									   \
    if (isCopy != NULL) {						   \
        *isCopy = JNI_TRUE;						   \
    }								   	   \
    CVMD_gcUnsafeExec(ee, {						   \
	CVMArrayOf##jelemType* directArray =				   \
	    (CVMArrayOf##jelemType*)CVMID_icellDirect(ee, array);	   \
									   \
	/*								   \
	 * Make sure the java-side array and the passed in buffer    	   \
	 * have the same size elements.					   \
	 */								   \
	CVMassert(CVMarrayElemSize(CVMobjectGetClass(directArray)) == \
		  sizeof(nativeType));					   \
        CVMD_arrayReadBody##jelemType(buf, directArray, 0, arrLen);	   \
    });									   \
    return buf;								   \
}									   \
									   \
void JNICALL							   	   \
CVMjniRelease##jelemType##ArrayElements(JNIEnv *env, arrType array,	   \
                                        nativeType* buf, jint mode)   	   \
{									   \
    CVMExecEnv* ee;							   \
    if (mode == JNI_ABORT) {						   \
        free(buf);						   	   \
	return;								   \
    }									   \
    ee = CVMjniEnv2ExecEnv(env);				   	   \
    CVMD_gcUnsafeExec(ee, {						   \
	CVMArrayOf##jelemType* directArray =				   \
	    (CVMArrayOf##jelemType*)CVMID_icellDirect(ee, array);	   \
	CVMUint32 arrLen = CVMD_arrayGetLength(directArray);		   \
									   \
	/*								   \
	 * Make sure the java-side array and the passed in buffer    	   \
	 * have the same size elements.					   \
	 */								   \
	CVMassert(CVMarrayElemSize(CVMobjectGetClass(directArray)) == \
		  sizeof(nativeType));					   \
        CVMD_arrayWriteBody##jelemType(buf, directArray, 0, arrLen);	   \
    });									   \
    if (mode != JNI_COMMIT) {						   \
        CVMassert(mode == 0);						   \
        free(buf);						   	   \
    }									   \
}									   \

CVM_DEFINE_JNI_ARRAY_ELEMENTS_ACCESSORS(jbooleanArray, Boolean, jboolean)
CVM_DEFINE_JNI_ARRAY_ELEMENTS_ACCESSORS(jbyteArray,    Byte,    jbyte)
CVM_DEFINE_JNI_ARRAY_ELEMENTS_ACCESSORS(jcharArray,    Char,    jchar)
CVM_DEFINE_JNI_ARRAY_ELEMENTS_ACCESSORS(jshortArray,   Short,   jshort)
CVM_DEFINE_JNI_ARRAY_ELEMENTS_ACCESSORS(jintArray,     Int,     jint)
CVM_DEFINE_JNI_ARRAY_ELEMENTS_ACCESSORS(jfloatArray,   Float,   jfloat)
CVM_DEFINE_JNI_ARRAY_ELEMENTS_ACCESSORS(jlongArray,    Long,    jlong)
CVM_DEFINE_JNI_ARRAY_ELEMENTS_ACCESSORS(jdoubleArray,  Double,  jdouble)

/*
 * The unconditionally copying versions of
 * CVMjni{Get,Release}PrimitiveArrayCritical().
 */
void* JNICALL
CVMjniGetPrimitiveArrayCritical(JNIEnv *env, jarray array, jboolean *isCopy)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMJavaInt  arrLen;
    void*       buf;
    /* Just use an array of bytes -- we don't statically know what
       kind of array we have */
    CVMArrayOfByteICell* arrayCell = (CVMArrayOfByteICell*)array; 
    CVMClassBlock*       arrayCb;
    CVMJavaInt           elemSize; /* The size of an element of 'array' */

    CVMID_arrayGetLength(ee, arrayCell, arrLen);
    CVMID_objectGetClass(ee, (CVMObjectICell*)arrayCell, arrayCb);
    elemSize = CVMarrayElemSize(arrayCb);
    buf = malloc(arrLen * elemSize);
    if (buf == NULL) {
	CVMthrowOutOfMemoryError(ee, NULL);
        return NULL;
    }
    if (isCopy != NULL) {
	*isCopy = JNI_TRUE;
    }
    CVMD_gcUnsafeExec(ee, {
	CVMArrayOfAnyType* directArray =
	    (CVMArrayOfAnyType*)CVMID_icellDirect(ee, array);
	CVMClassBlock* arrCB = CVMobjectGetClass((CVMObject*)directArray);
	CVMBasicType   base = CVMarrayBaseType(arrCB);
	/*
	 * This had better be an array of primitives
	 */
	CVMassert(base != CVM_T_CLASS);
	CVMassert(CVMarrayDepth(arrCB) == 1);
	/* %comment rt015 */
	switch(base) {
	    case CVM_T_BOOLEAN: {
		CVMArrayOfBoolean* dirArr = (CVMArrayOfBoolean*)directArray;
		CVMD_arrayReadBodyBoolean(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_INT: {
		CVMArrayOfInt* dirArr = (CVMArrayOfInt*)directArray;
		CVMD_arrayReadBodyInt(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_FLOAT: {
		CVMArrayOfFloat* dirArr = (CVMArrayOfFloat*)directArray;
		CVMD_arrayReadBodyFloat(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_BYTE: {
		CVMArrayOfByte* dirArr = (CVMArrayOfByte*)directArray;
		CVMD_arrayReadBodyByte(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_CHAR: {
		CVMArrayOfChar* dirArr = (CVMArrayOfChar*)directArray;
		CVMD_arrayReadBodyChar(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_SHORT: {
		CVMArrayOfShort* dirArr = (CVMArrayOfShort*)directArray;
		CVMD_arrayReadBodyShort(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_LONG: {
		CVMArrayOfLong* dirArr = (CVMArrayOfLong*)directArray;
		CVMD_arrayReadBodyLong(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_DOUBLE: {
		CVMArrayOfDouble* dirArr = (CVMArrayOfDouble*)directArray;
		CVMD_arrayReadBodyDouble(buf, dirArr, 0, arrLen);
		break;
	    }
	    default: CVMassert(CVM_FALSE); /* can't happen */
	}
    });
    return buf;
}

void JNICALL
CVMjniReleasePrimitiveArrayCritical(JNIEnv *env, jarray array,
				    void* buf, jint mode)
{
    CVMExecEnv* ee;
    CVMJavaInt  arrLen;
    /* Just use an array of bytes -- we don't statically know what
       kind of array we have */
    CVMArrayOfByteICell* arrayCell;
    CVMClassBlock*       arrayCb;

    if (mode == JNI_ABORT) {
        free(buf);
	return;
    }
    arrayCell = (CVMArrayOfByteICell*)array;
    ee = CVMjniEnv2ExecEnv(env);

    CVMID_arrayGetLength(ee, arrayCell, arrLen);
    CVMID_objectGetClass(ee, (CVMObjectICell*)arrayCell, arrayCb);
    CVMD_gcUnsafeExec(ee, {
	CVMArrayOfAnyType* directArray =
	    (CVMArrayOfAnyType*)CVMID_icellDirect(ee, array);
	CVMClassBlock* arrCB = CVMobjectGetClass((CVMObject*)directArray);
	CVMBasicType   base = CVMarrayBaseType(arrCB);
	/*
	 * This had better be an array of primitives
	 */
	CVMassert(base != CVM_T_CLASS);
	CVMassert(CVMarrayDepth(arrCB) == 1);
	/* %comment rt015 */
	switch(base) {
	    case CVM_T_BOOLEAN: {
		CVMArrayOfBoolean* dirArr = (CVMArrayOfBoolean*)directArray;
		CVMD_arrayWriteBodyBoolean(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_INT: {
		CVMArrayOfInt* dirArr = (CVMArrayOfInt*)directArray;
		CVMD_arrayWriteBodyInt(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_FLOAT: {
		CVMArrayOfFloat* dirArr = (CVMArrayOfFloat*)directArray;
		CVMD_arrayWriteBodyFloat(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_BYTE: {
		CVMArrayOfByte* dirArr = (CVMArrayOfByte*)directArray;
		CVMD_arrayWriteBodyByte(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_CHAR: {
		CVMArrayOfChar* dirArr = (CVMArrayOfChar*)directArray;
		CVMD_arrayWriteBodyChar(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_SHORT: {
		CVMArrayOfShort* dirArr = (CVMArrayOfShort*)directArray;
		CVMD_arrayWriteBodyShort(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_LONG: {
		CVMArrayOfLong* dirArr = (CVMArrayOfLong*)directArray;
		CVMD_arrayWriteBodyLong(buf, dirArr, 0, arrLen);
		break;
	    }
	    case CVM_T_DOUBLE: {
		CVMArrayOfDouble* dirArr = (CVMArrayOfDouble*)directArray;
		CVMD_arrayWriteBodyDouble(buf, dirArr, 0, arrLen);
		break;
	    }
	    default: CVMassert(CVM_FALSE); /* can't happen */
	}
    });
    if (mode != JNI_COMMIT) {
        CVMassert(mode == 0);
        free(buf);
    }
}

#define CVM_DEFINE_JNI_ARRAY_ALLOCATOR(jType_, typeName_, typeCode_)	\
jType_ JNICALL								\
CVMjniNew##typeName_##Array(JNIEnv* env, jsize length)			\
{									\
    CVMExecEnv*     ee         = CVMjniEnv2ExecEnv(env);		\
    CVMObjectICell* resultCell = CVMjniCreateLocalRef(ee);		\
    CVMClassBlock*  arrayCb;						\
									\
    if (resultCell == NULL) {						\
	return NULL;							\
    }									\
									\
    arrayCb = (CVMClassBlock*)CVMbasicTypeArrayClassblocks[typeCode_];	\
    CVMID_allocNewArray(ee, typeCode_, arrayCb, length, resultCell);	\
    if (CVMID_icellIsNull(resultCell)) {				\
        CVMjniDeleteLocalRef(env, resultCell);				\
	CVMthrowOutOfMemoryError(ee, NULL);    				\
	return NULL;							\
    } else {								\
	return resultCell;						\
    }									\
}									\

CVM_DEFINE_JNI_ARRAY_ALLOCATOR(jbooleanArray, Boolean, CVM_T_BOOLEAN)
CVM_DEFINE_JNI_ARRAY_ALLOCATOR(jbyteArray,    Byte,    CVM_T_BYTE)
CVM_DEFINE_JNI_ARRAY_ALLOCATOR(jcharArray,    Char,    CVM_T_CHAR)
CVM_DEFINE_JNI_ARRAY_ALLOCATOR(jshortArray,   Short,   CVM_T_SHORT)
CVM_DEFINE_JNI_ARRAY_ALLOCATOR(jintArray,     Int,     CVM_T_INT)
CVM_DEFINE_JNI_ARRAY_ALLOCATOR(jfloatArray,   Float,   CVM_T_FLOAT)
CVM_DEFINE_JNI_ARRAY_ALLOCATOR(jlongArray,    Long,    CVM_T_LONG)
CVM_DEFINE_JNI_ARRAY_ALLOCATOR(jdoubleArray,  Double,  CVM_T_DOUBLE)

jobject JNICALL
CVMjniGetObjectArrayElement(JNIEnv* env, jarray arrArg, jsize index)
{
    CVMExecEnv*     ee         = CVMjniEnv2ExecEnv(env);
    CVMObjectICell* resultCell;
    CVMArrayOfAnyTypeICell* array   = (CVMArrayOfAnyTypeICell*)arrArg;
    jsize arrayLen;

    CVMassert(array != NULL);
    CVMassert(!CVMID_icellIsNull(array));

    CVMID_arrayGetLength(ee, array, arrayLen);
    if ((CVMUint32)index >= (CVMUint32)arrayLen) {
	CVMthrowArrayIndexOutOfBoundsException(ee, NULL);
	return NULL;
    }

    resultCell = CVMjniCreateLocalRef(ee);
    if (resultCell == NULL) {
	return NULL;
    }

    CVMID_arrayReadRef(ee, array, index, resultCell);
    if (CVMID_icellIsNull(resultCell)) {
	CVMjniDeleteLocalRef(env, resultCell);
	return NULL;
    } else {
	return resultCell;
    }
}

/*
 * Just like 'aastore'
 */
void JNICALL
CVMjniSetObjectArrayElement(JNIEnv* env, jarray arrArg,
			    jsize index, jobject value)
{
    CVMExecEnv*     ee         = CVMjniEnv2ExecEnv(env);
    CVMArrayOfAnyTypeICell* array   = (CVMArrayOfAnyTypeICell*)arrArg;
    jsize arrayLen;

    CVMassert(array != NULL);
    CVMassert(!CVMID_icellIsNull(array));

    CVMID_arrayGetLength(ee, array, arrayLen);
    if ((CVMUint32)index >= (CVMUint32)arrayLen) {
	CVMthrowArrayIndexOutOfBoundsException(ee, NULL);
	return;
    }

    if ((value != NULL) && !CVMID_icellIsNull(value)) {
	CVMBool assignmentOk;
	/* Added for reference checking
	 * 	Pratik & Kunal 
	 **/
	CVMBool ret = CVM_TRUE;
	extern int JVMCeiInitialized;
	if(JVMCeiInitialized) {
		CVMD_gcUnsafeExec( ee,
			{
				ret = CVMJTrcAllowReferenceAssignment(ee, CVMID_icellDirect(ee,arrArg),
				                                      CVMID_icellDirect(ee,value));
			});
	}

	if(ret == CVM_FALSE){
		CVMthrowIllegalAssignmentError(ee,"CVMSetObjectArrayElement(): 
				                            IllegalAssignment Error \n");
		return;
	}

	CVMD_gcUnsafeExec(ee, {
	    /* Check assignability of 'value' into 'array' */
	    CVMClassBlock* arrayType = 
		CVMobjectGetClass(CVMID_icellDirect(ee, array));
	    CVMClassBlock* elemType =
		CVMarrayElementCb(arrayType);
	    CVMClassBlock* rhsType =
		CVMobjectGetClass(CVMID_icellDirect(ee, value));

	    if (rhsType != elemType) { /* Try a quick check first */
		assignmentOk = CVMisAssignable(ee, rhsType, elemType);
	    } else {
		assignmentOk = CVM_TRUE;
	    }
	});
	if (!assignmentOk) {
	    CVMthrowArrayStoreException(ee, NULL);
	    return;
	}
    }
    CVMID_arrayWriteRef(ee, array, index, CVMjniNonNullICellPtrFor(value));
}

/*
 * Just like anewarray, except for the initialization part.
 */
jarray JNICALL
CVMjniNewObjectArray(JNIEnv* env, jsize length,
		     jclass elementClass, jobject initialElement)
{
    CVMExecEnv*     ee         = CVMjniEnv2ExecEnv(env);
    CVMObjectICell* resultCell = CVMjniCreateLocalRef(ee);
    CVMClassBlock*  arrayCb;
    CVMClassBlock*  elementClassCb;

    if (resultCell == NULL) {
	return NULL;
    }

    elementClassCb = CVMjniGcSafeRef2Class(ee, elementClass);
    arrayCb = CVMclassGetArrayOf(ee, elementClassCb);
    if (arrayCb == NULL) {
	CVMjniDeleteLocalRef(env, resultCell);
	return NULL;
    }

    CVMID_allocNewArray(ee, CVM_T_CLASS, arrayCb, length, resultCell);
    if (CVMID_icellIsNull(resultCell)) {
	CVMjniDeleteLocalRef(env, resultCell);
        CVMthrowOutOfMemoryError(ee, NULL);
	return NULL;
    }
    if ((initialElement == NULL) ||
	CVMID_icellIsNull(initialElement)) {
	/* No need to initialize to NULL. The allocation already did that */
	return resultCell;
    } else {
	/* Check for assignability before initialization */
	CVMBool assignmentOk = CVM_TRUE;
	CVMD_gcUnsafeExec(ee, {
	    CVMObject* directInitObj = CVMID_icellDirect(ee, initialElement);
	    CVMClassBlock*   rhsType = CVMobjectGetClass(directInitObj);
	    
	    if (rhsType != elementClassCb) { /* Try a quick check first */
		assignmentOk = CVMisAssignable(ee, rhsType, elementClassCb);
	    } else {
		assignmentOk = CVM_TRUE;
	    }
	    if (assignmentOk) {
		CVMArrayOfRef* refArray =
		    (CVMArrayOfRef*)CVMID_icellDirect(ee, resultCell);
		int i;

		/* %comment: rt016 */
		for (i = 0; i < length; i++) {
		    CVMD_arrayWriteRef(refArray, i, directInitObj);
		}
	    } else {
		CVMjniDeleteLocalRef(env, resultCell);
		resultCell = NULL;
		CVMthrowArrayStoreException(ee, NULL);
	    }
	});
	return resultCell;
    }
}

jsize JNICALL
CVMjniGetArrayLength(JNIEnv* env, jarray arrArg)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMArrayOfAnyTypeICell* array   = (CVMArrayOfAnyTypeICell*)arrArg;
    jsize arrayLen;

    CVMID_arrayGetLength(ee, array, arrayLen);
    return arrayLen;
}

/*
 * Class operations
 */
jclass JNICALL
CVMjniGetSuperclass(JNIEnv* env, jclass clazz)
{
    CVMExecEnv*    ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = CVMjniGcSafeRef2Class(ee, clazz);
    CVMClassBlock* superCb;

    if (CVMcbIs(cb, INTERFACE)) {
	return NULL;
    }
    superCb = CVMcbSuperclass(cb);
    if (superCb == NULL) {
	return NULL;
    }
    return CVMjniNewLocalRef(env, CVMcbJavaInstance(superCb));
}


jboolean JNICALL
CVMjniIsAssignableFrom(JNIEnv* env, jclass clazz1, jclass clazz2)
{
    CVMExecEnv*    ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* srcCb;
    CVMClassBlock* dstCb;

    CVMD_gcUnsafeExec(ee, {
	srcCb = CVMjniGcUnsafeRef2Class(ee, clazz1);
	dstCb = CVMjniGcUnsafeRef2Class(ee, clazz2);
    });
    return CVMisAssignable(ee, srcCb, dstCb);
}

/*
 * Reflection operations
 */

/*
 * NOTE: The following routines are conditionally compiled:
 * CVMjniToReflectedMethod
 * CVMjniToReflectedField
 * CVMjniFromReflectedMethod
 * CVMjniFromReflectedField
 */

#ifdef CVM_REFLECT

jobject JNICALL
CVMjniToReflectedMethod(JNIEnv *env, jclass clazz,
			jmethodID methodID, jboolean isStatic)
{
    CVMExecEnv*     ee = CVMjniEnv2ExecEnv(env);
    CVMMethodBlock* mb = (CVMMethodBlock *)methodID;
    CVMObjectICell* resultCell; /* The resulting java.lang.reflect.Method */

    resultCell = CVMjniCreateLocalRef(ee);
    if (resultCell != NULL) {
	CVMreflectMethodBlockToNewJavaMethod(ee, mb, resultCell);
    }
    return resultCell; /* Exception already thrown if resultCell == 0 */
}

jobject JNICALL
CVMjniToReflectedField(JNIEnv *env, jclass clazz, jfieldID fieldID,
		       jboolean isStatic)
{
    CVMExecEnv*     ee = CVMjniEnv2ExecEnv(env);
    CVMFieldBlock*  fb = (CVMFieldBlock *)fieldID;
    CVMObjectICell* resultCell; /* The resulting java.lang.reflect.Field */

    resultCell = CVMjniCreateLocalRef(ee);
    if (resultCell != NULL) {
	CVMreflectNewJavaLangReflectField(ee, fb, resultCell);
    }
    return resultCell; /* Exception already thrown if resultCell == 0 */
}

jmethodID JNICALL
CVMjniFromReflectedMethod(JNIEnv* env, jobject method)
{
    CVMExecEnv*     ee = CVMjniEnv2ExecEnv(env);
    return CVMreflectGCSafeGetMethodBlock(ee, method, NULL);
}

jfieldID JNICALL
CVMjniFromReflectedField(JNIEnv* env, jobject field)
{
    CVMExecEnv*     ee = CVMjniEnv2ExecEnv(env);
    return CVMreflectGCSafeGetFieldBlock(ee, field);
}

#else
# define CVMjniToReflectedMethod NULL
# define CVMjniToReflectedField NULL
# define CVMjniFromReflectedMethod NULL
# define CVMjniFromReflectedField NULL
#endif /* CVM_REFLECT */

/*
 * String operations
 */
#ifdef CVM_DEBUG
static void
CVMjniSanityCheckStringAccess(CVMExecEnv* ee, jstring string)
{
    CVMClassBlock* cb;
    CVMassert(string != NULL);
    CVMID_objectGetClass(ee, string, cb);
    CVMassert(cb == CVMsystemClass(java_lang_String));
}
#else
#define CVMjniSanityCheckStringAccess(ee, string)
#endif

jsize JNICALL
CVMjniGetStringLength(JNIEnv* env, jstring string)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    jsize strLen;

    CVMjniSanityCheckStringAccess(ee, string);
    CVMID_fieldReadInt(ee, string, 
		       CVMoffsetOfjava_lang_String_count,
		       strLen);
    return strLen;
}

jstring JNICALL
CVMjniNewString(JNIEnv *env, const jchar* unicodeChars, jsize len)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMStringICell* stringICell = CVMjniCreateLocalRef(ee);
    if (stringICell == NULL) {
	return NULL;
    }

    CVMnewString(ee, stringICell, unicodeChars, len);
    if (CVMID_icellIsNull(stringICell)) {
	CVMjniDeleteLocalRef(env, stringICell);
	CVMthrowOutOfMemoryError(ee, NULL);
	return NULL;
    } else {
	return stringICell;
    }
}

jsize JNICALL
CVMjniGetStringUTFLength(JNIEnv *env, jstring string)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    jsize length;

    CVMjniSanityCheckStringAccess(ee, string);

    CVMID_localrootBegin(ee); {
	CVMID_localrootDeclare(CVMArrayOfCharICell, valueICellPtr);
	CVMObjectICell *valueAsObject = (CVMObjectICell *)valueICellPtr;
	CVMJavaChar c;
	CVMInt32 i, offset, len, readLen;

	CVMID_fieldReadRef(ee, string, 
			   CVMoffsetOfjava_lang_String_value,
			   valueAsObject);
	CVMID_fieldReadInt(ee, string, 
			   CVMoffsetOfjava_lang_String_offset,
			   offset);
	CVMID_fieldReadInt(ee, string, 
			   CVMoffsetOfjava_lang_String_count,
			   len);
	/*
	 * The repeated entries to CVMD_gcUnsafeExec are
	 * somewhat slow, but give GC a chance to
	 * run for very long strings.
	 */
	length = 0;
	while ( len > 0 ){
	    readLen = MIN( len, MAX_CHARS_PER_ACCESS );
	    CVMD_gcUnsafeExec(ee,{
		CVMArrayOfChar* theChars;
		theChars = CVMID_icellDirect(ee, valueICellPtr);
		for ( i = 0; i < readLen; i++ ){
		    CVMD_arrayReadChar( theChars, offset, c );
		    /*
		     * Because CVMunicodeChar2UtfLength is such a
		     * trivial function (a macro actually), I don't
		     * feel too bad about invoking it within this
		     * unsafe section. The alternative is to copy into an
		     * array and then iterate over it with the length
		     * function.
		     */
		    length += CVMunicodeChar2UtfLength(c);
		    offset++;
		}

	    });
	    len -= readLen;
	}
    } CVMID_localrootEnd();
    return length;
}

void JNICALL
CVMjniGetStringRegion(JNIEnv *env,
		      jstring string, jsize start, jsize len, jchar *buf)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);

    CVMjniSanityCheckStringAccess(ee, string);

    CVMID_localrootBegin(ee); {
	CVMID_localrootDeclare(CVMArrayOfCharICell, valueICellPtr);
	CVMObjectICell *valueAsObject = (CVMObjectICell *)valueICellPtr;
	CVMInt32 offset, str_len;

	CVMID_fieldReadRef(ee, string, 
			   CVMoffsetOfjava_lang_String_value,
			   valueAsObject);
	CVMID_fieldReadInt(ee, string, 
			   CVMoffsetOfjava_lang_String_offset,
			   offset);
	CVMID_fieldReadInt(ee, string, 
			   CVMoffsetOfjava_lang_String_count,
			   str_len);

	if ((start >= 0) && (len >= 0) &&
	    ((CVMUint32)start + (CVMUint32)len <= (CVMUint32)str_len)) {
	    CVMD_gcUnsafeExec(ee, {
		CVMArrayOfChar* charArray =
		    (CVMArrayOfChar*)CVMID_icellDirect(ee, valueAsObject);
		CVMD_arrayReadBodyChar(buf, charArray, offset + start, len);
	    });
	} else {
	    CVMthrowStringIndexOutOfBoundsException(ee, NULL);
	}
    } CVMID_localrootEnd();
}

void JNICALL
CVMjniGetStringUTFRegion(JNIEnv *env,
			 jstring string, jsize start, jsize len, char *buf)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);

    CVMjniSanityCheckStringAccess(ee, string);

    CVMID_localrootBegin(ee); {
	CVMID_localrootDeclare(CVMArrayOfCharICell, valueICellPtr);
	CVMObjectICell *valueAsObject = (CVMObjectICell *)valueICellPtr;
	CVMJavaChar charBuffer[MAX_CHARS_PER_ACCESS];
	CVMInt32 offset, str_len, bufferLen;

	CVMID_fieldReadRef(ee, string, 
			   CVMoffsetOfjava_lang_String_value,
			   valueAsObject);
	CVMID_fieldReadInt(ee, string, 
			   CVMoffsetOfjava_lang_String_offset,
			   offset);
	CVMID_fieldReadInt(ee, string, 
			   CVMoffsetOfjava_lang_String_count,
			   str_len);

	if ((start >= 0) && (len >= 0) &&
	    ((CVMUint32)start + (CVMUint32)len <= (CVMUint32)str_len)) {
	    /*
	     * This is somewhat slow, but gives GC a chance to
	     * run for very long strings.
	     */
	    offset += start;
	    while ( len > 0 ){
		bufferLen = MIN( len, MAX_CHARS_PER_ACCESS );
		CVMD_gcUnsafeExec(ee,{
		    CVMArrayOfChar* theChars;

		    theChars = CVMID_icellDirect(ee, valueICellPtr);
		    CVMD_arrayReadBodyChar( charBuffer, theChars, offset, bufferLen );

		});
		buf = CVMutfCopyFromCharArray(charBuffer, buf, bufferLen);
		len -= bufferLen;
		offset += bufferLen;
	    }
	    *buf = '\0';
	} else {
	    CVMthrowStringIndexOutOfBoundsException(ee, NULL);
	}
    } CVMID_localrootEnd();
}

const char* JNICALL
CVMjniGetStringUTFChars(JNIEnv *env, jstring string, jboolean *isCopy)
{
    char *utf;
    jsize strLen = CVMjniGetStringLength(env, string);
    jsize utfLen = CVMjniGetStringUTFLength(env, string);

    utf = malloc(utfLen + 1);
    if (utf == NULL) {
        CVMthrowOutOfMemoryError(CVMjniEnv2ExecEnv(env), NULL);
	return NULL;
    }
    CVMjniGetStringUTFRegion(env, string, 0, strLen, utf);
    if (isCopy) {
        *isCopy = JNI_TRUE;
    }
    return (const char*)utf;
}

void JNICALL
CVMjniReleaseStringUTFChars(JNIEnv *env, jstring str, const char *chars)
{
    free((void *)chars);
}

const jchar* JNICALL
CVMjniGetStringChars(JNIEnv *env, jstring string, jboolean *isCopy)
{
    jchar *chars;
    jsize strLen = CVMjniGetStringLength(env, string);

    chars = malloc(sizeof(jchar) * strLen);
    if (chars == NULL) {
        CVMthrowOutOfMemoryError(CVMjniEnv2ExecEnv(env), NULL);
	return NULL;
    }
    CVMjniGetStringRegion(env, string, 0, strLen, chars);
    if (isCopy) {
        *isCopy = JNI_TRUE;
    }
    return (const jchar*)chars;
}

void JNICALL
CVMjniReleaseStringChars(JNIEnv *env, jstring str, const jchar *chars)
{
    free((void *)chars);
}

/*
 * Constructs a new java.lang.String object from an array of UTF-8 characters. 
 *
 * We could avoid the temporary buffer if we copied directly into the
 * char[], but that would require going through the indirect layer.
 */
jstring JNICALL
CVMjniNewStringUTF(JNIEnv* env, const char* utf8Bytes)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMStringICell* stringICell = CVMjniCreateLocalRef(ee);
    if (stringICell == NULL) {
	return NULL;
    }

    CVMnewStringUTF(ee, stringICell, utf8Bytes);
    if (CVMID_icellIsNull(stringICell)) {
	CVMjniDeleteLocalRef(env, stringICell);
	CVMthrowOutOfMemoryError(ee, NULL);
	return NULL;
    } else {
	return stringICell;
    }
}

/* ROBBINS SEZ:
 * Mod'd to allow self posting of AI Exceptions ala athrow.
 */

jint JNICALL
CVMjniThrow(JNIEnv *env, jthrowable obj)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMBool     isThrowable;

    static jclass aieClass = NULL;

    if(aieClass == NULL)
    {
      // Modified by Pratik so that aieClass is a global ref and not a 
      // local ref.
      jclass tempAIE = (*env)->FindClass(env,
			  "javax/realtime/AsynchronouslyInterruptedException");
      if(tempAIE == NULL){
	CVMjniFatalError(env,
			 "jniThrow unable to load javax.realtime.AsynchronouslyInterruptedException");
      }
      aieClass = (*env)->NewGlobalRef(env, tempAIE);
    }

    CVMD_gcUnsafeExec(ee, {
	isThrowable = CVMgcUnsafeIsInstanceOf(
             ee, CVMID_icellDirect(ee, obj), 
	     CVMsystemClass(java_lang_Throwable));
    });
    if (!isThrowable) {
	CVMjniFatalError(env, 
	    "throw an object that is not an instance of java/lang/Throwable");
    }

#ifdef JDK12_BUG
    /* Fill in the backtrace into the exception object. */
    CVMfillInStackTrace(ee, obj);
#endif
 
    if((*env)->IsInstanceOf(env,obj,aieClass)){
      if(((*env)->MonitorEnter(env, (jobject)CVMcurrentThreadICell(ee))) != JNI_OK)
	CVMjniFatalError(env,"unable to enter current thread object monitor while posting AIE from jniThrow");
      
      /* ROBBINS SEZ: follow the same rules as the athrow opcode */
      if(!CVMpendingAIException(ee)){
	CVMgcUnsafeSetPendingAIException(ee,ee,CVMID_icellDirect(ee,obj));
      }

      if(((*env)->MonitorEnter(env, (jobject)CVMcurrentThreadICell(ee))) != JNI_OK)
	CVMjniFatalError(env,"unable to exit current thread object monitor while posting AIE from jniThrow");
    }
    else{
      CVMgcSafeThrowLocalException(CVMjniEnv2ExecEnv(env), obj);
    }
    return JNI_OK;
}

jint JNICALL
CVMjniThrowNew(JNIEnv *env, jclass clazz, const char *message)
{
    jint result;
    jstring msg = NULL;
    jobject obj = NULL;
    jmethodID methodID;

    if (CVMjniEnsureLocalCapacity(env, 2) != JNI_OK) {
        return JNI_ENOMEM;
    }

    methodID = CVMjniGetMethodID(env, clazz, "<init>",
	"(Ljava/lang/String;)V");

    if (methodID == NULL) {
	/* Exception already thrown */
	result = JNI_ERR;
	goto done;
    }

    if (message != NULL) {
	msg = CVMjniNewStringUTF(env, message);
	if (msg == NULL) {
	    result = JNI_ERR;
	    goto done;
	}
    } else {
	msg = NULL;
    }

    obj = CVMjniNewObject(env, clazz, methodID, msg);
    if (obj == NULL) {
        result = JNI_ERR;
	goto done;
    }

    result = CVMjniThrow(env, obj);

done:
    if (msg != NULL) {
	CVMjniDeleteLocalRef(env, msg);
    }
    if (obj != NULL) {
	CVMjniDeleteLocalRef(env, obj);
    }
    return result;
}

/* ROBBINS SEZ:
 * also check for pending AIE's */

jthrowable JNICALL
CVMjniExceptionOccurred(JNIEnv *env)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMThrowableICell* exceptionICell = NULL;

    if (CVMlocalExceptionOccurred(ee)) {
	exceptionICell = CVMlocalExceptionICell(ee);
    } else if (CVMremoteExceptionOccurred(ee)) {
	exceptionICell = CVMremoteExceptionICell(ee);
    }

    if (exceptionICell != NULL) {
        /* turn into local ref */
        return CVMjniNewLocalRef(env, exceptionICell);
    } else {
	return NULL;
    }
}

/* ROBBINS SEZ:
 * also check for pending AIE's */

jboolean JNICALL
CVMjniExceptionCheck(JNIEnv *env)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);

    if (CVMexceptionOccurred(ee)) {
	return JNI_TRUE;
    } else {
	return JNI_FALSE;
    }
}

void JNICALL
CVMjniExceptionClear(JNIEnv *env)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    /* %comment: rt017 */
    CVMclearLocalException(ee);
    CVMclearRemoteException(ee);
}

/* NOTE:  If we decide later to get rid of the prototypes just
   added to jni_impl.h, we will need to forward declare
   CVMjniCallVoidMethod here because CVMjniExceptionDescribe uses it
   before it's defined in this file. */

void JNICALL
CVMjniExceptionDescribe(JNIEnv *env)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    jobject localException = NULL;
    jobject remoteException = NULL;
    jobject throwable;
    jclass clazz;
    jmethodID method;

    if ((!CVMexceptionOccurred(ee))) {
	return;
    }

    if (CVMjniEnsureLocalCapacity(env, 3) != JNI_OK) {
        return;
    }

    if (CVMlocalExceptionOccurred(ee)) {
	localException =
	    (*env)->NewLocalRef(env, CVMlocalExceptionICell(ee));
	if (localException == NULL)
	    return;
    }

    if (CVMremoteExceptionOccurred(ee)) {
	remoteException =
	    (*env)->NewLocalRef(env, CVMremoteExceptionICell(ee));
	if (remoteException == NULL) {
	    if (localException != NULL) {
		CVMjniDeleteLocalRef(env, localException);
	    }
	    return;
	}
    }
    
    CVMclearLocalException(ee);
    CVMclearRemoteException(ee);

    if (localException != NULL) {
	throwable = localException;
    } else {
	CVMassert(remoteException != NULL);
	throwable = remoteException;
    }

    /*
    if (CVMlocalExceptionOccurred(ee)) 
      printf("[CVMjniExceptionDescribe] exception before getObjectclass\n");
    */

    clazz = CVMjniGetObjectClass(env, throwable);

    /*
    if (CVMlocalExceptionOccurred(ee)) 
      printf("[CVMjniExceptionDescribe] exception after getObjectclass\n");
    */

    CVMassert(clazz != NULL);
    method = CVMjniGetMethodID(env, clazz, "printStackTrace", "()V");
    CVMjniDeleteLocalRef(env, clazz);
    CVMassert(method != NULL);

    /*
    if (CVMlocalExceptionOccurred(ee)) 
      printf("[CVMjniExceptionDescribe] exception before before callvoid method\n");
    */

   CVMjniCallVoidMethod(env, throwable, method);

    /* If printStackTrace threw an exception, then print a message. */
    if (CVMlocalExceptionOccurred(ee)) {
	/* NOTE: this has happened in the past because of static
           initializers failing to get run, causing the system class
           not to be initialized and System.out being NULL, causing a
           NullPointerException while printing the stack trace. */
	CVMconsolePrintf("CVMjniExceptionDescribe failed: "
			 "couldn't print stack trace!\n");
	CVMclearLocalException(ee);
    }

    /* Rethrow the original exception(s). */
    if (localException != NULL) {
	CVMgcSafeThrowLocalException(ee, localException);
	CVMjniDeleteLocalRef(env, localException);
    }
    if (remoteException != NULL) {
	CVMgcSafeThrowRemoteException(ee, ee, remoteException);
	CVMjniDeleteLocalRef(env, remoteException);
    }
}

/*
 * Methods calls
 */

#undef SAFE_COUNT
#define SAFE_COUNT 32	/* number of arguments to push between GC checkpoint */ 

static char
CVMjniPushArgumentsVararg(JNIEnv *env,
			  CVMterseSigIterator *terse_signature,
			  CVMFrame *current_frame, void *a)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMStackVal32 *topOfStack = current_frame->topOfStack;
    int safeCount = SAFE_COUNT;
    va_list *args = (va_list *)a;
    int typeSyllable;

    while ((typeSyllable=CVM_TERSE_ITER_NEXT(*terse_signature))!=CVM_TYPEID_ENDFUNC) {
	/* Perform GC checkpoint every so often */
	if (safeCount-- == 0) {
	    CVMD_gcSafeCheckPoint(ee, {
		current_frame->topOfStack = topOfStack;
	    }, {});
	    safeCount = SAFE_COUNT;
	}

        switch (typeSyllable) {
	case CVM_TYPEID_BOOLEAN:
	case CVM_TYPEID_SHORT:
	case CVM_TYPEID_BYTE:
	case CVM_TYPEID_CHAR:
	case CVM_TYPEID_INT:
            (topOfStack++)->j.i = va_arg(*args, jint);
            continue;
	case CVM_TYPEID_FLOAT:
            (topOfStack++)->j.f = va_arg(*args, jdouble);
            continue;
        case CVM_TYPEID_OBJ: {
	    jobject obj = va_arg(*args, jobject);
	    CVMID_icellAssignDirect(ee, &topOfStack->j.r,
		CVMjniNonNullICellPtrFor(obj));
	    topOfStack++;
	    continue;
	}
	case CVM_TYPEID_LONG: {
	    jlong l = va_arg(*args, jlong);
	    CVMlong2Jvm(&topOfStack->j.raw, l);
            topOfStack += 2;
            continue;
	}
	case CVM_TYPEID_DOUBLE: {
	    jdouble d = va_arg(*args, jdouble);
	    CVMdouble2Jvm(&topOfStack->j.raw, d);
            topOfStack += 2;
            continue;
	}
	default:
	    /* This should never happen */
	    CVMassert(CVM_FALSE);
	    return 0;
        }
    }
    /* fell out of while loop. Return return type */
    current_frame->topOfStack = topOfStack;
    va_end(*args);
    return CVM_TERSE_ITER_RETURNTYPE(*terse_signature);
}

static char
CVMjniPushArgumentsArray(JNIEnv *env,
			 CVMterseSigIterator *terse_signature,
			 CVMFrame *current_frame, void *a)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMStackVal32 *topOfStack = current_frame->topOfStack;
    int safeCount = SAFE_COUNT;
    jvalue *args = (jvalue *)a;
    int typeSyllable;
    
    while ((typeSyllable=CVM_TERSE_ITER_NEXT(*terse_signature))!=CVM_TYPEID_ENDFUNC) {
	/* Perform GC checkpoint every so often */
	if (safeCount-- == 0) {
	    CVMD_gcSafeCheckPoint(ee, {
		current_frame->topOfStack = topOfStack;
	    }, {});
	    safeCount = SAFE_COUNT;
	}

        switch (typeSyllable) {
	case CVM_TYPEID_BOOLEAN:
            (topOfStack++)->j.i = (*args++).z;
            continue;
	case CVM_TYPEID_SHORT:
            (topOfStack++)->j.i = (*args++).s;
            continue;
	case CVM_TYPEID_BYTE:
            (topOfStack++)->j.i = (*args++).b;
            continue;
	case CVM_TYPEID_CHAR:
            (topOfStack++)->j.i = (*args++).c;
            continue;
	case CVM_TYPEID_INT:
            (topOfStack++)->j.i = (*args++).i;
            continue;
	case CVM_TYPEID_FLOAT:
	    (topOfStack++)->j.f = (*args++).f;
            continue;
	case CVM_TYPEID_OBJ: {
	    jobject obj = (*args++).l;
	    CVMID_icellAssignDirect(ee, &topOfStack->j.r,
		CVMjniNonNullICellPtrFor(obj));
	    topOfStack++;
	    continue;
	}
	case CVM_TYPEID_LONG: {
	    jlong l = (*args++).j;
	    CVMlong2Jvm(&topOfStack->j.raw, l);
            topOfStack += 2;
            continue;
	}
	case CVM_TYPEID_DOUBLE: {
	    jdouble d = (*args++).d;
	    CVMdouble2Jvm(&topOfStack->j.raw, d);
            topOfStack += 2;
            continue;
	}
	default:
            CVMassert(CVM_FALSE);
	    return 0;
        }
    }
    /* fell out of loop. return return type */
    current_frame->topOfStack = topOfStack;
    return CVM_TERSE_ITER_RETURNTYPE(*terse_signature);
}

static void
CVMjniInvoke(JNIEnv *env, jobject obj, jmethodID methodID,
    JNI_PushArguments_t pushArguments, void *args,
    CVMUint16 info, jvalue *retValue)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock *cb;
    CVMFrame*     frame;
    CVMInt32 invokeType = (info & CVM_INVOKE_TYPE_MASK);
    char retType = (info & CVM_INVOKE_RETTYPE_MASK);
    CVMBool isStatic = (invokeType == CVM_INVOKE_STATIC);
    CVMBool isVirtual = (invokeType == CVM_INVOKE_VIRTUAL);
    
    CVMassert(CVMframeIsFreelist(CVMeeGetCurrentFrame(ee)));
    
    if (CVMexceptionOccurred(ee)) {
#ifdef CVM_DEBUG
	CVMconsolePrintf("Pending exception entering CVMjniInvoke!\n");
	CVMjniExceptionDescribe(env);
	CVMassertHook();
#endif
	return;
    }

    if (retType == CVM_TYPEID_OBJ) {
	/* Caller should have ensured capacity */
	retValue->l = CVMjniCreateLocalRef(ee);
	/* We should never reach here if the ref creation fails */
	CVMassert(retValue->l != NULL);
    }

    /* Create the transition frame.*/
    CVMD_gcUnsafeExec(ee, {
	if (isStatic) {
	    cb = CVMjniGcUnsafeRef2Class(ee, obj);
	} else {
	    cb = CVMobjectGetClass(CVMID_icellDirect(ee, obj));
	}
	frame = (CVMFrame*)CVMpushTransitionFrame(ee, methodID);
    });
    if (frame == NULL) {
	if (retType == CVM_TYPEID_OBJ) {
	    CVMjniDeleteLocalRef(env, retValue->l);
	}
	CVMassert(CVMexceptionOccurred(ee));
	return;  /* exception already thrown */
    }

    // if (CVMlocalExceptionOccurred(ee)) printf("[jni_invoke] exception 1\n");

    CVMassert(!CVMlocalExceptionOccurred(ee));

    /* Push arguments and invoke the method. */
    CVMD_gcUnsafeExec(ee, {
	CVMBool resultIsNullObjectRef = CVM_FALSE;
	/* Push the "this" argument if the method is not static. */
	if (!isStatic) {
	    CVMID_icellAssignDirect(ee, &frame->topOfStack->j.r, obj);
	    frame->topOfStack++;
	}

	/* Push the all the other arguments. */
	{
	    CVMterseSigIterator terseSig;
	    char       	methodRetType;
	    CVMMethodBlock *mb = methodID;

	    CVMtypeidGetTerseSignatureIterator(CVMmbNameAndTypeID(mb), 
					       &terseSig);

	    methodRetType = (*pushArguments)(env, &terseSig, frame, args);
	    if (methodRetType != retType) {
		CVMjniFatalError(env, "Native code expects wrong return type"
				 "from Java callback");
	    }
	}

	CVMassert(!CVMlocalExceptionOccurred(ee));

	/* Call the java method. */
    CVMgcUnsafeExecuteJavaMethod(ee, methodID, isStatic, isVirtual);

	/* 
	 * Copy the result. frame->topOfStack will point after the result.
	 */
	if (!CVMlocalExceptionOccurred(ee)) {
	    switch (retType) {
	    case CVM_TYPEID_VOID:
		break;
	    case CVM_TYPEID_BOOLEAN:
		retValue->z = frame->topOfStack[-1].j.i;
		break;
	    case CVM_TYPEID_BYTE:
		retValue->b = frame->topOfStack[-1].j.i;
		break;
	    case CVM_TYPEID_CHAR:
		retValue->c = frame->topOfStack[-1].j.i;
		break;
	    case CVM_TYPEID_SHORT:
		retValue->s = frame->topOfStack[-1].j.i;
		break;
	    case CVM_TYPEID_INT:
		retValue->i = frame->topOfStack[-1].j.i;
		break;
	    case CVM_TYPEID_LONG: {
		retValue->j = CVMjvm2Long(&frame->topOfStack[-2].j.raw);
		break;
	    }
	    case CVM_TYPEID_FLOAT:
		retValue->f = (jfloat)frame->topOfStack[-1].j.f;
		break;
	    case CVM_TYPEID_DOUBLE: {
		retValue->d = CVMjvm2Double(&frame->topOfStack[-2].j.raw);
		break;
	    }
	    case CVM_TYPEID_OBJ:
		CVMID_icellAssignDirect(ee, retValue->l,
					&frame->topOfStack[-1].j.r);
		break;
	    }
	}
	/*
	else
	  {
	    printf("[jni_invoke] exception  after CVMgcUnsafeExecuteJavaMethod \n");
	  }
	*/

	/* Regardless of whether an exception occurred, we must not
	   allow an ICell containing NULL to escape to the caller. */
	if ((retType == CVM_TYPEID_OBJ) &&
	    CVMID_icellIsNull(retValue->l)) {
	    resultIsNullObjectRef = CVM_TRUE;
	}

	/* Pop the transition frame. */
	// if (CVMlocalExceptionOccurred(ee)) printf("[jni_invoke] exception 2\n");

	CVMpopFrame(&ee->interpreterStack,
		    ee->interpreterStack.currentFrame);
	/* Do not delete the local ref until the transition frame is
           popped! */
	// if (CVMlocalExceptionOccurred(ee)) printf("[jni_invoke] exception 3\n");
	if (resultIsNullObjectRef) {
	    CVMjniDeleteLocalRef(env, retValue->l);
	    retValue->l = NULL;
	}
    });
}

#undef CONCAT
#undef EVALCONCAT
#undef CONCAT4
#define CONCAT(x,y)		x##y
#define EVALCONCAT(x,y)		C(x,y)
#define CONCAT4(a,b,c,d)	a##b##c##d
#undef N_NONVIRTUAL
#undef N_VIRTUAL
#undef N_STATIC
#define N_NONVIRTUAL	Nonvirtual
#define N_VIRTUAL
#define N_STATIC	Static
#undef SIG_NONVIRTUAL
#undef SIG_VIRTUAL
#undef SIG_STATIC
#define SIG_NONVIRTUAL	jobject obj, jclass clazz
#define SIG_VIRTUAL	jobject obj
#define SIG_STATIC	jclass clazz
#undef OBJ_NONVIRTUAL
#undef OBJ_VIRTUAL
#undef OBJ_STATIC
#define OBJ_NONVIRTUAL	obj
#define OBJ_VIRTUAL	obj
#define OBJ_STATIC	clazz
#undef IF_Yes
#undef IF_No
#undef IF1_Yes
#undef IF1_No
#undef IF
#define IF_Yes(x,y)	x
#define IF_No(x,y)	y
#define IF1_Yes(x)	x
#define IF1_No(x)
#define IF(YN,x,y)	CONCAT(IF_,YN)(x,y)
#define IFY(YN,x)	CONCAT(IF1_,YN)(x)
#define IFN(YN,x)	CONCAT(IF1_,YN)(x)
#undef CALL_NAME_DOTDOTDOT
#undef CALL_NAME_VA_LIST
#undef CALL_NAME_JVALUE_ARRAY
#define CALL_NAME_DOTDOTDOT	Method
#define CALL_NAME_VA_LIST	MethodV
#define CALL_NAME_JVALUE_ARRAY	MethodA
#undef ARGS_DOTDOTDOT
#undef ARGS_VA_LIST
#undef ARGS_JVALUE_ARRAY
#define ARGS_DOTDOTDOT		&args
#define ARGS_VA_LIST		&args
#define ARGS_JVALUE_ARRAY	args
#undef PUSHARGSFUNC_DOTDOTDOT
#undef PUSHARGSFUNC_VA_LIST
#undef PUSHARGSFUNC_JVALUE_ARRAY
#define PUSHARGSFUNC_DOTDOTDOT		CVMjniPushArgumentsVararg
#define PUSHARGSFUNC_VA_LIST		CVMjniPushArgumentsVararg
#define PUSHARGSFUNC_JVALUE_ARRAY	CVMjniPushArgumentsArray
#undef ARG_DECL_DOTDOTDOT
#undef ARG_DECL_VA_LIST
#undef ARG_DECL_JVALUE_ARRAY
#define ARG_DECL_DOTDOTDOT	...
#define ARG_DECL_VA_LIST	va_list args
#define ARG_DECL_JVALUE_ARRAY	jvalue *args
#undef IS_DOTDOTDOT_DOTDOTDOT
#undef IS_DOTDOTDOT_VA_LIST
#undef IS_DOTDOTDOT_JVALUE_ARRAY
#define IS_DOTDOTDOT_DOTDOTDOT		Yes
#define IS_DOTDOTDOT_VA_LIST		No
#define IS_DOTDOTDOT_JVALUE_ARRAY	No

#define NAME(a,b,c,d)	CONCAT4(a,b,c,d)

#undef DEFINE_CALLMETHODXY
#define DEFINE_CALLMETHODXY(MethodType,					    \
			    ResultType, Result, unionType,		    \
			    ReturnCode,					    \
			    isValue, argType)				    \
									    \
ResultType JNICALL							    \
NAME(CVMjniCall,N_##MethodType,Result,CALL_NAME_##argType)(JNIEnv *env,	    \
			 SIG_##MethodType,				    \
			 jmethodID methodID, ARG_DECL_##argType)	    \
{									    \
    IFY(isValue,jvalue result;)						    \
    IFY(IS_DOTDOTDOT_##argType,va_list args;)				    \
									    \
    IFY(IS_DOTDOTDOT_##argType,va_start(args, methodID));		    \
    CVMjniInvoke(env, OBJ_##MethodType, methodID,			    \
		 PUSHARGSFUNC_##argType,				    \
		 ARGS_##argType,					    \
		 ReturnCode | CVM_INVOKE_##MethodType,			    \
		 IF(isValue,&result,0));				    \
    IFY(IS_DOTDOTDOT_##argType,va_end(args));				    \
    IF(isValue,return result.unionType,return);				    \
}									    \

#undef DEFINE_CALLMETHODX
#define DEFINE_CALLMETHODX(MT, RT, R, UT, RC, VD)			    \
DEFINE_CALLMETHODXY(MT,RT,R,UT,RC,VD,DOTDOTDOT)	/* ... */		    \
DEFINE_CALLMETHODXY(MT,RT,R,UT,RC,VD,VA_LIST) 	/* va_list args */	    \
DEFINE_CALLMETHODXY(MT,RT,R,UT,RC,VD,JVALUE_ARRAY)	/* jvalue *args */

#undef DEFINE_CALLMETHOD
#define DEFINE_CALLMETHOD(RT, R, UT, RC, VD)				    \
DEFINE_CALLMETHODX(NONVIRTUAL,RT,R,UT,RC,VD)	/* Nonvirtual */	    \
DEFINE_CALLMETHODX(VIRTUAL,RT,R,UT,RC,VD)	/* Virtual */		    \
DEFINE_CALLMETHODX(STATIC,RT,R,UT,RC,VD)	/* Static */

DEFINE_CALLMETHOD(jobject,Object,l,CVM_TYPEID_OBJ,Yes)		/* Object */
DEFINE_CALLMETHOD(jboolean,Boolean,z,CVM_TYPEID_BOOLEAN,Yes)	/* Boolean */
DEFINE_CALLMETHOD(jbyte,Byte,b,CVM_TYPEID_BYTE,Yes)		/* Byte */
DEFINE_CALLMETHOD(jchar,Char,c,CVM_TYPEID_CHAR,Yes)		/* Char */
DEFINE_CALLMETHOD(jshort,Short,s,CVM_TYPEID_SHORT,Yes)		/* Short */
DEFINE_CALLMETHOD(jint,Int,i,CVM_TYPEID_INT,Yes)			/* Int */
DEFINE_CALLMETHOD(jlong,Long,j,CVM_TYPEID_LONG,Yes)		/* Long */
DEFINE_CALLMETHOD(jfloat,Float,f,CVM_TYPEID_FLOAT,Yes)		/* Float */
DEFINE_CALLMETHOD(jdouble,Double,d,CVM_TYPEID_DOUBLE,Yes)	/* Double */
DEFINE_CALLMETHOD(void,Void,X,CVM_TYPEID_VOID,No)		/* Void */

jint JNICALL
CVMjniMonitorEnter(JNIEnv *env, jobject obj)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMBool ok = CVMgcSafeObjectLock(ee, obj);
    if (!ok) {
        CVMthrowOutOfMemoryError(ee, "Couldn't allocate memory for monitor");
    }
    return ok ? JNI_OK : JNI_ERR;
}

jint JNICALL
CVMjniMonitorExit(JNIEnv *env, jobject obj)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    CVMBool ok = CVMgcSafeObjectUnlock(ee, obj);
    if (!ok) {
        CVMthrowIllegalMonitorStateException(ee, "current thread not owner");
    }
    return ok ? JNI_OK : JNI_ERR;
}

jint JNICALL
CVMjniGetJavaVM(JNIEnv *env, JavaVM **p_jvm)
{
    *p_jvm = &CVMglobals.javaVM.vector;
    return JNI_OK;
}

/* Added to support TimeSys JVMCei
 * Octuber 13, 2000
 */

extern struct jvmcei JVMCei;

JNIEXPORT struct jvmcei * CVMjniGetJVMCeiInterface()
{
#ifdef CVM_DEBUG
  printf("Inside GetJVMCei returning JVMCei pointer\n");
#endif
  /* jvmcei_initialize(); */
  return &JVMCei;
}


/* NOTE: This can not be made const any more, because the JVMDI
   will instrument it if necessary, and the "checked" version of the
   JNI may do so also if we add that later. */

static struct JNINativeInterface CVMmainJNIfuncs = {
    NULL,
    NULL,
    NULL,

    NULL,

    CVMjniGetVersion,

    CVMjniDefineClass,
    CVMjniFindClass,

    CVMjniFromReflectedMethod,
    CVMjniFromReflectedField,

    CVMjniToReflectedMethod,

    CVMjniGetSuperclass,
    CVMjniIsAssignableFrom,

    CVMjniToReflectedField,

    CVMjniThrow,
    CVMjniThrowNew,
    CVMjniExceptionOccurred,
    CVMjniExceptionDescribe,
    CVMjniExceptionClear,
    CVMjniFatalError,

    CVMjniPushLocalFrame,
    CVMjniPopLocalFrame,

    CVMjniNewGlobalRef,
    CVMjniDeleteGlobalRef,
    CVMjniDeleteLocalRef,
    CVMjniIsSameObject,

    CVMjniNewLocalRef,
    CVMjniEnsureLocalCapacity,

    CVMjniAllocObject,
    CVMjniNewObject,
    CVMjniNewObjectV,
    CVMjniNewObjectA,

    CVMjniGetObjectClass,
    CVMjniIsInstanceOf,

    CVMjniGetMethodID,

    CVMjniCallObjectMethod,
    CVMjniCallObjectMethodV,
    CVMjniCallObjectMethodA,
    CVMjniCallBooleanMethod,
    CVMjniCallBooleanMethodV,
    CVMjniCallBooleanMethodA,
    CVMjniCallByteMethod,
    CVMjniCallByteMethodV,
    CVMjniCallByteMethodA,
    CVMjniCallCharMethod,
    CVMjniCallCharMethodV,
    CVMjniCallCharMethodA,
    CVMjniCallShortMethod,
    CVMjniCallShortMethodV,
    CVMjniCallShortMethodA,
    CVMjniCallIntMethod,
    CVMjniCallIntMethodV,
    CVMjniCallIntMethodA,
    CVMjniCallLongMethod,
    CVMjniCallLongMethodV,
    CVMjniCallLongMethodA,
    CVMjniCallFloatMethod,
    CVMjniCallFloatMethodV,
    CVMjniCallFloatMethodA,
    CVMjniCallDoubleMethod,
    CVMjniCallDoubleMethodV,
    CVMjniCallDoubleMethodA,
    CVMjniCallVoidMethod,
    CVMjniCallVoidMethodV,
    CVMjniCallVoidMethodA,

    CVMjniCallNonvirtualObjectMethod,
    CVMjniCallNonvirtualObjectMethodV,
    CVMjniCallNonvirtualObjectMethodA,
    CVMjniCallNonvirtualBooleanMethod,
    CVMjniCallNonvirtualBooleanMethodV,
    CVMjniCallNonvirtualBooleanMethodA,
    CVMjniCallNonvirtualByteMethod,
    CVMjniCallNonvirtualByteMethodV,
    CVMjniCallNonvirtualByteMethodA,
    CVMjniCallNonvirtualCharMethod,
    CVMjniCallNonvirtualCharMethodV,
    CVMjniCallNonvirtualCharMethodA,
    CVMjniCallNonvirtualShortMethod,
    CVMjniCallNonvirtualShortMethodV,
    CVMjniCallNonvirtualShortMethodA,
    CVMjniCallNonvirtualIntMethod,
    CVMjniCallNonvirtualIntMethodV,
    CVMjniCallNonvirtualIntMethodA,
    CVMjniCallNonvirtualLongMethod,
    CVMjniCallNonvirtualLongMethodV,
    CVMjniCallNonvirtualLongMethodA,
    CVMjniCallNonvirtualFloatMethod,
    CVMjniCallNonvirtualFloatMethodV,
    CVMjniCallNonvirtualFloatMethodA,
    CVMjniCallNonvirtualDoubleMethod,
    CVMjniCallNonvirtualDoubleMethodV,
    CVMjniCallNonvirtualDoubleMethodA,
    CVMjniCallNonvirtualVoidMethod,
    CVMjniCallNonvirtualVoidMethodV,
    CVMjniCallNonvirtualVoidMethodA,

    CVMjniGetFieldID,

    CVMjniGetObjectField,
    CVMjniGetBooleanField,
    CVMjniGetByteField,
    CVMjniGetCharField,
    CVMjniGetShortField,
    CVMjniGetIntField,
    CVMjniGetLongField,
    CVMjniGetFloatField,
    CVMjniGetDoubleField,

    CVMjniSetObjectField,
    CVMjniSetBooleanField,
    CVMjniSetByteField,
    CVMjniSetCharField,
    CVMjniSetShortField,
    CVMjniSetIntField,
    CVMjniSetLongField,
    CVMjniSetFloatField,
    CVMjniSetDoubleField,

    CVMjniGetStaticMethodID,

    CVMjniCallStaticObjectMethod,
    CVMjniCallStaticObjectMethodV,
    CVMjniCallStaticObjectMethodA,
    CVMjniCallStaticBooleanMethod,
    CVMjniCallStaticBooleanMethodV,
    CVMjniCallStaticBooleanMethodA,
    CVMjniCallStaticByteMethod,
    CVMjniCallStaticByteMethodV,
    CVMjniCallStaticByteMethodA,
    CVMjniCallStaticCharMethod,
    CVMjniCallStaticCharMethodV,
    CVMjniCallStaticCharMethodA,
    CVMjniCallStaticShortMethod,
    CVMjniCallStaticShortMethodV,
    CVMjniCallStaticShortMethodA,
    CVMjniCallStaticIntMethod,
    CVMjniCallStaticIntMethodV,
    CVMjniCallStaticIntMethodA,
    CVMjniCallStaticLongMethod,
    CVMjniCallStaticLongMethodV,
    CVMjniCallStaticLongMethodA,
    CVMjniCallStaticFloatMethod,
    CVMjniCallStaticFloatMethodV,
    CVMjniCallStaticFloatMethodA,
    CVMjniCallStaticDoubleMethod,
    CVMjniCallStaticDoubleMethodV,
    CVMjniCallStaticDoubleMethodA,
    CVMjniCallStaticVoidMethod,
    CVMjniCallStaticVoidMethodV,
    CVMjniCallStaticVoidMethodA,

    CVMjniGetStaticFieldID,

    CVMjniGetStaticObjectField,
    CVMjniGetStaticBooleanField,
    CVMjniGetStaticByteField,
    CVMjniGetStaticCharField,
    CVMjniGetStaticShortField,
    CVMjniGetStaticIntField,
    CVMjniGetStaticLongField,
    CVMjniGetStaticFloatField,
    CVMjniGetStaticDoubleField,

    CVMjniSetStaticObjectField,
    CVMjniSetStaticBooleanField,
    CVMjniSetStaticByteField,
    CVMjniSetStaticCharField,
    CVMjniSetStaticShortField,
    CVMjniSetStaticIntField,
    CVMjniSetStaticLongField,
    CVMjniSetStaticFloatField,
    CVMjniSetStaticDoubleField,

    CVMjniNewString,
    CVMjniGetStringLength,
    CVMjniGetStringChars,
    CVMjniReleaseStringChars,

    CVMjniNewStringUTF,
    CVMjniGetStringUTFLength,
    CVMjniGetStringUTFChars,
    CVMjniReleaseStringUTFChars,

    CVMjniGetArrayLength,
 
    CVMjniNewObjectArray,
    CVMjniGetObjectArrayElement,
    CVMjniSetObjectArrayElement,

    CVMjniNewBooleanArray,
    CVMjniNewByteArray,
    CVMjniNewCharArray,
    CVMjniNewShortArray,
    CVMjniNewIntArray,
    CVMjniNewLongArray,
    CVMjniNewFloatArray,
    CVMjniNewDoubleArray,

    CVMjniGetBooleanArrayElements,
    CVMjniGetByteArrayElements,
    CVMjniGetCharArrayElements,
    CVMjniGetShortArrayElements,
    CVMjniGetIntArrayElements,
    CVMjniGetLongArrayElements,
    CVMjniGetFloatArrayElements,
    CVMjniGetDoubleArrayElements,

    CVMjniReleaseBooleanArrayElements,
    CVMjniReleaseByteArrayElements,
    CVMjniReleaseCharArrayElements,
    CVMjniReleaseShortArrayElements,
    CVMjniReleaseIntArrayElements,
    CVMjniReleaseLongArrayElements,
    CVMjniReleaseFloatArrayElements,
    CVMjniReleaseDoubleArrayElements,

    CVMjniGetBooleanArrayRegion,
    CVMjniGetByteArrayRegion,
    CVMjniGetCharArrayRegion,
    CVMjniGetShortArrayRegion,
    CVMjniGetIntArrayRegion,
    CVMjniGetLongArrayRegion,
    CVMjniGetFloatArrayRegion,
    CVMjniGetDoubleArrayRegion,

    CVMjniSetBooleanArrayRegion,
    CVMjniSetByteArrayRegion,
    CVMjniSetCharArrayRegion,
    CVMjniSetShortArrayRegion,
    CVMjniSetIntArrayRegion,
    CVMjniSetLongArrayRegion,
    CVMjniSetFloatArrayRegion,
    CVMjniSetDoubleArrayRegion,

    CVMjniRegisterNatives,
    CVMjniUnregisterNatives,

    CVMjniMonitorEnter,
    CVMjniMonitorExit,

    CVMjniGetJavaVM,

    CVMjniGetStringRegion,
    CVMjniGetStringUTFRegion,

    CVMjniGetPrimitiveArrayCritical,
    CVMjniReleasePrimitiveArrayCritical,

    CVMjniGetStringChars,     /* CVMjniGetStringCritical */
    CVMjniReleaseStringChars, /* CVMjniReleaseStringCritical */

    CVMjniNewWeakGlobalRef,
    CVMjniDeleteWeakGlobalRef,

    CVMjniExceptionCheck,
    CVMjniGetJVMCeiInterface,
};


/* %comment: kbr036 */

typedef jint (JNICALL *JVM_OnLoad_t)(JavaVM *, char *, void *);

/* This takes in the entire -Xrun string and takes care of 
 * loading the library, 
 *
 * finding the JVM_OnUnload symbol and storing the function pointer of 
 * it, if any, and the loaded native library object handle in Xrun table.
 *
 * finding the JVM_Onload symbol, and calling it with the option string, 
 * if any, at the tail end of the argument. 
 *
 * Note that this just reuses the ClassLoader's native library loading 
 * code, so it automatically looks in java.library.path and 
 * sun.boot.library.path for the named library.
 */
#ifdef CVM_DYNAMIC_LINKING
static CVMBool
handleXrunArgument(JNIEnv* env, char* arg)
{
    jclass classLoaderClass = NULL;
    jclass nativeLibraryClass = NULL;
    jmethodID loadLibraryInternalID;
    jmethodID findID;
    char* libraryNamePtr;
    char* separatorPtr;
    char* optionsPtr;
    jstring jLibraryName = NULL;
    jobject nativeLibrary = NULL;
    JVM_OnLoad_t onLoadFunc;
    jstring jOnLoadName = NULL;
    JVM_OnUnload_t onUnloadFunc;
    jstring jOnUnloadName = NULL;
    JavaVM* vm;
    CVMBool result = CVM_FALSE;

    classLoaderClass = (*env)->FindClass(env, "java/lang/ClassLoader");
    if (classLoaderClass == NULL) {
	CVMconsolePrintf("Cannot start VM "
			 "(couldn't find class java.lang.ClassLoader)\n");
	goto done;
    }
    nativeLibraryClass =
	(*env)->FindClass(env, "java/lang/ClassLoader$NativeLibrary");
    if (nativeLibraryClass == NULL) {
	CVMconsolePrintf("Cannot start VM "
			 "(couldn't find class java.lang.ClassLoader$NativeLibrary)\n");
	goto done;
    }

    /* Get method IDs */
    loadLibraryInternalID =
	(*env)->GetStaticMethodID(env, classLoaderClass,
				  "loadLibraryInternal",
				  "(Ljava/lang/Class;Ljava/lang/String;Z)Ljava/lang/Object;");
    if (loadLibraryInternalID == NULL) {
	CVMconsolePrintf("Cannot start VM "
			 "(couldn't find method ClassLoader.loadLibraryInternal)\n");
	goto done;
    }
    findID =
	(*env)->GetMethodID(env, nativeLibraryClass,
			    "find",
			    "(Ljava/lang/String;)J");
    if (findID == NULL) {
	CVMconsolePrintf("Cannot start VM "
			 "(couldn't find method ClassLoader$NativeLibrary.find)\n");
	goto done;
    }

    libraryNamePtr = arg + 5;
    separatorPtr = strchr(libraryNamePtr, ':');
    if (separatorPtr == NULL) {
	/* No options. */
	optionsPtr = NULL;
	jLibraryName = (*env)->NewStringUTF(env, libraryNamePtr);
    } else {
	char *tmpLibraryNamePtr;
	int len;

	/* Have to trim options. */
	optionsPtr = separatorPtr + 1;
	/* Leave room for null terminator */
	len = separatorPtr - libraryNamePtr + 1;
	tmpLibraryNamePtr =
	    malloc(len * sizeof(char));
	memcpy(tmpLibraryNamePtr, libraryNamePtr, len - 1);
	tmpLibraryNamePtr[len - 1] = 0;
	jLibraryName = (*env)->NewStringUTF(env, tmpLibraryNamePtr);
	free(tmpLibraryNamePtr);
    }


    /* Call library loading function */
    nativeLibrary = (*env)->CallStaticObjectMethod(env, classLoaderClass,
						   loadLibraryInternalID,
						   NULL, jLibraryName, 0);

    if ((*env)->ExceptionCheck(env)) {
	/* Exception already thrown, I believe */
	CVMconsolePrintf("Could not find helper library for argument %s\n",
			 arg);
	goto done;
    }

    /* Look up JVM_OnLoad symbol */
    jOnLoadName = (*env)->NewStringUTF(env, "JVM_OnLoad");
    if (jOnLoadName == NULL) {
	CVMconsolePrintf("Could not call JVM_OnLoad of helper library for argument %s\n",
			 arg);
	goto done;
    }
    {
	jlong onloadVal =
	    (*env)->CallLongMethod(env, nativeLibrary, findID, jOnLoadName);
	if ((*env)->ExceptionCheck(env)) {
	    goto done;
	}
	onLoadFunc = (JVM_OnLoad_t)CVMlong2VoidPtr(onloadVal);
    }

    /* Look up JVM_OnUnload symbol
     * Store away NativeLibrary object ref and function pointer for calling
     * JVM_onUnload.
     */
    jOnUnloadName = (*env)->NewStringUTF(env, "JVM_OnUnload");
    if (jOnUnloadName == NULL) {
        CVMconsolePrintf("Could not call JVM_OnUnload of helper library for "
	    "argument %s\n", arg);
        goto done;
    }
    {
        jlong onunloadVal =
            (*env)->CallLongMethod(env, nativeLibrary, findID, jOnUnloadName);
        jobject globalRef = NULL;

        if ((*env)->ExceptionCheck(env)) {
            goto done;
        }
        globalRef = (*env)->NewGlobalRef(env, nativeLibrary);
        if (globalRef == NULL) {
             goto done;
        }

        onUnloadFunc = (JVM_OnUnload_t)CVMlong2VoidPtr(onunloadVal);
        CVMXrunAppendToTable(&CVMglobals.onUnloadTable, globalRef, 
			     onUnloadFunc);
    }

    /* Call JVM_OnLoad after the lookup and store away
     * of JVM_OnUnload so we won't call JVM_OnUnload 
     * at VM shutdown if JVM_OnLoad was never called.
     * This also avoid a failure after createing the globalRef,
     * after adding entry to the Xrun table, and after
     * calling the JVM_OnLoad function.
     */
    if (onLoadFunc != NULL) {
        (*env)->GetJavaVM(env, &vm);
        /* %comment: rt018 */
        (*onLoadFunc)(vm, optionsPtr, NULL);
    }

    result = CVM_TRUE;

    done:
    (*env)->DeleteLocalRef(env, classLoaderClass);
    (*env)->DeleteLocalRef(env, nativeLibraryClass);
    (*env)->DeleteLocalRef(env, jLibraryName);
    (*env)->DeleteLocalRef(env, nativeLibrary);
    (*env)->DeleteLocalRef(env, jOnLoadName);
    (*env)->DeleteLocalRef(env, jOnUnloadName);
    return result; 
}
#endif

JNIEXPORT jint JNICALL
JNI_GetDefaultJavaVMInitArgs(void *args_)
{
    return JNI_ERR;
}

static int numJVMs;

/*
 * Initialize shutdown related methods at VM startup time
 */
static CVMBool
initShutdownMethods(JNIEnv* env)
{
    jclass c;
    jmethodID m;
    
    c = CVMcbJavaInstance(CVMsystemClass(java_lang_Shutdown));
    m = CVMjniGetStaticMethodID(env, c, "shutdown", "()V");
    if (m == NULL) {
	return CVM_FALSE;
    }
    CVMglobals.java_lang_Shutdown_shutdown = m;

    c = CVMcbJavaInstance(CVMsystemClass(sun_misc_SystemThread));
    m = CVMjniGetStaticMethodID(env, c, "waitAllSystemThreadsExit", "()V");
    if (m == NULL) {
	return CVM_FALSE;
    }
    CVMglobals.sun_misc_SystemThread_waitAllSystemThreadsExit = m;
    return CVM_TRUE;
}

jint
JNI_CreateJavaVM(JavaVM **p_jvm, void **p_env, void *args)
{
    CVMExecEnv *ee;
    JNIEnv *env;
    int argCtr;
    CVMOptions options;
    JavaVMInitArgs* initArgs = (JavaVMInitArgs*) args;
    CVMInt32 numUnrecognizedOptions = 0;
    CVMInt32 numXrunArguments = 0;
    char **xrunArguments = NULL;

    if (initArgs == NULL) {
	CVMconsolePrintf("Cannot start VM "
			 "(arguments passed via Invocation API were NULL)\n");
	return JNI_ERR;
    }

    memset(&options, 0, sizeof options);
    
#ifdef CVM_CLASSLOADING
    /*
     * The default verification mode is CVM_VERIFY_REMOTE
     */
    options.classVerificationLevel = CVM_VERIFY_REMOTE;

#define CVM_CLASSLOADING_OPTIONS \
    " [-Xverify:<all | remote | none>] [-Xbootclasspath[/a | /p]:<path>]" 
#else
#define CVM_CLASSLOADING_OPTIONS
#endif
#ifdef CVM_DEBUG
#define CVM_DEBUG_OPTIONS " [-Xtrace:<val>]"
#else
#define CVM_DEBUG_OPTIONS
#endif
#ifdef CVM_JVMDI
#define CVM_JVMDI_OPTIONS " [-Xdebug] [-Xrunjdwp...]"
#else
#define CVM_JVMDI_OPTIONS
#endif

#ifdef CVM_GCIMPL_GC_OPTIONS
#define CVM_GC_OPTIONS CVM_GCIMPL_GC_OPTIONS
#else
#define CVM_GC_OPTIONS
#endif

#define CVM_NATIVE_OPTIONS "[-Xms<size>] [-Xss<size>] " \
        CVM_GC_OPTIONS \
	CVM_JVMDI_OPTIONS \
	CVM_DEBUG_OPTIONS \
	CVM_CLASSLOADING_OPTIONS

    for (argCtr = 0; argCtr < initArgs->nOptions; argCtr++) {
	const char *str = initArgs->options[argCtr].optionString;
	if (!strncmp(str, "-Xms", 4)) {
	    options.heapSizeStr = str + 4;
	} else if (!strncmp(str, "-Xss", 4)) {
	    options.nativeStackSizeStr = str + 4;
	} else if (!strncmp(str, "-Xgc:", 5)) {
	    options.gcAttributesStr = str + 5;
	}
#ifdef CVM_CLASSLOADING
	else if (!strncmp(str,
			  "-Xverify", 8)) {
	    const char* kind = str + 8;
	    if (*kind == 0)
	        options.classVerificationLevel = CVM_VERIFY_ALL;
	    else if (!strcmp(kind, ":all"))
	        options.classVerificationLevel = CVM_VERIFY_ALL;
	    else if (!strcmp(kind, ":remote"))
	        options.classVerificationLevel = CVM_VERIFY_REMOTE;
	    else if (!strcmp(kind, ":none"))
	        options.classVerificationLevel = CVM_VERIFY_NONE;
	    else {
	        CVMconsolePrintf("Illegal option: %s\n", str);
		return JNI_EINVAL;
	    }
	}
#endif
	/* NOTE that -Xdebug must currently be handled before
           CVMpreloaderGenerateAllStackmaps() is called */
	else if (!strcmp(str, "-Xdebug")) {
#ifdef CVM_JVMDI
	    options.debugging = CVM_TRUE;
#else
	    CVMconsolePrintf("Cannot start VM "
			     "(-Xdebug specified, but debugging support "
			     "not compiled in)\n");
	    return JNI_ERR;
#endif
	} else if (!strncmp(str, "-Xtrace:", 8)) {
#ifdef CVM_DEBUG
	    options.traceFlagsStr = str + 8;
#endif
	} else if (!strcmp(str, "vfprintf")) {
	    options.vfprintfHook = initArgs->options[argCtr].extraInfo;
	} else if (!strcmp(str, "exit")) {
	    options.exitHook = initArgs->options[argCtr].extraInfo;
	} else if (!strcmp(str, "abort")) {
	    options.abortHook = initArgs->options[argCtr].extraInfo;
	} else if (!strcmp(str, "-XsafeExit")) {
	    options.safeExitHook = initArgs->options[argCtr].extraInfo;
	} else if (!strcmp(str, "-XfullShutdown")) {
#ifdef CVM_HAVE_PROCESS_MODEL
	    options.fullShutdownFlag = CVM_TRUE;
#endif
#ifdef CVM_JVMDI
	} else if (!strncmp(str, "-Xrun", 5)) {
	    ++numXrunArguments;
	    /* Don't set the optionString to NULL, so we can take care
               of this on the next pass */
	    continue;
#endif
	} else {
	    /* Unrecognized option, pass to Java */
	    ++numUnrecognizedOptions;
	    continue;
	}
	/* Recognized options fall through to here */
	initArgs->options[argCtr].optionString = NULL;
    }

    /* Now snag -Xrun arguments so we can run them after initializing
       the VM */
    if (numXrunArguments > 0) {
	int i = 0;
	xrunArguments = malloc(numXrunArguments * sizeof(char *));
	for (argCtr = 0; argCtr < initArgs->nOptions; argCtr++) {
	    const char *str = initArgs->options[argCtr].optionString;
	    if ((str != NULL) && (!strncmp(str, "-Xrun", 5))) {
		xrunArguments[i] = (char *) str;
		initArgs->options[argCtr].optionString = NULL;		
		++i;
	    }
	}
	CVMassert(i == numXrunArguments);
    }

    /* %comment: rt021 */
    if (numJVMs == 0) {
	CVMinitStaticState();
    }

    if (!CVMinitVMGlobalState(&CVMglobals, &options)) {
        if (xrunArguments != NULL) {
            free(xrunArguments);
        }
	return JNI_ERR;
    }

    ee = &CVMglobals.mainEE;

    *p_jvm = &CVMglobals.javaVM.vector;
    env = CVMexecEnv2JniEnv(ee);
    *p_env = (void *)env;

    if (!initShutdownMethods(env)) {
	CVMconsolePrintf("Cannot start VM (shutdown method init failed)\n");
	if (xrunArguments != NULL) {
            free(xrunArguments);
        }
	return JNI_ERR;
    }

    if (!CVMpreloaderDisambiguateAllMethods()) {
	CVMconsolePrintf("Cannot start VM (method disambiguation failed)\n");
	if (xrunArguments != NULL) {
            free(xrunArguments);
        }
	return JNI_ERR;
    }
#ifdef CVM_INSTRUCTION_COUNTING
    /*
     * Initialize byte-code statistics gathering
     */
    CVMinitStats();
#endif /* CVM_INSTRUCTION_COUNTING */

    /* *** RTSJ *** */
    // Trying to initialize the JVMCei interface over here
    //  - Pratik
//    CVMdebugPrintf(("JNI_CreateJavaVM: Installing filters\n"));
    jvmcei_installAllFilters(CVMexecEnv2JniEnv(ee));
//    CVMdebugPrintf(("JNI_CreateJavaVM: Finished installing filters\n"));
    /* *** JSTR *** */

    /* JCC preallocates Class and String objects, so we need to run
     * their static initializers as if they had been created at runtime.
     */
    if (CVMclassInit(ee, (CVMClassBlock*)CVMsystemClass(java_lang_Class))) {
	CVMclassInit(ee, (CVMClassBlock*)CVMsystemClass(java_lang_String));
    }
//    CVMdebugPrintf(("JNI_CreateJavaVM: Finished loading Class and String\n"));
    if (CVMlocalExceptionOccurred(ee)) {
	CVMconsolePrintf("Cannot start VM (initialization failed)\n");
	if (xrunArguments != NULL) {
            free(xrunArguments);
        }
	return JNI_ERR;
    }

    /*
     * Create system ThreadGroup, main ThreadGroup, and main Thread object.
     * Is there some way we can do this in Java?
     */
    {
	jclass threadGroupClass =
	    CVMcbJavaInstance(CVMsystemClass(java_lang_ThreadGroup));
	jclass threadClass =
	    CVMcbJavaInstance(CVMsystemClass(java_lang_Thread));
	jstring mainString = CVMjniNewStringUTF(env, "main");
	jobject systemThreadGroup = CVMjniNewObject(env, threadGroupClass,
	    CVMjniGetMethodID(env, threadGroupClass, "<init>", "()V"));
	jobject mainThreadGroup = CVMjniNewObject(env, threadGroupClass,
	    CVMjniGetMethodID(env, threadGroupClass,
		"<init>", "(Ljava/lang/ThreadGroup;Ljava/lang/String;)V"),
		systemThreadGroup, mainString);
	jobject mainThread;
	jlong eetop;

	CVMglobals.mainThreadGroupICell = CVMID_getGlobalRoot(ee);

	if (systemThreadGroup == NULL || mainThreadGroup == NULL ||
	    CVMglobals.mainThreadGroupICell == NULL) {
	    /*
	     * Can't start this VM. Sorry!
	     */
	    CVMconsolePrintf("Cannot start VM "
			     "(out of memory while initializing)\n");
            if (xrunArguments != NULL) {
                free(xrunArguments);
            }
	    return JNI_ERR;
	}

	CVMID_icellAssign(ee, CVMglobals.mainThreadGroupICell,
			  mainThreadGroup);
#ifdef CVM_JVMDI
	{
	    /* %comment: rt023 */
	    CVMglobals.systemThreadGroupICell = CVMID_getGlobalRoot(ee);
	    if (CVMglobals.systemThreadGroupICell == NULL) {
		/*
		 * Can't start this VM. Sorry!
		 */
		CVMconsolePrintf("Cannot start VM "
				 "(out of memory while initializing)\n");
                if (xrunArguments != NULL) {
                    free(xrunArguments);
                }
		return JNI_ERR;
	    }
	    CVMID_icellAssign(ee, CVMglobals.systemThreadGroupICell,
			      systemThreadGroup);
	}
#endif
       	CVMjniDeleteLocalRef(env, systemThreadGroup);

	eetop = CVMvoidPtr2Long(ee);

	mainThread = CVMjniNewObject(env, threadClass,
	    CVMjniGetMethodID(env, threadClass,
			      "<init>", 
			      "(Ljava/lang/ThreadGroup;Ljava/lang/String;IJ)V"),
	    mainThreadGroup, mainString, java_lang_Thread_NORM_PRIORITY, eetop);

	CVMjniDeleteLocalRef(env, mainThreadGroup);
	CVMjniDeleteLocalRef(env, mainString);

	if (mainThread == NULL || (*env)->ExceptionCheck(env)) {
            if (xrunArguments != NULL) {
                free(xrunArguments);
            }
	    return JNI_ERR;
	}

	CVMjniDeleteLocalRef(env, mainThread);

	CVMassert(!CVMID_icellIsNull(CVMcurrentThreadICell(ee)));
    }

    {
	/* Call System.initializeSystemClass() via the JNI. This sets
	   up the global properties table which
	   CVM.parseCommandLineOptions(), below, will modify with
	   user-defined property definitions. This was at one point
	   called from a method called CVM.initialize() via
	   reflection, but we have to be able to configure without 
	   reflection. */
	jclass systemClass;
	jmethodID initMethodID;
    
	systemClass = CVMcbJavaInstance(CVMsystemClass(java_lang_System));
	initMethodID = (*env)->GetStaticMethodID(env, systemClass,
						 "initializeSystemClass",
						 "()V");
    
	if (initMethodID == NULL) {
	    CVMconsolePrintf("Cannot start VM "
			     "(couldn't find method "
			     "void java.lang.System."
			     "initializeSystemClass())\n");
            if (xrunArguments != NULL) {
                free(xrunArguments);
            }
	    return JNI_ERR;
	}
    
    //Very bad programming!!!!  This is a special case for System class
    ExecuteInImmortal(ee, {
	(*env)->CallStaticVoidMethod(env, systemClass, initMethodID);
    });
	if ((*env)->ExceptionCheck(env)) {
	    CVMconsolePrintf("Cannot start VM "
			     "(error during initialization)\n");
	    (*env)->ExceptionDescribe(env);
            if (xrunArguments != NULL) {
                free(xrunArguments);
            }
	    return JNI_ERR;
	}
    }


    /* Allocate an array of Strings and pass it as argument to
       CVM.parseCommandLineOptions(), which will add any user-defined
       properties to the global properties table and optionally find
       the main class name and arguments.

       %comment: rt024

       NOTE we skip any options that were recognized above.
    */
    {
	jarray jInitArgs;
	CVMInt32 i, index;
	jclass cvmClass;
	jmethodID initMethodID;
	jint result;
	jstring nativeOptions;

	jInitArgs = (*env)->NewObjectArray(
            env, numUnrecognizedOptions,
	    CVMcbJavaInstance(CVMsystemClass(java_lang_String)),
	    NULL);
	if (jInitArgs == NULL) {
	    CVMconsolePrintf("Cannot start VM "
			     "(out of memory while allocating initArgs)\n");
            if (xrunArguments != NULL) {
                free(xrunArguments);
            }
	    return JNI_ERR;
	}
	index = 0;
	for (i = 0; i < initArgs->nOptions; i++) {
	    if (initArgs->options[i].optionString != NULL) {
		jobject jstr =
		    (*env)->NewStringUTF(env,
					 initArgs->options[i].optionString);
		if (jstr == NULL) {
		    CVMconsolePrintf("Cannot start VM "
				     "(out of memory while allocating initArg)\n");
                    if (xrunArguments != NULL) {
                        free(xrunArguments);
                    }
		    return JNI_ERR;
		}
		(*env)->SetObjectArrayElement(env, jInitArgs, index, jstr);
		++index;
		(*env)->DeleteLocalRef(env, jstr);
	    }
	}
	CVMassert(index == numUnrecognizedOptions);

	nativeOptions = (*env)->NewStringUTF(env, CVM_NATIVE_OPTIONS);
	if (nativeOptions == NULL) {
            if (xrunArguments != NULL) {
                free(xrunArguments);
            }
	    return JNI_ERR;
	}

	cvmClass = CVMcbJavaInstance(CVMsystemClass(sun_misc_CVM));
	initMethodID = (*env)->GetStaticMethodID(
            env, cvmClass,
	    "parseCommandLineOptions",
	    "([Ljava/lang/String;Ljava/lang/String;Z)I");
	if (initMethodID == NULL) {
	    CVMconsolePrintf("Cannot start VM "
			     "(couldn't find method "
			     "int sun.misc.CVM.parseCommandLineOptions"
			     "(String[], boolean))\n");
            if (xrunArguments != NULL) {
                free(xrunArguments);
            }
	    return JNI_ERR;
	}
	result = (*env)->CallStaticIntMethod(env, cvmClass, initMethodID,
					     jInitArgs, nativeOptions,
					     initArgs->ignoreUnrecognized);
	if ((*env)->ExceptionCheck(env)) {
	    /*
	     * If an exception occurs, the result from the JNI call
	     * cannot be trusted. Make it an error.  
	     */
	    result = sun_misc_CVM_ARG_PARSE_ERR;
	}
	if (result == sun_misc_CVM_ARG_PARSE_ERR) {
	    CVMconsolePrintf("Cannot start VM "
			     "(error during initialization)\n");
	    if ((*env)->ExceptionCheck(env)) {
		(*env)->ExceptionDescribe(env);
	    }
            if (xrunArguments != NULL) {
                free(xrunArguments);
            }
	    return JNI_ERR;
	} else if (result == sun_misc_CVM_ARG_PARSE_EXITVM) {
	    jint retVal = JNI_OK;
#ifdef CVM_CLASSLOADING
	    /* to make sure we can exit */
	    if (!CVMclassBootClassPathInit(env)) { 
		CVMconsolePrintf("Failed to initialize bootclasspath");
		retVal = JNI_ERR; 
	    }
#endif /* CVM_CLASSLOADING */
            if (xrunArguments != NULL) {
                free(xrunArguments);
            }
	    return retVal; /* Caller will deal with it. */
	} else {
	    CVMassert(result == sun_misc_CVM_ARG_PARSE_OK);
	}

	(*env)->DeleteLocalRef(env, nativeOptions);
	(*env)->DeleteLocalRef(env, jInitArgs);
    }

#ifdef CVM_CLASSLOADING
    if (!CVMclassBootClassPathInit(env)) {
	CVMdebugPrintf(("Cannot start VM "
			"(allocating bootclasspath data failed.)\n"));
        if (xrunArguments != NULL) {
            free(xrunArguments);
        }
	return JNI_ERR;
    }
#endif /* CVM_CLASSLOADING */

    /* Now that the VM is initialized, may need to handle "helper"
       libraries for JVMDI/PI */
    if (xrunArguments != NULL) {

        /* Initialize Xrun list */
        if (!CVMXrunInitTable(&CVMglobals.onUnloadTable, numXrunArguments)) {
	    free(xrunArguments);
	    return JNI_ERR;
	}

 	for (argCtr = 0; argCtr < numXrunArguments; argCtr++) {
#ifdef CVM_DYNAMIC_LINKING
	    if (!handleXrunArgument(env, xrunArguments[argCtr])) {
		CVMconsolePrintf("Cannot start VM "
				 "(error handling -Xrun argument %s)\n",
				 xrunArguments[argCtr]);
		if ((*env)->ExceptionCheck(env)) {
		    (*env)->ExceptionDescribe(env);
		}
	        free(xrunArguments);
		return JNI_ERR;
	    }
#else
	    CVMconsolePrintf("Can not handle -Xrun argument "
			     "(dynamic linking not built into VM): %s\n",
			     xrunArguments[argCtr]);
	    free(xrunArguments);
	    return JNI_ERR;
#endif
	}
	free(xrunArguments);
    }

#ifdef CVM_JVMDI
    if (CVMglobals.jvmdiDebuggingEnabled) {
	CVMjvmdiNotifyDebuggerOfVmInit(ee);    
	/*
	 * The debugger event hook might have thrown an exception
	 */
	if ((*env)->ExceptionCheck(env)) {
	    CVMconsolePrintf("Cannot start VM "
			 "(error during JVMDI_EVENT_VM_INIT handling)\n");
	    (*env)->ExceptionDescribe(env);
	    return JNI_ERR;
	}
    }
#endif

    /* In the future JVMPI will need to get a VM init event here as
       well */

#ifdef CVM_JVMDI
    if (CVMglobals.jvmdiDebuggingEnabled) {
	CVMjvmdiNotifyDebuggerOfThreadStart(ee, CVMcurrentThreadICell(ee));
	/*
	 * The debugger event hook might have thrown an exception
	 */
	if ((*env)->ExceptionCheck(env)) {
	    CVMconsolePrintf("Cannot start VM "
			     "(error during "
			     "JVMDI_EVENT_THREAD_START handling)\n");
	    (*env)->ExceptionDescribe(env);
	    return JNI_ERR;
	}
    }
#endif


    /* ROBBINS SEZ:
       I'm not completly sure where this should go, but this seems as
       good a place as any. We need to get the dispatcher thread up and
       running before user code starts.
    */
    {
	jclass dispatcherClass;
	jmethodID startDispatcher;

	dispatcherClass = (*env)->FindClass(env,"com/timesys/Dispatcher");
	if(!dispatcherClass)
	    CVMconsolePrintf("Cannot load AE Dispatcher class\nContinuing without it\n");
	else{
	    startDispatcher = (*env)->GetStaticMethodID(env,dispatcherClass,
							"startDispatcher","()V");
	    if(!startDispatcher)
		CVMconsolePrintf("Cannot load AE Dispatcher start method\nContinuing without it\n");
	    else{
		/* the dispatcher thread goes into immortal */
		ExecuteInImmortal(ee, {
		    (*env)->CallStaticVoidMethod(env,dispatcherClass,startDispatcher);
		});
		if((*env)->ExceptionCheck(env)){
		    CVMconsolePrintf("Exception starting dispatcher, cannot start VM\n");
		    (*env)->ExceptionDescribe(env);
		    return JNI_ERR;
		}
	    }
	}
    }


    /* In the future JVMPI will need to get a thread start event here
       as well */
    CVMdebugPrintf(("JNI_CreateJavaVM: Finished everything!!!!\n"));
    ++numJVMs;

    return JNI_OK;
}

JNIEXPORT jint JNICALL
JNI_GetCreatedJavaVMs(JavaVM **vm_buf, jsize bufLen, jsize *numVMs)
{
#ifdef JDK12
    if (VM_created) {
	if (numVMs)
	    *numVMs = 1;
	if (bufLen > 0)
	    *vm_buf = (JavaVM *)(&main_vm);
    } else
	*numVMs = 0;
    return JNI_OK;
#else
    if (numJVMs > 0) {
	if (numVMs != NULL) {
	    *numVMs = 1;
	}
	if (bufLen > 0) {
	    *vm_buf = &CVMglobals.javaVM.vector;
	}
    } else {
	if (numVMs != NULL) {
	    *numVMs = 0;
	}
    }
    return JNI_OK;
#endif
}

static jint
CVMjniDestroyJavaVM(JavaVM *vm)
{
    JNIEnv *env;
    CVMExecEnv *ee;
    jint res;

    /* %comment: rt026 */
    res = (*vm)->AttachCurrentThread(vm, (void **)&env, NULL);
    if (res < 0) {
	return res;
    }
    ee = CVMjniEnv2ExecEnv(env);

    CVMwaitForUserThreads(ee);
    /* %comment dlong001 */

    /* Call shutdown hooks: java.lang.Shutdown.shutdown() */
    CVMjniCallStaticVoidMethod(env,
	CVMcbJavaInstance(CVMsystemClass(java_lang_Shutdown)),
	CVMglobals.java_lang_Shutdown_shutdown);

    if (CVMlocalExceptionOccurred(ee)) {
	/*
	 * Ignore and clear the exception
	 */
	CVMconsolePrintf("Exception occurred during Shutdown.shutdown()\n");
	CVMjniExceptionDescribe(env);
	CVMclearLocalException(ee);
    }

    /* Free nativeLibrary object refs and call JVM_onUnload
     * function for each if any.
     */
    CVMXrunProcessTable(&CVMglobals.onUnloadTable, vm, env);

    if (CVMglobals.fullShutdown) {
       
      
      /* ROBBINS SEZ:
	 We would shut downthe AE Dispatcher here, but it's taken down by
	 the posix thread model for us*/
      /*
      {
	jclass dispatcherClass;
	jmethodID killDispatcher;
	
	dispatcherClass = (*env)->FindClass(env,"com/timesys/Dispatcher"); 
	if(!dispatcherClass) 
	  CVMconsolePrintf("Cannot load AE Dispatcher class\nContinuing without it\n"); 
	else{ 
	  killDispatcher = (*env)->GetStaticMethodID(env,dispatcherClass, 
						     "killDispatcher","()V"); 
	  if(!killDispatcher) 
	    CVMconsolePrintf("Cannot load AE Dispatcher kill method\nContinuing without it\n"); 
	  else{ 
	    ExecuteInImmortal(ee, {
	      (*env)->CallStaticVoidMethod(env,dispatcherClass,killDispatcher); 
	    }); 
	    if((*env)->ExceptionCheck(env)){ 
	      CVMconsolePrintf("Exception killing dispatcher, cannot start VM\n"); 
 	      (*env)->ExceptionDescribe(env);		   
	    } 
	  }  
	} 
      } 
      */

	/*
	 * Call sun.misc.SystemThread.waitAllSystemThreadsExit()
	 * to shutdown system threads
	 */
	CVMjniCallStaticVoidMethod(env,
	    CVMcbJavaInstance(CVMsystemClass(sun_misc_SystemThread)),
	    CVMglobals.sun_misc_SystemThread_waitAllSystemThreadsExit);

	if (CVMlocalExceptionOccurred(ee)) {
	    /*
	     * Ignore and clear the exception
	     */
	    CVMconsolePrintf("Exception occurred during "
			     "SystemThread.waitAllSystemThreadsExit()\n");
	    CVMjniExceptionDescribe(env);
	    CVMclearLocalException(ee);
	}
	CVMwaitForAllThreads(ee);
    }

#ifdef CVM_JVMDI
    if (CVMglobals.jvmdiDebuggingEnabled) {
	CVMjvmdiNotifyDebuggerOfThreadEnd(ee, CVMcurrentThreadICell(ee));
    }
#endif

    CVMprepareToExit();

#ifdef CVM_INSTRUCTION_COUNTING
    /*
     * We are at the end of the VM run. Dump byte-code stats.
     */
    CVMdumpStats();
#endif

    /*
     * If we have a process model, then we cannot destroy the VM data
     * structures, otherwise the daemon threads may crash.
     */
    /* %comment dlong002 */

    if (CVMglobals.fullShutdown) {
	CVMunloadApplicationclasses(ee);

#ifdef CVM_CLASSLOADING
	CVMclassBootClassPathDestroy();
	CVMpackagesDestroy();
#endif
	CVMdestroyVMGlobalState(ee, &CVMglobals);
 
	/* 
	 * We should probably call DetachCurrentThread, but
         * CVMdestroyVMGlobalState has already destroyed the EE,
         * so free it here.
	 */

	if (ee != &CVMglobals.mainEE) {
	    free(ee);
#ifdef CVM_DEBUG
	    ee = NULL;
#endif
	}


	/* %comment: rt028 */
	if (numJVMs == 1) {
	    CVMdestroyStaticState();
	}

    }

    --numJVMs;

    return JNI_ERR;
}

static jint JNICALL
CVMjniAttachCurrentThread(JavaVM *vm, void **penv, void *_args)
{
    JavaVMAttachArgs *args = (JavaVMAttachArgs *)_args;
    /* %comment: rt029 */
    CVMExecEnv *ee = CVMgetEE();
    JNIEnv *env;

    if (ee != NULL) {
	/* already attached */
        env = CVMexecEnv2JniEnv(ee);
        *(JNIEnv **)penv = env;
	/* %comment: rt030 */
	return JNI_OK;
    } else {
	ee = (CVMExecEnv *)calloc(1, sizeof *ee);
	if (ee == NULL) {
	    return JNI_ERR;
	}
	if (!CVMinitExecEnv(ee, CVM_TRUE)) {
	    free(ee);
	    return JNI_ERR;
	}
	CVMaddThread(ee, CVM_TRUE); /* user thread */
        env = CVMexecEnv2JniEnv(ee);
        *(JNIEnv **)penv = env;

	{
	    /* allocate thread object */
	    jclass threadClass =
		CVMcbJavaInstance(CVMsystemClass(java_lang_Thread));
	    jobject threadGroup = CVMglobals.mainThreadGroupICell;
	    jobject threadName = NULL;
	    jobject thread;
	    jlong eetop;

	    eetop = CVMvoidPtr2Long(ee);
	    if (args && args->version == JNI_VERSION_1_2) {
		if (args->group != NULL) {
		    threadGroup = args->group;
		}
		if (args->name != NULL) {
		    threadName = CVMjniNewStringUTF(env, args->name);
		}
	    }

	    if (threadName != NULL) {
		thread = CVMjniNewObject(env, threadClass,
		    CVMjniGetMethodID(env, threadClass,
			"<init>", 
			"(Ljava/lang/ThreadGroup;Ljava/lang/String;IJ)V"),
		    threadGroup, threadName, java_lang_Thread_NORM_PRIORITY,
			eetop);
		CVMjniDeleteLocalRef(env, threadName);
	    } else {
		thread = CVMjniNewObject(env, threadClass,
		    CVMjniGetMethodID(env, threadClass,
			"<init>", 
			"(Ljava/lang/ThreadGroup;IJ)V"),
		    threadGroup, java_lang_Thread_NORM_PRIORITY, eetop);
	    }

	    if (CVMexceptionOccurred(ee)) {
                CVMjniExceptionClear(env);
		CVMdestroyExecEnv(ee);
		free(ee);
                return JNI_ERR;
            }

	    CVMassert(thread != NULL);

	    CVMID_icellAssign(ee, CVMcurrentThreadICell(ee), thread);
	    CVMjniDeleteLocalRef(env, thread);
	}

	/* JVMPI event */
#ifdef CVM_JVMDI
	if (CVMglobals.jvmdiDebuggingEnabled) {
	    CVMjvmdiNotifyDebuggerOfThreadStart(ee, CVMcurrentThreadICell(ee));
	}
#endif

	return JNI_OK;
    }
}

static jint JNICALL
CVMjniDetachCurrentThread(JavaVM *vm)
{
    CVMExecEnv *ee = CVMgetEE();
    if (ee == NULL) {
	return JNI_EDETACHED;
    } else {
	/* %comment: rt031 */
	JNIEnv *env = CVMexecEnv2JniEnv(ee);
	jclass threadClass =
	    CVMcbJavaInstance(CVMsystemClass(java_lang_Thread));
	jobject thread = CVMcurrentThreadICell(ee);
	jobject throwable = CVMjniExceptionOccurred(env);
	CVMjniExceptionClear(env);
	CVMjniCallNonvirtualVoidMethod(
            env, thread, threadClass,
	    CVMjniGetMethodID(env, threadClass,
			      "exit", 
			      "(Ljava/lang/Throwable;)V"),
	    throwable);

#ifdef CVM_JVMDI
	if (CVMglobals.jvmdiDebuggingEnabled) {
	    CVMjvmdiNotifyDebuggerOfThreadEnd(ee, CVMcurrentThreadICell(ee));
	}
#endif
	CVMjniDeleteLocalRef(env, throwable);

	/* %comment: rt032 */
	CVMdestroyExecEnv(ee);
	if (ee != &CVMglobals.mainEE) {
	    free(ee);
	}
	return JNI_OK;
    }
}

static jint JNICALL
CVMjniGetEnv(JavaVM *vm, void **penv, jint version)
{
    CVMExecEnv *ee = CVMgetEE();
    if (ee != NULL) {
	if (version == JNI_VERSION_1_1 || version == JNI_VERSION_1_2) {
	    *penv = (void *)CVMexecEnv2JniEnv(ee);
	    return JNI_OK;
#ifdef CVM_JVMPI
        } else if (version == JVMPI_VERSION_1) {
            *penv = jvmpi_interface();
            return JNI_OK;
#endif /* CVM_JVMPI */

#ifdef CVM_JVMDI
        } else if (version == JVMDI_VERSION_1) {
            *penv = (void *)CVMjvmdiGetInterface1(vm);
            return JNI_OK;
#endif /* CVM_JVMDI */
        } else {
	    *penv = NULL;
	    return JNI_EVERSION;
        }
    } else {
	*penv = NULL;
        return JNI_EDETACHED;
    }
}

static const struct JNIInvokeInterface CVMmainJVMfuncs = {
    NULL,
    NULL,
    NULL,

    CVMjniDestroyJavaVM,
    CVMjniAttachCurrentThread,
    CVMjniDetachCurrentThread,
    CVMjniGetEnv,
};

void
CVMinitJNIEnv(CVMJNIEnv *p_env)
{
    p_env->vector = &CVMmainJNIfuncs;
}

void
CVMdestroyJNIEnv(CVMJNIEnv *p_env)
{
}

struct JNINativeInterface *
CVMjniGetInstrumentableJNINativeInterface()
{
    return &CVMmainJNIfuncs;
}

void
CVMinitJNIJavaVM(CVMJNIJavaVM *p_javaVM)
{
    p_javaVM->vector = &CVMmainJVMfuncs;
}

void
CVMdestroyJNIJavaVM(CVMJNIJavaVM *p_javaVM)
{
}
