/*
 * -----------------------------------------------------------------
 *  sync.c :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)sync.c	1.56 00/10/31
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/defs.h"
#include "javavm/include/sync.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/localroots.h"
#include "javavm/include/common_exceptions.h"

#include "javavm/include/clib.h"

#include "javavm/export/jvmcei.h"
/*
 * Reentrant mutexes
 */

void
CVMreentrantMutexInitRK(CVMSysMonitor *m, CVMReentrantMutex * rm, CVMExecEnv *owner,
    CVMUint32 count)
{
    extern intFuncPtr JVMCeiFilters[];
    extern int JVMCeiInitialized;
    int cont = JVMCEI_CONTINUE_CURRENT_CALL;

    rm->count = count;
    rm->owner = owner;

    /* Dio: calling the filter */
    if (JVMCeiInitialized &&
	(JVMCeiFilters[JVMCEF_BEFORE_INIT_MONITOR] != NULL)
	&& (m!= NULL))
      {
	struct arg_t { CVMSysMonitor **pmon;} arg;
	arg.pmon = &m;
	
	cont = JVMCeiFilters[JVMCEF_BEFORE_INIT_MONITOR]((void *) &arg);
      }

    if (cont == JVMCEI_CONTINUE_CURRENT_CALL)
      CVMmutexInit(&rm->mutex);
    if (count > 0) {
	CVMBool success = CVMmutexTryLock(&rm->mutex);
	CVMassert(success); (void) success;
    }
}

void
CVMreentrantMutexInit(CVMReentrantMutex * rm, CVMExecEnv *owner,
    CVMUint32 count)
{
  CVMreentrantMutexInitRK(NULL, rm, owner, count);
}

void
CVMreentrantMutexDestroy(CVMReentrantMutex * rm)
{
    CVMmutexDestroy(&rm->mutex);
}

CVMBool
CVMreentrantMutexTryLock(CVMExecEnv * ee, CVMReentrantMutex * rm)
{
    if (rm->owner == ee) {
	++rm->count;
    } else {
	if (CVMmutexTryLock(&rm->mutex)) {
	    rm->owner = ee;
	    rm->count = 1;
	} else {
	    return CVM_FALSE;
	}
    }
    return CVM_TRUE;
}

void
CVMreentrantMutexLock(CVMExecEnv * ee, CVMReentrantMutex * rm)
{
    if (rm->owner == ee) {
	++rm->count;
    } else {
	CVMassert(CVMD_isgcSafe(ee));
	CVMmutexLock(&rm->mutex);
	rm->owner = ee;
	rm->count = 1;
    }
}

void
CVMreentrantMutexUnlock(CVMExecEnv * ee, CVMReentrantMutex * rm)
{
    CVMassert(rm->count > 0);
    CVMassert(rm->owner == ee);
    if (--rm->count == 0) {
	rm->owner = NULL;
	/* printf("mutexunlock: unlocking (%p)\n", rm);*/
	CVMmutexUnlock(&rm->mutex);
    }
}

CVMBool
CVMreentrantMutexIAmOwner(CVMExecEnv * ee, CVMReentrantMutex * rm)
{
    CVMassert(rm->owner != ee || rm->count > 0);
    return (rm->owner == ee);
}

/*
 * CVM_FALSE is returned to indicate that the wait was interrupted,
 * otherwise CVM_TRUE is returned.
 */
CVMBool
CVMreentrantMutexWait(CVMExecEnv *ee, CVMReentrantMutex *rm,
    CVMCondVar *c, CVMInt64 millis)
{
    CVMUint32 count;
    CVMBool result;

    CVMassert(CVMD_isgcSafe(ee));
    CVMassert(rm->owner == ee);

    count = rm->count;
    rm->owner = NULL;
    rm->count = 0;
    result = CVMcondvarWait(c, &rm->mutex, millis);
    rm->owner = ee;
    rm->count = count;
    return result;
}


/*
 * Raw Monitors
 */

void
CVMsysMonitorInit(CVMSysMonitor *m, CVMExecEnv *owner, CVMUint32 count)
{
    CVMreentrantMutexInitRK(m, &m->rmutex, owner, count);
    CVMcondvarInit(&m->condvar, &m->rmutex.mutex);
}

void
CVMsysMonitorDestroy(CVMSysMonitor *m)
{
    CVMcondvarDestroy(&m->condvar);
    CVMreentrantMutexDestroy(&m->rmutex);
}

/* Purpose: Notifies the next thread waiting on this monitor. */
/* Returns: CVM_FALSE if the current thread is not the owner of monitor. */
/* NOTE: The ee arg is expected to be same as CVMgetEE() i.e the ee for the
         current thread. */
CVMBool
CVMsysMonitorNotify(CVMExecEnv *ee, CVMSysMonitor* m)
{
    if (CVMsysMonitorOwner(m) != ee) {
        return CVM_FALSE;
    }
    CVMcondvarNotify(&m->condvar);
    return CVM_TRUE;
}

/* Purpose: Notifies all threads waiting on this monitor. */
/* Returns: CVM_FALSE if the current thread is not the owner of monitor. */
/* NOTE: The ee arg is expected to be same as CVMgetEE() i.e the ee for the
         current thread. */
CVMBool
CVMsysMonitorNotifyAll(CVMExecEnv *ee, CVMSysMonitor *m)
{
    if (CVMsysMonitorOwner(m) != ee) {
        return CVM_FALSE;
    }
    CVMcondvarNotifyAll(&m->condvar);
    return CVM_TRUE;
}

CVMWaitStatus
CVMsysMonitorWait(CVMExecEnv *ee, CVMSysMonitor *mon, CVMInt64 millis)
{
    CVMassert(CVMD_isgcSafe(ee));
    if (CVMsysMonitorOwner(mon) != ee) {
	return CVM_WAIT_NOT_OWNER;
    }
    if (CVMreentrantMutexWait(ee, &mon->rmutex, &mon->condvar, millis)) {
	return CVM_WAIT_OK;
    } else {
	return CVM_WAIT_INTERRUPTED;
    }
}


/*
 * VM mutexes
 */

void
CVMsysMutexInit(CVMSysMutex * m, const char * name, CVMUint8 rank)
{
    m->name = name;
    m->rank = rank;
    CVMreentrantMutexInit(&m->rmutex, NULL, 0);
#ifdef CVM_DEBUG
    m->nextOwned = NULL;
#endif
}

void
CVMsysMutexDestroy(CVMSysMutex * m)
{
    CVMreentrantMutexDestroy(&m->rmutex);
#ifdef CVM_DEBUG
    CVMassert(m->nextOwned == NULL && m->rmutex.owner == NULL);
#endif
}

CVMBool
CVMsysMutexTryLock(CVMExecEnv * ee, CVMSysMutex * m)
{
    CVMBool success;
    CVMtraceSysMutex(("CVMsysMutexTryLock(%s)\n", m->name));
#ifdef CVM_DEBUG
    CVMassert(ee->sysLocks == NULL || m->rmutex.owner == ee ||
	m->rank > ee->sysLocks->rank);
#endif
    success = CVMreentrantMutexTryLock(ee, &m->rmutex);
#ifdef CVM_DEBUG
    if (success && CVMsysMutexCount(m) == 1) {
	m->nextOwned = ee->sysLocks;
	ee->sysLocks = m;
    }
#endif
    return success;
}

void
CVMsysMutexLock(CVMExecEnv * ee, CVMSysMutex * m)
{
    CVMtraceSysMutex(("CVMsysMutexLock(\"%s\")\n", m->name));
#ifdef CVM_DEBUG
    if(!(ee->sysLocks == NULL || m->rmutex.owner == ee ||
    	m->rank > ee->sysLocks->rank))
    {
        CVMconsolePrintf("ee->sysLocks = %x, ee = %x, m->rmutex.owner = %x, m->rank = %d, ee->sysLocks->rank = %d, m->name = %s\n", ee->sysLocks, ee, m->rmutex.owner, m->rank, ee->sysLocks->rank, m->name);
    }
    CVMassert(ee->sysLocks == NULL || m->rmutex.owner == ee ||
	m->rank > ee->sysLocks->rank);
#endif
    CVMreentrantMutexLock(ee, &m->rmutex);
#ifdef CVM_DEBUG
    if (CVMsysMutexCount(m) == 1) {
	m->nextOwned = ee->sysLocks;
	ee->sysLocks = m;
    }
#endif
}

void
CVMsysMutexUnlock(CVMExecEnv * ee, CVMSysMutex * m)
{
    CVMtraceSysMutex(("CVMsysMutexUnlock(\"%s\")\n", m->name));
#ifdef CVM_DEBUG
    CVMassert(CVMsysMutexCount(m) > 1 || ee->sysLocks == m);
    if (CVMsysMutexCount(m) == 1) {
	ee->sysLocks = m->nextOwned;
	m->nextOwned = NULL;
    }
#endif
    CVMreentrantMutexUnlock(ee, &m->rmutex);
}

/*
 * CVM_FALSE is returned to indicate that the wait was interrupted,
 * otherwise CVM_TRUE is returned.
 */
CVMBool
CVMsysMutexWait(CVMExecEnv * ee, CVMSysMutex *m,
    CVMCondVar * c, CVMInt64 millis)
{
    CVMBool ret;
#ifdef CVM_DEBUG
    if (CVMsysMutexCount(m) == 1) {
	ee->sysLocks = ee->sysLocks->nextOwned;
    }
#endif
    ret = CVMreentrantMutexWait(ee, &m->rmutex, c, millis);
#ifdef CVM_DEBUG
    if (CVMsysMutexCount(m) == 1) {
	m->nextOwned = ee->sysLocks;
	ee->sysLocks = m;
    }
#endif
    return ret;
}

#ifdef CVM_DEBUG
void
CVMsysMicroLock(CVMExecEnv *ee, CVMSysMicroLock l)
{
    CVMassert(++ee->microLock == 1);
    CVMsysMicroLockImpl(l);
}

void
CVMsysMicroUnlock(CVMExecEnv *ee, CVMSysMicroLock l)
{
    CVMassert(--ee->microLock == 0);
    CVMsysMicroUnlockImpl(l);
}
#endif
