/*
 * @(#)xrun.c	1.4 00/09/08
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/classes.h"
#include "javavm/include/sync.h"
#include "javavm/include/globals.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/preloader.h"
#include "javavm/include/globalroots.h"
#include "javavm/include/directmem.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/stackmaps.h"
#include "javavm/include/utils.h"
#include "javavm/include/xrun.h"

CVMBool CVMXrunInitTable(CVMXrunTable *table, CVMInt32 numXrunArguments)
{
    CVMXrunSetCnt(table, 0);
    CVMXrunSetItemTable(table, (CVMXrunItem *)malloc(sizeof(struct CVMXrunItem) * numXrunArguments)); 
    return (CVMXrunGetItemTable(table) != NULL);
}

void CVMXrunAppendToTable(CVMXrunTable *table,
			  jobject libRef, JVM_OnUnload_t fptr) 
{
    CVMXrunSetNextItem(table, libRef, fptr); 
    CVMXrunBumpCnt(table);
}

/*
 * Call JVM_OnUnload function on each shared library.
 * Delete all nativeLibrary object refs. 
 * Free Xrun_table in CVMglobals.
 */
void CVMXrunProcessTable(CVMXrunTable *Xrun_table, 
			JavaVM *vm, JNIEnv *env)
{
    CVMXrunItem *itemPtr = CVMXrunGetItemTable(Xrun_table); 
    CVMInt32 i = 0;

    if (CVMXrunIsEmptyTable(Xrun_table)) {
	return;
    }
    while (i < CVMXrunGetCnt(Xrun_table)) {
	JVM_OnUnload_t JVM_OnUnload = itemPtr[i].onUnloadFunc;
	jobject libRef = itemPtr[i++].shareLibRef;
	/* call JVM_OnUnload function if any */
	if (JVM_OnUnload != NULL) {
	    (*JVM_OnUnload)(vm, NULL, NULL);
	}

	/* delete native library object ref */
	CVMassert(libRef != NULL);
	(*env)->DeleteGlobalRef(env, libRef); 
    }
    free(itemPtr);
}

