/*
 * @(#)JDKClass.java	1.3 00/11/07
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

package vm;
/*
 * JDK-VM-specific internal representation of
 * a class. Target-machine independent.
 * There is a references from each instance of components.ClassInfo to 
 * one of these, and a reference back as well.
 *
 * See also JDKVM for VM-specific info not associated directly with a class.
 */
import components.*;
import util.*;
import Const;
import Util;
import Str2ID;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.util.StringTokenizer;

public class
JDKClass extends ClassClass {
    protected int		typeCode = 0;
    public JDKMethodInfo	methods[];

    public JDKClass( ClassInfo c ){
	ci = c;
	c.vmClass = this;

	if ( c.methods != null ){
	    // per-method, JDK-specific info.
	    methods = new JDKMethodInfo[ c.methods.length ];
	    for( int i = 0; i < methods.length; i++ ){
		components.MethodInfo m = c.methods[i];
		methods[i] = new JDKMethodInfo( m );
		if (!hasStaticInitializer &&
		    m.isStaticMember() &&
		    m.name.string.equals(/*NOI18N*/"<clinit>")) {
		    hasStaticInitializer = true;
		}
	    }
	}
    }


    /*
     * In JDK1.1, the implementation of reflection expects
     * certain typecodes to be stuffed in the cbTypeCode field
     * of certain classes. In un-modified JDK, the reflection
     * implementation stuffs this code there itself. But this
     * requires writing in the classClass data structure, which
     * we'd like to make read-only! Thus we do the equivalent
     * job here, during image building.
     *
     */
    public int
    typecode(){
	return typeCode;
    }

    /* Inline all the methods in this code object */
    public void getInlining() { 
        for( int i = 0; i < methods.length; i++ ){
	    methods[i].getInlining();
	}
    }
}
