/*
 * @(#)Class.c	1.107 00/09/21
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*-
 *      Implementation of class Class
 *
 *      former threadruntime.c, Sun Sep 22 12:09:39 1991
 */

#ifdef JDK12
#include <string.h>
#include <stdlib.h>
#else
#include "javavm/include/clib.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/common_exceptions.h"
#include "generated/offsets/java_lang_Class.h"
#endif

#include "jni.h"
#include "jni_util.h"
#include "jvm.h"
#include "java_lang_Class.h"

#define OBJ "Ljava/lang/Object;"
#define CLS "Ljava/lang/Class;"
#define STR "Ljava/lang/String;"
#define JCL "Ljava/lang/ClassLoader;"
#define FLD "Ljava/lang/reflect/Field;"
#define MHD "Ljava/lang/reflect/Method;"
#define CTR "Ljava/lang/reflect/Constructor;"
#define PD  "Ljava/security/ProtectionDomain;"

#ifdef CVM_JAVALANGCLASS_NOT_ROMIZED
static JNINativeMethod methods[] = {
    {"getName",          "()" STR,          (void *)&JVM_GetClassName},
    {"getSuperclass",    "()" CLS,          NULL},
    {"getInterfaces",    "()[" CLS,         (void *)&JVM_GetClassInterfaces},
    {"getClassLoader0",  "()" JCL,          (void *)&JVM_GetClassLoader},
    {"getSigners",       "()[" OBJ,         (void *)&JVM_GetClassSigners},
    {"setSigners",       "([" OBJ ")V",     (void *)&JVM_SetClassSigners},
    {"isArray",          "()Z",             (void *)&JVM_IsArrayClass},
    {"isPrimitive",      "()Z",             (void *)&JVM_IsPrimitiveClass},
    {"getComponentType", "()" CLS,          (void *)&JVM_GetComponentType},
    {"getModifiers",     "()I",             (void *)&JVM_GetClassModifiers},
    {"getFields0",       "(I)[" FLD,        (void *)&JVM_GetClassFields},
    {"getMethods0",      "(I)[" MHD,        (void *)&JVM_GetClassMethods},
    {"getConstructors0", "(I)[" CTR,        (void *)&JVM_GetClassConstructors},
    {"getField0",        "(" STR "I)" FLD,  (void *)&JVM_GetClassField},
    {"getMethod0", "(" STR "[" CLS "I)" MHD,(void *)&JVM_GetClassMethod},
    {"getConstructor0",  "([" CLS "I)" CTR, (void *)&JVM_GetClassConstructor},
    {"getProtectionDomain0", "()" PD,       (void *)&JVM_GetProtectionDomain},
    {"setProtectionDomain0", "(" PD ")V",   (void *)&JVM_SetProtectionDomain},
    {"getDeclaredClasses0",  "()[" CLS,      (void *)&JVM_GetDeclaredClasses},
    {"getDeclaringClass",   "()" CLS,      (void *)&JVM_GetDeclaringClass},
};
#endif

#undef OBJ
#undef CLS
#undef STR
#undef JCL
#undef FLD
#undef MHD
#undef CTR
#undef PD

#undef gcSafeRef2Class
#define gcSafeRef2Class(ee, cl)	CVMgcSafeClassRef2ClassBlock(ee, cl)

JNIEXPORT jboolean JNICALL
Java_java_lang_Class_checkInitializedFlag(JNIEnv *env, jclass cls)
{
    CVMClassBlock* cb = gcSafeRef2Class(CVMjniEnv2ExecEnv(env), cls);
    return CVMcbCheckRuntimeFlag(cb, INITIALIZED);
}

/*
 * Return true if this class is currently being initialized. If me == false,
 * then don't return true if the class is being initialized by this thread.
 */
JNIEXPORT jboolean JNICALL
Java_java_lang_Class_checkInitializingFlag(JNIEnv *env, jclass cls, 
					   jboolean me)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = gcSafeRef2Class(ee, cls);
    CVMExecEnv* clinitEE = CVMcbClinitEE(cb);

    return (clinitEE != NULL) && ((clinitEE == ee) == me);
}

JNIEXPORT jboolean JNICALL
Java_java_lang_Class_checkErrorFlag(JNIEnv *env, jclass cls)
{
    CVMClassBlock* cb = gcSafeRef2Class(CVMjniEnv2ExecEnv(env), cls);
    return CVMcbCheckRuntimeFlag(cb, ERROR);
}

JNIEXPORT void JNICALL
Java_java_lang_Class_setInitializedFlag(JNIEnv *env, jclass cls)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = gcSafeRef2Class(ee, cls);
    CVMcbSetRuntimeFlag(cb, ee, INITIALIZED);
}

JNIEXPORT void JNICALL
Java_java_lang_Class_setInitializingFlag(JNIEnv *env, jclass cls)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = gcSafeRef2Class(ee, cls);
    CVMcbClinitEE(cb) = ee;
}

JNIEXPORT void JNICALL
Java_java_lang_Class_setErrorFlag(JNIEnv *env, jclass cls)
{
    CVMExecEnv* ee = CVMjniEnv2ExecEnv(env);
    CVMClassBlock* cb = gcSafeRef2Class(ee, cls);
    CVMcbSetRuntimeFlag(cb, ee, ERROR);
}

JNIEXPORT void JNICALL
Java_java_lang_Class_clearInitializingFlag(JNIEnv *env, jclass cls)
{
    CVMClassBlock* cb = gcSafeRef2Class(CVMjniEnv2ExecEnv(env), cls);
    CVMcbClinitEE(cb) = NULL;
}

JNIEXPORT void JNICALL
Java_java_lang_Class_registerNatives(JNIEnv *env, jclass cls)
{
#ifdef CVM_JAVALANGCLASS_NOT_ROMIZED
    methods[1].fnPtr = (void *)(*env)->GetSuperclass;
    (*env)->RegisterNatives(env, cls, methods, 
			    sizeof(methods)/sizeof(JNINativeMethod));
#endif
}

JNIEXPORT jclass JNICALL
Java_java_lang_Class_forName0(JNIEnv *env, jclass this, jstring classname,
			      jboolean initialize, jobject loader)
{
    char *clname;
    jclass cls = 0;
    char buf[128];
    int len;
    int unicode_len;

    if (classname == NULL) {
        JNU_ThrowNullPointerException(env, 0);
	return 0;
    }

    len = (*env)->GetStringUTFLength(env, classname);
    unicode_len = (*env)->GetStringLength(env, classname);
    if (len >= sizeof(buf)) {
        clname = malloc(len + 1);
        if (clname == NULL) {
            JNU_ThrowOutOfMemoryError(env, NULL);
            return NULL;
        }
    } else {
        clname = buf;
    }
    (*env)->GetStringUTFRegion(env, classname, 0, unicode_len, clname);

    if (VerifyFixClassname(clname) == JNI_TRUE) {
	/* slashes present in clname, use name b4 translation for exception */
	(*env)->GetStringUTFRegion(env, classname, 0, unicode_len, clname);
	JNU_ThrowClassNotFoundException(env, clname);
	goto done;
    }

    if (!VerifyClassname(clname, JNI_TRUE)) {  /* expects slashed name */
        JNU_ThrowClassNotFoundException(env, clname);
	goto done;
    }

    cls = JVM_FindClassFromClassLoader(env, clname, initialize,
				       loader, JNI_FALSE);
 done:
    if (clname != buf) {
	free(clname);
    }

    return cls;
}

JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isInstance(JNIEnv *env, jobject cls, jobject obj)
{
    if (obj == NULL) {
        return JNI_FALSE;
    }
    return (*env)->IsInstanceOf(env, obj, (jclass)cls);
}

JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isAssignableFrom(JNIEnv *env, jobject cls, jobject cls2)
{
    if (cls2 == NULL) {
	JNU_ThrowNullPointerException(env, 0);
	return JNI_FALSE;
    }
    return (*env)->IsAssignableFrom(env, cls2, cls);
}

JNIEXPORT jclass JNICALL
Java_java_lang_Class_getPrimitiveClass(JNIEnv *env,
				       jclass cls,
				       jstring name)
{
    const char *utfName;
    jclass result;

    if (name == NULL) {
	JNU_ThrowNullPointerException(env, 0);
	return NULL;
    }

    utfName = (*env)->GetStringUTFChars(env, name, 0);
    if (utfName == 0)
        return NULL;

    result = JVM_FindPrimitiveClass(env, utfName);

    (*env)->ReleaseStringUTFChars(env, name, utfName);

    return result;
}

#ifndef CVM_JAVALANGCLASS_NOT_ROMIZED
JNIEXPORT jclass JNICALL
Java_java_lang_Class_getSuperclass(JNIEnv *env, jobject this)
{
    return (*env)->GetSuperclass(env, this);
}
#endif
